﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Text;
using System.Windows.Forms;
using WobbrockLib.Windows;

namespace WobbrockLib.Controls
{
    /// <summary>
    /// 
    /// </summary>
    [ToolboxBitmap(typeof(System.Windows.Forms.ListBox))]
    public partial class DisabledItemsListBox : ListBox
    {
        private int _lastIdx;
        private char _chDisabled;

        /// <summary>
        /// 
        /// </summary>
        public DisabledItemsListBox()
			: this(null)
        {
            // do nothing
        }

		/// <summary>
		/// 
		/// </summary>
		/// <param name="container"></param>
		public DisabledItemsListBox(IContainer container)
		{
			container?.Add(this);
			InitializeComponent();
			base.DrawMode = DrawMode.OwnerDrawFixed;
			base.SelectionMode = SelectionMode.One;
			_chDisabled = '~';
			_lastIdx = -1;
		}

        /// <summary>
        /// 
        /// </summary>
        [DefaultValue(typeof(DrawMode), "OwnerDrawFixed")]
        public new DrawMode DrawMode
        {
            get { return DrawMode.OwnerDrawFixed; }
        }

        /// <summary>
        /// 
        /// </summary>
        [DefaultValue(typeof(SelectionMode), "One")]
        public new SelectionMode SelectionMode
        {
            get { return SelectionMode.One; }
        }

        /// <summary>
        /// 
        /// </summary>
        [DefaultValue('~')]
        [RefreshProperties(RefreshProperties.Repaint)]
        public char DisabledPrefix
        {
            get { return _chDisabled; }
            set { _chDisabled = value; }
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="e"></param>
        protected override void OnDrawItem(DrawItemEventArgs e)
        {
            string s = String.Empty;
            if (this.Items.Count > 0 && e.Index >= 0) // protect against the empty list
            {
                Brush brush = Brushes.Black;
                s = this.Items[e.Index].ToString();
                if (s[0] == _chDisabled)
                {
                    s = s.Substring(1);
                    brush = Brushes.Gray;
                }
                e.DrawBackground(); // draw the background for all items
                e.Graphics.DrawString(s, e.Font, brush, e.Bounds, StringFormat.GenericDefault); // draw all items
            }
            if ((e.State & DrawItemState.Selected) != 0) // turn selected item white
            {
                e.Graphics.DrawString(s, e.Font, Brushes.White, e.Bounds, StringFormat.GenericDefault);
                if (this.Focused) // draw focus rect around selected item
                {
                    Pen pen = new Pen(Color.DimGray, 1f);
                    pen.DashStyle = DashStyle.Dash;
                    Rectangle r = e.Bounds;
                    r.Width--;
                    r.Height--;
                    e.Graphics.DrawRectangle(pen, r);
                    pen.Dispose();
                }
            }
            base.OnDrawItem(e);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="index"></param>
        /// <returns></returns>
        public bool GetItemEnabled(int index)
        {
            return this.Items[index].ToString()[0] != _chDisabled;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="index"></param>
        /// <param name="enabled"></param>
        public void SetItemEnabled(int index, bool enabled)
        {
            string s = this.Items[index].ToString();
            if (enabled && s[0] == _chDisabled)
            {
                s = s.Substring(1);
                this.Items[index] = s;
            }
            else if (!enabled && s[0] != _chDisabled)
            {
                s = _chDisabled + s;
                this.Items[index] = s;
                if (index == this.SelectedIndex)
                    this.SelectedIndex = -1; // deselect if we've disabled the formerly selected item
            }
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="m"></param>
        protected override void WndProc(ref Message m)
        {
            if (    m.Msg == (int) Win32.WM.LBUTTONDOWN // left button clicks
                || (m.Msg == (int) Win32.WM.MOUSEMOVE && (m.WParam.ToInt32() & (int) Win32.MK.LBUTTON) != 0) // or left button drags
                )
            {
                int x = m.LParam.ToInt32() & 0xFF; // low word
                int y = m.LParam.ToInt32() >> 16; // high word
                int idx = this.IndexFromPoint(x, y);
                if (0 <= idx && idx < this.Items.Count && this.Items[idx].ToString()[0] == _chDisabled)
                    return; // do nothing, and disabled items won't be selected
            }
            base.WndProc(ref m);
        }

        /// <summary>
        /// 
        /// </summary>
        public override int SelectedIndex
        {
            get
            {
                return base.SelectedIndex;
            }
            set
            {
                int idx = value;
                while (0 <= idx && idx < this.Items.Count && this.Items[idx].ToString()[0] == _chDisabled)
                    idx++; // advance to the next
                if (idx < this.Items.Count)
                    base.SelectedIndex = idx;
            }
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="e"></param>
        protected override void OnSelectedIndexChanged(EventArgs e)
        {
            _lastIdx = this.SelectedIndex; // record for OnKeyDown
            base.OnSelectedIndexChanged(e);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="e"></param>
        protected override void OnKeyDown(KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Down || e.KeyCode == Keys.Right)
            {
                int nextIdx = _lastIdx + 1;
                while (0 <= nextIdx && nextIdx < Items.Count && Items[nextIdx].ToString()[0] == _chDisabled)
                    nextIdx++; // go to next enabled item
                if (0 <= nextIdx && nextIdx < Items.Count && Items[nextIdx].ToString()[0] != _chDisabled)
                    this.SelectedIndex = nextIdx;
                e.Handled = true;
            }
            else if (e.KeyCode == Keys.Up || e.KeyCode == Keys.Left)
            {
                if (_lastIdx >= 0)
                {
                    int prevIdx = _lastIdx - 1;
                    while (0 <= prevIdx && prevIdx < Items.Count && Items[prevIdx].ToString()[0] == _chDisabled)
                        prevIdx--; // go to previous enabled item
                    if (0 <= prevIdx && prevIdx < Items.Count && Items[prevIdx].ToString()[0] != _chDisabled)
                        this.SelectedIndex = prevIdx;
                }
                e.Handled = true;
            }
            else if (e.KeyCode == Keys.End || e.KeyCode == Keys.PageDown)
            {
                for (int i = Items.Count - 1; i >= 0; i--)
                {
                    if (Items[i].ToString()[0] != _chDisabled)
                    {
                        this.SelectedIndex = i;
                        break;
                    }
                }
                e.Handled = true;
            }
            else if (e.KeyCode == Keys.Home || e.KeyCode == Keys.PageUp)
            {
                for (int i = 0; i < Items.Count; i++)
                {
                    if (Items[i].ToString()[0] != _chDisabled)
                    {
                        this.SelectedIndex = i;
                        break;
                    }
                }
                e.Handled = true;
            }
            base.OnKeyDown(e);
        }
    }
}
