﻿using System;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;
using WobbrockLib.Types;
using WobbrockLib.Extensions;

namespace WobbrockLib.Controls
{
	/// <summary>
	/// Provides a button control whose text label can be explicitly positioned within it. This 
	/// is useful for certain fonts and font sizes who render "lower" or "higher" than we'd like
	/// on a regular button label. With this control, the label can be positioned precisely on 
	/// the button face.
	/// </summary>
	[ToolboxBitmap(typeof(System.Windows.Forms.Button))]
	public partial class PositionableTextButton : Button
	{
		#region Fields

		private Point _position;
		private string _text;
		private Color _disabledForeColor;

		#endregion

		#region Events

		/// <summary>
		/// Event that fires when the value of the TextPosition property has changed.
		/// </summary>
		public event EventHandler TextPositionChanged;

		/// <summary>
		/// Event that fires when the value of the DisabledForeColor property has changed.
		/// </summary>
		public event EventHandler DisabledForeColorChanged;

		#endregion

		#region Constructor

		/// <summary>
		/// Default constructor.
		/// </summary>
		public PositionableTextButton()
			: this(null)
		{
			// do nothing
		}

		/// <summary>
		/// Constructor with container.
		/// </summary>
		/// <param name="container">The container into which to place this control.</param>
		public PositionableTextButton(IContainer container)
		{
			container?.Add(this);
			InitializeComponent();

			_position = Point.Empty;
			_text = String.Empty;
			_disabledForeColor = Color.DarkGray;

			base.UseMnemonic = false; // changed default from base
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets the position of the label within the button.
		/// </summary>
		[DefaultValue(typeof(Point), "0,0")]
		[Category("Appearance")]
		[Description("Indicates the location of the text label within the button's face.")]
		[RefreshProperties(RefreshProperties.None)]
		public Point TextPosition
		{
			get
			{
				return _position;
			}
			set
			{
				if (value != _position)
				{
					_position = value;
					this.Invalidate();
					this.TextPositionChanged?.Invoke(this, EventArgs.Empty);
				}
			}
		}

		/// <summary>
		/// Gets or sets the color of the text label when the button is disabled.
		/// </summary>
		[DefaultValue(typeof(Color), "DarkGray")]
		[Category("Appearance")]
		[Description("Indicates the color of the text label when the button's Enabled property is false.")]
		[RefreshProperties(RefreshProperties.None)]
		public Color DisabledForeColor
		{
			get
			{
				return _disabledForeColor;
			}
			set
			{
				if (value != _disabledForeColor)
				{
					_disabledForeColor = value;
					if (!this.Enabled)
						Invalidate();
					this.DisabledForeColorChanged?.Invoke(this, EventArgs.Empty);
				}
			}
		}

		/// <summary>
		/// Gets or sets the text label drawn in the button. We use an underlying member
		/// variable so that we can temorarily clear it in the OnPaint override without
		/// re-triggering the OnPaint override.
		/// </summary>
		public override string Text
		{
			get
			{
				return _text;
			}
			set
			{
				if (value != _text)
				{
					_text = value;
					Invalidate();
					base.OnTextChanged(EventArgs.Empty);
				}
			}
		}

		/// <summary>
		/// Gets or sets whether the text is rendered from left-to-right from the Position
		/// location or right-to-left.
		/// </summary>
		[DefaultValue(typeof(RightToLeft), "No")]
		public override RightToLeft RightToLeft
		{
			get { return base.RightToLeft; }
			set { base.RightToLeft = value; }
		}

		/// <summary>
		/// Gets the text alignment, which is read-only.
		/// </summary>
		[DefaultValue(typeof(ContentAlignment), "TopLeft")]
		public override ContentAlignment TextAlign
		{
			get
			{
				return (this.RightToLeft == RightToLeft.Yes) ? ContentAlignment.TopRight : ContentAlignment.TopLeft;
			}
		}

		/// <summary>
		/// Gets the relationship between the text and any image on the control.
		/// </summary>
		[Browsable(false)]
		public new TextImageRelation TextImageRelation
		{
			get { return TextImageRelation.Overlay; }
		}

		/// <summary>
		/// Gets whether this button can use an ampersand (&) to indicate a mnemonic key.
		/// </summary>
		[Browsable(false)]
		[DefaultValue(false)]
		public new bool UseMnemonic
		{
			get { return false; }
		}

		/// <summary>
		/// Gets whether the text label in this button will render with ellipsis (...) if there
		/// is insufficient room to display it.
		/// </summary>
		[Browsable(false)]
		public new bool AutoEllipsis
		{
			get { return false; }
		}

		/// <summary>
		/// Gets whether this button renders text compatible with previous versions of Microsoft
		/// Windows Forms.
		/// </summary>
		[Browsable(false)]
		public new bool UseCompatibleTextRendering
		{
			get { return false; }
		}

		#endregion

		#region Paint

		/// <summary>
		/// Draw our positionable text label. We have to temporarily clear the underlying Text
		/// property member so that the base method call doesn't paint the Text label in the
		/// center of the button.
		/// </summary>
		/// <param name="e">The arguments for this event.</param>
		protected override void OnPaint(PaintEventArgs e)
		{
			string txtSave = _text; // save
			_text = String.Empty; // clear for painting
			base.OnPaint(e);
			_text = txtSave; // restore

			// now draw the label from the top-left corner, or if RTL is set, then from the top-right corner
			SolidBrush brush = new SolidBrush(this.Enabled ? this.ForeColor : _disabledForeColor);
			if (this.RightToLeft == RightToLeft.Yes)
			{
				SizeF sz = e.Graphics.MeasureString(_text, this.Font);
				e.Graphics.DrawString(_text, this.Font, brush, _position.X - sz.Width, _position.Y);
			}
			else
			{
				e.Graphics.DrawString(_text, this.Font, brush, _position.X, _position.Y);
			}
			brush.Dispose();
		}

		#endregion

		#region Methods

		/// <summary>
		/// Gets the position of the text label as a [0,1] percentage of the width or height of 
		/// the button face on which it is drawn. The position is from the top-left corner of 
		/// the text label, or if the RightToLeft property is set, then it is from the top-right 
		/// corner.
		/// </summary>
		/// <param name="dimension">Whether to get the horizontal dimension, vertical dimension, 
		/// or both. We use a ScrollBars enumerated type for convenience.</param>
		public SizeF GetTextPositionPercentage(ScrollBars dimension)
		{
			SizeF percentage = new SizeF(-1f, -1f);
			if (dimension == ScrollBars.Horizontal || dimension == ScrollBars.Both)
			{
				percentage.Width = (float) GeotrigEx.MapValue(_position.X, 0.0, this.Width, 0.0, 1.0);
			}
			if (dimension == ScrollBars.Vertical || dimension == ScrollBars.Both)
			{
				percentage.Height = (float) GeotrigEx.MapValue(_position.Y, 0.0, this.Height, 0.0, 1.0);
			}
			return percentage;
		}

		/// <summary>
		/// Sets the position of the text label on this button as a percentage of the button's width,
		/// height, or both. The position is from the top-left corner of the text label, or if the 
		/// RightToLeft property is set, then it is from the top-right corner.
		/// </summary>
		/// <param name="dimension">Whether to set the horizontal dimension, vertical dimension, 
		/// or both. We use a ScrollBars enumerated type for convenience.</param>
		/// <param name="percentage">A value in [0,1] for the position to set the label along the 
		/// given dimension(s).</param>
		public void SetTextPositionPercentage(ScrollBars dimension, float percentage)
		{
			percentage = (float) GeotrigEx.PinValue(percentage, 0.0, 1.0);
			if (dimension == ScrollBars.Horizontal || dimension == ScrollBars.Both)
			{
				_position.X = (int) Math.Round(GeotrigEx.MapValue(percentage, 0.0, 1.0, 0.0, this.Width));
			}
			if (dimension == ScrollBars.Vertical || dimension == ScrollBars.Both)
			{
				_position.Y = (int) Math.Round(GeotrigEx.MapValue(percentage, 0.0, 1.0, 0.0, this.Height));
			}
			Invalidate();
		}

		/// <summary>
		/// Centers the text label horizontally on the button face at a [0,1] percentage of the 
		/// way down the vertical face of the button. The position is from the top-left corner of 
		/// the text label, or if the RightToLeft property is set, then it is from the top-right 
		/// corner.
		/// </summary>
		/// <param name="percentage">A value in [0,1] for the position to set the label along 
		/// the given dimension.</param>
		public void CenterTextAtVerticalPercentage(float percentage)
		{
			percentage = (float) GeotrigEx.PinValue(percentage, 0.0, 1.0);
			_position.Y = (int) Math.Round(GeotrigEx.MapValue(percentage, 0.0, 1.0, 0.0, this.Height));

			Graphics g = Graphics.FromHwnd(this.Handle);
			SizeF sz = g.MeasureString(_text, this.Font);
			g.Dispose();

			_position.X = (int) Math.Round(this.Width / 2.0 - sz.Width / 2.0);
			Invalidate();
		}

		/// <summary>
		/// Centers the text label on the button face in both the horizontal and vertical directions.
		/// </summary>
		public void CenterText()
		{
			Graphics g = Graphics.FromHwnd(this.Handle);
			SizeF sz = g.MeasureString(_text, this.Font);
			g.Dispose();

			PointR pt = new PointR(
				this.Width / 2.0 - sz.Width / 2.0,
				this.Height / 2.0 - sz.Height / 2.0
				);

			_position = (Point) pt; // rounds to nearest (x,y) ints
			Invalidate();
		}

		#endregion
	}
}
