﻿using System;
using System.Drawing;
using System.ComponentModel;
using System.Windows.Forms;

namespace WobbrockLib.Controls
{
	/// <summary>
	/// <para>
	/// Implements a ToolTip variant that is only shown once automatically, and after being
	/// shown, will not be shown again unless SetToolTip or Show is called. This is useful
	/// when we want to show a ToolTip message and then not re-show it automatically after 
	/// the mouse leaves the control and later returns.
	/// </para>
	/// <para>
	/// Also, this ToolTip class implements three fixes: (1) When calling Show with either a 
	/// Point location or an (x,y) location and NO duration, the ToolTip persists until it is 
	/// clicked. Now, those versions of Show behave like the others, where the ToolTip disappears 
	/// if the mouse leaves the control. (2) When calling Show with either a Point location or 
	/// an (x,y) location AND a duration, the ToolTip shows for its entire duration, even when 
	/// the mouse leaves the control; this behavior was inconsistent with other versions of Show, 
	/// and is now remedied. (3) When using a ToolTip in the form of a Balloon, the first call to
	/// Show has a bug that treats the ToolTip like a traditional rectangular tip, and thus, its 
	/// Balloon pointer is in the wrong location. Immediate back-to-back calls of the Show method 
	/// resolve this problem. This is a known bug from Microsoft.
	/// </para>
	/// </summary>
	public partial class ShowOnceToolTip : ToolTip
	{
		#region Fields

		private bool _shown;

		#endregion

		#region Constructor

		/// <summary>
		/// Default constructor.
		/// </summary>
		public ShowOnceToolTip()
			: this(null)
		{
			// do nothing
		}

		/// <summary>
		/// Constructor with container.
		/// </summary>
		/// <param name="container">The container in which to place this control.</param>
		public ShowOnceToolTip(IContainer container)
		{
			container?.Add(this);
			InitializeComponent();
			_shown = false;
			this.Popup += ToolTip_Popup;
		}

		#endregion

		#region Popup Event

		/// <summary>
		/// Event handler for when this ToolTip is about to appear. We prevent the ToolTip
		/// from automatically showing again if it has already been shown. To show it again,
		/// another call to SetToolTip or to Show is required.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void ToolTip_Popup(object sender, PopupEventArgs e)
		{
			e.Cancel = _shown;
			if (!e.Cancel) // we're about to show the tip
				_shown = true;
		}

		#endregion

		#region Show

		/// <summary>
		/// Associates this ToolTip with the given control and displays the ToolTip with
		/// the given caption. The tip will automatically disappear if the mouse moves 
		/// off the given control.
		/// </summary>
		/// <param name="control">The control with which to associate this ToolTip.</param>
		/// <param name="caption">The string to display in this ToolTip.</param>
		public new void SetToolTip(Control control, string caption)
		{
			_shown = false;
			base.SetToolTip(control, caption);
		}

		/// <summary>
		/// Shows this ToolTip with the given text associated with the given window. The tip
		/// will automatically disappear if the mouse moves off the given window.
		/// </summary>
		/// <param name="text">The string to display in this ToolTip.</param>
		/// <param name="window">The window with which to associate this ToolTip. This is
		/// usually a Control.</param>
		public new void Show(string text, IWin32Window window)
		{
			_shown = false;
			base.Show(text, window);
		}

		/// <summary>
		/// Shows this ToolTip with the given text associated with the given window for the
		/// given duration. The tip will automatically disappear if the mouse moves off the 
		/// given window, even if the duration has not expired.
		/// </summary>
		/// <param name="text">The string to display in this ToolTip.</param>
		/// <param name="window">The window with which to associate this ToolTip. This is
		/// usually a Control.</param>
		/// <param name="duration">The time, in milliseconds, to show this ToolTip.</param>
		public new void Show(string text, IWin32Window window, int duration)
		{
			_shown = false;
			base.Show(text, window, duration);
		}

		/// <summary>
		/// Shows this ToolTip with the given text associated with the given window at the
		/// given location. The tip would normally show until it is clicked, even when the
		/// mouse moves off the window. Therefore, we implement our own MouseLeave functionality
		/// to hide the tip. Also, if the tip is in Balloon form, it will not point correctly
		/// to the given position on the first showing. Therefore, we must call Show twice to
		/// correct this problem. This is a known Microsoft bug.
		/// </summary>
		/// <param name="text">The string to display in this ToolTip.</param>
		/// <param name="window">The window with which to associate this ToolTip. This is
		/// usually a Control.</param>
		/// <param name="point">The location relative to the window's coordinates at which
		/// to show this ToolTip.</param>
		public new void Show(string text, IWin32Window window, Point point)
		{
			Control control = (Control) window;
			control.MouseLeave -= Control_MouseLeave;
			control.MouseLeave += Control_MouseLeave;
			if (this.IsBalloon)
				base.Show(text, window, point);

			_shown = false;
			base.Show(text, window, point);
		}

		/// <summary>
		/// Shows this ToolTip with the given text associated with the given window at the
		/// given (x,y) location. The tip would normally show until it is clicked, even when the
		/// mouse moves off the window. Therefore, we implement our own MouseLeave functionality
		/// to hide the tip. Also, if the tip is in Balloon form, it will not point correctly
		/// to the given position on the first showing. Therefore, we must call Show twice to
		/// correct this problem. This is a known Microsoft bug.
		/// </summary>
		/// <param name="text">The string to display in this ToolTip.</param>
		/// <param name="window">The window with which to associate this ToolTip. This is
		/// usually a Control.</param>
		/// <param name="x">The x-coordinate at which to show this ToolTip relative to the
		/// window's coordinates.</param>
		/// <param name="y">The y-coordinate at which to show this ToolTip relative to the
		/// window's coordinates.</param>
		public new void Show(string text, IWin32Window window, int x, int y)
		{
			Control control = (Control) window;
			control.MouseLeave -= Control_MouseLeave;
			control.MouseLeave += Control_MouseLeave;
			if (this.IsBalloon)
				base.Show(text, window, x, y);

			_shown = false;
			base.Show(text, window, x, y);
		}

		/// <summary>
		/// Shows this ToolTip with the given text associated with the given window at the
		/// given location. The tip would normally show for its full duration, even when the
		/// mouse moves off the window. Therefore, we implement our own MouseLeave functionality
		/// to hide the tip. Also, if the tip is in Balloon form, it will not point correctly
		/// to the given position on the first showing. Therefore, we must call Show twice to
		/// correct this problem. This is a known Microsoft bug. 
		/// </summary>
		/// <param name="text">The string to display in this ToolTip.</param>
		/// <param name="window">The window with which to associate this ToolTip. This is
		/// usually a Control.</param>
		/// <param name="point">The location relative to the window's coordinates at which
		/// to show this ToolTip.</param>
		/// <param name="duration">The time, in milliseconds, to show this ToolTip.</param>
		public new void Show(string text, IWin32Window window, Point point, int duration)
		{
			Control control = (Control) window;
			control.MouseLeave -= Control_MouseLeave;
			control.MouseLeave += Control_MouseLeave;
			if (this.IsBalloon)
				base.Show(text, window, point, duration);

			_shown = false;
			base.Show(text, window, point, duration);
		}

		/// <summary>
		/// Shows this ToolTip with the given text associated with the given window at the
		/// given (x,y) location. The tip would normally show for its full duration, even when
		/// the mouse moves off the window. Therefore, we implement our own MouseLeave functionality
		/// to hide the tip. Also, if the tip is in Balloon form, it will not point correctly
		/// to the given position on the first showing. Therefore, we must call Show twice to
		/// correct this problem. This is a known Microsoft bug. 
		/// </summary>
		/// <param name="text">The string to display in this ToolTip.</param>
		/// <param name="window">The window with which to associate this ToolTip. This is
		/// usually a Control.</param>
		/// <param name="x">The x-coordinate at which to show this ToolTip relative to the
		/// window's coordinates.</param>
		/// <param name="y">The y-coordinate at which to show this ToolTip relative to the
		/// window's coordinates.</param>
		/// <param name="duration">The time, in milliseconds, to show this ToolTip.</param>
		public new void Show(string text, IWin32Window window, int x, int y, int duration)
		{
			Control control = (Control) window;
			control.MouseLeave -= Control_MouseLeave;
			control.MouseLeave += Control_MouseLeave;
			if (this.IsBalloon)
				base.Show(text, window, x, y, duration);

			_shown = false;
			base.Show(text, window, x, y, duration);
		}

		/// <summary>
		/// The event handler for when the mouse leaves the control with which this ToolTip 
		/// becomes associated. When that happens, we hide the ToolTip.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void Control_MouseLeave(object sender, EventArgs e)
		{
			Control control = (Control) sender;
			control.MouseLeave -= Control_MouseLeave;
			this.Hide(control);
		}

		#endregion
	}
}
