﻿using System;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;
using WobbrockLib.Types;

namespace WobbrockLib.Controls
{
	/// <summary>
	/// This class implements a Windows Forms timer that can be given an object reference
	/// when it is started, and when its Tick event fires, that object state is an argument
	/// to the event.
	/// </summary>
	[ToolboxBitmap(typeof(System.Windows.Forms.Timer))]
	public partial class StateTimer : System.Windows.Forms.Timer
	{
		// Fields
		private object _state;

		/// <summary>
		/// Event that occurs whenever the specified interval time elapses.
		/// </summary>
		[Category("Behavior")]
		[Description("Occurs whenever the specified interval time elapses.")]
		[DefaultValue("")]
		[RefreshProperties(RefreshProperties.None)]
		public new event EventHandler<object> Tick;

		/// <summary>
		/// Constructs a default instance of this timer.
		/// </summary>
		public StateTimer()
			: this(null)
		{
			// do nothing
		}

		/// <summary>
		/// Constructs an instance of this timer to be added to the given container.
		/// </summary>
		/// <param name="container">The container to which to add this instance.</param>
		public StateTimer(IContainer container)
		{
			container?.Add(this);
			InitializeComponent();
		}

		/// <summary>
		/// Starts the timer with the given object state, which will be passed in as the 
		/// second argument to any Tick event handler registered by a client.
		/// </summary>
		/// <param name="state">The object state the caller wishes to be passed into their
		/// timer's Tick event handler</param>
		public void Start(object state)
		{
			_state = state;
			base.Start();
		}

		/// <summary>
		/// Override of the main event method for this class. The Tick event is invoked and 
		/// the second parameter is the object state that was passed into the Start method,
		/// or null if none was passed.
		/// </summary>
		/// <param name="e">The arguments for this event, which are unused.</param>
		/// <remarks>Note that no base.OnTick method is called, because for this kind of Timer
		/// subclass, only this new Tick event is fired, which contains the object state.</remarks>
		protected override void OnTick(EventArgs e)
		{
			Tick?.Invoke(this, _state);
		}

	}
}
