﻿using System;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;
using WobbrockLib.Windows;

namespace WobbrockLib.Controls
{
	/// <summary>
	/// This class implements a type of TrackBar that uses a ToolTip to show the
	/// current value whenever the TrackBar is scrolled with the mouse or arrow 
	/// keys, or whenever a mouse button goes down on the TrackBar thumb. The
	/// behavior of the ToolTip can be customized.
	/// </summary>
	[ToolboxBitmap(typeof(System.Windows.Forms.TrackBar))]
	public partial class ToolTipTrackBar : TrackBar
	{
		#region Fields 

		private readonly Size DefaultOffset = new Size(-20, -20);
		private const int DefaultDuration = 2000;
		private int _duration;

		#endregion

		#region Constructor

		/// <summary>
		/// Construct a default ToolTipTrackBar.
		/// </summary>
		public ToolTipTrackBar()
			: this(null)
		{
			// do nothing
		}

		/// <summary>
		/// Construct a ToolTipTrackBar contained by the given container.
		/// </summary>
		/// <param name="container">The container of this ToolTipTrackBar.</param>
		public ToolTipTrackBar(IContainer container)
		{
			container?.Add(this);
			InitializeComponent();
			MyInitialization();
		}

		/// <summary>
		/// Helper method for common initialization settings for the constructors to use.
		/// These values should match the defaults for these properties.
		/// </summary>
		private void MyInitialization()
		{
			this.ShowToolTip = true;
			this.TipOffset = DefaultOffset;
			this.TipDuration = DefaultDuration;
			this.TipAnimation = true;
			this.TipFading = true;
			this.TipShowAlways = false;
			this.TipBackColor = SystemColors.Info;
			this.TipForeColor = SystemColors.InfoText;
		}

		#endregion

		#region ToolTip Event Handlers

		/// <summary>
		/// Handles the draw event for the ToolTip, which has its OwnerDraw property set to true.
		/// Without doing this, the BackColor and ForeColor properties would not be honored.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void ttValue_Draw(object sender, DrawToolTipEventArgs e)
		{
			e.DrawBackground();
			e.DrawText();
			e.DrawBorder();
		}

		#endregion

		#region TrackBar Event Overrides

		/// <summary>
		/// When the thumb moves whether by the mouse dragging it or by the arrow keys on
		/// the keyboard, show the ToolTip.
		/// </summary>
		/// <param name="e">The arguments for this event.</param>
		protected override void OnScroll(EventArgs e)
		{
			base.OnScroll(e);
			if (this.ShowToolTip)
			{
				Rectangle rc = GetThumbRect();
				ShowTip(rc);
			}
		}

		/// <summary>
		/// If the mouse goes down inside the TrackBar thumb, show the ToolTip.
		/// </summary>
		/// <param name="e">The arguments for this event.</param>
		protected override void OnMouseDown(MouseEventArgs e)
		{
			base.OnMouseDown(e);
			if (this.ShowToolTip)
			{
				Rectangle rc = GetThumbRect();
				if (rc.Contains(e.X, e.Y))
					ShowTip(rc);
			}
		}

		/// <summary>
		/// When the mouse comes up, if we've been showing the ToolTip and it has
		/// a non-infinite duration, then hide the ToolTip.
		/// </summary>
		/// <param name="e">The arguments for this event.</param>
		protected override void OnMouseUp(MouseEventArgs e)
		{
			base.OnMouseUp(e);
			if (this.ShowToolTip && this.TipDuration != -1)
				ttValue.Hide(this);
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets whether the ToolTip associated with this TrackBar should be shown. If false, this 
		/// TrackBar will behave no differently than a Windows Forms TrackBar.
		/// </summary>
		[Category("ToolTip")]
		[Description("Determines whether the ToolTip that shows this TrackBar's changing value should be shown.")]
		[DefaultValue(true)]
		[RefreshProperties(RefreshProperties.None)]
		public bool ShowToolTip { get; set; }

		/// <summary>
		/// Gets or sets a negative or positive offset from the TrackBar thumb at which to position the top left of
		/// the ToolTip. A value of (0,0) would have the tip exactly overlap the TrackBar thumb. The default of
		/// (-20,-20) positions the tip further to the upper-left of the thumb, leaving the thumb visible.
		/// </summary>
		/// <remarks>
		/// A caller could "pin" the tip to the exact right end of the TrackBar's trough by subscribing to the Scroll 
		/// event and then using this code, assuming "tb" is the ToolTipTrackBar instance:
		/// <code>
		///		Rectangle rcThumb = tb.GetThumbRect();
		///		Rectangle rcTrough = tb.GetTroughRect();
		///		tb.TipOffset = new Size(rcTrough.Right - rcThumb.Left, 0);
		/// </code>
		/// </remarks>
		[Category("ToolTip")]
		[Description("Determines a negative or positive offset from the TrackBar thumb at which to position the top left of the ToolTip.")]
		[DefaultValue(typeof(Size), "-20,-20")]
		[RefreshProperties(RefreshProperties.None)]
		public Size TipOffset { get; set; }

		/// <summary>
		/// Gets or sets the duration in milliseconds of this ToolTip. A value of -1 means the ToolTip will remain
		/// persistent after being shown. The default duration is 2000 ms.
		/// </summary>
		[Category("ToolTip")]
		[Description("Determines the duration in milliseconds for the duration of the ToolTip. Use -1 for infinite duration.")]
		[DefaultValue(DefaultDuration)]
		[RefreshProperties(RefreshProperties.None)]
		public int TipDuration
		{
			get { return _duration; }
			set { _duration = (value > 0) ? value : -1; }
		}

		/// <summary>
		/// Gets or sets the ToolTip's UseAnimation property.
		/// </summary>
		[Category("ToolTip")]
		[Description("Determines whether the ToolTip uses animation when shown.")]
		[DefaultValue(true)]
		[RefreshProperties(RefreshProperties.None)]
		public bool TipAnimation
		{
			get { return ttValue.UseAnimation; }
			set { ttValue.UseAnimation = value; }
		}

		/// <summary>
		/// Gets or sets the ToolTip's UseFading property.
		/// </summary>
		[Category("ToolTip")]
		[Description("Determines whether the ToolTip uses fading when shown.")]
		[DefaultValue(true)]
		[RefreshProperties(RefreshProperties.None)]
		public bool TipFading
		{
			get { return ttValue.UseFading; }
			set { ttValue.UseFading = value; }
		}

		/// <summary>
		/// Gets or sets the ToolTip's ShowAlways property.
		/// </summary>
		[Category("ToolTip")]
		[Description("Determines whether the ToolTip will be shown always, even if the parent window is not active.")]
		[DefaultValue(false)]
		[RefreshProperties(RefreshProperties.None)]
		public bool TipShowAlways
		{
			get { return ttValue.ShowAlways; }
			set { ttValue.ShowAlways = value; }
		}

		/// <summary>
		/// Gets or sets the ToolTip's BackColor property.
		/// </summary>
		[Category("ToolTip")]
		[Description("Determines the background color of this ToolTip.")]
		[DefaultValue(typeof(Color), "Info")]
		[RefreshProperties(RefreshProperties.Repaint)]
		public Color TipBackColor
		{
			get { return ttValue.BackColor; }
			set { ttValue.BackColor = value; }
		}

		/// <summary>
		/// Gets or sets the ToolTip's ForeColor property.
		/// </summary>
		[Category("ToolTip")]
		[Description("Determines the foreground text color of this ToolTip.")]
		[DefaultValue(typeof(Color), "InfoText")]
		[RefreshProperties(RefreshProperties.Repaint)]
		public Color TipForeColor
		{
			get { return ttValue.ForeColor; }
			set { ttValue.ForeColor = value; }
		}

		#endregion

		#region Methods

		/// <summary>
		/// Gets the rectangle of the TrackBar's thumb, in coordinates relative to the TrackBar itself.
		/// </summary>
		/// <returns>A rectangle with the thumb's coordinates.</returns>
		/// <remarks>This method is public so that clients can use it to help position the ToolTip
		/// using the TipOffset property. See the code example for that property.</remarks>
		public Rectangle GetThumbRect()
		{
			Win32.RECT r = Win32.RECT.Empty;
			Win32.SendMessage(this.Handle, (uint) Win32.TBM.GETTHUMBRECT, 0u, ref r);
			return new Rectangle(r.left, r.top, r.right - r.left, r.bottom - r.top);
		}

		/// <summary>
		/// Gets the rectangle of the TrackBar's trough, also known as its channel, in coordinates relative to 
		/// the TrackBar itself.
		/// </summary>
		/// <returns>A rectangle with the trough's coordinates.</returns>
		/// <remarks>This method is public so that clients can use it to help position the ToolTip
		/// using the TipOffset property. See the code example for that property.</remarks>
		public Rectangle GetTroughRect()
		{
			Win32.RECT r = Win32.RECT.Empty;
			Win32.SendMessage(this.Handle, (uint) Win32.TBM.GETCHANNELRECT, 0u, ref r);
			return new Rectangle(r.left, r.top, r.right - r.left, r.bottom - r.top);
		}

		/// <summary>
		/// Helper method to show the ToolTip using the given thumb's position. The tip will
		/// be shown relative to the thumb's position taking into account the TipOffset 
		/// property.
		/// </summary>
		/// <param name="rcThumb">The current position of the TrackBar's thumb.</param>
		private void ShowTip(Rectangle rcThumb)
		{
			if (_duration != -1)
			{
				ttValue.Show(
					this.Value.ToString(),
					this,
					rcThumb.Left + TipOffset.Width,
					rcThumb.Top + TipOffset.Height,
					_duration);
			}
			else // infinite duration
			{
				ttValue.Show(
					this.Value.ToString(),
					this,
					rcThumb.Left + TipOffset.Width,
					rcThumb.Top + TipOffset.Height);
			}
		}

		#endregion

	}
}
