﻿using System;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;

namespace WobbrockLib.Controls
{
    /// <summary>
    /// A transparent label that is see through and also can be clicked through to the underlying
	/// user interface.
    /// </summary>
    [ToolboxBitmap(typeof(System.Windows.Forms.Label))]
    public partial class TransparentLabel : TransparentLayer
    {
		#region Constructor

		/// <summary>
		/// Default constructor.
		/// </summary>
		public TransparentLabel()
			: this(null)
        {
            // do nothing
        }

        /// <summary>
        /// Constructor with a container.
        /// </summary>
        /// <param name="container">The container in which to add this label.</param>
		public TransparentLabel(IContainer container)
            : base(container)
        {
			base.RefreshDelay = -1; // default for label
		}

		#endregion

		#region Paint

		/// <summary>
		/// Draw the label's text.
		/// </summary>
		/// <param name="e">The arguments for this event.</param>
		protected override void OnPaint(PaintEventArgs e)
		{
			base.OnPaint(e); // have our clients draw and our base class draw first

			using (SolidBrush brush = new SolidBrush(this.ForeColor))
			{
				if (this.RightToLeft == RightToLeft.Yes)
				{
					SizeF size = e.Graphics.MeasureString(this.Text, this.Font);
					e.Graphics.DrawString(this.Text, this.Font, brush, this.Width - size.Width, 0);
				}
				else
				{
					e.Graphics.DrawString(this.Text, this.Font, brush, -1, 0);
				}
			}
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets the interval, in milliseconds, at which this transparent label is 
		/// refreshed when transparent. This issue is moot when the label is not currently 
		/// transparent. A value less than or equal to zero means the parent's area beneath
		/// the label will not be refreshed periodically, which is usually fine. This property
		/// is overridden simply to change the default value.
		/// </summary>
		[DefaultValue(-1)] // infinite
        [Category("Behavior")]
		[Description("The rate, in milliseconds, for the parent of this label to redraw itself beneath this label, and for this label to redraw itself.")]
		[RefreshProperties(RefreshProperties.None)]
        public override int RefreshDelay
        {
            get { return base.RefreshDelay; }
            set { base.RefreshDelay = value; }
        }

        /// <summary>
        /// Gets or sets the bounds of this label. Establish a new default value.
        /// </summary>
        [DefaultValue(typeof(Size), "100,13")]
        [Category("Appearance")]
        [RefreshProperties(RefreshProperties.None)]
        public new Size Size
        {
            get { return base.Size; }
            set { base.Size = value; }
        }

        /// <summary>
        /// Gets or sets the text property of this label.
        /// </summary>
        public override string Text
        {
            get
            {
                return base.Text;
            }
            set
            {
                base.Text = value;
                this.RecreateHandle();
            }
        }

        /// <summary>
        /// Gets or sets the font property of this label.
        /// </summary>
        public override Font Font
        {
            get
            {
                return base.Font;
            }
            set
            {
                base.Font = value;
                this.RecreateHandle();
            }
        }

        /// <summary>
        /// Gets or sets the right-to-left reading order of this property.
        /// </summary>
        public override RightToLeft RightToLeft
        {
            get
            {
                return base.RightToLeft;
            }
            set
            {
                base.RightToLeft = value;
                this.RecreateHandle();
            }
        }

		#endregion
	}
}