﻿using System;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;
using WobbrockLib.Windows;

namespace WobbrockLib.Controls
{
    /// <summary>
    /// Provides a truly transparent layer that not only can be seen through but that can
	/// also be clicked through. It contains a timer that is optionally used to cause its
	/// parent to invalidate and repaint the region of itself that is beneath this 
	/// transparent layer, which is necessary for some applications.
    /// </summary>
    [ToolboxBitmap(typeof(System.Windows.Forms.Control))]
    public partial class TransparentLayer : Control
    {
		#region Fields

		protected const int DefaultRefreshDelay = 55;

		private Timer _tmrRefresh;
		protected int _refreshDelay;
		protected bool _transparent;
		protected Color _borderColor;
		protected bool _drawBorder;
		protected int _borderThickness;

		#endregion

		#region Constructor

		/// <summary>
		/// Constructs a new TransparentLayer.
		/// </summary>
		public TransparentLayer()
            : this(null)
        {
            // do nothing
        }

        /// <summary>
        /// Constructor with a container.
        /// </summary>
        /// <param name="container">The container in which to add this layer.</param>
		public TransparentLayer(IContainer container)
        {
			container?.Add(this);
            InitializeComponent();
            this.SetStyle(ControlStyles.Selectable, false);
			this.SetStyle(ControlStyles.SupportsTransparentBackColor, true);
			base.BackColor = Color.Transparent;

			// set defaults
			_drawBorder = false;
			_borderColor = Color.Black;
			_borderThickness = 1;
			_refreshDelay = DefaultRefreshDelay;

			_tmrRefresh = new System.Windows.Forms.Timer();
			_tmrRefresh.Tick += new EventHandler(RefreshTimer_Tick);
			_tmrRefresh.Interval = DefaultRefreshDelay;
			_tmrRefresh.Enabled = false;

			this.Transparent = true; // default
		}

		#endregion

		#region Input Transparency

		/// <summary>
		/// Give ourselves the transparent window style initially. This is input 
		/// transparency, not visual transparency. See also the Transparent property.
		/// </summary>
		protected override CreateParams CreateParams
		{
			get
			{
				CreateParams cp = base.CreateParams;
				cp.ExStyle |= (int) Win32.WS_EX.TRANSPARENT;
				return cp;
			}
		}

		/// <summary>
		/// Overrides the main message pump for this control. We look for non-client
		/// hit test message and produce a transparency value result, which causes
		/// the framework to pass the click through ourselves to anything under.
		/// </summary>
		/// <param name="m">The message sent to our message queue.</param>
		protected override void WndProc(ref Message m)
		{
			if (_transparent && m.Msg == (int) Win32.WM.NCHITTEST)
				m.Result = new IntPtr((int) Win32.HT.TRANSPARENT);
			else
				base.WndProc(ref m);
		}

		#endregion

		#region Painting

		/// <summary>
		/// Do not paint the background if we're using transparency.
		/// </summary>
		/// <param name="e">The arguments for this event.</param>
		protected override void OnPaintBackground(PaintEventArgs e)
		{
			if (!_transparent)
			{
				base.OnPaintBackground(e);
			}
		}

		/// <summary>
		/// Possibly paint a border around ourselves if our properties call for it.
		/// </summary>
		/// <param name="e">The arguments for this event.</param>
		protected override void OnPaint(PaintEventArgs e)
		{
			base.OnPaint(e); // raise the Paint event to clients

			if (_drawBorder) // draw our border on top of whatever a client does
			{
				Pen pen = new Pen(_borderColor, _borderThickness);
				Rectangle r = this.ClientRectangle;
				r.X += _borderThickness / 2;
				r.Y += _borderThickness / 2;
				r.Width -= _borderThickness;
				r.Height -= _borderThickness;
				e.Graphics.DrawRectangle(pen, r);
				pen.Dispose();
			}
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets whether a border will be drawn around the outside of this transparent
		/// layer. Without a border, the bounds of the transparent layer can be hard to discern.
		/// </summary>
		[DefaultValue(false)]
		[Category("Appearance")]
		[Description("Indicates whether to draw a border around the transparent layer.")]
		[RefreshProperties(RefreshProperties.None)]
		public bool DrawBorder
		{
			get
			{
				return _drawBorder;
			}
			set
			{
				_drawBorder = value;
				this.Invalidate();
			}
		}

		/// <summary>
		/// Gets or sets what the border color is if a border is drawn.
		/// </summary>
		[DefaultValue(typeof(Color), "Black")]
		[Category("Appearance")]
		[Description("Indicates the color of the border drawn around the transparent layer. The DrawBorder property must be true to show the border.")]
		[RefreshProperties(RefreshProperties.None)]
		public Color BorderColor
		{
			get
			{
				return _borderColor;
			}
			set
			{
				_borderColor = value;
				this.Invalidate();
			}
		}

		/// <summary>
		/// Gets or sets how thick a line is drawn around the border, in pixels. The minimum
		/// value is 1 and the maximum is one less than half of the smaller of the layer's
		/// current width or height.
		/// </summary>
		[DefaultValue(1)]
		[Category("Appearance")]
		[Description("Indicates the thickness, in pixels, of the border drawn around the transparent layer. The DrawBorder property must be true to show the border.")]
		[RefreshProperties(RefreshProperties.None)]
		public int BorderThickness
		{
			get
			{
				return _borderThickness;
			}
			set
			{
				if (value > 0)
				{
					int halfMin = Math.Min(this.Width, this.Height) / 2;
					_borderThickness = (value < halfMin) ? value : halfMin - 1;
					this.Invalidate();
				}
			}
		}

        /// <summary>
        /// Gets or sets whether we are currently using transparency or not.
        /// </summary>
        [DefaultValue(true)]
		[Category("Appearance")]
		[Description("Indicates whether this layer's background should be drawn transparently and whether mouse events should go through it.")]
		[RefreshProperties(RefreshProperties.None)]
		public virtual bool Transparent
		{
			get
			{
				return _transparent;
			}
			set
			{
				_transparent = value;
				if (value) // turning transparency ON
				{
					WindowEx.SetInputTransparent(this.Handle, true);
					this.Parent?.Refresh();
				}
				else // turning transparency OFF
				{
					WindowEx.SetInputTransparent(this.Handle, false);
					this.Parent?.Refresh();
					this.Refresh();
				}
				_tmrRefresh.Enabled = (value && _refreshDelay > 0);
			}
		}

		/// <summary>
		/// Gets or sets the interval, in milliseconds, at which this transparent layer is 
		/// refreshed when transparent. This issue is moot when the layer is not currently 
		/// transparent. A value less than or equal to zero means the control will not be 
		/// refreshed periodically.
		/// </summary>
		[DefaultValue(DefaultRefreshDelay)]
		[Category("Behavior")]
		[Description("The time, in milliseconds, between updates for the parent of this layer to redraw itself beneath this layer, and for this layer to redraw itself.")]
		[RefreshProperties(RefreshProperties.None)]
		public virtual int RefreshDelay
		{
			get
			{
				return _refreshDelay;
			}
			set
			{
				if (value <= 0) // infinite (no refresh)
				{
					_refreshDelay = -1;
					_tmrRefresh.Enabled = false;
				}
				else
				{
					_refreshDelay = value;
					_tmrRefresh.Interval = value;
					_tmrRefresh.Enabled = _transparent;
				}
			}
		}

		#endregion

		#region Events

		/// <summary>
		/// When our location changes, invalidate our bounds on our parent.
		/// </summary>
		/// <param name="e">The arguments for this event.</param>
		protected override void OnLocationChanged(EventArgs e)
		{
			base.OnLocationChanged(e);
			this.Parent?.Invalidate(this.Bounds, true); // update our parent and ourselves!
		}

		/// <summary>
		/// When our location changes, invalidate our bounds on our parent.
		/// </summary>
		/// <param name="e">The arguments for this event.</param>
		protected override void OnSizeChanged(EventArgs e)
		{
			base.OnSizeChanged(e);
			this.Parent?.Invalidate(this.Bounds, true); // update our parent and ourselves!
		}

		#endregion

		#region Refresh Timer

		/// <summary>
		/// Causes the parent of this layer to redraw itself and its children. This is the 
		/// way we are invalidated and redrawn when we're transparent, because a direct call
		/// to Invalidate doesn't repaint us, since we don't paint our background when we're
		/// transparent!
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void RefreshTimer_Tick(object sender, EventArgs e)
        {
			this.Parent?.Invalidate(this.Bounds, true); // update our parent and ourselves
		}

		#endregion
	}
}
