﻿using System;
using System.IO;
using System.IO.MemoryMappedFiles;
using System.Runtime.InteropServices;

namespace WobbrockLib.Data
{
	/// <summary>
	/// Abstract class for using a memory mapped file to represent an array. Sublcasses can build on this foundation for 1-D arrays,
	/// 2-D matrices, 3-D tensors, or structures of higher dimensionality. Memory mapped files can be useful for very large data structures 
	/// exceeding C#'s 2GB object limit.
	/// </summary>
	/// <typeparam name="T">The data type stored in this array. Note that the value must be a non-nullable struct, which includes 
	/// the atomic C# types like int, char, byte, etc. It cannot be an object reference.</typeparam>
	public abstract class AbsMemoryMappedArray<T> where T : struct
	{
		// Inheritable fields
		protected static readonly int _typeSize = Marshal.SizeOf(typeof(T));

		// Private fields
		private MemoryMappedFile _mmf;
		private MemoryMappedViewAccessor _mmva;

		/// <summary>
		/// Constructor for a memory mapped file representing an array. Before the array can be used, it must be initialized.
		/// </summary>
		/// <remarks>The size, in bytes, of the data type to be stored in this array is calculated and stored so that subclasses can
		/// calculate proper byte offsets when accessing items in the array.</remarks>
		public AbsMemoryMappedArray()
		{
			// do nothing
		}

		/// <summary>
		/// Destructor. Ensures any non-null and open members are disposed of.
		/// </summary>
		~AbsMemoryMappedArray()
		{
			_mmva?.Dispose();
			_mmf?.Dispose();
		}

		/// <summary>
		/// Initializes a memory mapped array in system memory with the given number of bytes. Note that these are raw bytes 
		/// and not an array length or dimension.
		/// </summary>
		/// <param name="bytes">The raw number of bytes to allocate for this array.</param>
		/// <returns>True if initialized successfully; false otherwise.</returns>
		/// <remarks>If this method is called on an already-initialized array, that array is disposed of and a new one
		/// is initialized.</remarks>
		protected bool Initialize(long bytes)
		{
			this.Dispose(); // possibly dispose of this instance's resources if already in use

			_mmf = MemoryMappedFile.CreateNew(null, bytes); // create a new file
			if (_mmf != null)
			{
				_mmva = _mmf.CreateViewAccessor(); // get a view on the file
				if (_mmva == null)
				{
					_mmf.Dispose();
					return false; // failed to get the view
				}
				return true; // success!
			}
			return false; // failed to create new
		}

		/// <summary>
		/// Initializes a memory mapped array from a file it creates on disk. The file is created with the given number
		/// of bytes, which are all initialized to be zero.
		/// </summary>
		/// <param name="file">The file to create on disk and make a memory mapped array for.</param>
		/// <param name="bytes">The number of zero bytes to write to the file.</param>
		/// <param name="useIfExists">If true and a file exists at the given location, it will be used. If false, a new file will
		/// be created.</param>
		/// <returns>True if the file is created and the memory mapped array is established successfully; false otherwise.</returns>
		protected bool Initialize(string file, long bytes, bool useIfExists)
		{
			this.Dispose(); // possibly dispose of this instance's resources if already in use

			bool success = true;
			if (!useIfExists || !File.Exists(file))
			{
				BinaryWriter writer = null;
				try
				{
					FileStream fs = File.Create(file);
					writer = new BinaryWriter(fs);
					for (long i = 0L; i < bytes; i++)
					{
						writer.Write((byte) 0);
					}
				}
				catch (IOException ioex)
				{
					Console.WriteLine(ioex);
					success = false;
				}
				catch (Exception ex)
				{
					Console.WriteLine(ex);
					success = false;
				}
				finally
				{
					writer?.Close();
				}
			}
			if (success)
			{
				try
				{
					_mmf = MemoryMappedFile.CreateFromFile(file, FileMode.OpenOrCreate);
					if (_mmf != null)
					{
						_mmva = _mmf.CreateViewAccessor(); // get a view on the file
						if (_mmva == null)
						{
							_mmf.Dispose();
							return false; // failed to get the view
						}
						return true; // success!
					}
				}
				catch (IOException ioex)
				{
					Console.WriteLine(ioex);
				}
				catch (Exception ex)
				{
					Console.WriteLine(ex);
				}
				return false; // failed to create memory mapped file
			}
			return false; // failed to create file on disk
		}

		/// <summary>
		/// Disposes of the resources associated with this memory mapped file and its view accessor, if any.
		/// This method should be called after the client is done using this instance.
		/// </summary>
		public void Dispose()
		{
			_mmva?.Dispose();
			_mmf?.Dispose();
		}

		/// <summary>
		/// Gets the number of bytes allocated to this memory mapped array. Note that this does not mean that
		/// meaningful values have been written into this array's full capacity.
		/// </summary>
		public long Capacity
		{
			get 
			{
				return (_mmva != null) ? _mmva.Capacity : 0L;
			}
		}

		/// <summary>
		/// Gets the value at the given byte offset in this memory mapped file representing an array.
		/// </summary>
		/// <param name="offset">The byte offset from which to read the value.</param>
		/// <returns>The array value stored at the given byte offset.</returns>
		/// <remarks>Subclasses should work with array indices, not byte offsets, as regular arrays do, and can
		/// use the type size to calculate the proper byte offset into the array.</remarks>
		protected T GetValue(long offset)
		{
			_mmva.Read<T>(offset, out T value);
			return value;
		}

		/// <summary>
		/// Sets the value at the given byte offset in this memory mapped file representing an array.
		/// </summary>
		/// <param name="offset">The byte offset at which to set the given value.</param>
		/// <param name="value">The value to set at the given byte offset.</param>
		/// <remarks>Subclasses should work with array indices, not byte offsets, as regular arrays do, and can
		/// use the type size to calculate the proper byte offset into the array.</remarks>
		protected void SetValue(long offset, T value)
		{
			_mmva.Write<T>(offset, ref value);
		}

		/// <summary>
		/// Indexer for this array.
		/// </summary>
		/// <param name="offset">The byte offset at which to get or set the given value.</param>
		/// <returns>The array value stored at the given byte offset.</returns>
		protected T this[long offset]
		{
			get { return this.GetValue(offset); }
			set { this.SetValue(offset, value); }
		}

		/// <summary>
		/// Copies a number of bytes from the source array to the destination array at the given offsets. 
		/// </summary>
		/// <param name="sourceArray">The array from which to copy bytes.</param>
		/// <param name="sourceOffset">The source byte offset from which to start copying bytes.</param>
		/// <param name="destinationArray">The array to which to copy bytes.</param>
		/// <param name="destinationOffset">The destination byte offset to which to start copying bytes.</param>
		/// <param name="bytes">The number of bytes to copy.</param>
		protected static void Copy(AbsMemoryMappedArray<T> sourceArray, long sourceOffset, AbsMemoryMappedArray<T> destinationArray, long destinationOffset, long bytes)
		{
			for (long i = 0L, j = 0L; i < bytes; i++, j++)
			{
				byte b = sourceArray._mmva.ReadByte(sourceOffset + i);
				destinationArray._mmva.Write(destinationOffset + j, b);
			}
		}
		
	}
}
