﻿using System;
using System.Text;
using System.Collections.Generic;
using System.Windows.Forms;
using WobbrockLib.Windows;

namespace WobbrockLib.Devices
{
	/// <summary>
	/// Offers functions similar to those in WobbrockLib.Extensions but for the keyboard,
	/// which is why it is in WobbrockLib.Devices. These static functions enable the user to 
	/// synthesize keystroke events programmatically.
	/// </summary>
	public static class KeyboardEx
	{
		/// <summary>
		/// Sends the given key, perhaps in combination with the Ctrl, Shift, and/or Alt modifier keys.
		/// Not all possible Keys enumeration values are supported.
		/// </summary>
		/// <param name="key">The key to send, perhaps in combination with the Ctrl, Shift, and/or Alt 
		/// modifier keys. For example, to send Ctrl+A, use (Keys.Control | Keys.A).</param>
		/// <param name="sendWait">If true, the SendKeys.SendWait method will be used. Otherwise, the 
		/// SendKeys.Send method will be used.</param>
		/// <see cref="https://docs.microsoft.com/en-us/dotnet/api/system.windows.forms.sendkeys?view=netframework-4.8"/>
		public static void Send(Keys key, bool sendWait)
		{
			bool shift = (key & Keys.Shift) != 0;
			bool ctrl = (key & Keys.Control) != 0;
			bool alt = (key & Keys.Alt) != 0;

			key &= ~Keys.Shift;
			key &= ~Keys.Control;
			key &= ~Keys.Alt;

			StringBuilder sb = new StringBuilder();
			if (shift)
				sb.Append('+');
			if (ctrl)
				sb.Append('^');
			if (alt)
				sb.Append('%');

			sb.Append('(');
			switch (key)
			{
				case Keys.Back:
					sb.Append("{BACKSPACE}");
					break;
				case Keys.Pause:
					sb.Append("{BREAK}");
					break;
				case Keys.CapsLock:
					sb.Append("{CAPSLOCK}");
					break;
				case Keys.Delete:
					sb.Append("{DELETE}");
					break;
				case Keys.Down:
					sb.Append("{DOWN}");
					break;
				case Keys.End:
					sb.Append("{END}");
					break;
				case Keys.Enter:
					sb.Append("{ENTER}");
					break;
				case Keys.Escape:
					sb.Append("{ESC}");
					break;
				case Keys.Help:
					sb.Append("{HELP}");
					break;
				case Keys.Home:
					sb.Append("{HOME}");
					break;
				case Keys.Insert:
					sb.Append("{INSERT}");
					break;
				case Keys.Left:
					sb.Append("{LEFT}");
					break;
				case Keys.NumLock:
					sb.Append("{NUMLOCK}");
					break;
				case Keys.PageDown:
					sb.Append("{PGDN}");
					break;
				case Keys.PageUp:
					sb.Append("{PGUP}");
					break;
				case Keys.PrintScreen:
					sb.Append("{PRTSC}");
					break;
				case Keys.Right:
					sb.Append("{RIGHT}");
					break;
				case Keys.Scroll:
					sb.Append("{SCROLLLOCK}");
					break;
				case Keys.Tab:
					sb.Append("{TAB}");
					break;
				case Keys.Up:
					sb.Append("{UP}");
					break;
				case Keys.F1:
					sb.Append("{F1}");
					break;
				case Keys.F2:
					sb.Append("{F2}");
					break;
				case Keys.F3:
					sb.Append("{F3}");
					break;
				case Keys.F4:
					sb.Append("{F4}");
					break;
				case Keys.F5:
					sb.Append("{F5}");
					break;
				case Keys.F6:
					sb.Append("{F6}");
					break;
				case Keys.F7:
					sb.Append("{F7}");
					break;
				case Keys.F8:
					sb.Append("{F8}");
					break;
				case Keys.F9:
					sb.Append("{F9}");
					break;
				case Keys.F10:
					sb.Append("{F10}");
					break;
				case Keys.F11:
					sb.Append("{F11}");
					break;
				case Keys.F12:
					sb.Append("{F12}");
					break;
				case Keys.F13:
					sb.Append("{F13}");
					break;
				case Keys.F14:
					sb.Append("{F14}");
					break;
				case Keys.F15:
					sb.Append("{F15}");
					break;
				case Keys.F16:
					sb.Append("{F16}");
					break;
				case Keys.F17:
					sb.Append("{F17}");
					break;
				case Keys.F18:
					sb.Append("{F18}");
					break;
				case Keys.F19:
					sb.Append("{F19}");
					break;
				case Keys.F20:
					sb.Append("{F20}");
					break;
				case Keys.F21:
					sb.Append("{F21}");
					break;
				case Keys.F22:
					sb.Append("{F22}");
					break;
				case Keys.F23:
					sb.Append("{F23}");
					break;
				case Keys.F24:
					sb.Append("{F24}");
					break;
				case Keys.Add:
					sb.Append("{ADD}");
					break;
				case Keys.Subtract:
					sb.Append("{SUBTRACT}");
					break;
				case Keys.Multiply:
					sb.Append("{MULTIPLY}");
					break;
				case Keys.Divide:
					sb.Append("{DIVIDE}");
					break;
				case Keys.Space:
					sb.Append(' ');
					break;
				case Keys.D0:
					sb.Append('0');
					break;
				case Keys.D1:
					sb.Append('1');
					break;
				case Keys.D2:
					sb.Append('2');
					break;
				case Keys.D3:
					sb.Append('3');
					break;
				case Keys.D4:
					sb.Append('4');
					break;
				case Keys.D5:
					sb.Append('5');
					break;
				case Keys.D6:
					sb.Append('6');
					break;
				case Keys.D7:
					sb.Append('7');
					break;
				case Keys.D8:
					sb.Append('8');
					break;
				case Keys.D9:
					sb.Append('9');
					break;
				case Keys.A:
					sb.Append('A');
					break;
				case Keys.B:
					sb.Append('B');
					break;
				case Keys.C:
					sb.Append('C');
					break;
				case Keys.D:
					sb.Append('D');
					break;
				case Keys.E:
					sb.Append('E');
					break;
				case Keys.F:
					sb.Append('F');
					break;
				case Keys.G:
					sb.Append('G');
					break;
				case Keys.H:
					sb.Append('H');
					break;
				case Keys.I:
					sb.Append('I');
					break;
				case Keys.J:
					sb.Append('J');
					break;
				case Keys.K:
					sb.Append('K');
					break;
				case Keys.L:
					sb.Append('L');
					break;
				case Keys.M:
					sb.Append('M');
					break;
				case Keys.N:
					sb.Append('N');
					break;
				case Keys.O:
					sb.Append('O');
					break;
				case Keys.P:
					sb.Append('P');
					break;
				case Keys.Q:
					sb.Append('Q');
					break;
				case Keys.R:
					sb.Append('R');
					break;
				case Keys.S:
					sb.Append('S');
					break;
				case Keys.T:
					sb.Append('T');
					break;
				case Keys.U:
					sb.Append('U');
					break;
				case Keys.V:
					sb.Append('V');
					break;
				case Keys.W:
					sb.Append('W');
					break;
				case Keys.X:
					sb.Append('X');
					break;
				case Keys.Y:
					sb.Append('Y');
					break;
				case Keys.Z:
					sb.Append('Z');
					break;
				case Keys.NumPad0:
					sb.Append('0');
					break;
				case Keys.NumPad1:
					sb.Append('1');
					break;
				case Keys.NumPad2:
					sb.Append('2');
					break;
				case Keys.NumPad3:
					sb.Append('3');
					break;
				case Keys.NumPad4:
					sb.Append('4');
					break;
				case Keys.NumPad5:
					sb.Append('5');
					break;
				case Keys.NumPad6:
					sb.Append('6');
					break;
				case Keys.NumPad7:
					sb.Append('7');
					break;
				case Keys.NumPad8:
					sb.Append('8');
					break;
				case Keys.NumPad9:
					sb.Append('9');
					break;
				case Keys.Separator:
					sb.Append('.');
					break;
				case Keys.OemOpenBrackets:
					sb.Append("{[}");
					break;
				case Keys.OemCloseBrackets:
					sb.Append("{]}");
					break;
				case Keys.Oemtilde:
					sb.Append('`');
					break;
				case Keys.OemMinus:
					sb.Append('-');
					break;
				case Keys.Oemplus:
					sb.Append("{+}");
					break;
				case Keys.OemBackslash:
					sb.Append('\\');
					break;
				case Keys.Oem5:
					sb.Append('\\');
					break;
				case Keys.Oem1:
					sb.Append(';');
					break;
				case Keys.Oem7:
					sb.Append('\'');
					break;
				case Keys.Oemcomma:
					sb.Append(',');
					break;
				case Keys.OemPeriod:
					sb.Append('.');
					break;
				case Keys.OemQuestion:
					sb.Append('/');
					break;
				default:
					throw new ArgumentException(String.Format("Key value {0} not supported in KeyboardEx.Send.", key));
			}
			sb.Append(')');

			if (sendWait)
				SendKeys.SendWait(sb.ToString());
			else
				SendKeys.Send(sb.ToString());
		}

		/// <summary>
		/// Tests whether the given key is currently physically down on the keyboard.
		/// </summary>
		/// <param name="vk">The Win32 virtual key code to test if it is down.</param>
		/// <returns>True if the given key is currently down; false otherwise.</returns>
		public static bool IsKeyDown(Win32.VK vk)
		{
			return AreKeysDown(new Win32.VK[1] { vk });
		}

		/// <summary>
		/// Tests whether the given key is currently the only key physically down on the keyboard.
		/// </summary>
		/// <param name="vk">The Win32 virtual key code to test if it is down.</param>
		/// <returns>True if the given key is the only key currently down; false otherwise.</returns>
		public static bool IsOnlyKeyDown(Win32.VK vk)
		{
			return AreOnlyKeysDown(new Win32.VK[1] { vk });
		}

		/// <summary>
		/// Tests whether all of the given keys are currently physically down on the keyboard.
		/// </summary>
		/// <param name="vks">An array of virtual key codes to test if they are down.</param>
		/// <returns>True if the given virtual keys are currently down; false otherwise.</returns>
		public static bool AreKeysDown(Win32.VK[] vks)
		{
			if (vks == null || vks.Length == 0)
				return false;

			bool down = true;
			for (int i = 0; i < vks.Length; i++)
			{
				short keyState = Win32.GetKeyState((int) vks[i]);
				if ((keyState & 0x80) != 0x80)
				{
					down = false;
					break;
				}
			}
			return down;
		}

		/// <summary>
		/// Tests whether all of the given keys are currently physically down on the keyboard, and no other keys 
		/// are also down.
		/// </summary>
		/// <param name="vks">An array of virtual key codes to test if they are down.</param>
		/// <returns>True if the given virtual keys are currently down and no other keys are down; false 
		/// otherwise.</returns>
		/// <remarks>We iterative from value 8 to 254, inclusive, because values less than 8 correspond go the 
		/// mouse buttons, and value 255 is unassigned, and yet sometimes returns with the most significant bit 
		/// set.</remarks>
		public static bool AreOnlyKeysDown(Win32.VK[] vks)
		{
			if (vks == null || vks.Length == 0)
				return false;

			byte[] keyStates = new byte[256];
			if (Win32.GetKeyboardState(keyStates))
			{
				// ensure all of our given keys are down
				for (int i = 0; i < vks.Length; i++)
				{
					if ((keyStates[(int) vks[i]] & 0x80) != 0x80)
						return false;
					else
						keyStates[(int) vks[i]] = 0x00; // clear
				}

				// if a modifier key is pressed, the flag will also be set 
				// for which specific modifier key it was. but we don't want 
				// these to thwart our testing of whether any other keys are 
				// currently down.
				keyStates[(int) Win32.VK.LCONTROL] = 0x00;
				keyStates[(int) Win32.VK.RCONTROL] = 0x00;
				keyStates[(int) Win32.VK.LSHIFT] = 0x00;
				keyStates[(int) Win32.VK.RSHIFT] = 0x00;
				keyStates[(int) Win32.VK.LMENU] = 0x00;
				keyStates[(int) Win32.VK.RMENU] = 0x00;

				// now ensure no other keys are down
				for (int j = 8; j < keyStates.Length - 1; j++) // skip mouse buttons and disregard slot [255]
				{
					if ((keyStates[j] & 0x80) == 0x80)
						return false;
				}
			}
			return true; // the vks were the only keys down
		}

		/// <summary>
		/// Tests whether any keys are down on the physical keyboard. "Down" means actively being held down, not 
		/// toggled as in the case of Caps Lock, Scroll Lock, Function Lock, or Number Lock. 
		/// </summary>
		/// <returns>True if any keys are currently physically down; false otherwise.</returns>
		/// <remarks>We iterative from value 8 to 254, inclusive, because values less than 8 correspond go the 
		/// mouse buttons, and value 255 is unassigned, and yet sometimes returns with the most significant bit 
		/// set.</remarks>
		public static bool AreAnyKeysDown()
		{
			bool down = false;
			byte[] keyStates = new byte[256];
			if (Win32.GetKeyboardState(keyStates))
			{
				for (int i = 8; i < keyStates.Length - 1; i++) // skip mouse buttons and disregard slot [255]
				{
					if ((keyStates[i] & 0x80) == 0x80)
					{
						down = true;
						break;
					}
				}
			}
			return down;
		}
	}
}
