﻿using System;
using System.Windows.Forms;
using System.Drawing;
using System.Diagnostics;
using System.Runtime.InteropServices;
using Microsoft.Win32;
using WobbrockLib.Windows;

namespace WobbrockLib.Devices
{
    /// <summary>
    /// 
    /// </summary>
    public class LowLevelMouseHook
    {
        /// <summary>
        /// Internal utility struct for tracking the possibility of double-clicks arising from low level mouse events.
        /// </summary>
        [StructLayout(LayoutKind.Sequential)]
        private struct DoubleClickState
        {
            public static readonly DoubleClickState Empty;
            public MouseButtons LastDownButton;
            public long LastDownTime;
            public Point LastDownPt;
            public bool IssueClickOnMouseUp;
            public bool IssueDoubleClickOnMouseUp;
        }

        /// <summary>
        /// Events fired by this RawInputMouse instance.
        /// </summary>
        public event EventHandler<MouseEventArgs> HookMouseMove;
        public event EventHandler<MouseEventArgs> HookMouseDown;
        public event EventHandler<MouseEventArgs> HookMouseUp;
        public event EventHandler<MouseEventArgs> HookMouseClick;
        public event EventHandler<MouseEventArgs> HookMouseDoubleClick;
        public event EventHandler<MouseEventArgs> HookMouseWheel;

        // double-click members
        private DoubleClickState _state;

        // hook members
        private IntPtr _hookID;
        private Win32.WindowsHookProc _hookProc;

		/// <summary>
		/// Constructor for this LowLevelMouseHook.
		/// </summary>
		/// <param name="name"></param>
		public LowLevelMouseHook(string name)
        {
            _hookID = IntPtr.Zero;
            _hookProc = HookCallback;
            _state = DoubleClickState.Empty;
        }

		/// <summary>
		/// Destructor for this LowLevelMouseHook instance.
		/// </summary>
		~LowLevelMouseHook()
        {
            Uninstall();
        }

		/// <summary>
		/// Name accessor for this low level mouse hook. 
		/// </summary>
		public string Name { get; }

        /// <summary>
        /// 
        /// </summary>
        public void Install()
        {
            Uninstall(); // stop the hook if it is already active first

            // then install a new hook
            Process procCurrent = Process.GetCurrentProcess();
            ProcessModule procMod = procCurrent.MainModule;
			_hookID = Win32.SetWindowsHookEx(
						(int) Win32.WH.MOUSE_LL,
						_hookProc,
						Win32.GetModuleHandle(procMod.ModuleName),
						0u);
            procMod.Dispose();
            procCurrent.Dispose();

            // set this special registry key so the hook won't be uninstalled during periods of high activity
            RegistryKey key = Registry.CurrentUser.OpenSubKey(@"Control Panel\Desktop\", true);
            if (key == null)
            {
                key = Registry.CurrentUser.CreateSubKey(@"Control Panel\Desktop\");
            }
            key.SetValue("LowLevelHooksTimeout", 10000, RegistryValueKind.DWord);
            key.Close();
        }

        /// <summary>
        /// 
        /// </summary>
        public void Uninstall()
        {
            if (_hookID != IntPtr.Zero)
            {
                Win32.UnhookWindowsHookEx(_hookID);
            }
            _hookID = IntPtr.Zero;
        }

        /// <summary>
        /// 
        /// </summary>
        public bool IsActive
        {
            get { return _hookID != IntPtr.Zero; }
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="nCode"></param>
        /// <param name="wParam"></param>
        /// <param name="lParam"></param>
        /// <returns></returns>
        private IntPtr HookCallback(int nCode, int wParam, IntPtr lParam)
        {
            long now = DateTime.Now.Ticks / TimeSpan.TicksPerMillisecond;
            if (nCode >= 0)
            {
                Win32.MSLLHOOKSTRUCT info = (Win32.MSLLHOOKSTRUCT) Marshal.PtrToStructure(lParam, typeof(Win32.MSLLHOOKSTRUCT));

                bool isDown = false;
                int nClicks = 1;
                MouseButtons button = MouseButtons.None;

                switch (wParam) // figure out which button was pressed, if any
                {
                    ////////// down events
                    case (int) Win32.WM.LBUTTONDOWN:
                        button = MouseButtons.Left;
                        isDown = true;
                        nClicks = HandleDoubleClickLogic(button, info.pt.x, info.pt.y, now);
                        break;
                    case (int) Win32.WM.RBUTTONDOWN:
                        button = MouseButtons.Right;
                        isDown = true;
                        nClicks = HandleDoubleClickLogic(button, info.pt.x, info.pt.y, now);
                        break;
                    case (int) Win32.WM.MBUTTONDOWN:
                        button = MouseButtons.Middle;
                        isDown = true;
                        nClicks = HandleDoubleClickLogic(button, info.pt.x, info.pt.y, now);
                        break;
                    case (int) Win32.WM.XBUTTONDOWN:
                        if (((int) info.mouseData >> 16) == 1) // high-order word specifies which button was pressed
                            button = MouseButtons.XButton1;
                        else if (((int) info.mouseData >> 16) == 2)
                            button = MouseButtons.XButton2;
                        else Debug.Fail("XButton message received without a button identifier.");
                        isDown = true;
                        nClicks = HandleDoubleClickLogic(button, info.pt.x, info.pt.y, now);
                        break;
                    
                    ////////// up events
                    case (int) Win32.WM.LBUTTONUP:
                        button = MouseButtons.Left;
                        nClicks = _state.IssueDoubleClickOnMouseUp ? 2 : 1;
                        break;
                    case (int) Win32.WM.RBUTTONUP:
                        button = MouseButtons.Right;
                        nClicks = _state.IssueDoubleClickOnMouseUp ? 2 : 1;
                        break;
                    case (int) Win32.WM.MBUTTONUP:
                        button = MouseButtons.Middle;
                        nClicks = _state.IssueDoubleClickOnMouseUp ? 2 : 1;
                        break;
                    case (int) Win32.WM.XBUTTONUP:
                        if (((int) info.mouseData >> 16) == 1) // high-order word specifies which button was pressed
                            button = MouseButtons.XButton1;
                        else if (((int) info.mouseData >> 16) == 2)
                            button = MouseButtons.XButton2;
                        else Debug.Fail("XButton message received without a button identifier.");
                        nClicks = _state.IssueDoubleClickOnMouseUp ? 2 : 1;
                        break;
                }

                short delta = -1; // wheel
                if (wParam == (int) Win32.WM.MOUSEWHEEL)
                {
                    delta = (short) ((info.mouseData >> 16) & 0xffff);
                    nClicks = 0;
                }

                // set up the event argument
                MouseEventArgs e = new MouseEventArgs(button, nClicks, info.pt.x, info.pt.y, delta);

                // fire the appropriate event
                if (button != MouseButtons.None)
                {
                    if (isDown)
                    {
						HookMouseDown?.Invoke(this, e);
					}
                    else // button came up
                    {
                        if (_state.IssueDoubleClickOnMouseUp)
                        {
							HookMouseDoubleClick?.Invoke(this, e);
                            _state = DoubleClickState.Empty; // unset
                        }
                        else if (_state.IssueClickOnMouseUp)
                        {
							HookMouseClick?.Invoke(this, e);
                            _state.IssueClickOnMouseUp = false; // unset
                        }
                        HookMouseUp?.Invoke(this, e);
                    }
                }
                else // no buttons -- wheel or mouse movement
                {
                    if (delta != -1)
                        HookMouseWheel?.Invoke(this, e);
                    else // mouse movement
                        HookMouseMove?.Invoke(this, e);
                }
            }
            return Win32.CallNextHookEx(_hookID, nCode, wParam, lParam);
        }

        /// <summary>
        /// Handler for double-click logic that should be called on any button-down event.
        /// </summary>
        /// <param name="button">The button that was pressed down.</param>
        /// <param name="x">The x-coordinate of the button press.</param>
        /// <param name="y">The y-coordinate of the button press.</param>
        /// <param name="time">The time, in milliseconds, of the button press.</param>
        /// <returns>The number of clicks, 1 or 2, that have occurred.</returns>
        private int HandleDoubleClickLogic(MouseButtons button, int x, int y, long time)
        {
            int nClicks;
            if (_state.LastDownButton == button // has to be same button
                && Math.Abs(_state.LastDownPt.X - x) <= SystemInformation.DoubleClickSize.Width // has to be nearly same X-coord
                && Math.Abs(_state.LastDownPt.Y - y) <= SystemInformation.DoubleClickSize.Height // has to be nearly same Y-coord
				&& time - _state.LastDownTime <= SystemInformation.DoubleClickTime) // has to happen fast enough
            {
                _state.IssueDoubleClickOnMouseUp = true; // dlbclk determination is made on the second down
                _state.IssueClickOnMouseUp = false; // no single click is now loaded
                nClicks = 2;
            }
            else // no double click here, so set this state
            {
                _state.LastDownButton = button;
				_state.LastDownPt = new Point(x, y);
				_state.LastDownTime = time;
                _state.IssueDoubleClickOnMouseUp = false; // no double click was determined
                _state.IssueClickOnMouseUp = true; // a single click is "loaded," and the first mouse-up of any button unloads it
                nClicks = 1;
            }
            return nClicks;
        }
    }
}
