﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;
using System.Runtime.InteropServices;
using Microsoft.Win32;
using WobbrockLib.Windows;

namespace WobbrockLib.Devices
{
    public class RawInputKeyboard
    {
        public event EventHandler<KeyEventArgs> RawKeyDown;
		public event EventHandler<KeyPressEventArgs> RawKeyPress;
		public event EventHandler<KeyEventArgs> RawKeyUp;

        // raw input keyboard members
        private string _name;

        /// <summary>
        /// Constructor of a RawInputKeyboard instance.
        /// </summary>
        /// <param name="name">A name to identify this instance.</param>
        public RawInputKeyboard(string name)
        {
            _name = name;
        }

        /// <summary>
        /// Destructor.
        /// </summary>
        ~RawInputKeyboard()
        {
            Unregister();
        }

        /// <summary>
        /// Gets the client-given name of this RawInputKeyboard instance.
        /// </summary>
        public string Name
        {
            get { return _name; }
        }

        /// <summary>
        /// Registers this RawInputKeyboard to receive raw input events in the form of WM_INPUT messages in
        /// a target window's WndProc function.
        /// </summary>
        /// <remarks>Registering a keyboard for raw input prevents any Low-level Keyboard Hook from functioning;
        /// the two cannot be run simultaneously.</remarks>
        public bool Register(IntPtr hwndTarget, bool background)
        {
            Win32.RAWINPUTDEVICE[] pRawInputDevices = new Win32.RAWINPUTDEVICE[1];
            pRawInputDevices[0].usUsagePage = 1; // generic desktop page
            pRawInputDevices[0].usUsage = 6; // keyboard
            if (background)
                pRawInputDevices[0].dwFlags = (uint) Win32.RIDEV.INPUTSINK;
            pRawInputDevices[0].hwndTarget = hwndTarget;
            return Win32.RegisterRawInputDevices(pRawInputDevices, 1u, (uint) Marshal.SizeOf(pRawInputDevices[0]));
        }

        /// <summary>
        /// Unregisters this RawInputKeyboard from receiving raw input events in the form of WM_INPUT messages.
        /// </summary>
        public bool Unregister()
        {
            Win32.RAWINPUTDEVICE[] pRawInputDevices = new Win32.RAWINPUTDEVICE[1];
            pRawInputDevices[0].usUsagePage = 1; // generic desktop page
            pRawInputDevices[0].usUsage = 6; // keyboard
            pRawInputDevices[0].dwFlags = (uint) Win32.RIDEV.REMOVE; // remove the device
            return Win32.RegisterRawInputDevices(pRawInputDevices, 1u, (uint) Marshal.SizeOf(pRawInputDevices[0]));
        }

        /// <summary>
        /// Clients owning a RawInputKeyboard should place this method call upon receiving WM_INPUT messages in 
        /// their overridden WndProc method. They should pass the lParam of such messages as the handle to
        /// the raw input device.
        /// </summary>
        /// <param name="hRawInput">The lParam of the WM_INPUT message sent to WndProc.</param>
        /// <returns>True if processed; false otherwise.</returns>
        public bool ProcessRawInput(IntPtr hRawInput)
        {
            bool processed = false;

            // first get the size of the data to acquire
            uint pcbSize = 0u;
            uint cbSizeHeader = (uint) Marshal.SizeOf(typeof(Win32.RAWINPUTHEADER));
            uint successIs0 = Win32.GetRawInputData(hRawInput, (uint) Win32.RID.INPUT, IntPtr.Zero, ref pcbSize, cbSizeHeader);

            if (successIs0 == 0u && pcbSize > 0) // then get the data itself
            {
                IntPtr pData = Marshal.AllocHGlobal((int) pcbSize); // claim memory for the data
                uint pDataSize = Win32.GetRawInputData(hRawInput, (uint) Win32.RID.INPUT, pData, ref pcbSize, cbSizeHeader);

                if (pDataSize > 0u)
                {
                    uint dwType = 0u, message = 0u;
                    ushort vKey = 0, makeCode = 0;
                    if (Win32.SizeOfIntPtr == 32) // x86
                    {
                        Win32.RAWINPUTx86 ri = (Win32.RAWINPUTx86) Marshal.PtrToStructure(pData, typeof(Win32.RAWINPUTx86)); // convert memory to struct
                        dwType = ri.header.dwType;
                        message = ri.keyboard.Message;
                        vKey = ri.keyboard.VKey;
                        makeCode = ri.keyboard.MakeCode;
                    }
                    else if (Win32.SizeOfIntPtr == 64) // x64
                    {
                        Win32.RAWINPUTx64 ri = (Win32.RAWINPUTx64) Marshal.PtrToStructure(pData, typeof(Win32.RAWINPUTx64)); // convert memory to struct
                        dwType = ri.header.dwType;
                        message = ri.keyboard.Message;
                        vKey = ri.keyboard.VKey;
                        makeCode = ri.keyboard.MakeCode;
                    }
                    else Debug.Fail("System is neither x86 nor x64 according to Win32.SizeOfIntPtr.");

                    if (dwType == (uint) Win32.RIM.TYPEKEYBOARD)
                    {
                        processed = true; // at this point, we can say this was raw input from a keyboard
                        //
                        // KeyDown
                        //
                        if (RawKeyDown != null && (message == (uint) Win32.WM.KEYDOWN || message == (uint) Win32.WM.SYSKEYDOWN)) // KeyDown
                        {
                            Keys data = (Keys) vKey;
                            Keys mod = GetModifierKey(data); // get if a modifier key was just pressed
                            
                            // test the modifier keys for whether they are down; doesn't apply to them being pressed themselves
                            bool shift = ((Win32.GetKeyState((int) Win32.VK.SHIFT) & 0x80) == 0x80 ? true : false);
                            bool ctrl = ((Win32.GetKeyState((int) Win32.VK.CONTROL) & 0x80) == 0x80 ? true : false);
                            bool alt = ((Win32.GetKeyState((int) Win32.VK.MENU) & 0x80) == 0x80 ? true : false);

                            // if a modifier key is down or was just pressed, add it to the data
                            if (shift || mod == Keys.ShiftKey)
                                data |= Keys.Shift;
                            if (ctrl || mod == Keys.ControlKey)
                                data |= Keys.Control;
                            if (alt || mod == Keys.Menu)
                                data |= Keys.Alt;

                            KeyEventArgs e = new KeyEventArgs(data);
                            RawKeyDown.Invoke(this, e);
                        }
                        //
                        // KeyPress
                        //
                        if (RawKeyPress != null && message == (uint) Win32.WM.KEYDOWN) // KeyPress
                        {
                            bool shift = ((Win32.GetKeyState((int) Win32.VK.SHIFT) & 0x80) == 0x80 ? true : false);
                            bool caps = ((Win32.GetKeyState((int) Win32.VK.CAPITAL) & 0x01) == 0x01 ? true : false);
                            byte[] state = new byte[256];
                            Win32.GetKeyboardState(state);
                            short lpChar;
                            if (Win32.ToAscii(vKey, makeCode, state, out lpChar, 0u) == 1)
                            {
                                char key = (char) (lpChar & 0xFF); // take only low-order byte
                                if ((shift ^ caps) && Char.IsLetter(key))
                                    key = Char.ToUpper(key);
                                KeyPressEventArgs e = new KeyPressEventArgs(key);
                                RawKeyPress.Invoke(this, e);
                            }
                        }
                        //
                        // KeyUp
                        //
                        if (RawKeyUp != null && (message == (uint) Win32.WM.KEYUP || message == (uint) Win32.WM.SYSKEYUP)) // KeyUp
                        {
                            Keys data = (Keys) vKey;
                            Keys mod = GetModifierKey(data); // get if a modifier key was just pressed
                            
                            // test the modifier keys for whether they are down
                            bool shift = ((Win32.GetKeyState((int) Win32.VK.SHIFT) & 0x80) == 0x80 ? true : false);
                            bool ctrl = ((Win32.GetKeyState((int) Win32.VK.CONTROL) & 0x80) == 0x80 ? true : false);
                            bool alt = ((Win32.GetKeyState((int) Win32.VK.MENU) & 0x80) == 0x80 ? true : false);

                            // if a modifier key is down, add it to the data
                            if (shift)
                                data |= Keys.Shift;
                            if (ctrl)
                                data |= Keys.Control;
                            if (alt)
                                data |= Keys.Alt;

                            // if a modifier key was just released, turn off saying that key is down
                            if (mod == Keys.ShiftKey)
                                data &= ~Keys.Shift;
                            if (mod == Keys.ControlKey)
                                data &= ~Keys.Control;
                            if (mod == Keys.Menu)
                                data &= ~Keys.Alt;

                            KeyEventArgs e = new KeyEventArgs(data);
                            RawKeyUp.Invoke(this, e);
                        }
                    }
                }
                uint sok = Win32.DefRawInputProc(pData, 1, cbSizeHeader);
                Marshal.FreeHGlobal(pData);
            }
            return processed;
        }

        /// <summary>
        /// The virtual key code we receive in the case of a modifier key event, when cast to a Keys type,
        /// can tell us which modifier key was pressed but it isn't a clean value. For example, when the
        /// shift button is pressed, Keys.ShiftKey is the value, but when the control key is
        /// pressed, Keys.LButton | Keys.ShiftKey is the value. This method therefore tells us
        /// which control key was pressed.
        /// </summary>
        /// <param name="data">The VKey field from the RAWKEYBOARD struct cast to a Keys enumeration.</param>
        /// <returns>The modifier key that was pressed, if any, or Keys.None.</returns>
        private Keys GetModifierKey(Keys data)
        {
            if (data == Keys.ShiftKey)
                return Keys.ShiftKey;
            else if (data == (Keys.LButton | Keys.ShiftKey))
                return Keys.ControlKey;
            else if (data == (Keys.RButton | Keys.ShiftKey))
                return Keys.Menu;
            else
                return Keys.None;
        }
    }
}
