﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Drawing;
using System.Diagnostics;
using System.Runtime.InteropServices;
using Microsoft.Win32;
using WobbrockLib.Windows;

namespace WobbrockLib.Devices
{
    /// <summary>
    /// Implements a RawInputMouse, which is capable of receiving raw input, optionally when the client
    /// window that owns this instance does not have the current focus. This class is similar to a 
    /// low level mouse hook, but it cannot be uninstalled by the system automatically as a low level
    /// mouse hook can be.
    /// </summary>
    public class RawInputMouse
    {
        /// <summary>
        /// Internal utility struct for tracking the possibility of double-clicks arising from low level mouse events.
        /// </summary>
        [StructLayout(LayoutKind.Sequential)]
        private struct DoubleClickState
        {
            public static readonly DoubleClickState Empty;
            public MouseButtons LastDownButton;
            public long LastDownTime;
            public Point LastDownPt;
            public bool IssueClickOnMouseUp;
            public bool IssueDoubleClickOnMouseUp;
        }
        
        /// <summary>
        /// Events fired by this RawInputMouse instance.
        /// </summary>
        public event EventHandler<MouseEventArgs> RawMouseMove;
        public event EventHandler<MouseEventArgs> RawMouseDown;
        public event EventHandler<MouseEventArgs> RawMouseUp;
        public event EventHandler<MouseEventArgs> RawMouseClick;
        public event EventHandler<MouseEventArgs> RawMouseDoubleClick;
        public event EventHandler<MouseEventArgs> RawMouseWheel;

        // double-click members
        private DoubleClickState _dblClkState;

        /// <summary>
		/// Constructor for this RawInputMouse.
		/// </summary>
		public RawInputMouse(string name)
        {
            _dblClkState = DoubleClickState.Empty;
        }

        /// <summary>
        /// Destructor for this RawInputMouse instance.
        /// </summary>
        ~RawInputMouse()
        {
            Unregister();
        }

		/// <summary>
		/// Name accessor for this raw input mouse.
		/// </summary>
		public string Name { get; }

		/// <summary>
		/// Register to receive raw inpu tdata from mice.
		/// </summary>
		/// <param name="hwndTarget">The target window whose WndProc will handle the WM_INPUT messages.</param>
		/// <param name="background">If true, this instance will receive raw input from mice even when the
		/// window whose handle was passed in the constructor does not have the input focus.</param>
		/// <returns>True if registration is successful; false otherwise.</returns>
		public bool Register(IntPtr hwndTarget, bool background)
        {
            Win32.RAWINPUTDEVICE[] pRawInputDevices = new Win32.RAWINPUTDEVICE[1];
            pRawInputDevices[0].usUsagePage = 1; // generic desktop page
            pRawInputDevices[0].usUsage = 2; // mouse
            if (background)
                pRawInputDevices[0].dwFlags = (uint) Win32.RIDEV.INPUTSINK;
            pRawInputDevices[0].hwndTarget = hwndTarget;
            return Win32.RegisterRawInputDevices(pRawInputDevices, 1u, (uint) Marshal.SizeOf(pRawInputDevices[0]));
        }

        /// <summary>
        /// Unregister raw input mice.
        /// </summary>
        /// <returns>True if successful; false otherwise.</returns>
        public bool Unregister()
        {
            Win32.RAWINPUTDEVICE[] pRawInputDevices = new Win32.RAWINPUTDEVICE[1];
            pRawInputDevices[0].usUsagePage = 1; // generic desktop page
            pRawInputDevices[0].usUsage = 2; // mouse
            pRawInputDevices[0].dwFlags = (uint) Win32.RIDEV.REMOVE; // remove the device
            return Win32.RegisterRawInputDevices(pRawInputDevices, 1u, (uint) Marshal.SizeOf(pRawInputDevices[0]));
        }

        /// <summary>
        /// Clients owning a RawInputMouse should place this method call upon receiving WM_INPUT messages in 
        /// their overridden WndProc method. They should pass the lParam of such messages as the handle to
        /// the raw input device.
        /// </summary>
        /// <param name="hRawInput">The lParam of the WM_INPUT message sent to WndProc.</param>
        /// <returns>True if processed; false otherwise.</returns>
        public bool ProcessRawInput(IntPtr hRawInput)
        {
            bool processed = false; 

            // first get the size of the data to acquire
            uint pcbSize = 0u;
            uint cbSizeHeader = (uint) Marshal.SizeOf(typeof(Win32.RAWINPUTHEADER));
            uint successIs0 = Win32.GetRawInputData(hRawInput, (uint) Win32.RID.INPUT, IntPtr.Zero, ref pcbSize, cbSizeHeader);

            if (successIs0 == 0u && pcbSize > 0) // then get the data itself
            {
                IntPtr pData = Marshal.AllocHGlobal((int) pcbSize); // claim memory for the data
                uint pDataSize = Win32.GetRawInputData(hRawInput, (uint) Win32.RID.INPUT, pData, ref pcbSize, cbSizeHeader);

                if (pDataSize > 0u)
                {
                    uint dwType = 0u;
					ushort usButtonFlags = 0;
					ushort usButtonData = 0;
                    if (Win32.SizeOfIntPtr == 32) // x86
                    {
                        Win32.RAWINPUTx86 ri = (Win32.RAWINPUTx86) Marshal.PtrToStructure(pData, typeof(Win32.RAWINPUTx86)); // convert memory to struct
                        dwType = ri.header.dwType;
                        usButtonFlags = ri.mouse.usButtonFlags;
                        usButtonData = ri.mouse.usButtonData;
                    }
                    else if (Win32.SizeOfIntPtr == 64) // x64
                    {
                        Win32.RAWINPUTx64 ri = (Win32.RAWINPUTx64) Marshal.PtrToStructure(pData, typeof(Win32.RAWINPUTx64)); // convert memory to struct
                        dwType = ri.header.dwType;
                        usButtonFlags = ri.mouse.usButtonFlags;
                        usButtonData = ri.mouse.usButtonData;
                    }
                    else Debug.Fail("System is neither x86 nor x64 according to Win32.SizeOfIntPtr.");
                    
                    if (dwType == (uint) Win32.RIM.TYPEMOUSE)
                    {
                        processed = true; // at this point, we can say this was raw input from a mouse

                        long now = DateTime.Now.Ticks / TimeSpan.TicksPerMillisecond;
                        Point pt = Cursor.Position;
                        bool isDown = false;
                        int nClicks = 1;
                        short delta = -1; // wheel
                        MouseButtons button = MouseButtons.None;

                        switch (usButtonFlags) // figure out which button was pressed, if any
                        {
                            ////////// down events
                            case (ushort) Win32.RI_MOUSE.LEFT_BUTTON_DOWN:
                                button = MouseButtons.Left;
                                isDown = true;
                                nClicks = HandleDoubleClickLogic(button, pt.X, pt.Y, now);
                                break;
                            case (ushort) Win32.RI_MOUSE.RIGHT_BUTTON_DOWN:
                                button = MouseButtons.Right;
                                isDown = true;
                                nClicks = HandleDoubleClickLogic(button, pt.X, pt.Y, now);
                                break;
                            case (ushort) Win32.RI_MOUSE.MIDDLE_BUTTON_DOWN:
                                button = MouseButtons.Middle;
                                isDown = true;
                                nClicks = HandleDoubleClickLogic(button, pt.X, pt.Y, now);
                                break;
                            case (ushort) Win32.RI_MOUSE.BUTTON_4_DOWN:
                                button = MouseButtons.XButton1;
                                isDown = true;
                                nClicks = HandleDoubleClickLogic(button, pt.X, pt.Y, now);
                                break;
                            case (ushort) Win32.RI_MOUSE.BUTTON_5_DOWN:
                                button = MouseButtons.XButton2;
                                isDown = true;
                                nClicks = HandleDoubleClickLogic(button, pt.X, pt.Y, now);
                                break;

                            ////////// up events
                            case (ushort) Win32.RI_MOUSE.LEFT_BUTTON_UP:
                                button = MouseButtons.Left;
                                nClicks = _dblClkState.IssueDoubleClickOnMouseUp ? 2 : 1;
                                break;
                            case (ushort) Win32.RI_MOUSE.RIGHT_BUTTON_UP:
                                button = MouseButtons.Right;
                                nClicks = _dblClkState.IssueDoubleClickOnMouseUp ? 2 : 1;
                                break;
                            case (ushort) Win32.RI_MOUSE.MIDDLE_BUTTON_UP:
                                button = MouseButtons.Middle;
                                nClicks = _dblClkState.IssueDoubleClickOnMouseUp ? 2 : 1;
                                break;
                            case (ushort) Win32.RI_MOUSE.BUTTON_4_UP:
                                button = MouseButtons.XButton1;
                                nClicks = _dblClkState.IssueDoubleClickOnMouseUp ? 2 : 1;
                                break;
                            case (ushort) Win32.RI_MOUSE.BUTTON_5_UP:
                                button = MouseButtons.XButton2;
                                nClicks = _dblClkState.IssueDoubleClickOnMouseUp ? 2 : 1;
                                break;

                            ////////// wheel events
                            case (ushort) Win32.RI_MOUSE.WHEEL:
                            case (ushort) Win32.RI_MOUSE.HORIZONTAL_WHEEL:
                                delta = (short) usButtonData;
                                nClicks = 0;
                                break;
                        }

                        // set up the event argument
                        MouseEventArgs e = new MouseEventArgs(button, nClicks, pt.X, pt.Y, delta);

                        // fire the appropriate event
                        if (button != MouseButtons.None)
                        {
                            if (isDown)
                            {
								RawMouseDown?.Invoke(this, e);
							}
                            else // button came up
                            {
                                if (_dblClkState.IssueDoubleClickOnMouseUp)
                                {
									RawMouseDoubleClick?.Invoke(this, e);
									_dblClkState = DoubleClickState.Empty; // unset
                                }
                                else if (_dblClkState.IssueClickOnMouseUp)
                                {
									RawMouseClick?.Invoke(this, e);
									_dblClkState.IssueClickOnMouseUp = false; // unset
                                }
								RawMouseUp?.Invoke(this, e);
							}
                        }
                        else // no button events -- wheel or mouse movement -- note: a button could be held down
                        {
                            if (delta != -1)
								RawMouseWheel?.Invoke(this, e);
                            else // mouse movement
								RawMouseMove?.Invoke(this, e);
                        }
                    }
                }
                uint sok = Win32.DefRawInputProc(pData, 1, cbSizeHeader);
                Marshal.FreeHGlobal(pData);
            }
            return processed;
        }

        /// <summary>
        /// Handler for double-click logic that should be called on any button-down event.
        /// </summary>
        /// <param name="button">The button that was pressed down.</param>
        /// <param name="x">The x-coordinate of the button press.</param>
        /// <param name="y">The y-coordinate of the button press.</param>
        /// <param name="time">The time, in milliseconds, of the button press.</param>
        /// <returns>The number of clicks, 1 or 2, that have occurred.</returns>
        private int HandleDoubleClickLogic(MouseButtons button, int x, int y, long time)
        {
            int nClicks;
            if (_dblClkState.LastDownButton == button // has to be same button
				&& Math.Abs(_dblClkState.LastDownPt.X - x) <= SystemInformation.DoubleClickSize.Width // has to be nearly same X-coord
				&& Math.Abs(_dblClkState.LastDownPt.Y - y) <= SystemInformation.DoubleClickSize.Height // has to be nearly same Y-coord
				&& time - _dblClkState.LastDownTime <= SystemInformation.DoubleClickTime) // has to happen fast enough
			{
                _dblClkState.IssueDoubleClickOnMouseUp = true; // dlbclk determination is made on the second down
                _dblClkState.IssueClickOnMouseUp = false; // no single click is now loaded
                nClicks = 2;
            }
            else // no double click here, so set this state
            {
                _dblClkState.LastDownButton = button;
				_dblClkState.LastDownPt = new Point(x, y);
				_dblClkState.LastDownTime = time;
                _dblClkState.IssueDoubleClickOnMouseUp = false; // no double click was determined
                _dblClkState.IssueClickOnMouseUp = true; // a single click is "loaded," and the first mouse-up of any button unloads it
                nClicks = 1;
            }
            return nClicks;
        }
    }
}
