using System;
using System.IO;
using System.Reflection;
using WobbrockLib.Windows;

namespace WobbrockLib.Devices
{
    /// <summary>
    /// Static class of functions that make it easy to load, play, and loop *.WAV sound files.
	/// These files can either be files specified by a string filepath, or files that are 
	/// embedded resources within an application.
    /// </summary>
	public static class SoundEx
    {
        #region Sounds from File Paths

        /// <summary>
        /// Loads a *.wav file from the full path specified. Returns the bytes of the *.wav file.
        /// </summary>
        /// <param name="fullpath">The full path to the *.wav file.</param>
        /// <returns>The bytes of the sound, or null if an error occurs.</returns>
        public static byte[] LoadSound(string fullpath)
        {
            byte[] wav = null;
            if (File.Exists(fullpath))
            {
                FileStream fs = null;
                BinaryReader br = null;
                try
                {
                    fs = File.Open(fullpath, FileMode.Open, FileAccess.Read);
                    br = new BinaryReader(fs);
                    wav = new byte[fs.Length];
                    br.Read(wav, 0, wav.Length);
                }
                catch (Exception) { }
                finally
                {
                    if (br != null)
                        br.Close();
                    if (fs != null)
                        fs.Close();
                }
            }
            return wav;
        }

        /// <summary>
        /// Loads and plays a *.wav file from the full path specified. The sound does not stay loaded
        /// into memory.
        /// </summary>
        /// <param name="fullpath">The full path to the *.wav file.</param>
        /// <returns>True if the sound is loaded and played successfully; false otherwise.</returns>
        public static bool LoadPlaySound(string fullpath)
        {
            bool played = false;
            if (File.Exists(fullpath))
            {
                FileStream fs = null;
                BinaryReader br = null;
                try
                {
                    fs = File.Open(fullpath, FileMode.Open, FileAccess.Read);
                    br = new BinaryReader(fs);
                    byte[] wav = new byte[fs.Length];
                    br.Read(wav, 0, wav.Length);
                    played = PlaySound(wav);
                }
                catch (Exception) { }
                finally
                {
                    if (br != null)
                        br.Close();
                    if (fs != null)
                        fs.Close();
                }
            }
            return played;
        }

        /// <summary>
        /// Loads and loops a *.wav file from the full path specified.
        /// </summary>
        /// <param name="fullpath">The full path to the *.wav file.</param>
        /// <returns>True if the sound is loaded and looped successfully; false otherwise.</returns>
        public static bool LoadLoopSound(string fullpath)
        {
            bool played = false;
            if (File.Exists(fullpath))
            {
                FileStream fs = null;
                BinaryReader br = null;
                try
                {
                    fs = File.Open(fullpath, FileMode.Open, FileAccess.Read);
                    br = new BinaryReader(fs);
                    byte[] wav = new byte[fs.Length];
                    br.Read(wav, 0, wav.Length);
                    played = LoopSound(wav);
                }
                catch (Exception) { }
                finally
                {
                    if (br != null)
                        br.Close();
                    if (fs != null)
                        fs.Close();
                }
            }
            return played;
        }

        #endregion

        #region Sounds from Embedded Resources

        /// <summary>
		/// Loads an embedded *.wav resource with the given resource filename. Returns the bytes of the 
		/// *.wav file.
		/// </summary>
		/// <param name="type">The type identifier for the calling application. For example, if the caller
		/// is within the MainForm class of an application, use typeof(MainForm) for this parameter.</param>
		/// <param name="rsc">The name of the embedded resource file. If it is contained in an rsc folder,
		/// for example, then the name will be something like "rsc.name.ext".</param>
		/// <returns>The bytes of the sound, or null if an error occurs.</returns>
        public static byte[] LoadSoundResource(Type type, string rsc)
		{
			byte[] wav = null;
			Stream s = Assembly.GetCallingAssembly().GetManifestResourceStream(type, rsc);
			if (s != null)
			{
				wav = new byte[s.Length];
				s.Read(wav, 0, wav.Length);
				s.Close();
			}
			return wav;
		}

		/// <summary>
		/// Loads and plays an embedded *.wav resource with the given name. The sound does not stay loaded into 
		/// memory.
		/// </summary>
		/// <param name="type">The type identifier for the calling application. For example, if the caller
		/// is within the MainForm class of an application, use typeof(MainForm) for this parameter.</param>
		/// <param name="rsc">The name of the embedded resource file. If it is contained in an rsc folder,
		/// for example, then the name will be something like "rsc.name.ext".</param>
		/// <returns>True if the sound is loaded and played successfully; false otherwise.</returns>
		public static bool LoadPlaySoundResource(Type type, string rsc)
		{
			bool played = false;
            Stream s = Assembly.GetCallingAssembly().GetManifestResourceStream(type, rsc);
			if (s != null)
			{
				byte[] wav = new byte[s.Length];
				s.Read(wav, 0, wav.Length);
				s.Close();
				played = PlaySound(wav);
			}
			return played;
		}

		/// <summary>
		/// Loads and loops an embedded *.wav resource with the given name.
		/// </summary>
		/// <param name="type">The type identifier for the calling application. For example, if the caller
		/// is within the MainForm class of an application, use typeof(MainForm) for this parameter.</param>
		/// <param name="rsc">The name of the embedded resource file. If it is contained in an rsc folder,
		/// for example, then the name will be something like "rsc.name.ext".</param>
		/// <returns>True if the sound is loaded and looped successfully; false otherwise.</returns>
		public static bool LoadLoopSoundResource(Type type, string rsc)
        {
            bool played = false;
            Stream s = Assembly.GetCallingAssembly().GetManifestResourceStream(type, rsc);
            if (s != null)
            {
                byte[] wav = new byte[s.Length];
                s.Read(wav, 0, wav.Length);
                s.Close();
                played = LoopSound(wav);
            }
            return played;
        }

        #endregion

        #region Sounds as *.wav Bytes

        /// <summary>
        /// Plays a *.wav sound already loaded into memory.
        /// </summary>
        /// <param name="wav">The raw bytes of a *.wav file.</param>
        /// <returns>True if played successfully; false otherwise.</returns>
		public static bool PlaySound(byte[] wav)
		{
			bool played = false;
			if (wav != null && wav.Length > 0)
			{
                played = Win32.PlaySound( // play the wav sound bytes
					wav, 
					IntPtr.Zero, 
					(uint) (Win32.SND.ASYNC | Win32.SND.MEMORY | Win32.SND.NOWAIT)
					);
			}
			return played;
		}

        /// <summary>
        /// Loops a *.wav sound already loaded into memory.
        /// </summary>
        /// <param name="wav">The raw bytes of a *.wav file.</param>
        /// <returns>True if played successfully; false otherwise.</returns>
        public static bool LoopSound(byte[] wav)
        {
            bool played = false;
            if (wav != null && wav.Length > 0)
            {
                played = Win32.PlaySound( // loop the wav sound bytes
                    wav,
                    IntPtr.Zero,
                    (uint) (Win32.SND.LOOP | Win32.SND.ASYNC | Win32.SND.MEMORY | Win32.SND.NOWAIT)
                    );
            }
            return played;
        }

        #endregion
    }
}
