﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using WobbrockLib.Windows;

namespace WobbrockLib.Devices
{
    /// <summary>
    /// Implements a PictureBox control that displays images from a captured Web camera.
    /// For a more sophisticated approach, see http://www.codeproject.com/Articles/20519/Webcamera-Multithreading-and-VFW.
    /// </summary>
    [ToolboxBitmap(typeof(System.Windows.Forms.PictureBox))]
    public partial class WebcamPictureBox : System.Windows.Forms.PictureBox
    {
        #region Types

        /// <summary>
        /// Type indicating which of the built-in Web camera dialogs to show.
        /// </summary>
        public enum WebCamDialogType
        {
            VideoFormat = 0,
            VideoSource,
            VideoDisplay,
            VideoCompression
        }

        /// <summary>
        /// Type indicating whether the images captured from the Web cam should be displayed in Preview
        /// mode or repeatedly written to the clipboard and retrieved and set from there. In Preview mode,
        /// the Web cam image is just blitted directly by the Web cam API onto the surface of this control
        /// using GDI+. Thus, the Image property of this control is never set. Images can still be periodically
        /// grabbed using the Snapshot method, which goes through the clipboard but does not set
        /// the Image property of this control. Alternatively, the built-in Preview mode can be avoided,
        /// and instead, images can be obtained from the Web cam repeatedly using a timer we run, and
        /// each image can be set as this control's Image property. Preview mode is preferred if we do not 
        /// need to set the Image property of this control, and if occasional snapshots suffice. Also, 
        /// in Preview mode, the Web cam image can only be displayed in its native resolution, so SizeMode
        /// settings do not have any effect. Also, no custom image changing/changed events will fire, as again,
        /// the Image property is not being set. If SizeMode settings are desired, and programmatic use of the 
        /// clipboard is acceptable, or if image changing/changed events are desired, do not use Preview mode,
        /// but instead use the manual timer-copying-to-clipboard approach.
        /// </summary>
        public enum WebCamImageConnectionType
        {
            /// <summary>
            /// Uses the built-in Web cam API's Preview mode to blit the Web cam image directly
            /// to the surface of this control using GDI+. The Image property of this control
            /// will not be set, and SizeMode options therefore will have no effect.
            /// </summary>
            PreviewViaGDIPlusNoSetImage = 0,

            /// <summary>
            /// Does not use the built-in Web cam API's Preview mode, but instead runs a timer
            /// to snarf the Web cam image, which is periodically dumped onto the clipboard, retrieved
            /// from there, and then set as the Image property of this control. SizeMode options therefore 
            /// take effect, and clients can also modify the image to be displayed in their image-changing 
            /// event handler before that image is displayed.
            /// </summary>
            CaptureViaClipboardSetImage = 1
        }

        #endregion

        #region Custom Events

        /// <summary>
        /// <para>
        /// This event fires when the Image property of this control is about to change due to
        /// capturing a Web cam image. (It does not fire if the Image property is programmatically 
        /// changed by the owner of this control.) The data for the event is the new image to which
        /// the Image property is about to be set. The new image can be modified as desired by the 
        /// caller. If it is set to <b>null</b>, then the Image property is not updated and the 
        /// WebCamImageChanged event will not be fired. The old image can be retrieved from the Image 
        /// property directly, as it has not yet changed, but is about to.
        /// </para>
        /// <para>
        /// This event will only fire if the WebCamImageConnectionType.CaptureViaClipboardSetImage value is 
        /// set for the ImageConnection property, as the WebCamImageConnectionType.PreviewViaGDIPlusNoSetImage value 
        /// does not actually cause the Image property to be set for this control, but blits the preview image 
        /// directly onto the surface of the control using GDI+.
        /// </para>
        /// </summary>
        public event EventHandler<Image> WebCamImageChanging;

        /// <summary>
        /// <para>
        /// This event fires after the Image property of this control has changed due to capturing
        /// a Web cam image. (It does not fire if the Image property is programmatically changed by
        /// the owner of this control.) The data for this event is the old image that was just replaced 
        /// when the new image was set. The new image can be retieved from the Image property directly,
        /// as it has just been changed.
        /// </para>
        /// <para>
        /// This event will only fire if the WebCamImageConnectionType.CaptureViaClipboardSetImage value is 
        /// set for the ImageConnection property, as the WebCamImageConnectionType.PreviewViaGDIPlusNoSetImage value 
        /// does not actually cause the Image property to be set for this control, but blits the preview image 
        /// directly onto the surface of the control using GDI+.
        /// </para>
        /// </summary>
        public event EventHandler<Image> WebCamImageChanged;

        #endregion

        #region Constants and Fields

        private const int FILM_FRAMERATE = 42; // 24 fps = 42 ms between frames
        private const int RESOLUTION_WIDTH = 320;
        private const int RESOLUTION_HEIGHT = 240;

        private IntPtr _hCapture;
        private Size _resolution;
        private System.Windows.Forms.Timer _timer;
        private WebCamImageConnectionType _connection;

        #endregion

        #region Constructor, Finalizer

        /// <summary>
        /// Constructor.
        /// </summary>
        public WebcamPictureBox()
        {
            InitializeComponent();

            _timer = new System.Windows.Forms.Timer();
            _timer.Interval = FILM_FRAMERATE;
            _timer.Tick += new EventHandler(OnTimerTick);

            _resolution = new Size(RESOLUTION_WIDTH, RESOLUTION_HEIGHT);
            _connection = WebCamImageConnectionType.PreviewViaGDIPlusNoSetImage;
        }

        /// <summary>
        /// Finalizer that stops our connection to the Web cam driver.
        /// </summary>
        ~WebcamPictureBox()
        {
            this.Disconnect();
        }

        #endregion

        #region Public API: Property-Page Properties

        /// <summary>
        /// Gets or sets the connection type for capturing Web cam images.
        /// </summary>
        /// <remarks>If a recording to an *.avi file is underway, the connection type
        /// cannot be changed until the recording is stopped, and setting this property
        /// is a no-op.</remarks>
        [DefaultValue(typeof(WebCamImageConnectionType), "0")] // default is enum value 0
        public WebCamImageConnectionType ImageConnection
        {
            get
            {
                return _connection;
            }
            set
            {
                if (value != _connection && !this.IsRecording)
                {
                    if (this.IsConnected) // if we're actively capturing
                    {
                        this.Disconnect(); // stop the capture
                        _connection = value; // set the new value
                        this.Connect(); // restart the capture
                    }
                    else _connection = value; // just set the flag
                }
            }
        }

        /// <summary>
        /// Gets or sets the interval, in milliseconds, between successive Web cam 
        /// image frames. Applies to either connection type.
        /// </summary>
        /// <remarks>If a recording to an *.avi file is underway, the capture interval
        /// cannot be changed until the recording is stopped, and setting this property
        /// is a no-op.</remarks>
        [DefaultValue(FILM_FRAMERATE)]
        public int ImageCaptureInterval
        {
            get
            { 
                return _timer.Interval;
            }
            set
            {
                if (value != _timer.Interval && !this.IsRecording)
                {
                    if (this.IsConnected) // if we're actively capturing
                    {
                        this.Disconnect(); // stop the capture
                        _timer.Interval = value; // set the new value
                        this.Connect(); // restart the capture
                    }
                    else _timer.Interval = value; // just set the value
                }
            }
        }

        /// <summary>
        /// Gets or sets the resolution of the Web cam image. The default is 640 x 480.
        /// </summary>
        /// <remarks>If a recording to an *.avi file is underway, the resolution
        /// cannot be changed until the recording is stopped, and setting this property
        /// is a no-op.</remarks>
        [DefaultValue(typeof(Size), "new Size(640, 480)")]
        public Size ImageResolution
        {
            get 
            { 
                return _resolution; 
            }
            set
            {
                if (value != _resolution && !this.IsRecording)
                {
                    if (this.IsConnected) // if we're actively capturing
                    {
                        this.Disconnect(); // stop the capture
                        _resolution = value; // set the new value
                        this.Connect(); // restart the capture
                    }
                    else _resolution = value; // just set the value
                }
            }
        }

        #endregion

        #region Public API: Connecting to Images

        /// <summary>
        /// Gets whether this control is currently capturing images from the Web cam.
        /// </summary>
        public bool IsConnected
        {
            get
            {
                Win32.CAPSTATUS c = Win32.CAPSTATUS.Empty;
                return  // if we have a capture window established
                        _hCapture != IntPtr.Zero
                        &&
                        (   // if we're in preview mode for showing images
                            (_connection == WebCamImageConnectionType.PreviewViaGDIPlusNoSetImage
                            && // and we successfully get our status
                            Win32.SendMessage(_hCapture, (uint) Win32.WM.CAP_GET_STATUS, (uint) Marshal.SizeOf(c), ref c) == 1
                            && // and we have a live preview-mode capture window
                            c.fLiveWindow)
                        || // or we're in manual timer-based connection mode
                            (_connection == WebCamImageConnectionType.CaptureViaClipboardSetImage
                            && // we have an image in our picturebox
                            this.Image != null
                            && // and our timer is running
                            _timer.Enabled)
                        );
            }
        }

        /// <summary>
        /// Starts capturing Web cam images and displaying them using this control.
        /// </summary>
        /// <returns>True if successful; false otherwise.</returns>
        public bool Connect()
        {
            // The flags we use to create the capture window include the child window flag, but the
            // flag for whether we show this capture window depends on the image connection we're using.
            // If we're in Preview mode blitting straight from the Web cam API using GDI+, then we
            // need to make this capture window visible. But if we're manually setting the Image
            // property using a timer, then we do not want the capture window visible, becuase our
            // picture box itself is showing the image. Doing the latter allows for modifying the
            // picture, e.g., drawing to it first, flipping it over the X-axis, or whatever else a client
            // may like to do, before the image is displayed.
            Win32.WS flags = 
                (_connection == WebCamImageConnectionType.PreviewViaGDIPlusNoSetImage)
                ? 
                Win32.WS.CHILD | Win32.WS.VISIBLE // if images are coming from preview mode, show the capture window
                : 
                Win32.WS.CHILD; // if images are coming manually from our timer, do not show the capture window
            try
            {
                _hCapture = Win32.capCreateCaptureWindow(
                        "WebCamPictureBox",
                        (int) flags,
                        0, 0,
                        _resolution.Width,
                        _resolution.Height,
                        this.Handle, 0);
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex);
                return false;
            }
            if (_hCapture != IntPtr.Zero)
            {
                if (_connection == WebCamImageConnectionType.PreviewViaGDIPlusNoSetImage)
                {
                    return Win32.SendMessage(_hCapture, (uint) (Win32.WM.CAP_DRIVER_CONNECT), 0u, 0) == 1 // connect to the Web cam
                        && Win32.SendMessage(_hCapture, (uint) (Win32.WM.CAP_SET_PREVIEWRATE), (uint) _timer.Interval, 0) == 1 // set the preview interval
                        && Win32.SendMessage(_hCapture, (uint) (Win32.WM.CAP_SET_PREVIEW), 1u, 0) == 1; // turn on preview mode
                }
                else if (_connection == WebCamImageConnectionType.CaptureViaClipboardSetImage)
                {
                    if (Win32.SendMessage(_hCapture, (uint) (Win32.WM.CAP_DRIVER_CONNECT), 0u, 0) == 1 // connect to the Web cam
                        && Win32.SendMessage(_hCapture, (uint) (Win32.WM.CAP_SET_PREVIEW), 0u, 0) == 1) // turn off preview mode 
                    {
                        OnTimerTick(this, EventArgs.Empty); // as if our timer went off immediately--will start timer also
                        return true;
                    }
                    else return false;
                }
            }
            return false;
        }

        /// <summary>
        /// Stops this control from capturing Web cam images.
        /// </summary>
        /// <returns>True if successful; false otherwise.</returns>
        public bool Disconnect()
        {
            _timer.Stop();
            bool success =
                Win32.SendMessage(_hCapture, (uint) (Win32.WM.CAP_DRIVER_DISCONNECT), 0u, 0) == 1
                &&
                Win32.DestroyWindow(_hCapture);
            _hCapture = IntPtr.Zero;
            this.Image = null;
            return success;
        }

        /// <summary>
        /// Return the image currently being displayed.
        /// </summary>
        /// <remarks>
        /// <para>If we are displaying the Web cam image in Preview mode, the image is not itself
        /// actually being set as the Image property of this control, but just being blitted
        /// directly to its surface by GDI+. Thus, the easiest way to capture this image is to ask
        /// that it be dumped to the clipboard, and retrieved from there.</para>
        /// <para>If we are already capturing images via the clipboard for display in our Image property, 
        /// then we simply need to return that Image property as it currently stands.</para>
        /// </remarks>
        /// <returns>The image currently being captured from the Web camera by this control.</returns>
        public Image Snapshot()
        {
            if (_connection == WebCamImageConnectionType.PreviewViaGDIPlusNoSetImage) // our Image property will not be set
            {
                return
                    Win32.SendMessage(_hCapture, (uint) (Win32.WM.CAP_EDIT_COPY), 0u, 0) == 1 // if successful copying to clipboard
                    ?
                    Clipboard.GetImage() // retrieve from clipboard
                    :
                    null; // otherwise, return null
            }
            else if (_connection == WebCamImageConnectionType.CaptureViaClipboardSetImage)
            {
                return this.Image;
            }
            return null;
        }

        /// <summary>
        /// This timer event is only relevant when our image connection type is CaptureViaClipboardSetImage. On each 
        /// timer interval, the Web cam is asked for its current frame. That frame is copied to the clipboard so that 
        /// we have access to it as an image. We then set this control's Image property to display that image.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        /// <remarks>When we're recording to an *.avi file, the if-test atop this method's code will not
        /// succeed, so this picturebox control will not be updated and its image will freeze. But the recording
        /// process is still working. If the picturebox must remain active while recording, use the PreviewViaGDIPlusNoSetImage
        /// image connection type.</remarks>
        private void OnTimerTick(object sender, EventArgs e)
        {
            _timer.Stop(); // stop

            if (Win32.SendMessage(_hCapture, (uint) (Win32.WM.CAP_GRAB_FRAME), 0u, 0) == 1 // put current frame in buffer
                && Win32.SendMessage(_hCapture, (uint) (Win32.WM.CAP_EDIT_COPY), 0u, 0) == 1 // copy it to the clipboard
                && Clipboard.ContainsImage()) // verify it worked
            {
                Image newImage = Clipboard.GetImage(); // retrieve

                WebCamImageChanging?.Invoke(this, newImage); // fire

                Image oldImage = this.Image; // keep for event
                this.Image = newImage; // update! note the event-subscriber could have changed "newImage"

                WebCamImageChanged?.Invoke(this, oldImage); // fire
            }

            _timer.Start(); // restart
        }

        #endregion

        #region Public API: Recording to *.avi Files

        /// <summary>
        /// Gets whether we're currently recording our video capture stream to an *.avi file.
        /// </summary>
        public bool IsRecording
        {
            get
            {
                Win32.CAPSTATUS c = Win32.CAPSTATUS.Empty;
                return // we have a capture window
                    _hCapture != IntPtr.Zero
                    && // and we successfully get our status
                    Win32.SendMessage(_hCapture, (uint) Win32.WM.CAP_GET_STATUS, (uint) Marshal.SizeOf(c), ref c) == 1
                    && // and our status says we're recording to video right now
                    c.fCapturingNow;
            }
        }

        /// <summary>
        /// Starts recording video to a fully-specified *.avi file.
        /// </summary>
        /// <param name="path">A fully-specified filename path.</param>
        /// <param name="seconds">The number of seconds for which to record the video. If this value is
        /// 0u, then the video will be recorded until Stop or Disconnect is called.</param>
        /// <param name="async">If true, a background thread will be spawned to do the recording
        /// so that the main user interface thread is not locked up and the UI remains
        /// responsive. Recommended if seconds is 0u.</param>
        /// <param name="audio">If true, audio will be recorded.</param>
        /// <param name="okCancel">If true, an OK/Cancel dialog will be shown to allow the user to
        /// manually trigger when recording begins (or is canceled).</param>
        /// <returns>True if successful; false otherwise.</returns>
        public bool Record(string path, uint seconds, bool async, bool audio, bool okCancel)
        {
            // get the current streaming to file capabilities
            Win32.CAPTUREPARMS c = Win32.CAPTUREPARMS.Empty;
            if (Win32.SendMessage(_hCapture, (uint) Win32.WM.CAP_GET_SEQUENCE_SETUP, (uint) Marshal.SizeOf(c), ref c) == 0)
                return false; // failure -- can't get current video streaming capabilities

            // set options for this recording session
            if (seconds != 0u) // set a time limit if necessary
            {
                c.fLimitEnabled = true;
                c.wTimeLimit = seconds; 
            }
            c.fYield = async; // do we want the user interface to remain responsive?
            c.fCaptureAudio = audio; // do we want audio to be recorded?
            c.fMakeUserHitOKToCapture = okCancel; // prompt the user to hit OK to start?
            
            return // make the new settings
                Win32.SendMessage(_hCapture, (uint) Win32.WM.CAP_SET_SEQUENCE_SETUP, (uint) Marshal.SizeOf(c), ref c) == 1
                && // set the filename path to write to
                Win32.SendMessage(_hCapture, (uint) Win32.WM.CAP_FILE_SET_CAPTURE_FILE, 0u, path) == 1
                && // begin streaming to the file
                Win32.SendMessage(_hCapture, (uint) Win32.WM.CAP_SEQUENCE, 0u, 0) == 1;
        }

        /// <summary>
        /// Sends the message to stop streaming video to an *.avi file. 
        /// </summary>
        /// <returns>True if successful; false otherwise.</returns>
        public bool Stop()
        {
            return Win32.SendMessage(_hCapture, (uint) Win32.WM.CAP_STOP, 0u, 0) == 1;
        }

        #endregion

        #region Public API: Video and Audio Settings

        /// <summary>
        /// Gets the main four parameters of the video settings currently in use.
        /// </summary>
        /// <param name="biWidth">The width of the video.</param>
        /// <param name="biHeight">The height of the video.</param>
        /// <param name="biBitCount">The bit-depth of the video.</param>
        /// <param name="biCompression">The compression value of the video. Some values are
        /// defined as BI_* values in the Win32.BI enumerated type.</param>
        /// <returns>True if successful and the out values can be trusted; false otherwise.</returns>
        /// <see cref="http://msdn.microsoft.com/en-us/library/windows/desktop/dd183376%28v=vs.85%29.aspx"/>
        public bool GetVideoFormat(out int biWidth, out int biHeight, out ushort biBitCount, out uint biCompression)
        {
            // get the current video format settings
            Win32.BITMAPINFO bmi = Win32.BITMAPINFO.Empty;
            bmi.bmiHeader.biSize = (uint) Marshal.SizeOf(bmi.bmiHeader);
            int retval = Win32.SendMessage(_hCapture, (uint) Win32.WM.CAP_GET_VIDEOFORMAT, (uint) Marshal.SizeOf(bmi), ref bmi);
            biWidth = bmi.bmiHeader.biWidth;
            biHeight = bmi.bmiHeader.biHeight;
            biBitCount = bmi.bmiHeader.biBitCount;
            biCompression = bmi.bmiHeader.biCompression;
            return retval != 0;
        }

        /// <summary>
        /// Sets the four main parameters of the video settings.
        /// </summary>
        /// <param name="biWidth">The width of the video.</param>
        /// <param name="biHeight">The height of the video.</param>
        /// <param name="biBitCount">The bit-depth of the video.</param>
        /// <param name="biCompression">The compression value of the video. Some values are
        /// defined as BI_* values in the Win32.BI enumerated type.</param>
        /// <returns>True if successful; false otherwise.</returns>
        /// <see cref="http://msdn.microsoft.com/en-us/library/windows/desktop/dd183376%28v=vs.85%29.aspx"/>
        public bool SetVideoFormat(int biWidth, int biHeight, ushort biBitCount, uint biCompression)
        {
            // get the current video format settings
            Win32.BITMAPINFO bmi = Win32.BITMAPINFO.Empty;
            bmi.bmiHeader.biSize = (uint) Marshal.SizeOf(bmi.bmiHeader);
            if (Win32.SendMessage(_hCapture, (uint) Win32.WM.CAP_GET_VIDEOFORMAT, (uint) Marshal.SizeOf(bmi), ref bmi) == 0)
                return false;
            
            // make the desired settings
            bmi.bmiHeader.biWidth = biWidth;
            bmi.bmiHeader.biHeight = biHeight;
            bmi.bmiHeader.biBitCount = biBitCount;
            bmi.bmiHeader.biCompression = biCompression;

            // attempt to set and return whether successful
            return Win32.SendMessage(_hCapture, (uint) Win32.WM.CAP_SET_VIDEOFORMAT, (uint) Marshal.SizeOf(bmi), ref bmi) != 0;
        }

        /// <summary>
        /// Gets the three main parameters of the audio recording settings.
        /// </summary>
        /// <param name="nSamplesPerSec">Frequency in Hz of the sample rate. Values should be 8000, 11025, 22050, or 44100.</param>
        /// <param name="wBitsPerSample">The number of bits to encode a sample. Values should be 8 or 16.</param>
        /// <param name="nChannels">Monaural uses 1 channel; stero uses 2 channels.</param>
        /// <returns>True if successful; false otherwise.</returns>
        /// <see cref="http://msdn.microsoft.com/en-us/library/windows/desktop/dd757713%28v=vs.85%29.aspx"/>
        public bool GetAudioFormat(out uint nSamplesPerSec, out ushort wBitsPerSample, out ushort nChannels)
        {
            // get the current audio format settings
            Win32.WAVEFORMATEX wf = Win32.WAVEFORMATEX.Empty;
            int retval = Win32.SendMessage(_hCapture, (uint) Win32.WM.CAP_GET_AUDIOFORMAT, (uint) Marshal.SizeOf(wf), ref wf);
            nSamplesPerSec = wf.nSamplesPerSec;
            wBitsPerSample = wf.wBitsPerSample;
            nChannels = wf.nChannels;
            return retval != 0;
        }

        /// <summary>
        /// Sets the three main parameters of the audio recording settings.
        /// </summary>
        /// <param name="nSamplesPerSec">Frequency in Hz of the sample rate. Values should be 8000, 11025, 22050, or 44100.</param>
        /// <param name="wBitsPerSample">The number of bits to encode a sample. Values should be 8 or 16.</param>
        /// <param name="nChannels">Monaural uses 1 channel; stero uses 2 channels.</param>
        /// <returns>True if successful; false otherwise.</returns>
        /// <see cref="http://msdn.microsoft.com/en-us/library/windows/desktop/dd757713%28v=vs.85%29.aspx"/>
        public bool SetAudioFormat(uint nSamplesPerSec, ushort wBitsPerSample, ushort nChannels)
        {
            // get the current settings
            Win32.WAVEFORMATEX wf = Win32.WAVEFORMATEX.Empty;
            if (Win32.SendMessage(_hCapture, (uint) Win32.WM.CAP_GET_AUDIOFORMAT, (uint) Marshal.SizeOf(wf), ref wf) == 0)
                return false;

            // make the desired settings
            wf.nChannels = nChannels;
            wf.nSamplesPerSec = nSamplesPerSec;
            wf.wBitsPerSample = wBitsPerSample;
            wf.nBlockAlign = (ushort) ((nChannels * wBitsPerSample) / 8u); // calculate
            wf.nAvgBytesPerSec = nSamplesPerSec * wf.nBlockAlign; // calculate
            
            // attempt to set and return whether successful
            return Win32.SendMessage(_hCapture, (uint) Win32.WM.CAP_SET_AUDIOFORMAT, (uint) Marshal.SizeOf(wf), ref wf) != 0;
        }

        #endregion

        #region Built-in Dialogs

        /// <summary>
        /// Shows one of the built-in dialogs provided by the Web cam driver.
        /// </summary>
        /// <param name="type">The type of dialog to be shown.</param>
        /// <returns>True if successful; false otherwise. The method may be successful and yet the driver
        /// may nonetheless not supply a particular dialog box.</returns>
        public bool ShowDialog(WebCamDialogType type)
        {
            int result;
            switch (type)
            {
                case WebCamDialogType.VideoFormat:
                    result = Win32.SendMessage(_hCapture, (uint) Win32.WM.CAP_DLG_VIDEOFORMAT, 0u, 0);
                    break;
                case WebCamDialogType.VideoSource:
                    result = Win32.SendMessage(_hCapture, (uint) Win32.WM.CAP_DLG_VIDEOSOURCE, 0u, 0);
                    break;
                case WebCamDialogType.VideoDisplay:
                    result = Win32.SendMessage(_hCapture, (uint) Win32.WM.CAP_DLG_VIDEODISPLAY, 0u, 0);
                    break;
                case WebCamDialogType.VideoCompression:
                    result = Win32.SendMessage(_hCapture, (uint) Win32.WM.CAP_DLG_VIDEOCOMPRESSION, 0u, 0);
                    break;
                default:
                    result = 0;
                    break;
            }
            return result != 0;
        }

        #endregion
    }
}