﻿using System;
using System.Collections.Generic;
using System.Text;

namespace WobbrockLib.Extensions
{
    /// <summary>
    /// Useful pacing functions for implementing animations.
    /// </summary>
    public static class AnimationEx
    {
		/// <summary>
		/// Produces a percentage of distance [0,1] that is uniform along the steps used to animate. Callers can
		/// use this percentage to determine where along the animation path the animated object is to be placed
		/// or drawn.
		/// </summary>
		/// <param name="index">The current "step" we're on. This is usually the current index in a for-loop.</param>
		/// <param name="ubound">The total number of animation steps that will be taken. This is usually the upper bound in a for-loop.</param>
		/// <returns>A percentage of the distance to animate the object.</returns>
		/// <remarks>
		/// This function does not produce very satisfying results for the human eye; it is included
		/// for comparisons. One can graph the function "y = 0.5x + 0.5" at https://www.desmos.com/calculator.
		/// </remarks>
		public static double Uniform(int index, int ubound)
		{
			double x = GeotrigEx.MapValue(index + 1, 0, ubound, -1.00, +1.00); // map our step into x=[-1,+1]
			double y = (0.5 * x + 0.5);
			return GeotrigEx.PinValue(y, 0.00, 1.00);
		}

		/// <summary>
		/// Produces a percentage of distance [0,1] that is uniform in the time used to animate. Callers can use
		/// this function with an animation timer that goes off periodically.
		/// </summary>
		/// <param name="tStart">The start time stamp, usually in milliseconds, of this animation. Consider using TimeEx.NowMs.</param>
		/// <param name="tNow">The current time stamp, usually in milliseconds. Consider using TimeEx.NowMs.</param>
		/// <param name="tDuration">The total time duration, usually in milliseconds, for the desired animation.</param>
		/// <returns>A percentage of the distance to animate the object.</returns>
		public static double Uniform(long tStart, long tNow, int tDuration)
		{
			int elapsed = (int) (tNow - tStart);
			return Uniform(elapsed, tDuration);
		}

		/// <summary>
		/// Produces a percentage of distance (0,1) that creates a slow-in, fast-out effect for animation.
		/// Two functions are used: a logistic one from x=[-1,-.2) and a linear one from x=[-.2,+1]. Callers can
		/// use this percentage to determine where along the animation path the animated object is to be placed
		/// or drawn.
		/// </summary>
		/// <param name="index">The current "step" we're on. This is usually the current index in a for-loop.</param>
		/// <param name="ubound">The total number of animation steps that will be taken. This is usually the upper bound in a for-loop.</param>
		/// <returns>A percentage of the distance to animate to the target.</returns>
		/// <remarks>
		/// One can graph the functions "y = 1 / (1 + e^(-4.2x)) - 0.015" and "y = 0.6x + 0.4" at https://www.desmos.com/calculator.
		/// </remarks>
		public static double SlowIn(int index, int ubound)
		{
			double y;
			double x = GeotrigEx.MapValue(index + 1, 0, ubound, -1.00, +1.00); // map our step into x=[-1,+1]
			if (x < -0.2)
				y = 1.0 / (1.0 + Math.Exp(-4.2 * x)) - 0.015;
			else
				y = 0.6 * x + 0.4;
			return GeotrigEx.PinValue(y, 0.00, 1.00);
		}

		/// <summary>
		/// Produces a percentage of distance [0,1] that starts slowly and then speeds up. Callers can use
		/// this function with an animation timer that goes off periodically.
		/// </summary>
		/// <param name="tStart">The start time stamp, usually in milliseconds, of this animation. Consider using TimeEx.NowMs.</param>
		/// <param name="tNow">The current time stamp, usually in milliseconds. Consider using TimeEx.NowMs.</param>
		/// <param name="tDuration">The total time duration, usually in milliseconds, for the desired animation.</param>
		/// <returns>A percentage of the distance to animate the object.</returns>
		public static double SlowIn(long tStart, long tNow, int tDuration)
		{
			int elapsed = (int) (tNow - tStart);
			return SlowIn(elapsed, tDuration);
		}

		/// <summary>
		/// Produces a percentage of distance (0,1) that creates a fast-in, slow-out effect for animation.
		/// Two functions are used: a uniform one from x=[-1,+.2) and a logistic one from x=[+.2,+1]. Callers can
		/// use this percentage to determine where along the animation path the animated object is to be placed
		/// or drawn.
		/// </summary>
		/// <param name="index">The current "step" we're on. This is usually the current index in a for-loop.</param>
		/// <param name="ubound">The total number of animation steps that will be taken. This is usually the upper bound in a for-loop.</param>
		/// <returns>A percentage of the distance to animate the object.</returns>
		/// <remarks>
		/// One can graph the functions "y = 0.6x + 0.6" and "y = 1 / (1 + e^(-4.2x)) + 0.015" at https://www.desmos.com/calculator.
		/// </remarks>
		public static double SlowOut(int index, int ubound)
        {
            double y;
            double x = GeotrigEx.MapValue(index + 1, 0, ubound, -1.00, +1.00); // map our step into x=[-1,+1]
			if (x < +0.2)
                y = 0.6 * x + 0.6;
            else
                y = 1.0 / (1.0 + Math.Exp(-4.2 * x)) + 0.015;
			return GeotrigEx.PinValue(y, 0.00, 1.00);
		}

		/// <summary>
		/// Produces a percentage of distance [0,1] that starts quickly and then slows down. Callers can use
		/// this function with an animation timer that goes off periodically.
		/// </summary>
		/// <param name="tStart">The start time stamp, usually in milliseconds, of this animation. Consider using TimeEx.NowMs.</param>
		/// <param name="tNow">The current time stamp, usually in milliseconds. Consider using TimeEx.NowMs.</param>
		/// <param name="tDuration">The total time duration, usually in milliseconds, for the desired animation.</param>
		/// <returns>A percentage of the distance to animate the object.</returns>
		public static double SlowOut(long tStart, long tNow, int tDuration)
		{
			int elapsed = (int) (tNow - tStart);
			return SlowOut(elapsed, tDuration);
		}

		/// <summary>
		/// <para>
		/// Produces a percentage of distance (0,1) that creates a slow-in, slow-out effect for animation.
		/// The function used is a logistic function tuned so that its useful domain is (-1,+1),
		/// and useful range is (0,1) over that domain. That is, at x = -1, the y-value is 0. 
		/// At x = +1, the y-value is +1. We actually use two slightly y-shifted curves, one for x=[-1,0)
		/// and one for x=[0,+1]. They are equal at x=0, where y=0.505.
		/// </para>
		/// <para>
		/// Callers can use this percentage to determine where along the animation path the animated object 
		/// is to be placed or drawn.
		/// </para>
		/// </summary>
		/// <param name="index">The current "step" we're on. This is usually the current index in a for-loop.</param>
		/// <param name="ubound">The total number of animation steps that will be taken. This is usually the upper bound in a for-loop.</param>
		/// <returns>A percentage of the distance to animate the object.</returns>
		/// <remarks>
		/// One can graph the functions "y = 1 / (1 + e^(-5.5x)) - 0.005" and "y = 1 / (1 + e^(-5.5x)) + 0.005" at https://www.desmos.com/calculator.
		/// </remarks>
		public static double SlowInOut(int index, int ubound)
		{
			double y;
			double x = GeotrigEx.MapValue(index + 1, 0, ubound, -1.00, +1.00); // map our step into x=[-1,+1]
			if (x < 0)
				y = 1.0 / (1.0 + Math.Exp(-5.5 * x)) - 0.005; // return our y=[0,1]
			else
				y = 1.0 / (1.0 + Math.Exp(-5.5 * x)) + 0.005;
			return GeotrigEx.PinValue(y, 0.00, 1.00);
		}

		/// <summary>
		/// Produces a percentage of distance [0,1] that starts slowly, speeds up, and then ends slowly. Callers can use
		/// this function with an animation timer that goes off periodically.
		/// </summary>
		/// <param name="tStart">The start time stamp, usually in milliseconds, of this animation. Consider using TimeEx.NowMs.</param>
		/// <param name="tNow">The current time stamp, usually in milliseconds. Consider using TimeEx.NowMs.</param>
		/// <param name="tDuration">The total time duration, usually in milliseconds, for the desired animation.</param>
		/// <returns>A percentage of the distance to animate the object.</returns>
		public static double SlowInOut(long tStart, long tNow, int tDuration)
		{
			int elapsed = (int) (tNow - tStart);
			return SlowInOut(elapsed, tDuration);
		}
	}
}
