﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using WobbrockLib.Types;

namespace WobbrockLib.Extensions
{
    /// <summary>
    /// Utility functions for clustering 2-D data.
    /// </summary>
    public static class ClusterEx
    {
        #region k-means

        /// <summary>
        /// Implements the k-means algorithm with starting cluster centers C. Traditionally, C is determined in Lloyd's algorithm
        /// by using random centers, but k-means++ provides better initial centers. Unless implementing one's own initialization,
        /// a caller is expected to call KMeansPlusPlus and pass the result C to this function.
        /// </summary>
        /// <param name="points">The (x,y) points to cluster.</param>
        /// <param name="C">The cluster centers to begin with. While random is traditional, k-means++ provides better results. The
        /// number of initial centers is also taken to be the value of k, the number of clusters to return. Finally, upon return,
        /// C will have been updated to reflect the final cluster centers.</param>
        /// <returns>A list of point clusters, i.e., a list of point lists. From the original unclustered set of points, 
        /// the returned list has grouped points into clusters whose centers are returned in C.</returns>
        /// <remarks>Traditional k-means is often called Lloyd's algorithm based on the following publication:
        /// Lloyd, S. (1982). Least squares quantization in PCM. IEEE Transactions on Information Theory 28 (2), pp. 129-137.
        /// </remarks>
        /// <see cref="http://en.wikipedia.org/wiki/K-means_clustering" />
        public static List<List<PointR>> KMeans(List<PointR> points, List<PointR> C)
        {
            List<List<PointR>> X = new List<List<PointR>>(C.Count);
            for (int i = 0; i < C.Count; i++) // one cluster per center
            {
                X.Add(new List<PointR>()); // X will hold point clusters
            }

            while (true) // loop until assignments to clusters don't change, as judged by centers not changing
            {
                // Assignment step: Assign each observation points[i] to the cluster X[i] with the closest center C[j].
                for (int i = 0; i < X.Count; i++)
                {
                    X[i].Clear();
                }
                for (int i = 0; i < points.Count; i++)
                {
                    double dbest = double.MaxValue;
                    int jbest = -1;
                    for (int j = 0; j < C.Count; j++)
                    {
                        double d = GeotrigEx.Distance(points[i], C[j]);
                        if (d < dbest)
                        {
                            dbest = d;
                            jbest = j;
                        }
                    }
                    X[jbest].Add(points[i]);
                }
                // If any clusters are empty, discard them. This means no points were closest to that center. 
                // This can happen if all points in X are the same. This has the effect of avoiding duplicate centers.
                for (int i = X.Count - 1; i >= 0; i--)
                {
                    if (X[i].Count == 0)
                    {
                        X.RemoveAt(i);
                        C.RemoveAt(i);
                    }
                }

                // Update step: Calculate the new means to be the centroid of the observations in the cluster.
                List<PointR> Clast = new List<PointR>(C); // copy old centers for comparison later to see if we've converged
                C.Clear(); // clear old centers
                for (int i = 0; i < X.Count; i++)
                {
                    C.Add(GeotrigEx.Centroid(X[i]));
                }

                // The algorithm is deemed to have converged when the assignments no longer change.
                bool anyNewClusters = false;
                for (int i = 0; i < C.Count; i++)
                {
                    if (!Clast.Contains(C[i]))
                    {
                        anyNewClusters = true;
                        break;
                    }
                }
                if (!anyNewClusters)
                    break; // break from while (true)
            }
            return X;
        }

        /// <summary>
        /// Gets the best initial centers for the k-means algorithm. This algorithm is called k-means++. Despite its name,
        /// it just gets the initial centers for k-means; calling k-means itself is left to a separate function.
        /// </summary>
        /// <param name="points">The (x,y) points to later cluster by k-means.</param>
        /// <param name="k">The number of cluster centers to get, for k >= 1. If k == 1, just the nonparametric centroid, or
        /// midpoint, of all points is returned.</param>
        /// <returns>A list of k points representing centers to initialize k-means with.</returns>
        /// <remarks>k-means++ was published in the following article on which this implementation draws:
        /// Arthur, D. and Vassilvitskii, S. (2007). k-means++: The advantages of careful seeding. Proceedings of
        ///     the 18th Annual ACM-SIAM Symposium on Discrete Algorithms (SODA '07). Philadelphia, PA: Society for 
        ///     Industrial and Applied Mathematics, pp. 1027-1035.
        /// </remarks>
        public static List<PointR> KMeansPlusPlus(List<PointR> points, int k)
        {
            if (k > points.Count)
                k = points.Count; // can't have more 'k' than points themselves!

            List<PointR> C = new List<PointR>(); // centers
            if (k == 1) // global centroid
            {
                PointR cxy = GeotrigEx.MidPoint(points); // "nonparametric centroid"
                C.Add(cxy);
            }
            else if (k > 1) // k-means++
            {
                //  1. Choose one point uniformly at random from (x_1, x_2, ..., x_n), and add it to C.
                int idx = RandomEx.Integer(0, points.Count - 1);
                C.Add(points[idx]);

                double[] Dx = new double[points.Count]; // distances
                while (C.Count < k)
                {
                    // 2. For each point x_i, set D(x_i) to be the distance between x_i and the nearest center in C.
                    for (int i = 0; i < points.Count; i++) // for each point x_i
                    {
                        double dbest = double.MaxValue;
                        for (int j = 0; j < C.Count; j++) // find nearest center in C for points[i]
                        {
                            double d = GeotrigEx.Distance(points[i], C[j]);
                            if (d < dbest)
                            {
                                dbest = d;
                            }
                        }
                        Dx[i] = dbest; // set Dx[i] for points[i]
                    }

                    // 3. Choose a real number y uniformly at random between 0 and D(x_1)^2 + D(x_2)^2 + ... + D(x_n)^2.
                    double dssq = 0.0;
                    for (int i = 0; i < Dx.Length; i++)
                    {
                        dssq += Dx[i] * Dx[i]; // distance sum of squares
                    }
                    double y = RandomEx.Double(0.0, dssq);

                    // 4. Find the unique integer i so that D(x_1)^2 + D(x_2)^2 + ... D(x_i)^2 >= y > D(x_1)^2 + D(x_2)^2 + ... + D(x_(i-1))^2.
                    double dssqgt = 0.0; // sum of squares greater than y
                    double dssqlt = 0.0; // sum of squares less than y
                    for (int i = 0; i < Dx.Length; i++)
                    {
                        double d = Dx[i] * Dx[i];
                        dssqgt += d;
                        if ((dssqlt < y && y <= dssqgt) || (dssq == 0.0))
                        {
                            C.Add(points[i]); // 5. Add x_i to C.
                            break;
                        }
                        dssqlt += d;
                    }
                } // Repeat Steps 2-5 until we have chosen k centers.
            }
            return C;
        }

        #endregion

        #region G-means

        /// <summary>
        /// An advanced algorithm that is like k-means but does not require being given an explicit k. It does require a
        /// statistical significance level alpha. The published G-means algorithm has been modified here for use with 
        /// k-means++ to pick initial centers and center-splits.
        /// </summary>
        /// <param name="points">The points to cluster.</param>
        /// <param name="alpha">Statistical significance level to be used in testing for Gaussian-ness. Use .10, .05, .025, 
        /// .01 or .005 are recommended.</param>
        /// <param name="C">Out parameter returns the centers of the returned point clusters.</param>
        /// <returns>A list of point clusters, i.e., a list of point lists. From the original unclustered set of points, 
        /// the returned list has grouped points into clusters whose index-matching centers are returned in C.</returns>
        /// <remarks>G-means was published in the following article on which this implementation draws:
        ///   Hamerly, G. and Elkan, C. (2004). Learning the k in k-means. In Advances in Neural Information Processing 
        ///     Systems 16, Thrun, S., Saul, L. K. and Schölkopf, B. (eds.). Cambridge, MA: The MIT Press, pp. 281-288.
        ///     [This book resulted from the Proceedings of the 17th Annual Conference on Neural Information Processing 
        ///     Systems (NIPS '03). Vancouver and Whistler, British Columbia (December 8-13, 2003).]
        /// </remarks>
        public static List<List<PointR>> GMeans(List<PointR> points, double alpha, out List<PointR> C)
        {
            List<List<PointR>> X = null; // the point clusters
            C = KMeansPlusPlus(points, 1); // initialize with one center

            while (true) // loop until centers found by G-means do not change
            {
                // Each iteration of the algorithm splits into two those centers whose data appear not to come from a Gaussian distribution.
                // Between each round of splitting, we run k-means on the entire dataset and all the centers to refine the current solution.
                int nCenters = C.Count;
                X = KMeans(points, C);
                for (int i = X.Count - 1; i >= 0; i--) // reverse loop so the addition of new centers doesn't mess up bounds
                {
                    if (X[i].Count > 1) // in the case of a 1-point cluster, keep the center, which already matches the point itself
                    {
                        // Test points in the ith cluster for Gaussian-ness. If the points are Gaussian, keep their center C[i].
                        // Initialize two "child" centers c1 and c2 of C[i] for this cluster using k-means++.
                        List<PointR> cc = KMeansPlusPlus(X[i], 2); // split

                        // Run k-means on these two centers in this cluster. We do not need the return values, just setting c1c2.
                        KMeans(X[i], cc); // c1 and c2 now have centers chosen by k-means
                        if (cc.Count > 1) // if cc.Count is 1, k-means decided only one center was necessary because all points in X[i] are the same
                        {
                            // Create a vector v connecting the split centers, and compute is length squared (its module).
                            PointR v = new PointR(cc[1].X - cc[0].X, cc[1].Y - cc[0].Y);
                            double vlensq = v.X * v.X + v.Y * v.Y; // ||v||^2

                            // Create X', which is the 1-D projection of X onto the vector v.
                            double[] Xp = new double[X[i].Count]; // new x_i' values
                            for (int j = 0; j < X[i].Count; j++)
                            {
                                Xp[j] = (X[i][j].X * v.X + X[i][j].Y * v.Y) / vlensq; // project X[i][j] onto v
                            }
                            // Perform the Anderson-Darling test of normality on X'. If significant, reject normality and keep the split.
                            if (StatsEx.AndersonDarling(Xp, alpha, out double A2, out double CV))
                            {
                                C.RemoveAt(i); // sig. test, so data is non-Gaussian; reject old center and 
                                C.AddRange(cc); // accept new centers (add to end).
                            }
                        }
                    }
                }
                if (C.Count == nCenters)
                    break; // no more centers were added this iteration
            }
            return X;
        }

        /// <summary>
        /// TODO: Not yet implemented.
        /// </summary>
        /// <param name="points"></param>
        /// <param name="C"></param>
        /// <returns></returns>
        /// <remarks>PG-means was published in the following article on which this implementation draws:
        /// Feng, Y. and Hamerly, G. (2007). PG-means: Learning the number of clusters in data. In Advances in Neural Information 
        /// Processing Systems 19, Schölkopf, B., Platt, J. and Hofmann, T. (eds.). Cambridge, MA: The MIT Press, pp. 393-400.
        /// [This book resulted from the Proceedings of the 20th Annual Conference on Neural Information Processing 
        /// Systems (NIPS '06). Vancouver and Whistler, British Columbia (December 4-9, 2006).]
        /// </remarks>
        public static List<List<PointR>> PGMeans(List<PointR> points, out List<PointR> C)
        {
            throw new NotImplementedException("ClusterEx.PGMeans has not been implemented.");
        }

        #endregion

        #region CC-Pivot

        /// <summary>
        /// Implements the CC-Pivot algorithm of Ailon et al. (2005) but, given that algorithm relies on some
        /// randomness, different calls to it can produce different results for the same data. Accordingly, we
        /// can make multiple calls to it, taking the best results. I call that approach "IterativeCCPivot" and
        /// it requires a scoring function to know what "best" is. 
        /// </summary>
        /// <param name="points">The points to cluster.</param>
        /// <param name="iterations">The number of iterations over which to run CC-Pivot.</param>
        /// <param name="centers">Out parameter returning the centers of each index-matched point cluster.</param>
        /// <returns>The point clusters.</returns>
        public static List<List<PointR>> IterativeCCPivot(List<PointR> points, int iterations, out List<PointR> centers)
        {
            RectangleR box = GeotrigEx.FindBoundingBox(points);
            double diagonal = box.Diagonal;

            double bestDist = double.MaxValue;
            List<List<PointR>> bestClusters = null;
            List<PointR> bestCenters = null;

            for (int i = 0; i < iterations; i++)
            {
                List<List<PointR>> theseClusters = CCPivot(points, diagonal, out List<PointR> theseCenters); // CC-Pivot
                double dist = CCPivotDistance(theseClusters, diagonal); // score the result
                if (dist < bestDist) // possibly update best
                {
                    bestDist = dist;
                    bestClusters = theseClusters;
                    bestCenters = theseCenters;
                }
            }

            centers = bestCenters;
            return bestClusters;
        }

        /// <summary>
        /// Computes a score indicating how well clustered the points are in the given point clusters.
        /// Conceptually, good groupings have very short distances between all points within their clusters,
        /// and very long distances between all points across clusters. Of course, the points might not themselves
        /// be arranged that way, but a distance metric reflecting this state indicates how well clustered the data is.
        /// The shorter the distance, the better, for the same set of points.
        /// </summary>
        /// <param name="clusters">A list of points lists, each representing a point cluster.</param>
        /// <returns>A distance indicating how well clustered the points are. The lower the distance,
        /// the better, for the same set of points.</returns>
        private static double CCPivotDistance(List<List<PointR>> clusters, double diagonal)
        {
            double dist = 0.0; // sum of normalized distances
            int count = 0; // number of sums

            //
            // sum normalized distances WITHIN clusters
            //
            for (int i = 0; i < clusters.Count; i++) // step through each cluster
            {
                for (int j = 0; j < clusters[i].Count; j++) // step through each point in that cluster
                {
                    PointR pFrom = clusters[i][j];
                    for (int k = j + 1; k < clusters[i].Count; k++) // step through each other point in that cluster
                    {
                        PointR pTo = clusters[i][k];
                        double d = GeotrigEx.Distance(pFrom, pTo); // raw distance
						//dist += Math.Sqrt(d / diagonal);
						dist += (d / diagonal);
						count++;
                    }
                }
            }

            // DEBUG: check count: the number of comparisons in a cluster should be the cluster indices added together
            //int nWithin = 0;
            //for (int i = 0; i < clusters.Count; i++)
            //    for (int j = clusters[i].Count - 1; j >= 0; j--)
            //        nWithin += j;
            //Debug.Assert(nWithin == count, "Number of distance sums within clusters is incorrect.");

            //
            // sum converse normalized distances ACROSS clusters
            //
            for (int i = 0; i < clusters.Count; i++) // step through each cluster
            {
                for (int j = 0; j < clusters[i].Count; j++) // step through each point in that cluster
                {
                    PointR pFrom = clusters[i][j];
                    for (int k = i + 1; k < clusters.Count; k++) // step through each other cluster
                    {
						for (int m = 0; m < clusters[k].Count; m++) // step through each point in that other cluster
						{
							PointR pTo = clusters[k][m];
							double d = GeotrigEx.Distance(pFrom, pTo); // raw distance
							//dist += (1.0 - Math.Pow(d / diagonal, 2.0));
							//dist += (1.0 - d / diagonal);
							dist -= (d / diagonal);
							count++;
						}
                    }
                }
            }

            // DEBUG: count check
            //int nAcross = 0;
            //for (int i = 0; i < clusters.Count; i++)
            //    for (int j = i + 1; j < clusters.Count; j++)
            //        nAcross += (clusters[i].Count * clusters[j].Count);
            //Debug.Assert(nAcross == (count - nWithin), "Number of distance sums across clusters is incorrect.");

			// normalize distance by cluster count but add a penalty for more clusters
            return dist / clusters.Count + 0.5 * clusters.Count; // lower is better
        }

        /// <summary>
        /// Uses the CC-Pivot method of correlation clustering to cluster points.
        /// </summary>
        /// <param name="points">The points to cluster.</param>
        /// <param name="centers">Out parameter returns the centers of the returned point clusters.</param>
        /// <returns>A list of point clusters. From the original unclustered set of points, the returned list has 
        /// grouped points into clusters whose index-matching centers are returned in centers.</returns>
        /// <remarks>Ailon, N., Charikar, M. and Newman, A. (2005). Aggregating inconsistent information: Ranking
        /// and clustering. Proceedings of the ACM Symposium on the Theory of Computing (STOC '05). Baltimore, 
        /// Maryland (May 22-24, 2005). New York: ACM Press, pp. 684-693. DOI: 10.1145/1060590.1060692</remarks>
        /// <see cref="https://people.mpi-inf.mpg.de/~alantha/papers/aggregating.pdf"/>
        private static List<List<PointR>> CCPivot(List<PointR> points, double diagonal, out List<PointR> centers)
        {
            List<List<PointR>> clusters = new List<List<PointR>>(points.Count); // the point clusters, empty initially

			CCPivotHelper(clusters, points, diagonal); // "clusters" will be filled after this recursive call

			// compute the centers of the clusters as their 2-D midpoints
			centers = new List<PointR>(clusters.Count);
            for (int i = 0; i < clusters.Count; i++)
            {
                PointR cxy = GeotrigEx.MidPoint(clusters[i]); // "nonparametric centroids"
                centers.Add(cxy);
            }

			return clusters;
        }

        /// <summary>
        /// Recursive helper function that implements CC-Pivot. Points are successively clustered probabilistically
        /// based on their 2-D Euclidean distance from each other. The closer points are together, the higher the
        /// chance they will be put into the same cluster. Distances are normalized by the diagonal of the bounding
        /// box surrounding all points. 
        /// </summary>
        /// <param name="clusters">The clusters forming thus far upon each recursion.</param>
        /// <param name="points">The dwindling set of points considered upon each recursion.</param>
        /// <param name="diagonal">The diagonal surrounding all original points to be clustered.</param>
        /// <remarks>Ailon, N., Charikar, M. and Newman, A. (2005). Aggregating inconsistent information: Ranking
        /// and clustering. Proceedings of the ACM Symposium on the Theory of Computing (STOC '05). Baltimore, 
        /// Maryland (May 22-24, 2005). New York: ACM Press, pp. 684-693. DOI: 10.1145/1060590.1060692</remarks>
        /// <see cref="https://people.mpi-inf.mpg.de/~alantha/papers/aggregating.pdf"/>
        private static void CCPivotHelper(List<List<PointR>> clusters, List<PointR> points, double diagonal)
        {
            if (points.Count == 0)
                return; // base case: all points clustered!

            int i = RandomEx.Integer(0, points.Count - 1);
            PointR pivot = points[i]; // uniform random pivot point

            clusters.Add(new List<PointR>(points.Count)); // C in CC-Pivot
            clusters[clusters.Count - 1].Add(pivot); // start a new cluster for this recursion

            List<PointR> subpoints = new List<PointR>(); // V' in CC-Pivot

            for (int j = 0; j < points.Count; j++) // for all other nodes in the graph
            {
				if (j != i)
				{
					bool added = false;
					double d = GeotrigEx.Distance(pivot, points[j]); // distance
					double x = d / diagonal; // normalized distance

					// various probabilistic models from x=[0..1] to get a decreasing y=[0..1]
					// graph them at https://www.desmos.com/calculator
					//double y = -x + 1.0;
					//double y = Math.Pow(1.0 - x, 2.0);
					double y = 0.3760 * StatsEx.NormalDistribution(x, 0.0, 0.15);
					//double y = -0.5 + StatsEx.ExponentialDistribution(x, 1.5);

					double p = RandomEx.Double(0.00, 1.00); // probability
					if (p < y)
					{
						clusters[clusters.Count - 1].Add(points[j]); // add to C in CC-Pivot
						added = true;
					}
					if (!added)
						subpoints.Add(points[j]); // add to V' in CC-Pivot
				}
            }
            CCPivotHelper(clusters, subpoints, diagonal); // recursive step: subpoints is fewer than points
        }

        #endregion
    }
}
