﻿using System;
using System.Collections.Generic;
using System.Text;

namespace WobbrockLib.Extensions
{
    /// <summary>
    /// Utility functions for working with sets.
    /// </summary>
    public static class SetsEx
    {
        /// <summary>
        /// 
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="set1"></param>
        /// <param name="set2"></param>
        /// <returns></returns>
        public static List<T> Intersect<T>(List<T> set1, List<T> set2)
        {
            List<T> newset = new List<T>(Math.Min(set1.Count, set2.Count) + 1);
            List<T> shortest = (set1.Count < set2.Count) ? set1 : set2;
            List<T> longest = (set1.Count >= set2.Count) ? set1 : set2;
            for (int i = 0; i < shortest.Count; i++) // now go through shortest
            {
                if (longest.Contains(shortest[i])) // keep any items contained by longest
                {
                    newset.Add(shortest[i]);
                }
            }
            return newset;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="set1"></param>
        /// <param name="set2"></param>
        /// <returns></returns>
        public static List<T> Union<T>(List<T> set1, List<T> set2)
        {
            List<T> shortest = (set1.Count < set2.Count) ? set1 : set2;
            List<T> longest = (set1.Count >= set2.Count) ? set1 : set2;
            List<T> newset = new List<T>(longest); // copy over the longest set
            for (int i = 0; i < shortest.Count; i++) // now go through shortest
            {
                if (!longest.Contains(shortest[i]))
                {
                    newset.Add(shortest[i]);
                }
            }
            return newset;
        }

        /// <summary>
        /// Produces the powerset of a given set. The powerset contains 2^n items
        /// for a set with n items. It includes the empty set and the original set
        /// itself.
        /// </summary>
        /// <typeparam name="T">The type of objects stored in the set.</typeparam>
        /// <param name="set">The set on which to computer the powerset.</param>
        /// <returns>The powerset of the given set.</returns>
        public static List<List<T>> PowerSet<T>(List<T> set)
        {
            long size = (int) Math.Pow(2, set.Count);
            List<List<T>> pset = new List<List<T>>((int) size);
            for (long i = 0; i < size; i++)
            {
                List<T> oset = new List<T>(set.Count);
                for (int j = 0; j < set.Count; j++)
                {
                    // check jth bit of i
                    if (((i >> j) & 0x01) == 1)
                    {
                        oset.Add(set[j]);
                    }
                }
                pset.Add(oset);
            }
            return pset;
        }

        /// <summary>
        /// Produces a space-free string representing the powerset of the given set,
        /// using the standard curly-brace notation for powersets. (Spaces may exist
        /// if they were in the original object representations themselves. Objects
        /// whose string representations contain curly braces will have those curly
        /// braces removed to avoid confusion.)
        /// </summary>
        /// <param name="set">A set of objects whose powerset string representation will be returned.</param>
        /// <returns>A string representing the powerset of the given set.</returns>
        public static string PowerSetString<T>(List<T> set)
        {
            string str = "{"; // open whole powerset
            List<List<T>> pset = PowerSet<T>(set);
            for (int i = 0; i < pset.Count; i++)
            {
                str += "{"; // open one set
                for (int j = 0; j < pset[i].Count; j++)
                {
                    string s = pset[i][j].ToString();
                    s = s.Trim(new char[2] { '{', '}' });
                    s += ",";
                    str += s;
                }
                str = str.TrimEnd(new char[1] { ',' });
                str += "},"; // close one set
            }
            str = str.TrimEnd(new char[1] { ',' });
            return str + "}"; // close whole powerset
        }
    }
}
