using System;
using System.Net;
using System.Net.Sockets;
using System.Windows.Forms;
using System.Threading;
using System.Text;

namespace WobbrockLib.Net
{
    /// <summary>
    /// Implements a TCP server that listens for client messages.
    /// </summary>
	public class NetListener
	{
		#region NetEventArgs

        /// <summary>
        /// Event arguments that deliver to an owner control a string message
        /// that has been received by a client over the network.
        /// </summary>
		public class NetEventArgs : System.EventArgs
		{
			private string _s;
			public NetEventArgs()
			{
				_s = String.Empty;
			}
			public NetEventArgs(string s)
			{
				_s = s;
			}
			public string Data
			{
				get { return _s; }
			}
		}

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
		public delegate void NetEventHandler(object sender, NetEventArgs e);

		#endregion
		
		public virtual event EventHandler	 NetEvent;
		public virtual event NetEventHandler NetDataEvent;
		
		protected IPAddress					_ip;
        protected int						_port;

        protected Control                   _owner;
        protected TcpListener               _server;
        protected NetworkStream             _clientStream;
        protected Thread			        _thread;

        /// <summary>
        /// Constructs a server that listens for a client to send messages over TCP.
        /// </summary>
        /// <param name="owner"></param>
		public NetListener(System.Windows.Forms.Control owner)
		{
			_owner = owner;
			_ip = IPAddress.Parse("127.0.0.1");
			_port = 9080;
		}

        /// <summary>
        /// The IP address to which this server is listening.
        /// </summary>
		public virtual string IP
        {
            get
            {
                return _ip.ToString();
            }
            set
            {
                try
                {
                    _ip = IPAddress.Parse(value);
                }
                catch (Exception ex)
                {
                    Console.Write(ex.Message);
                }
            }
        }

        /// <summary>
        /// The Port on which this server is listening.
        /// </summary>
		public virtual int Port
        {
            get
            {
                return _port;
            }
            set
            {
                if (value > 0)
                {
                    _port = value;
                }
            }
        }

        /// <summary>
        /// Gets whether a client is currently connected to this listener (server).
        /// </summary>
		public virtual bool IsConnected
		{
			get
			{
				return (_clientStream != null);
			}
		}

        /// <summary>
        /// Gets whether this server is listening for a client connection.
        /// </summary>
		public virtual bool IsListening
        {
            get
            {
                return (_thread != null && _thread.IsAlive);
            }
        }

        /// <summary>
        /// Starts or restarts this server listening for a client connection.
        /// </summary>
		public virtual void Start()
        {
            if (IsListening)
                this.Stop();

            // start a worker thread to listen in a separate thread
            _thread = new Thread(new ThreadStart(ThreadListen));
            _thread.Name = String.Format("NetListener {0}", DateTime.Now);

            // the CLR will close this thread once all foreground threads exit
            _thread.IsBackground = true;
            _thread.Start();
        }

        /// <summary>
        /// Stops this server from listening for a client connection.
        /// </summary>
		public virtual void Stop()
        {
            if (_thread != null && _thread.IsAlive)
            {
                if (_clientStream != null)
                {
                    _clientStream.Close();
                    _clientStream = null;
                }
                if (_server != null)
                {
                    _server.Stop();
                    _server = null;
                }
                _thread.Abort();
                _thread = null;
            }
        }

        /// <summary>
        /// Send a message back to the client, assuming a client is connected to this listener.
        /// </summary>
        /// <param name="msg">The message to send.</param>
        /// <returns>True if the message is sent successfully; false otherwise.</returns>
		public virtual bool Send(string msg)
		{
			if (_clientStream == null) 
				return false;

			bool success = true;
			try
			{
				// encode our message for sending
				byte[] buffer = Encoding.UTF8.GetBytes(msg);
				
				// send the message over the stream to the client
				_clientStream.Write(buffer, 0, buffer.Length);
			}
			catch (SocketException sex) 
			{
				Console.WriteLine(sex.Message);
				success = false;
			}
			catch (Exception ex)
			{
				Console.WriteLine(ex.Message);
				success = false;
			}
			return success;
		}

        /// <summary>
        /// The private routine for the background thread listening for a client connection and
        /// receiving data from the TCP client. This routine fires events to the owner control of
        /// this listener.
        /// </summary>
		protected virtual void ThreadListen()
		{
            try
			{
				_server = new TcpListener(_ip, _port);
				_server.Start(); // start listening

                while (true) // no break from this loop -- CLR will close background thread
                {
                    TcpClient client = _server.AcceptTcpClient(); // blocking until a connection
                    _clientStream = client.GetStream(); // connection!

                    while (!_clientStream.DataAvailable)
                        Thread.Sleep(0); // yield
                    byte[] buffer = new byte[client.Available];
                    string data = String.Empty;

                    // loop to receive all the data sent by the client
                    int i = 0;
                    while ((i = _clientStream.Read(buffer, 0, buffer.Length)) != 0) // blocks until it can read
                    {
                        // translate the raw bytes into an UTF-8 string
                        data = Encoding.UTF8.GetString(buffer, 0, i);

                        if (NetEvent != null)
                            _owner.BeginInvoke(NetEvent); // fire event
                        if (NetDataEvent != null)
                            _owner.BeginInvoke(NetDataEvent, new object[2] { this, new NetEventArgs(data) }); // fire event

                        // yield the processor until we get new data
                        while (!_clientStream.DataAvailable)
                            Thread.Sleep(0); // yield
                        buffer = new byte[client.Available];
                        data = String.Empty;
                    }

                    client.Close();
                    _clientStream.Close();
                    _clientStream = null;
                }
			}
			catch (ThreadAbortException taex)
			{
				Console.WriteLine("ThreadAbortException: {0}", taex.Message);
			}
			catch (SocketException sex)
			{
				Console.WriteLine("SocketException: {0}", sex.Message);
			}
			catch (Exception ex)
			{
				Console.WriteLine("Exception: {0}", ex.Message);
			}
			finally
			{
                if (_clientStream != null)
                {
                    _clientStream.Close();
                    _clientStream = null;
                }
				if (_server != null)
				{
					_server.Stop();
					_server = null;
				}
			}
		}

	}
}