﻿using System;
using System.Collections.Generic;
using WobbrockLib.Extensions;
using WobbrockLib.Optimization;

namespace WobbrockLib.Optimization.Genetic
{
    /// <summary>
    /// This class simulates an ecosystem in which simulated organisms mutate from
    /// one generation to the next, surviving according to a fitness measurement
    /// and thereby converging to a solution.
    /// </summary>
    public class Ecosystem
    {
        /// <summary>
        /// Constructs an empty ecosystem capable of evolving organisms over generations.
        /// </summary>
        public Ecosystem()
        {
            // do nothing
        }

        /// <summary>
        /// Populates the ecosystem with an initial randomly generated population. A single parent ("eve") 
        /// is used to birth random organisms from the same species.
        /// </summary>
        /// <param name="population">The number of organisms to establish in the initial population.</param>
        /// <param name="eve">A single parent that serves as a prototype for creating the 
        /// initial population. The parent will be birthing random organisms to start the population.</param>
        /// <param name="state">If the fitness can be calculated and stored as a static value thereafter, the 
        /// world state can be passed in to do so, saving future computations of fitness and speeding up evolution 
        /// considerably.</param>
        /// <returns>A population of organisms birthed randomly from Eve.</returns>
        public List<Organism> Populate(int population, Organism eve, object state = null)
        {
            List<Organism> organisms = new List<Organism>(population);
            for (int i = 0; i < population; i++)
            {
                Organism o = eve.BirthRandom(state);
                organisms.Add(o);
            }
            return organisms;
        }

        /// <summary>
        /// Evolves generations of organisms to converge to a solution encoded in their DNA. 
        /// </summary>
        /// <param name="organisms">The initial population of organisms to evolve. These can be created by 
        /// having a single "Eve" organism use the BirthRandom() method.</param>
        /// <param name="comparer">The fitness comparer able to compare the fitness of two organisms in light 
        /// of the world state.</param>
        /// <param name="maxGeneration">The maximum generation to which to evolve the organisms.</param>
        /// <param name="survivalRate">The percentage [0..1] of those fittest organisms that survive in a 
        /// generation to spawn the next generation.</param>
        /// <param name="exceptionRate">The percentage [0..1] of remaining unfit organisms chosen at random 
        /// to survive and spawn. Useful for escaping local maxima.</param>
        /// <param name="convergences">The number of successive generations the fittest solution must remain 
        /// converged to short-circuit the evolution; or non-positive if no convergence short-circuiting is 
        /// desired.</param>
        /// <param name="state">The world state that can be used to evaluate the fitness of the organisms.</param>
        /// <param name="fn">An optional callback function indicating the generation number that has just completed 
        /// its evolution. Note that the callback is made synchronously on this method's calling thread, so if the 
        /// subscriber intends to carry out extensive operations or update UI controls as a result, a new thread should 
        /// be spawned to handle that.</param>
        /// <returns>An organism deemed "fit enough" for evolution to short-circuit, or whose solution has 
        /// converged successfully, or if no such organisms are found, the most fit organism after all requested 
        /// generations have evolved.</returns>
        public Organism Evolve(
            List<Organism> organisms, 
            FitnessComparer comparer, 
            int maxGeneration, 
            double survivalRate, 
            double exceptionRate = 0.00,
            int convergences = 0,
            object state = null, 
            OptimizationCallback fn = null)
        {
            if (organisms == null || organisms.Count == 0)
                return null; // error code

            // calculate loop variables
            int population = organisms.Count;
            int numFittest = (int) Math.Max(1.0, Math.Round(survivalRate * population));
            int numExceptions = (int) Math.Round(exceptionRate * (population - numFittest));
            int spawnPerParent = (int) Math.Max(1.0, Math.Round((double) (population - numFittest - numExceptions) / (numFittest + numExceptions)));

            // initialize convergence variables
            int nConverge = 0;
            Organism prevFittest = null;

            // main loop over generations
            while (organisms[0].Generation < maxGeneration)
            {
                // (1) sort by descending order of fitness, so fittest parent is "on top" of this list at [0].
                // this is a time-consuming operation unless fitness can be computed and stored at birth.
                organisms.Sort(comparer);

                // (2) look for solution convergence, if a convergence count was supplied
                if (convergences > 0 && prevFittest != null)
                {
                    if (prevFittest == organisms[0] ||  // if the reference variable is itself the same, then trivially there must be convergence!
                        prevFittest.HasConvergedWith(organisms[0], state)) // or if not, then test explicitly for convergence of the solutions
                    {
                        if (++nConverge >= convergences) // increment and test
                            return organisms[0]; // solution has converged repeatedly
                    }
                    else nConverge = 0; // reset
                }
                prevFittest = organisms[0]; // update reference to fittest

                // this will hold the new generation, which includes new spawn, this
                // gen's fittest parents, and this gen's dumb-lucky unfit exceptions
                List<Organism> newOrganisms = new List<Organism>(population + numFittest + numExceptions);

                // (3) loop over the fittest parents, spawning a new generation
                for (int i = 0; i < numFittest; i++)
                {
                    Organism parent = organisms[i];
                    if (parent.IsFitEnough(state))
                        return parent; // solution found!

                    // mutate more in early generations, less near final generations
                    int nMutations = (int) GeotrigEx.MapValue(parent.Generation, 1.0, maxGeneration, parent.Size, 1.0, true);
                    for (int j = 0; j < spawnPerParent; j++) // birth new offspring
                    {
                        Organism spawn = parent.BirthMutated(nMutations, state);
                        if (spawn.IsFitEnough(state))
                            return spawn; // solution found!
                        newOrganisms.Add(spawn);
                    }
                    parent.Generation++; // update the parent's age...
                    newOrganisms.Add(parent); // ...and add parent to next generation
                }
                // (4) choose any unfit exceptions that get lucky enough to survive
                for (int i = 0; i < numExceptions; i++)
                {
                    int r = RandomEx.Integer(numFittest, organisms.Count - 1); // random index from the unfit masses
                    Organism parent = organisms[r]; // random unfit parent can't be fit enough
                    int nMutations = (int) GeotrigEx.MapValue(parent.Generation, 1.0, maxGeneration, parent.Size, 1.0, true);
                    for (int j = 0; j < spawnPerParent; j++)
                    {
                        Organism spawn = parent.BirthMutated(nMutations, state);
                        newOrganisms.Add(spawn); // no fitness check; just add with chance it could sort atop next go-around
                    }
                    parent.Generation++;
                    newOrganisms.Add(parent);
                }
                // (5) update with next generation and notify onlookers
                organisms = newOrganisms; 
                fn?.Invoke(organisms[0].Generation - 1); // last generation complete
            }
            // (6) no good-enough solution found after all generations, so return fittest we bred
            int fittest = -1;
            double fitness = double.MinValue;
            for (int i = 0; i < organisms.Count; i++)
            {
                double fiti = organisms[i].FitnessAssignedAtBirth ? organisms[i].GetFitness(null) : organisms[i].GetFitness(state);
                if (fiti > fitness)
                {
                    fittest = i; // update
                    fitness = fiti; // update
                }
            }
            return organisms[fittest];
        }

    }
}