﻿using System;
using System.Collections.Generic;

namespace WobbrockLib.Optimization.Genetic
{
    /// <summary>
    /// This class is capable of comparing two Organsisms' fitness for the
    /// purposes of sorting organisms based on fitness. The reason this class is
    /// necessary instead of making Organism an IComparable is that the comparator
    /// may require knowledge of the world state when performing comparisons.
    /// This class encapsulates that knowledge.
    /// </summary>
    public class FitnessComparer : IComparer<Organism>
    {
        private object _state;

        /// <summary>
        /// Constructs a FitnessComparer.
        /// </summary>
        /// <param name="state">The world state used when computing fitness. 
        /// May be null if such state is unnecessary.</param>
        public FitnessComparer(object state)
        {
            _state = state;
        }

        /// <summary>
        /// Compares two organisms, returning -1, 0, and +1 such that organisms,
        /// when sorted, will be sorted from highest fitness to least fitness in descending
        /// order.
        /// </summary>
        /// <param name="x">The first organism to compare.</param>
        /// <param name="y">The second organism to compare.</param>
        /// <returns>-1 if x is fitter than y; +1 if y is fitter than x; 0 if the fitness
        /// of x and y are equal.</returns>
        public virtual int Compare(Organism x, Organism y)
        {
            double xFitness;
            double yFitness;

            if (x.FitnessAssignedAtBirth && y.FitnessAssignedAtBirth) // just look up fitness (fast)
            {
                xFitness = x.GetFitness(null);
                yFitness = y.GetFitness(null);
            }
            else // compute fitness afresh (slow)
            {
                xFitness = x.ComputeFitness(_state);
                yFitness = y.ComputeFitness(_state);
            }

            if (xFitness > yFitness)
                return -1;
            else if (xFitness < yFitness)
                return +1;
            else
                return 0;
        }
    }
}
