﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace WobbrockLib.Optimization.Genetic
{
    public partial class GeneticForm : Form
    {
        /// <summary>
        /// Options class for this form.
        /// </summary>
        public class Options
        {
            public int Population { get; set; }
            public int Generations { get; set; }
            public double SurvivalRate { get; set; }
            public double ExceptionRate { get; set; }
            public int Convergences { get; set; }

            /// <summary>
            /// Construct a new set of options with their default values.
            /// </summary>
            public Options() // defaults
            {
                this.Population = 100;
                this.Generations = 1000;
                this.SurvivalRate = 0.20;
                this.ExceptionRate = 0.05;
                this.Convergences = 100;
            }

            /// <summary>
            /// Copy the current options and return a deep copy.
            /// </summary>
            /// <returns></returns>
            public Options Copy()
            {
                Options o = new Options();
                o.Population = this.Population;
                o.Generations = this.Generations;
                o.SurvivalRate = this.SurvivalRate;
                o.ExceptionRate = this.ExceptionRate;
                o.Convergences = this.Convergences;
                return o;
            }
        }

        private GeneticForm.Options _o;

        /// <summary>
        /// Constructor with the default options.
        /// </summary>
        public GeneticForm()
        {
            InitializeComponent();
            _o = new GeneticForm.Options();
        }

        /// <summary>
        /// Create the form with the options given.
        /// </summary>
        /// <param name="o"></param>
        public GeneticForm(GeneticForm.Options o)
        {
            InitializeComponent();
            _o = o.Copy();
        }

        /// <summary>
        /// Loads the form and sets the values of the widgets.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        /// <remarks>This is the "copy in" procedure.</remarks>
        private void GeneticForm_Load(object sender, EventArgs e)
        {
            numPopulation.Value = _o.Population;
            numGenerations.Value = _o.Generations;
            numSurvival.Value = (decimal) (_o.SurvivalRate * 100.0);
            numException.Value = (decimal) (_o.ExceptionRate * 100.0);
            numConverge.Value = _o.Convergences;
        }

        /// <summary>
        /// Gets or sets the options from the controls on the form.
        /// </summary>
        /// <remarks>This is the "copy out" procedure.</remarks>
        public GeneticForm.Options FormOptions
        {
            get
            {
                GeneticForm.Options o = new GeneticForm.Options();
                o.Population = (int) numPopulation.Value;
                o.Generations = (int) numGenerations.Value;
                o.SurvivalRate = (double) numSurvival.Value / 100.0;
                o.ExceptionRate = (double) numException.Value / 100.0;
                o.Convergences = (int) numConverge.Value;
                return o;
            }
        }

        /// <summary>
        /// Enables or disables the exception percentage widgets.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void chkException_CheckedChanged(object sender, EventArgs e)
        {
            lblException.Enabled = chkException.Checked;
            numException.Enabled = chkException.Checked;
            lblExceptionRange.Enabled = chkException.Checked;
            numException.Value = chkException.Checked ? 5 : 0;
        }

        /// <summary>
        /// Unchecks the exception percentage box when the percentage reaches zero.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void numException_ValueChanged(object sender, EventArgs e)
        {
            chkException.Checked = (numException.Value != 0);
        }

        /// <summary>
        /// Enables or disables the convergence criterion widgets.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void chkConverge_CheckedChanged(object sender, EventArgs e)
        {
            lblConverge.Enabled = chkConverge.Checked;
            numConverge.Enabled = chkConverge.Checked;
            lblConvergeRange.Enabled = chkConverge.Checked;
            numConverge.Value = chkConverge.Checked ? 100 : 0; // default
        }

        /// <summary>
        /// Unchecks the convergence criterion box when the criterion reaches zero.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void numConverge_ValueChanged(object sender, EventArgs e)
        {
            chkConverge.Checked = (numConverge.Value != 0);
        }
    }
}
