﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace WobbrockLib.Optimization
{
    /// <summary>
    /// Defines a class capable of executing the Shotgun Hill Climbing optimization algorithm. Shotgun
    /// hill climbing is like regular hill climbing but with an arbitrary number of randomly started 
    /// climbers. It therefore degenerates into regular hill climbing if the number of climbers is 1.
    /// </summary>
    public class ShotgunClimbers : HillClimber
    {
        /// <summary>
        /// Default constants.
        /// </summary>
        protected const int CLIMBERS = 100;

        /// <summary>
        /// Fields.
        /// </summary>
        protected int _climbers;

        /// <summary>
        /// Constructs an instance of a shotgun hill climber, which places any number of hill climbers into 
        /// a solution space and has them each climb to the top of their respective peak. Then, the climber 
        /// on the highest peak is taken as the final solution. Degenerates into regular hill climbing if the 
        /// number of climbers is 1.
        /// </summary>
        /// <param name="climbers">The initial number of climbers to use. Defaults to 100.</param>
        public ShotgunClimbers(int climbers = CLIMBERS)
        {
            _climbers = climbers;
        }

        /// <summary>
        /// Gets or sets the number of climbers in this instance of a shotgun hill climber.
        /// </summary>
        public int Climbers
        {
            get
            {
                return _climbers;
            }
            set
            {
                if (value >= 1)
                    _climbers = value;
                else
                    throw new ArgumentOutOfRangeException("Argument out of range in ShotgunClimber.Climbers.");
            }
        }

        /// <summary>
        /// A convenience method that sets the number of climbers prior to starting the optimization process.
        /// </summary>
        /// <param name="start">The starting solution from which to optimize. Used to generated random starting points
        /// for each of the climbers.</param>
        /// <param name="climbers">The number of climbers to "shotgun" into random starting positions within the
        /// solution space. The default is 100.</param>
        /// <param name="state">Optional world state to inform the calculation of cost during the optimization
        /// process.</param>
        /// <param name="fn">An optional callback function that reports the climbers who have completed their
        /// climbs thus far. Note that the callback is made synchronously on the calling thread, so if the 
        /// subscriber intends to carry out extensive operations or update UI controls as a result, a
        /// new background thread should be spawned to handle that.</param>
        /// <returns>The best solution found by any climber.</returns>
        public Optimization Optimize(Optimization start, int climbers = CLIMBERS, object state = null, OptimizationCallback fn = null)
        {
            this.Climbers = climbers;
            return this.Optimize(start, state, fn);
        }

        /// <summary>
        /// Searches a solution space for the optimal solution using a number of randomly placed hill climbers
        /// "shotgunned" into the solution space and then who climb whatever hill they initially land upon. Of 
        /// course, the more climbers that are placed in the solution space to begin with, the better the odds of 
        /// finding the global optimum. If any climber finds a perfect or satisficing solution along the way, the 
        /// method is short-circuited and returns with that solution.
        /// </summary>
        /// <param name="start">The starting solution from which to optimize. Used to generated random starting 
        /// points for each of the climbers.</param>
        /// <param name="state">Optional world state to inform the calculation of cost during the optimization
        /// process.</param>
        /// <param name="fn">An optional callback function that reports the climbers who have completed their
        /// climbs thus far. Note that the callback is made synchronously on the calling thread, so if the 
        /// subscriber intends to carry out extensive operations or update UI controls as a result, a
        /// new background thread should be spawned to handle that.</param>
        /// <returns>The best solution found by any climber.</returns>
        public override Optimization Optimize(Optimization start, object state = null, OptimizationCallback fn = null)
        {
            ulong allMoves = 0uL; // sum of all climbers' moves
            _nMoves = 0uL; // reset

            // see if the starting position happens to satisfice
            if (IsSolution(start, state))
                return start;

            // track the lowest-cost climber so we don't have to find it later
            Optimization sBest = null; // best solution
            double cBest = double.MaxValue; // lowest cost summit

            // loop over climbers and have them each hill climb
            for (int i = 0; i < _climbers; i++)
            {
                // shoot a randomly-placed climber into the solution space
                // and have it climb its hill using our base class hill climber
                Optimization s = base.Optimize(start.GetRandom(), state, null); // one hill climber

                allMoves += _nMoves; // add the moves from each climber to all moves by all climbers
                _nMoves = allMoves; // update here in case we satisfice and short-circuit

                fn?.Invoke(i + 1); // the (i+1)th climber completed its climb

                double cost = s.GetCost(state);
                if (IsSolution(cost)) // perfect or good enough?
                {
                    return s;
                }
                else if (cost < cBest) // track and update the lowest-cost climber
                {
                    cBest = cost; // new lowest cost peak
                    sBest = s; // new lowest cost solution
                }
            }
            return sBest; // best solution found by any climber            
        }

    }
}
