﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Diagnostics;
using WobbrockLib.Extensions;

namespace WobbrockLib.Optimization
{
    /// <summary>
    /// Defines a class capable of executing the Simulated Annealing iterative improvement algorithm
    /// to find the best possible solution within a given solution space.
    /// </summary>
    public class SimulatedAnnealer : Optimizer
    {
        /// <summary>
        /// Default constants.
        /// </summary>
        protected const double AUTO_SET_TEMP = -1.00; // flag for auto-setting start and min temps
        protected const double START_TEMP = 1.00;
        protected const double MIN_TEMP = 1E-5; // 1 / 100,000th of the start temp
        protected const double COOLING_RATE = 0.03;
        protected const int ITERATIONS_PER_TEMP = 100;
        protected const double INITIAL_ACCEPTANCE_PROB = 0.99; // used for auto-setting start temperature

        /// <summary>
        /// Fields.
        /// </summary>
        protected double _startTemp;
        protected double _minTemp;
        protected double _coolingRate;
        protected int _iterPerTemp;

        /// <summary>
        /// Constructs an instance capable of running the Simulated Annealing algorithm. Note
        /// that the default constants for starting temperature and minimum temperature might be 
        /// quite inappropriate depending on the problem space being tackled. Automatic values
        /// for these temperatures can be set with the AutoSetTemperatures method.
        /// </summary>
        /// <param name="startTemp">The starting temperature for the annealer. The default is 1.00.</param>
        /// <param name="minTemp">The minimum temperature for the annealer. The default is 0.00001, i.e., 1E-5.</param>
        /// <param name="coolingRate">The rate at which the annealer cools. The default is 0.03.</param>
        /// <param name="iterationsPerTemp">The number of iterations per temperature setting. The default is 100.</param>
        public SimulatedAnnealer(
            double startTemp = START_TEMP, 
            double minTemp = MIN_TEMP, 
            double coolingRate = COOLING_RATE, 
            int iterationsPerTemp = ITERATIONS_PER_TEMP)
        {
            _startTemp = startTemp;
            _minTemp = minTemp;
            _coolingRate = coolingRate;
            _iterPerTemp = iterationsPerTemp;
        }

        /// <summary>
        /// Gets or sets the starting temperature for this annealing process.
        /// </summary>
        /// <remarks>If setting, the caller might wish instead to have the annealer automatically calculate 
        /// the temperature using AutoSetTemperatures, as the algorithm in general is very sensitive to the 
        /// temperature settings.</remarks>
        public double StartTemperature
        {
            get
            {
                return _startTemp;
            }
            set
            {
                if (value > 0.00)
                    _startTemp = value;
                else
                    throw new ArgumentOutOfRangeException("Argument out of range in SimulatedAnnealer.StartTemperature.");
            }
        }

        /// <summary>
        /// Gets or sets the minimum temperature for this annealing process. Note that this setting must be
        /// less than the StartTemperature property, so consider setting it first before setting this one.
        /// A good rule of thumb is that this temperature should be 1 / 100,000th of the starting temperature,
        /// i.e., the start temperature multiplied by 1E-5.
        /// </summary>
        /// <remarks>If setting, the caller might wish instead to have the annealer automatically calculate 
        /// the temperature using AutoSetTemperatures, as the algorithm in general is very sensitive to the 
        /// temperature settings.</remarks>
        public double MinimumTemperature
        {
            get
            {
                return _minTemp;
            }
            set
            {
                if (0.00 < value && value < _startTemp)
                    _minTemp = value;
                else
                    throw new ArgumentOutOfRangeException("Argument out of range in SimulatedAnnealer.MinimumTemperature.");
            }
        }

        /// <summary>
        /// Automatically calculate and set the starting and minimum temperatures used in the annealing process. 
        /// These are both set together, as they are inter-dependent. The starting temperature is set using a basic 
        /// sampling method adapted from Kirkpatrick et al. (1983). The minimum temperature is set to 1 / 100,000th 
        /// of the start temperature to guarantee a full cooling schedule of acceptance probabilities decreasing from 
        /// ~100% to ~0%.
        /// </summary>
        /// <param name="sln">A solution within the solution space. This solution is used to generate other random 
        /// solutions within the space, which serve as samples for testing the acceptance probability of adopting
        /// a worse move.</param>
        /// <param name="state">Optional world state for use in computing the cost of solutions.</param>
        /// <remarks>The performance of simulated annealing depends a great deal on the starting temperature used and
        /// the minimum temperature paired with it. Ideally, we want the initial acceptance rate to be near ~100%, and 
        /// to decrase from there according to the cooling schedule to reach near ~0% before terminating.</remarks>
        /// <reference>
        /// Kirkpatrick, S., Gelatt, C.D. and Vecchi, M.P. (1983). Optimization by simulated annealing. Science 220 (4598), 
        ///   pp. 671-680. DOI: 10.1126/science.220.4598.671
        /// </reference>
        /// <seealso cref="http://science.sciencemag.org/content/220/4598/671"/>
        public void AutoSetTemperatures(Optimization sln, object state = null)
        {
            // sample the acceptance probability empirically. if the estimated probability is <=1%, then increase 
            // the temperature by an order of magnitude (10x). if not, then continue to double it (2x) until it 
            // reaches our desired initial acceptance probability.
            double temperature = START_TEMP; // seed
            while (true)
            {
                double prob = EstimateAcceptanceProbability(sln, temperature, state);
                if (prob < INITIAL_ACCEPTANCE_PROB)
                    temperature *= (prob <= 0.01) ? 10.0 : 2.0; // 10x or 2x
                else break;
            }
            _startTemp = temperature; // auto-set, yay!

            // now set the minimum temperature in view of the start temperature just set. setting it to be
            // 1 / 100,000th of the just-set start temp should allow a full cooling schedule to be run from
            // ~100% acceptance rate initially down to ~0% acceptance rate at the end.
            _minTemp = _startTemp * MIN_TEMP;
        }

        /// <summary>
        /// Estimate the acceptance probability in this search space for the given temperature by sampling
        /// one hundred random transitions and averaging their acceptance probability, i.e., the probability
        /// that a possible move to a neighbor will, in fact, result in a move.
        /// </summary>
        /// <param name="sln">A solution from the space of possible solutions to be sampled.</param>
        /// <param name="temperature">The temperature to use when estimating the acceptance probability.</param>
        /// <param name="state">Optional world state used to compute the cost of solutions.</param>
        /// <returns>The estimated average acceptance probability for moves within this space using the given
        /// temperature.</returns>
        protected double EstimateAcceptanceProbability(Optimization sln, double temperature, object state = null)
        {
            double psum = 0.00;
            int samples = 100;

            for (int i = 0; i < samples; i++) // average N randomly sampled moves
            {
                Optimization s = sln.GetRandom(); // get random solution
                List<Optimization> moves = s.GetMoves(); // get its neighbors
                int idx = RandomEx.Integer(0, moves.Count - 1); // pick one neighbor at random
                psum += AcceptanceProbability(s.GetCost(state), moves[idx].GetCost(state), temperature);
            }

            return (psum / samples);
        }

        /// <summary>
        /// Gets or sets the cooling rate for this annealing process, usually in [0.01, 0.20]. The rate is subtracted 
        /// from 1.00  and then that quantity is multiplied by the current temperature to produce the next lower 
        /// temperature. The default is 0.03, so each new temperature setting is 97% of the value of the previous 
        /// temperature.
        /// </summary>
        public double CoolingRate
        {
            get
            {
                return _coolingRate;
            }
            set
            {
                if (0.00 < value && value < 1.00)
                    _coolingRate = value;
                else
                    throw new ArgumentOutOfRangeException("Argument out of range in SimulatedAnnealer.CoolingRate.");
            }
        }

        /// <summary>
        /// Gets or sets the number of possible moves at each temperature. Usually between 100 and 1000. If 1, then 
        /// the number of potential moves is just the number of iterations produced by the lowering of the initial 
        /// temperature until it is below the minimum temperature based on the cooling rate. The default value for 
        /// this property is 100.
        /// </summary>
        /// <see cref="http://katrinaeg.com/simulated-annealing.html"/> 
        public int IterationsPerTemperature
        {
            get
            {
                return _iterPerTemp;
            }
            set
            {
                if (value >= 1)
                    _iterPerTemp = value;
                else
                    throw new ArgumentOutOfRangeException("Argument out of range in SimulatedAnnealer.IterationsPerTemperature.");
            }
        }

        /// <summary>
        /// A convenience method that sets the start temperature, minimum temperature, cooling rate, and iterations
        /// at each temperature prior to starting the optimization search using simulated annealing.
        /// </summary>
        /// <param name="start">A starting solution from which to begin to search the solution space.</param>
        /// <param name="startTemp">The starting temperature for this annealing process. To automatically calculate
        /// the start temperature, pass -1 for this value and for the parameter <i>minTemp</i>.</param>
        /// <param name="minTemp">The minimum temperature for this annealing process. To automatically calculate
        /// the minimum temperature, pass -1 for this value and for the parameter <i>startTemp</i>.</param>
        /// <param name="coolingRate">The cooling rate for this annealing process, usually [0.01 to 0.20]. The rate
        /// is subtracted from 1.00 and then that quantity is multiplied by the current temperature to produce the
        /// new lower temperature. So a cooling rate of 0.03 results in each new temperature being 97% of the prior
        /// temperature.</param>
        /// <param name="iterPerTemp">For each temperature setting, the number of attempts at moving that will be made. 
        /// Usually, a value between 100 and 1000, or simply 1 if the only iterations desired are driven by the cooling 
        /// schedule, i.e., at each temperature, there is a single chance to make a move. A move is made when, at any time,
        /// either (a) the move is to a better solution (i.e., one with lower cost), or (b) the move is to a worse solution 
        /// (i.e., one with higher cost) but, by chance, an acceptance probability, which declines with progressive cooling, 
        /// was met.</param>
        /// <param name="state">Optional world state for helping to compute the cost of a solution.</param>
        /// <param name="fn">An optional callback function that reports the current temperature during the cooling process
        /// of the simulated annealer. The current temperature is not reported as a raw temperature value but as the percentage
        /// of the start temperature, so decreasing from 100% to 0%. Note that the callback is made synchronously on this method's 
        /// calling thread, so if the subscriber intends to carry out extensive operations or update UI controls as a result, a new 
        /// thread should be spawned to handle that.</param>
        /// <returns>The final move made, a local or possibly global optimum.</returns>
        /// <see cref="http://katrinaeg.com/simulated-annealing.html"/>
        /// <see cref="http://www.theprojectspot.com/tutorial-post/simulated-annealing-algorithm-for-beginners/6"/> 
        /// <see cref="https://link.springer.com/article/10.1023/B:COAP.0000044187.23143.bd"/> 
        public Optimization Optimize(
            Optimization start,
            double startTemp = AUTO_SET_TEMP, 
            double minTemp = AUTO_SET_TEMP,  
            double coolingRate = COOLING_RATE, 
            int iterPerTemp = ITERATIONS_PER_TEMP,
            object state = null,
            OptimizationCallback fn = null)
        {
            if (startTemp < 0.0 && minTemp < 0.0)
            {
                this.AutoSetTemperatures(start, state);
            }
            else if (startTemp > 0.0 && minTemp > 0.0)
            {
                this.StartTemperature = startTemp;
                this.MinimumTemperature = minTemp;
            }
            else throw new ArgumentException("Arguments 'startTemp' and 'minTemp' must carry the same sign in SimulatedAnnealer.Optimize.");

            this.CoolingRate = coolingRate;
            this.IterationsPerTemperature = iterPerTemp;

            return this.Optimize(start, state, fn);
        }

        /// <summary>
        /// Implements simulated annealing as a refinement to hill climbing for iterative improvement optimization search. 
        /// </summary>
        /// <param name="start">A starting solution from which to begin to search the solution space.</param>
        /// <param name="state">Optional world state for helping to compute the cost of a solution.</param>
        /// <param name="fn">An optional callback function that reports the current temperature during the cooling process
        /// of the simulated annealer. The current temperature is not reported as a raw temperature value but as the percentage
        /// of the start temperature, so decreasing from 100% to 0%. Note that the callback is made synchronously on this method's 
        /// calling thread, so if the subscriber intends to carry out extensive operations or update UI controls as a result, a new 
        /// thread should be spawned to handle that.</param>
        /// <returns>The final move made, a local or possibly global optimum.</returns>
        /// <remarks>The first link below is nicely written but has the equation wrong. It is missing the negation in the
        /// numerator of the exponent.</remarks>
        /// <see cref="http://katrinaeg.com/simulated-annealing.html"/>
        /// <see cref="http://www.theprojectspot.com/tutorial-post/simulated-annealing-algorithm-for-beginners/6"/> 
        /// <see cref="https://link.springer.com/article/10.1023/B:COAP.0000044187.23143.bd"/> 
        public override Optimization Optimize(Optimization start, object state = null, OptimizationCallback fn = null)
        {
            _nMoves = 0uL; // reset

            // see if the starting position happens to satisfice
            if (IsSolution(start, state))
                return start;

            double thisCost = start.GetCost(state); // cost of the current solution

            double temperature = _startTemp;
            fn?.Invoke(100.0); // 100%

            while (temperature > _minTemp)
            {
                for (int i = 0; i < _iterPerTemp; i++) // no. of move opportunities before reducing temp
                {
                    List<Optimization> moves = start.GetMoves(); // get neighbors, i.e., possible moves
                    int idx = RandomEx.Integer(0, moves.Count - 1); // choose a random neighbor

                    // calculate the acceptance probability of making a move to this neighbor. the probability
                    // will always be 100% that we make this move if the cost of the new move is less than the
                    // cost of the current solution, i.e., if the new move is better than where we are.
                    // if the new move is worse, then we still might move according to the probability
                    // generated by the cooling schedule, i.e., the current temperature.
                    double prob = AcceptanceProbability(thisCost, moves[idx].GetCost(state), temperature);
                    double rand = RandomEx.Double(0.00, 1.00);
                    if (prob >= rand)
                    {
                        start = moves[idx]; // this move becomes next round's start position
                        thisCost = start.GetCost(state); // cost of the new solution
                        _nMoves++; // made a move

                        if (IsSolution(thisCost)) // do we have a solution?
                            return start;
                    }
                }
                temperature *= (1.00 - _coolingRate);
                fn?.Invoke(temperature / _startTemp * 100.0); // report as [0, 100] percentage of initial temperature
            }
            return start;
        }

        /// <summary>
        /// Calculate the acceptance probability based on the old cost, new cost, and current temperature.
        /// </summary>
        /// <param name="oldCost">The cost of the current solution.</param>
        /// <param name="newCost">The cost of the possible new move.</param>
        /// <param name="temperature">The current temperature, as per the simulated annealing metaphor.</param>
        /// <returns>The probability in [0.00, 1.00] of accepting the new move.</returns>
        /// <remarks>The link below is nicely written but has the equation wrong. It is missing the negation 
        /// in the numerator of the exponent.</remarks>
        /// <see cref="http://katrinaeg.com/simulated-annealing.html"/>
        private double AcceptanceProbability(double oldCost, double newCost, double temperature)
        {
            double prob;
            if (newCost < oldCost)
                prob = 1.00;
            else
                prob = Math.Exp(-(newCost - oldCost) / temperature); // e^(-(new - old) / T)
            return prob;
        }
    }
}
