﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace WobbrockLib.Optimization
{
    public partial class SimulatedAnnealingForm : Form
    {
        /// <summary>
        /// Options class for this form.
        /// </summary>
        public class Options
        {
            public bool AutoSetTemperatures { get; set; }
            public double StartTemperature { get; set; }
            public double MinimumTemperature { get; set; }
            public double CoolingRate { get; set; }
            public int IterationsPerTemperature { get; set; }

            /// <summary>
            /// Construct a new set of options with their default values.
            /// </summary>
            public Options() // defaults
            {
                this.AutoSetTemperatures = true;
                this.StartTemperature = 1.00;
                this.MinimumTemperature = 1E-5;
                this.CoolingRate = 0.03;
                this.IterationsPerTemperature = 100;
            }

            /// <summary>
            /// Copy the current options and return a deep copy.
            /// </summary>
            /// <returns></returns>
            public Options Copy()
            {
                Options o = new Options();
                o.AutoSetTemperatures = this.AutoSetTemperatures;
                o.StartTemperature = this.StartTemperature;
                o.MinimumTemperature = this.MinimumTemperature;
                o.CoolingRate = this.CoolingRate;
                o.IterationsPerTemperature = this.IterationsPerTemperature;
                return o;
            }
        }

        private SimulatedAnnealingForm.Options _o;

        /// <summary>
        /// Constructor with the default options.
        /// </summary>
        public SimulatedAnnealingForm()
        {
            InitializeComponent();
            _o = new SimulatedAnnealingForm.Options();
        }

        /// <summary>
        /// Construct the form with the options given.
        /// </summary>
        /// <param name="o"></param>
        public SimulatedAnnealingForm(SimulatedAnnealingForm.Options o)
        {
            InitializeComponent();
            _o = o.Copy();
        }

        /// <summary>
        /// Loads the form and sets the values of the widgets.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        /// <remarks>This is the "copy in" procedure.</remarks>
        private void AnnealingForm_Load(object sender, EventArgs e)
        {
            chkAutoSetTemps.Checked = _o.AutoSetTemperatures;
            numStartTemp.Value = (decimal) _o.StartTemperature;
            numMinTemp.Value = (decimal) _o.MinimumTemperature;
            numCooling.Value = (decimal) (_o.CoolingRate * 100.0);
            numIterationsPerTemp.Value = _o.IterationsPerTemperature;
            chkAutoSetTemps_CheckedChanged(this, EventArgs.Empty); // handle graying out
        }

        /// <summary>
        /// Gets or sets the options from the controls on the form.
        /// </summary>
        /// <remarks>This is the "copy out" procedure.</remarks>
        public SimulatedAnnealingForm.Options FormOptions
        {
            get
            {
                SimulatedAnnealingForm.Options o = new SimulatedAnnealingForm.Options();
                o.AutoSetTemperatures = chkAutoSetTemps.Checked;
                o.StartTemperature = (double) numStartTemp.Value;
                o.MinimumTemperature = (double) numMinTemp.Value;
                o.CoolingRate = (double) numCooling.Value / 100.0;
                o.IterationsPerTemperature = (int) numIterationsPerTemp.Value;
                return o;
            }
        }

        /// <summary>
        /// Enables or disables the exception percentage widgets.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void chkAutoSetTemps_CheckedChanged(object sender, EventArgs e)
        {
            lblStartTemp.Enabled = !chkAutoSetTemps.Checked;
            numStartTemp.Enabled = !chkAutoSetTemps.Checked;
            lblStartTempRange.Enabled = !chkAutoSetTemps.Checked;
            lblMinTemp.Enabled = !chkAutoSetTemps.Checked;
            numMinTemp.Enabled = !chkAutoSetTemps.Checked;
            lblMinTempRange.Enabled = !chkAutoSetTemps.Checked;
            cmdCalcMinTemp.Enabled = !chkAutoSetTemps.Checked;
        }

        /// <summary>
        /// Calculates 1/100,000th of the start temperature for setting the minimum temperature.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void cmdCalcMinTemp_Click(object sender, EventArgs e)
        {
            numMinTemp.Value = numStartTemp.Value * 1E-5m;
        }

    }
}
