﻿using System;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Threading;
using System.Windows.Forms;
using WobbrockLib.Devices;
using WobbrockLib.Windows;

namespace Test
{
	/// <summary>
	/// <para>
	/// A form that creates a semi-transparent circular area cursor, but with the regular
	/// system point cursor drawn at its center. Although the point cursor is always on
	/// top of the area cursor, and therefore its events act on it, we demonstrate how to
	/// send those events "through" ourselves, synthesizing them for the UI below, whatever
	/// it is. This app supports all mouse down, up, click, and double click events, as well
	/// as dragging and dropping.
	/// </para>
	/// </summary>
	public partial class ClickThruForm : Form
	{
		#region Fields

		private const int ClickTime = 250; // a Down-Up pair must occur within this time to issue a Click
		private const int SleepTime = 110; // the time after a Click or Double Click that we sleep before regaining "input opacity"

		private RawInputMouse _rim; // raw input mouse feed for synchronizing our area cursor with the system cursor
		private RawInputKeyboard _rik; // raw input keyboard feed for the Escape key to exit our app
		private CursorWrapper _pointer; // encapsulated system pointer so we can easily position and freeze it

		private bool _mouseDown; // true when a mouse button is down
		private bool _synthesized; // true for events we create versus those that actually come from the user

		#endregion

		#region Start/Stop

		/// <summary>
		/// Constructor.
		/// </summary>
		public ClickThruForm()
		{
			InitializeComponent();
			this.SetStyle(ControlStyles.OptimizedDoubleBuffer | 
				ControlStyles.AllPaintingInWmPaint | 
				ControlStyles.UserPaint, 
				true);
			
			_rim = new RawInputMouse("ClickThruForm Raw Input Mouse");
			_rim.RawMouseMove += new EventHandler<MouseEventArgs>(RawInputMouse_MouseMove);
			_rim.RawMouseUp += new EventHandler<MouseEventArgs>(RawInputMouse_MouseUp);

			_rik = new RawInputKeyboard("ClickThruForm Raw Input Keyboard");
			_rik.RawKeyDown += new EventHandler<KeyEventArgs>(RawInputKeyboard_KeyDown);

			_pointer = new CursorWrapper();
		}

		/// <summary>
		/// Main Windows message pump for this application. How we process raw input events.
		/// </summary>
		/// <param name="m"></param>
		protected override void WndProc(ref Message m)
		{
			if (m.Msg == (int) Win32.WM.INPUT)
			{
				bool processed = _rim.ProcessRawInput(m.LParam); // mouse
				if (!processed)
					processed = _rik.ProcessRawInput(m.LParam); // keyboard
			}
			base.WndProc(ref m);
		}

		/// <summary>
		/// Form's load event. We set some styles, give ourselves a circular shape, and register for 
		/// the raw input feeds.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void ClickThruForm_Load(object sender, EventArgs e)
		{
			// set some extended window styles (optional)
			uint exStyle = Win32.GetWindowLong(this.Handle, (int) Win32.GWL.EXSTYLE);
			uint retval = Win32.SetWindowLong(this.Handle, (int) Win32.GWL.EXSTYLE, exStyle | (uint) (Win32.WS_EX.NOACTIVATE | Win32.WS_EX.TOOLWINDOW));
			
			// make ourselves circular
			GraphicsPath path = new GraphicsPath();
			path.AddEllipse(0, 0, this.Width, this.Height);
			this.Region = new Region(path);
			path.Dispose();

			// initialize our form's position based on the system cursor
			this.Location = new Point(_pointer.Position.X - this.Width / 2, _pointer.Position.Y - this.Height / 2);

			// register for the raw input feeds
			_rim.Register(this.Handle, true);
			_rik.Register(this.Handle, true);
		}

		/// <summary>
		/// Form's closed event. We stop all timers and unregister for the raw input feeds.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void ClickThruForm_FormClosed(object sender, FormClosedEventArgs e)
		{
			tmrDown.Stop();
			tmrClick.Stop();

			_rim.Unregister();
			_rik.Unregister();
		}

		/// <summary>
		/// Form's paint event. Draw a crosshairs at the center of our form.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void ClickThruForm_Paint(object sender, PaintEventArgs e)
		{
			//Point c = new Point(this.Width / 2, this.Height / 2);
			//e.Graphics.DrawLine(Pens.Black, c.X - 15, c.Y, c.X + 15, c.Y);
			//e.Graphics.DrawLine(Pens.Black, c.X, c.Y - 15, c.X, c.Y + 15);
		}

		/// <summary>
		/// Raw keyboard event. If the Escape key is ever pressed, we exit our application.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void RawInputKeyboard_KeyDown(object sender, KeyEventArgs e)
		{
			if (e.KeyCode == Keys.Escape)
			{
				this.Close();
			}
		}

		#endregion

		#region Mouse Events

		/// <summary>
		/// Raw mouse move. Keep our form's position centered on the system cursor, whose 
		/// screen position is delievered in 'e'.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void RawInputMouse_MouseMove(object sender, MouseEventArgs e)
		{
			// keep ourselves position synchronized with the system cursor
			this.Location = new Point(e.X - this.Width / 2, e.Y - this.Height / 2);
		}

		/// <summary>
		/// Handles when the mouse goes down on our form. This could be the beginning of
		/// a drag, in which case the mouse stays down, a click, or a double click. If we 
		/// were to synthesize a Down immediately through ourselves, we couldn't ever
		/// get a Click or Double Click, because we'd be "input transparent" for the 
		/// following Up. So, we require a short dwell before synthesizing the Down action 
		/// through ourselves. That way, if an Up happens first, we can actually form clicks 
		/// and double clicks.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void ClickThruForm_MouseDown(object sender, MouseEventArgs e)
		{
			if (_synthesized)
				return;

			tmrClick.Stop(); // we might be Down here a 2nd time for a Double Click

			Console.WriteLine("ClickThruForm_MouseDown; _mouseDown={0}", _mouseDown);
			_mouseDown = true;

			// set the down point's screen location for use in starting a drag
			Point pt = PointToScreen(e.Location);
			_pointer.Freeze(pt);

			// now start the down timer; maybe an Up (i.e., Click) will occur in the meantime?
			tmrDown.Interval = ClickTime;
			tmrDown.Start(e);
		}

		/// <summary>
		/// When the mouse has been held down this long without going up, then this timer goes off
		/// and we can synthesize a Down event through ourselves. Note that we can programmatically
		/// reach this function also when the mouse moves far enough so as to start a drag. Either
		/// way, we synthesize a Down event through us. We must remain "input transparent" while the
		/// mouse is down; this "input transparency" is reverted when the mouse comes up (and after 
		/// a short delay to let drag-and-drop operations complete).
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void tmrDown_Tick(object sender, object e)
		{
			tmrDown.Stop();

			Console.WriteLine("tmrDown_Tick; _mouseDown={0}", _mouseDown);

			MouseEventArgs m = (MouseEventArgs) e;
			Synthesize(MouseEx.ButtonAction.Up, m.Button); // release the physical Down on us

			Point pt = PointToScreen(m.Location);
			_pointer.Freeze(pt); // re-freeze after Up, since the Up will break the freeze

			SetInputTransparent(true);
			bool success = WindowEx.FocusWindowAt(pt, WindowEx.FocusFns.Foreground, false);
			Synthesize(MouseEx.ButtonAction.Down, m.Button);
			// NOTE: no need for Thread.Sleep(SleepTime) since we're not going "input 
			// opaque" immediately here like in the Click and Double Click cases! the
			// crucial Thread.Sleep call correspondingly is in RawMouseInput_MouseUp.

			_pointer.Unfreeze();
		}

		/// <summary>
		/// Handles when the mouse comes up anywhere, including when we're "input transparent," 
		/// and thus, cannot get the event on our form. We use this event to unset our down
		/// and dragging flags, because we can't possibly be down and/or dragging at this point.
		/// We also set ourselves to no longer be "input transparent."
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void RawInputMouse_MouseUp(object sender, MouseEventArgs e)
		{
			if (_synthesized)
				return;

			tmrDown.Stop();
			Console.WriteLine("RawInputMouse_MouseUp; _mouseDown={0}", _mouseDown);

			_mouseDown = false;
			Thread.Sleep(SleepTime); // crucial! must-do so drag-and-drop operations complete correctly
			SetInputTransparent(false);
		}

		/// <summary>
		/// When a click occurs, we can't immediatley synthesize a Click through ourselves, 
		/// because if we do, it would require us to become "input transparent," and therefore 
		/// unable to catch what might be the second half of a Double Click. So, we set a timer
		/// to go off 1 ms after the system's Double Click time. If the timer goes off, we can 
		/// issue this Click. If a Double Click happens before then, we stop the timer and process 
		/// it.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		/// <remarks>
		/// Note that the order of events for a Click is:
		/// 
		///		ClickThruForm_MouseDown			// starts Down timer
		///		RawInputMouse_MouseUp			// stops Down timer
		///		ClickThruForm_MouseClick		// starts Click timer
		///		tmrClick_Tick					// stops Click timer, issues Click!
		/// </remarks>
		private void ClickThruForm_MouseClick(object sender, MouseEventArgs e)
		{
			if (_synthesized)
				return;

			Console.WriteLine("ClickThruForm_MouseClick; _mouseDown={0}", _mouseDown);

			// freeze the cursor at the click position so that the synthesized Click occurs at the right place
			Point pt = PointToScreen(e.Location);
			_pointer.Freeze(pt);

			// now start the Click timer; maybe a Double Click occurs in the meantime?
			tmrClick.Interval = SystemInformation.DoubleClickTime + 1;
			tmrClick.Start(e);
		}

		/// <summary>
		/// We can't handle a click directly in ClickThruForm_MouseClick because if we
		/// do, we can't catch a double-click event because of our transparency. So 
		/// there we set a timer, and if a double-click doesn't happen in time, then
		/// we know we can execute the click.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event, which are the click's 
		/// MouseEventArgs.</param>
		private void tmrClick_Tick(object sender, object e)
		{
			tmrClick.Stop();
			Console.WriteLine("tmrClick_Tick; _mouseDown={0}", _mouseDown);
			MouseEventArgs m = (MouseEventArgs) e;

			SetInputTransparent(true);
			Point pt = PointToScreen(m.Location);
			bool success = WindowEx.FocusWindowAt(pt, WindowEx.FocusFns.Foreground, false);
			Synthesize(MouseEx.ButtonAction.Click, m.Button);
			Thread.Sleep(SleepTime); // Crucial! must-do before turning off input transparency
			SetInputTransparent(false);

			_pointer.Unfreeze();
		}

		/// <summary>
		/// We can catch a double-click as usual, and synthesize one to go "through" us.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		/// <remarks>
		/// Note that the order of events for a Double Click is:
		/// 
		///		ClickThruForm_MouseDown			// starts Down timer
		///		RawInputMouse_MouseUp			// stops Down timer
		///		ClickThruForm_MouseClick		// starts Click timer
		///		ClickThruForm_MouseDown			// stops Click timer, starts Down timer
		///		RawInputMouse_MouseUp			// stops Down timer
		///		ClickThruForm_MouseDoubleClick	// issues Double Click!
		/// </remarks>
		private void ClickThruForm_MouseDoubleClick(object sender, MouseEventArgs e)
		{
			if (_synthesized)
				return;
			
			Console.WriteLine("ClickThruForm_MouseDoubleClick; _mouseDown={0}", _mouseDown);
			_mouseDown = false;

			SetInputTransparent(true);
			Point pt = PointToScreen(e.Location);
			bool success = WindowEx.FocusWindowAt(pt, WindowEx.FocusFns.Foreground, false);
			Synthesize(MouseEx.ButtonAction.DoubleClick, e.Button);
			Thread.Sleep(SleepTime); // Crucial! must-do before turning off input transparency
			SetInputTransparent(false);

			_pointer.Unfreeze();
		}

		#endregion

		#region Helpers

		/// <summary>
		/// Sets our form to be "input transparent" or not. We reflect our current setting in
		/// the color of the form, with blue being "input opaque" and light gray being 
		/// "input transparent."
		/// </summary>
		/// <param name="transparent">If true, we will be set to "input transparent," unless we 
		/// already are. If false, we will be set to "input opaque," unless we already are.</param>
		private void SetInputTransparent(bool transparent)
		{
			if (WindowEx.SetInputTransparent(this.Handle, transparent)) // if succeeds
			{
				this.BackColor = transparent ? Color.LightGray : Color.DodgerBlue;
			}
		}

		/// <summary>
		/// Synthesize a given mouse button action. Because we want our event handlers to only 
		/// handle events that come from the user, rather than are synthesized, we set flags
		/// to prevent event re-entry. Also, because we want the event we synthesize to occur right
		/// upon the TakeAction call, we must include Application.DoEvents to pump the synthesized
		/// event through the message queue.
		/// </summary>
		/// <param name="action">The action to take.</param>
		/// <param name="button">The button to take it with.</param>
		private void Synthesize(MouseEx.ButtonAction action, MouseButtons button)
		{
			_synthesized = true; // protect against event re-entry for synthesized events
			MouseEx.TakeAction(action, button); // the magic!
			Application.DoEvents(); // causes the action's events to be processed
			_synthesized = false; // only after they're processed, can we unset the flag
		}

		#endregion

		#region Menu Detection

		/// <summary>
		/// 
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void tmrMenu_Tick(object sender, object e)
		{
			tmrDown.Stop();

			// check for old-style pre-.NET 4.0 menus
			if (Menu32Ex.IsMenuOpen())
			{
				SomeWindow_MenuOpened(null, (IntPtr) e);
			}
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void SomeWindow_MenuOpened(object sender, IntPtr e)
		{
			Console.WriteLine("Menu OPENED on hWnd={0}", e);
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void SomeWindow_MenuClosed(object sender, IntPtr e)
		{
			Console.WriteLine("Menu CLOSED on hWnd={0}", e);
		}

		#endregion
	}
}
