﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Diagnostics;
using WobbrockLib.Extensions;

namespace WobbrockLib.Types
{
    /// <summary>
    /// A structure representing a fraction with a numerator and denominator of integers.
    /// </summary>
    public struct Fraction
    {
        public static readonly Fraction Empty;
        private int _numerator;
        private int _denominator;

        /// <summary>
        /// Constructor for the Fraction structure. Fractions comprise a numerator and a denominator.
        /// </summary>
        /// <param name="numerator">The numerator for this fraction.</param>
        /// <param name="denominator">The non-zero denominator for this fraction.</param>
        public Fraction(int numerator, int denominator)
        {
            _numerator = numerator;
            if (denominator != 0)
                _denominator = denominator;
            else
                throw new ArgumentException("Cannot set denominator to zero in Fraction constructor.");
        }

        /// <summary>
        /// Copy constructor for the Fraction structure.
        /// </summary>
        /// <param name="f">A Fraction to copy.</param>
        public Fraction(Fraction f)
        {
            _numerator = f._numerator;
            _denominator = f._denominator;
        }

        /// <summary>
        /// Creates a printable string for this fraction. Fraction strings are printed simply as "X/Y", 
        /// where "X" and "Y" are integers and "Y" is non-zero.
        /// </summary>
        /// <returns>A printable string for this fraction.</returns>
        public override string ToString()
        {
            return String.Format("{0}/{1}", _numerator, _denominator);
        }

        /// <summary>
        /// Parses the given string to create a fraction.
        /// </summary>
        /// <param name="s">The given string to parse. It should be in the format of X/Y, where "X" and "Y"
        /// are integers and "Y" is non-zero.</param>
        /// <returns>If successful, a fraction structure representing the parsed string.</returns>
        public static Fraction FromString(string s)
        {
            Fraction f = Fraction.Empty;
            try
            {
                int pos = s.IndexOf('/');
                int num = int.Parse(s.Substring(0, pos));
                int denom = int.Parse(s.Substring(pos + 1, s.Length - (pos + 1)));
                f = new Fraction(num, denom);
            }
            catch { }
            return f;
        }

        /// <summary>
        /// Gets or sets the numerator for this fraction.
        /// </summary>
        public int Numerator
        {
            get { return _numerator; }
            set { _numerator = value; }
        }

        /// <summary>
        /// Gets or sets the denominator for this fraction. Note that denominators cannot be set to zero.
        /// </summary>
        public int Denominator
        {
            get
            {
                return _denominator;
            }
            set
            {
                if (value != 0)
                    _denominator = value;
                else
                    throw new ArgumentException("Cannot set denominator to zero in Fraction.Denominator.");
            }
        }

        /// <summary>
        /// Gets the double-precision floating-point value of this fraction.
        /// </summary>
        public double Value
        {
            get { return (double) _numerator / _denominator; }
        }

        /// <summary>
        /// Calculates the greatest common denominator (gcd) of this fraction's numerator and denominator.
        /// </summary>
        /// <returns>The greatest common denominator (gcd) of this fraction.</returns>
        /// <seealso cref="https://en.wikipedia.org/wiki/Greatest_common_divisor#Binary_method"/> 
        public int GCD()
        {
            int a = _numerator;
            int b = _denominator;
            int d = 0;

            while ((a % 2 == 0) && (b % 2 == 0))
            {
                a /= 2;
                b /= 2;
                d++;
            }
            while (a != b)
            {
                if (a % 2 == 0)
                    a /= 2;
                else if (b % 2 == 0)
                    b /= 2;
                else if (a > b)
                    a = (a - b) / 2;
                else
                    b = (b - a) / 2;
            }
            return (a * (int) Math.Pow(2, d));
        }

        /// <summary>
        /// Attempts to reduce this fraction to its simplest terms using the greatest common denominator. 
        /// For example, 12/48 reduces to 1/4 but no simpler.
        /// </summary>
        /// <returns>True if the fraction was reducible; false otherwise.</returns>
        public bool Reduce()
        {
            int gcd = this.GCD();
            if (gcd > 1)
            {
                _numerator /= gcd;
                _denominator /= gcd;
                return true;
            }
            return false;
        }

        /// <summary>
        /// Equals operator for fractions. The test is independent of the specific numerator and
        /// denominator values, and tests whether the proportions identified by the fractions
        /// are equal.
        /// </summary>
        /// <remarks>For a test that considers the specific numerator and denominator values, use
        /// the Equals method.</remarks>
        /// <param name="f1">The left-hand side of the operator.</param>
        /// <param name="f2">The right- side of the operator.</param>
        /// <returns>True if the fractions represent the same proportions; false otherwise.</returns>
        public static bool operator ==(Fraction f1, Fraction f2)
        {
            double d1 = f1.Value;
            double d2 = f2.Value;
            return (d1 == d2);
        }

        /// <summary>
        /// Not-equals operator for fractions. The test is independent of the specific numerator and
        /// denominator values, and tests whether the proportions identified by the fractions
        /// are unequal.
        /// </summary>
        /// <remarks>For a test that considers the specific numerator and denominator values, use
        /// the Equals method.</remarks>
        /// <param name="f1">The left-hand side of the operator.</param>
        /// <param name="f2">The right- side of the operator.</param>
        /// <returns>True if the fractions represent different proportions; false otherwise.</returns>
        public static bool operator !=(Fraction f1, Fraction f2)
        {
            double d1 = f1.Value;
            double d2 = f2.Value;
            return (d1 != d2);
        }

        /// <summary>
        /// Tests whether the given object is equal to this fraction. Unlike the "==" operator, this
        /// method requires the specific numerator and denominator values to agree, not just represent
        /// the same proportion.
        /// </summary>
        /// <param name="obj">The object to test.</param>
        /// <returns>True if the given object is a fraction equal to this fraction; false otherwise.</returns>
        public override bool Equals(object obj)
        {
            if (obj is Fraction)
            {
                Fraction f = (Fraction) obj;
                return (this.Numerator == f.Numerator && this.Denominator == f.Denominator);
            }
            return false;
        }

        /// <summary>
        /// Gets a hashcode for this fraction.
        /// </summary>
        /// <returns>A hashcode for this fraction.</returns>
        public override int GetHashCode()
        {
            return this.ToString().GetHashCode();
        }
    }
}
