﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Drawing.Drawing2D;
using WobbrockLib.Extensions;

namespace WobbrockLib.Types
{
    /// <summary>
    /// A structure representing a triangle with double-precision floating-point vertices.
    /// </summary>
    public struct TriangleR
    {
        public static readonly TriangleR Empty;
        private PointR _v1;
        private PointR _v2;
        private PointR _v3;

        /// <summary>
        /// Constructs a triangle given each vertex.
        /// </summary>
        /// <param name="v1">The first vertex of this triangle.</param>
        /// <param name="v2">The second vertex of this triangle.</param>
        /// <param name="v3">The third vertex of this triangle.</param>
        public TriangleR(PointR v1, PointR v2, PointR v3)
        {
            _v1 = v1;
            _v2 = v2;
            _v3 = v3;
        }

        /// <summary>
        /// Constructs a triangle based on each of the values of its vertices.
        /// </summary>
        /// <param name="v1x">The x-coordinate of the first vertex.</param>
        /// <param name="v1y">The y-coordinate of the first vertex.</param>
        /// <param name="v2x">The x-coordinate of the second vertex.</param>
        /// <param name="v2y">The y-coordinate of the second vertex.</param>
        /// <param name="v3x">The x-coordinate of the third vertex.</param>
        /// <param name="v3y">The y-coordinate of the third vertex.</param>
        public TriangleR(double v1x, double v1y, double v2x, double v2y, double v3x, double v3y)
        {
            _v1 = new PointR(v1x, v1y);
            _v2 = new PointR(v2x, v2y);
            _v3 = new PointR(v3x, v3y);
        }

        /// <summary>
        /// Constructs an equilateral triangle whose base (v1, v3) is centered on the given point 
        /// and whose top (v2) is up the -y axis from the given point.
        /// </summary>
        /// <param name="baseCenter">The centerpoint of the base of the equilateral triangle.</param>
        /// <param name="sideLength">The length of a side of this equilateral triangle.</param>
        public TriangleR(PointR baseCenter, double sideLength)
            : this(baseCenter, sideLength, Math.Sqrt((sideLength * sideLength) - ((0.5 * sideLength) * (0.5 * sideLength))))
        {
            // do nothing
        }

        /// <summary>
        /// Constructs an isoceles triangle whose base (v1, v3) is centered on the given point and 
        /// whose top (v2) is 'height' units up from the given point.
        /// </summary>
        /// <param name="baseCenter">The centerpoint of the base of the isoceles triangle.</param>
        /// <param name="baseLength">The length of the base of the isoceles triangle.</param>
        /// <param name="height">The height of the isoceles triangle.</param>
        public TriangleR(PointR baseCenter, double baseLength, double height)
        {
            _v1 = new PointR(baseCenter.X - baseLength / 2.0, baseCenter.Y);
            _v2 = new PointR(baseCenter.X, baseCenter.Y - height);
            _v3 = new PointR(baseCenter.X + baseLength / 2.0, baseCenter.Y);
        }

        /// <summary>
        /// Copy constructor for this triangle.
        /// </summary>
        /// <param name="t">The triangle whose values to copy.</param>
        public TriangleR(TriangleR t)
        {
            _v1 = t._v1;
            _v2 = t._v2;
            _v3 = t._v3;
        }

        /// <summary>
        /// Gets a string representation of this triangle.
        /// </summary>
        /// <returns>A string representation of this triangle.</returns>
        public override string ToString()
        {
            return String.Format("{{V1=({0},{1}), V2=({2},{3}), V3=({4},{5})}}", _v1.X, _v1.Y, _v2.X, _v2.Y, _v3.X, _v3.Y);
        }

        /// <summary>
        /// Creates a triangle from its string representation.
        /// </summary>
        /// <param name="s">A string representation of a triangle.</param>
        /// <returns>The TriangleR representing the data from the string.</returns>
        public static TriangleR FromString(string s)
        {
            TriangleR t = TriangleR.Empty;
            PointR v1 = PointR.Empty;
            PointR v2 = PointR.Empty;
            PointR v3 = PointR.Empty;

            try
            {
                int pos = s.IndexOf("V1"); // "V1"
                int comma = s.IndexOf(',', pos + 4);
                v1.X = double.Parse(s.Substring(pos + 4, comma - (pos + 4)));

                pos = comma + 1;
                comma = s.IndexOf(')', pos);
                v1.Y = double.Parse(s.Substring(pos, comma - pos));

                pos = s.IndexOf("V2"); // "V2"
                comma = s.IndexOf(',', pos + 4);
                v2.X = double.Parse(s.Substring(pos + 4, comma - (pos + 4)));

                pos = comma + 1;
                comma = s.IndexOf(')', pos);
                v2.Y = double.Parse(s.Substring(pos, comma - pos));

                pos = s.IndexOf("V3"); // "V3"
                comma = s.IndexOf(',', pos + 4);
                v3.X = double.Parse(s.Substring(pos + 4, comma - (pos + 4)));

                pos = comma + 1;
                comma = s.IndexOf(')', pos);
                v3.Y = double.Parse(s.Substring(pos, comma - pos));

                t = new TriangleR(v1, v2, v3);
            }
            catch { }
            return t;
        }

        /// <summary>
        /// Gets or sets the first vertex of this triangle.
        /// </summary>
        public PointR V1
        {
            get { return _v1; }
            set { _v1 = value; }
        }

        /// <summary>
        /// Gets or sets the second vertex of this triangle.
        /// </summary>
        public PointR V2
        {
            get { return _v2; }
            set { _v2 = value; }
        }

        /// <summary>
        /// Gets or sets the third vertex of this triangle.
        /// </summary>
        public PointR V3
        {
            get { return _v3; }
            set { _v3 = value; }
        }

        /// <summary>
        /// Gets the average x and y coordinates of this triangle's three vertices.
        /// </summary>
        public PointR Centroid
        {
            get
            {
                return new PointR(
                    (_v1.X + _v2.X + _v3.X) / 3.0,
                    (_v1.Y + _v2.Y + _v3.Y) / 3.0
                    );
            }
        }

        /// <summary>
        /// Gets the leftmost x-value of this triangle.
        /// </summary>
        public double Left
        {
            get { return Math.Min(Math.Min(_v1.X, _v2.X), _v3.X); }
        }

        /// <summary>
        /// Gets the topmost y-value of this triangle.
        /// </summary>
        public double Top
        {
            get { return Math.Min(Math.Min(_v1.Y, _v2.Y), _v3.Y); }
        }

        /// <summary>
        /// Gets the rightmost x-value of this triangle.
        /// </summary>
        public double Right
        {
            get { return Math.Max(Math.Max(_v1.X, _v2.X), _v3.X); }
        }

        /// <summary>
        /// Gets the bottommost y-value of this triangle.
        /// </summary>
        public double Bottom
        {
            get { return Math.Max(Math.Max(_v1.Y, _v2.Y), _v3.Y); }
        }

        /// <summary>
        /// Gets the difference between the rightmost and leftmost x-values of this triangle.
        /// </summary>
        public double Width
        {
            get { return this.Right - this.Left; }
        }

        /// <summary>
        /// Gets the difference between the bottommost and topmost y-values of this triangle.
        /// </summary>
        public double Height
        {
            get { return this.Bottom - this.Top; }
        }

        /// <summary>
        /// Gets a bounding rectangle that will exactly encompass the leftmost, topmost, rightmost, and bottommost
        /// extents of this triangle.
        /// </summary>
        public RectangleR Bounds
        {
            get { return new RectangleR(this.Left, this.Top, this.Width, this.Height); }
        }

        /// <summary>
        /// Creates a graphics path representing this triangle. Useful for drawing or
        /// creating Region objects.
        /// </summary>
        public GraphicsPath Path
        {
            get
            {
                return new GraphicsPath(
                    new PointF[3] { (PointF) _v1, (PointF) _v2, (PointF) _v3 },
                    new byte[3] {
					    (byte) PathPointType.Start, 
					    (byte) PathPointType.Line, 
					    (byte) (PathPointType.Line | PathPointType.CloseSubpath)
                    });
            }
        }

        /// <summary>
        /// Translates all coordinates of this triangle by the given deltas.
        /// </summary>
        /// <param name="dx">The amount of x-coordinate to translate by.</param>
        /// <param name="dy">The amount of y-coordinate to translate by.</param>
        public void TranslateBy(double dx, double dy)
        {
            _v1.X += dx;
            _v2.X += dx;
            _v3.X += dx;

            _v1.Y += dy;
            _v2.Y += dy;
            _v3.Y += dy;
        }

        /// <summary>
        /// Translates the centroid of this triangle to the given location.
        /// </summary>
        /// <param name="cx">The x-coordinate of the location to translate the centroid to.</param>
        /// <param name="cy">The y-coordinate of the location to translate the centroid to.</param>
        public void TranslateTo(double x, double y)
        {
            PointR c = this.Centroid;
            TranslateBy(x - c.X, y - c.Y);
        }

        /// <summary>
        /// Rotates this triangle around its centroid by the given radians.
        /// </summary>
        /// <param name="radians">The radians by which to rotate this triangle.</param>
        public void RotateBy(double radians)
        {
            this.RotateBy(this.Centroid, radians);
        }

        /// <summary>
        /// Rotate the triangle around the given point by the given radians.
        /// </summary>
        /// <param name="around">The point around which to rotate this triangle.</param>
        /// <param name="radians">The radians by which to rotate this triangle.</param>
        public void RotateBy(PointR around, double radians)
        {
            _v1 = GeotrigEx.RotatePoint(_v1, around, radians);
            _v2 = GeotrigEx.RotatePoint(_v2, around, radians);
            _v3 = GeotrigEx.RotatePoint(_v3, around, radians);
        }

        /// <summary>
        /// Scales the triangle by the given percent. A percent of 0.50 will make the triangle half
        /// as big, while a percent of 2.00 will make the triangle twice as big. The centroid of the
        /// triangle is preserved.
        /// </summary>
        /// <param name="percent">The percent by which to scale the triangle.</param>
        public void ScaleBy(float percent)
        {
            PointR centroid = this.Centroid; // save
            this.TranslateTo(0.0, 0.0);
            _v1.X *= percent;
            _v1.Y *= percent;
            _v2.X *= percent;
            _v2.Y *= percent;
            _v3.X *= percent;
            _v3.Y *= percent;
            this.TranslateTo(centroid.X, centroid.Y); // restore
        }

        /// <summary>
        /// Hit-tests whether the given point falls within this triangle. This method uses 
        /// Barycentric coordinates to do the hit-testing.
        /// </summary>
        /// <param name="pt">The point to hit-test.</param>
        /// <returns>True if the point is strictly contained by this triangle; false otherwise.</returns>
        /// <see cref="http://mathworld.wolfram.com/BarycentricCoordinates.html"/>
        public bool Contains(PointR pt)
        {
            double b0, b1, b2, b3;

            b0 = (_v2.X - _v1.X) * (_v3.Y - _v1.Y) - (_v3.X - _v1.X) * (_v2.Y - _v1.Y);
            b1 = ((_v2.X - pt.X) * (_v3.Y - pt.Y) - (_v3.X - pt.X) * (_v2.Y - pt.Y)) / b0;
            b2 = ((_v3.X - pt.X) * (_v1.Y - pt.Y) - (_v1.X - pt.X) * (_v3.Y - pt.Y)) / b0;
            b3 = ((_v1.X - pt.X) * (_v2.Y - pt.Y) - (_v2.X - pt.X) * (_v1.Y - pt.Y)) / b0;

            return (b1 > 0f && b2 > 0f && b3 > 0f);
        }

        /// <summary>
        /// Equals operator for TriangleR.
        /// </summary>
        /// <param name="t1">The left-hand side of the operator.</param>
        /// <param name="t2">The right-hand side of the operator.</param>
        /// <returns>True if the three vertices of the triangles are equal; false otherwise.</returns>
        public static bool operator ==(TriangleR t1, TriangleR t2)
        {
            return (
                   t1._v1 == t2._v1 
                && t1._v2 == t2._v2 
                && t1._v3 == t2._v3
                );
        }
        
        /// <summary>
        /// Not-equals operator for TriangleR.
        /// </summary>
        /// <param name="t1">The left-hand side of the operator.</param>
        /// <param name="t2">The right-hand side of the operator.</param>
        /// <returns>True if any of the vertices of the triangles disagree; false otherwise.</returns>
        public static bool operator !=(TriangleR t1, TriangleR t2)
        {
            return (
                   t1._v1 != t2._v1
                || t1._v2 != t2._v2 
                || t1._v3 != t2._v3
                );
        }

        /// <summary>
        /// Tests whether the given object is equal to this triangle.
        /// </summary>
        /// <param name="obj">The object to test.</param>
        /// <returns>True if the given object is a triangle equal to this triangle; false otherwise.</returns>
        public override bool Equals(object obj)
        {
            if (obj is TriangleR)
            {
                TriangleR t = (TriangleR) obj;
                return (this.V1 == t.V1 && this.V2 == t.V2 && this.V3 == t.V3);
            }
            return false;
        }

        /// <summary>
        /// Gets a hashcode for this triangle.
        /// </summary>
        /// <returns>A hashcode for this triangle.</returns>
        public override int GetHashCode()
        {
            return this.ToString().GetHashCode();
        }
    }
}
