﻿using System;
using System.IO;

namespace WobbrockLib.Windows
{
    /// <summary>
    /// Useful utility functions for console applications.
    /// </summary>
    public static class ConsoleEx
    {
        /// <summary>
        /// In a console application, gets a full file path guaranteed to point to an existing file.
        /// </summary>
        /// <param name="prompt">The prompt text displayed to the user.</param>
        /// <param name="sDefault">The default answer for when the user just presses Enter.</param>
        /// <returns>The full path of an existing file.</returns>
        public static string GetFileFromUser(string prompt, string sDefault)
        {
            string file = String.Empty;
            Console.Write(prompt);
            if (sDefault != String.Empty)
                Console.WriteLine(" [{0}]", sDefault);
            else
                Console.WriteLine();
            while (true)
            {
                Console.Write("> ");
                file = Console.ReadLine();
                file = file.Trim(new char[2] { ' ', '\"' });
                if (file == String.Empty)
                {
                    file = sDefault;
                }
                if (!File.Exists(file))
                {
                    Console.WriteLine("File not found. Please try again.");
                }
                else break;
            }
            return file;
        }

        /// <summary>
        /// In a console application, gets a full path guaranteed to point to an existing file or
        /// directory. If the path ends in a backslash, it is examined as a directory. If it does
        /// not, it is first examined as a file, and if that does not exist, then it is examined
        /// as a directory.
        /// </summary>
        /// <param name="prompt">The prompt text displayed to the user.</param>
        /// <param name="sDefault">The default answer for when the user just presses Enter.</param>
        /// <returns>The full path of an existing file or directory.</returns>
        public static string GetFileOrDirectoryFromUser(string prompt, string sDefault)
        {
            string path = String.Empty;
            Console.Write(prompt);
            if (sDefault != String.Empty)
                Console.WriteLine(" [{0}]", sDefault);
            else
                Console.WriteLine();
            while (true)
            {
                Console.Write("> ");
                path = Console.ReadLine();
                path = path.Trim(new char[2] { ' ', '\"' });
                if (path == String.Empty)
                {
                    path = sDefault;
                }
                if (path[path.Length - 1] == '\\')
                {
                    if (Directory.Exists(path))
                        break;
                    else 
                        Console.WriteLine("Directory not found. Please try again.");
                }
                else if (File.Exists(path))
                    break;
                else if (Directory.Exists(path))
                    break;
                else
                    Console.WriteLine("File or directory not found. Please try again.");
            }
            return path;
        }

        /// <summary>
        /// In a console application, gets a yes or no response from the user.
        /// </summary>
        /// <param name="prompt">The prompt text displayed to the user.</param>
        /// <param name="sDefault">The default answer for when the user just presses Enter.</param>
        /// <returns>A Boolean indicating whether the user expressed Yes or No.</returns>
        public static bool GetYesNoFromUser(string prompt, bool bDefault)
        {
            string yn = String.Empty;
            Console.Write(prompt);
            Console.WriteLine(" [{0}]", bDefault ? "y" : "n");
            while (true)
            {
                Console.Write("> ");
                yn = Console.ReadLine();
                yn = yn.Trim().ToLower();
                if (yn == String.Empty)
                {
                    yn = bDefault ? "y" : "n";
                }
                if (yn[0] != 'y' && yn[0] != 'n')
                {
                    Console.WriteLine("Please enter [y]es or [n]o. Try again.");
                }
                else break;
            }
            return (yn[0] == 'y');
        }

        /// <summary>
        /// In a console application, gets an integer value from the user between low and high, inclusive.
        /// </summary>
        /// <param name="prompt">The prompt text displayed to the user.</param>
        /// <param name="low">The low value in the range of acceptable integers.</param>
        /// <param name="high">The high value in the range of acceptable integers.</param>
        /// <param name="iDefault">The default answer for when the user just presses Enter.</param>
        /// <returns>An integer between low and high, inclusive.</returns>
        public static int GetIntFromUser(string prompt, int low, int high, int iDefault)
        {
            string sint = String.Empty;
            Console.Write(prompt);
            Console.WriteLine(" [{0}]", iDefault);
            while (true)
            {
                Console.Write("> ");
                sint = Console.ReadLine();
                sint = sint.Trim().ToLower();
                int iint;
                if (sint == String.Empty)
                    return iDefault;
                else if (Int32.TryParse(sint, out iint) && (low <= iint && iint <= high))
                    return iint;
                else
                    Console.WriteLine("Please enter an integer between {0} and {1}, inclusive.", low, high);
            }
        }
    }
}
