﻿using System;
using System.Drawing;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Globalization;
using System.Threading;
using System.Windows.Forms;
using WobbrockLib.Types;
using WobbrockLib.Windows;

namespace WobbrockLib.Windows
{
    /// <summary>
    /// This static class provides an easy-to-use interface to Control Panel settings.
    /// </summary>
    public static class ControlPanelEx
    {
        #region Mouse Control Panel

        /// <summary>
        /// Gets or sets the gain on the Mouse Control Panel. Legitimate values will be from [1u, 20u], inclusive.
        /// The Windows default is a value of 10u.
        /// </summary>
        public static uint MouseGain
        {
            get
            {
                uint gain = 0;
                Win32.SystemParametersInfo((uint) Win32.SPI.GETMOUSESPEED, 0u, ref gain, 0u);
                return gain;
            }
            set
            {
                if (value < 1u)
                    value = 1u;
                else if (value > 20u)
                    value = 20u;
                Win32.SystemParametersInfo((uint) Win32.SPI.SETMOUSESPEED, 0u, value, 0u);
            }
        }

        /// <summary>
        /// Gets or sets the checkbox on the Mouse Control Panel for enhancing pointing precision.
        /// If true, this setting causes the mouse transfer function to be non-linear, making
        /// fine pointing movements easier while still providing high cursor velocities when the
        /// mouse is moved quickly. If false, a linear transfer function is used instead.
        /// </summary>
        /// <remarks>
        /// Gets or sets the uint[3] pointer acceleration parameters as defined here:
        /// http://msdn.microsoft.com/en-us/library/ms724947(VS.85).aspx. The defaults are
        /// [6,10,1] for acceleration on, and [0,0,0] for acceleration off.
        /// </remarks>
        public static bool EnhancePointingPrecision
        {
            get
            {
                uint[] u = new uint[3];
                Win32.SystemParametersInfo((uint) Win32.SPI.GETMOUSE, 0u, u, 0u);
                return u[2] == 1;
            }
            set
            {
                uint[] u = value ? new uint[3] { 6, 10, 1 } : new uint[3] { 0, 0, 0 };
                Win32.SystemParametersInfo((uint) Win32.SPI.SETMOUSE, 0u, u, 0u);
            }
        }

        /// <summary>
        /// Opens the Mouse Control Panel set to the given 0-based tab index. This is useful for applications
        /// that want to inspect or alter the mouse settings. The Pointer Options tab is index 2.
        /// </summary>
        /// <param name="tab">The 0-based tab index to open.</param>
        public static void OpenMouseControlPanel(int tab)
        {
            Process.Start("control", String.Format("main.cpl,,{0}", tab));
        }

        #endregion

        #region Accessibility Control Panel

        /// <summary>
        /// Gets whether or not the FilterKeys keyboard accessibility feature is turned on. FilterKeys can 
        /// be used to disable key repeat rate and other similar things.
        /// </summary>
        /// <param name="f">This ref parameter will contain a structure describing the filter keys settings.</param>
        /// <returns>True if FilterKeys is on; false otherwise.</returns>
        public static bool GetFilterKeys(ref Win32.FILTERKEYS f)
        {
            f = Win32.FILTERKEYS.Empty;
            f.cbSize = (uint) Marshal.SizeOf(f);
            Win32.SystemParametersInfo((uint) Win32.SPI.GETFILTERKEYS, f.cbSize, ref f, 0u);
            return ((f.dwFlags & (uint) Win32.FKF.FILTERKEYSON) == 0x01u);
        }

        /// <summary>
        /// Sets the FilterKeys accessibility feature to whatever parameters are given.
        /// </summary>
        /// <param name="f">The FilterKeys settings to set.</param>
        /// <returns>True if FilterKeys are set successfully; false otherwise.</returns>
        public static bool SetFilterKeys(Win32.FILTERKEYS f)
        {
            f.cbSize = (uint) Marshal.SizeOf(f);
            return Win32.SystemParametersInfo((uint) Win32.SPI.SETFILTERKEYS, f.cbSize, ref f, 0u);
        }

        /// <summary>
        /// Uses FilterKeys to disable key repeats, filtering them out.
        /// </summary>
        /// <returns>True if key repeats are disabled successfully; false otherwise.</returns>
        public static bool FilterKeyRepeats()
        {
            Win32.FILTERKEYS f = Win32.FILTERKEYS.Empty;
            f.cbSize = (uint) Marshal.SizeOf(f);
            f.dwFlags = (uint) Win32.FKF.FILTERKEYSON; // turn on Filter Keys
            f.iWaitMSec = 0u;	// key-down time until key is sent (0 means instantly)
            f.iDelayMSec = 0u;	// time from key-down to key beginning to repeat (0 means no repeat)
            f.iRepeatMSec = 0u;	// time interval between key repeats (0 means no repeat)
            f.iBounceMSec = 0u;	// req'd time after key-up before same key is accepted (0 means debouncing is off)
            return Win32.SystemParametersInfo((uint) Win32.SPI.SETFILTERKEYS, f.cbSize, ref f, 0u);
        }

        #endregion

        #region Region Control Panel

        /// <summary>
        /// Gets or sets the decimal separator used to parse numbers. Note that the setting only applies to the
        /// current thread, not the permanent control panel setting.
        /// </summary>
        /// <see cref="https://stackoverflow.com/questions/24785689/trying-to-set-the-decimal-separator-for-the-current-language-getting-instance"/>
        public static string DecimalSeparator
        {
            get
            {
                return Thread.CurrentThread.CurrentCulture.NumberFormat.NumberDecimalSeparator; // == NumberFormatInfo.CurrentInfo.NumberDecimalSeparator;
            }
            set
            {
                CultureInfo ci = (CultureInfo) Thread.CurrentThread.CurrentCulture.Clone();
                if (ci.NumberFormat.NumberDecimalSeparator != value)
                {
                    ci.NumberFormat.NumberDecimalSeparator = value;
                    Thread.CurrentThread.CurrentCulture = ci;
                }
            }
        }

		#endregion

		#region Display Control Panel

        /// <summary>
        /// Gets the display scaling factor for the primary display screen. In a multimon setup,
        /// this is the display screen indicated as "1" in display settings. To see the display
        /// scaling, right click on the desktop and choose "Display settings..."
        /// </summary>
        /// <returns>The display scaling factor, such as 1.00 for 100%, 1.50 for 150%, and so on.
        /// If the function fails, -1 is returned.</returns>
        public static double GetDisplayScaling()
		{
            Screen screen = Screen.PrimaryScreen;
            return GetDisplayScaling(screen);
		}

        /// <summary>
        /// Gets the display scaling factor for the screen containing the given screen coordinate.
        /// To see the display scaling, right click on the desktop and choose "Display settings..."
        /// </summary>
        /// <param name="screenPt">The point in screen coordinates for which to get the display
        /// scaling factor.</param>
        /// <returns>The display scaling factor, such as 1.00 for 100%, 1.50 for 150%, and so on.
        /// If the function fails, -1 is returned.</returns>
        /// <remarks>In a multimon setup, different displays can have different display scalings.</remarks>
        public static double GetDisplayScaling(Point screenPt)
        {
            Screen screen = Screen.FromPoint(screenPt);
            return GetDisplayScaling(screen);
        }

        /// <summary>
        /// Gets the display scaling factor for the screen containing the given (x,y) coordinate.
        /// To see the display scaling, right click on the desktop and choose "Display settings..."
        /// </summary>
        /// <param name="screenX">The X screen coordinate for which to get the display scaling factor.</param>
        /// <param name="screenY">The Y screen coordinate for which to get the display scaling factor.</param>
        /// <returns>The display scaling factor, such as 1.00 for 100%, 1.50 for 150%, and so on.
        /// If the function fails, -1 is returned.</returns>
        /// <remarks>In a multimon setup, different displays can have different display scalings.</remarks>
        public static double GetDisplayScaling(int screenX, int screenY)
        {
            Screen screen = Screen.FromPoint(new Point(screenX, screenY));
            return GetDisplayScaling(screen);
        }

        /// <summary>
        /// Gets the display scaling factor for the given screen. To see the display scaling, right click 
        /// on the desktop and choose "Display settings..."
        /// </summary>
        /// <param name="screen">The screen for which to retrieve the display scaling factor.</param>
        /// <remarks>In a multimon setup, different displays can have different display scalings.</remarks>
        /// <returns>The display scaling factor, such as 1.00 for 100%, 1.50 for 150%, and so on.
        /// If the function fails, -1.00 is returned.</returns>
        public static double GetDisplayScaling(Screen screen)
		{
            Win32.DEVMODE dm = new Win32.DEVMODE();
            dm.dmSize = (ushort) Marshal.SizeOf(typeof(Win32.DEVMODE));

            // for debugging
            //IntPtr i = System.Runtime.InteropServices.Marshal.OffsetOf(typeof(Win32.DEVMODE), "dmFormName");
            //Debug.WriteLine("dmFormName Field offset: {0}", i.ToInt32()); // should be 70
            //i = System.Runtime.InteropServices.Marshal.OffsetOf(typeof(Win32.DEVMODE), "dmLogPixels");
            //Debug.WriteLine("dmLogPixels Field offset: {0}", i.ToInt32()); // should be 102

            if (Win32.EnumDisplaySettings(screen.DeviceName, -1, ref dm))
                return (double) Math.Round(Decimal.Divide(dm.dmPelsWidth, screen.Bounds.Width), 2);
            else
                return -1.00;
        }

        /// <summary>
        /// Gets the global screen bounds as a single rectangle for all the displays. Note that depending 
        /// on the multimon setup, this does not mean that every addressable (x,y) value within this 
        /// global rectangle will be visible on a monitor.
        /// </summary>
        /// <returns>A rectangle spanning the minimum and maximum extents of all displays.</returns>
        public static Rectangle GetGlobalDisplayBounds()
		{
            int left = Int32.MaxValue;
            int top = Int32.MaxValue;
            int right = Int32.MinValue;
            int bottom = Int32.MinValue;

            for (int i = 0; i < Screen.AllScreens.Length; i++)
            {
                Rectangle r = Screen.AllScreens[i].Bounds;
                if (r.Left < left)
                    left = r.Left;
                if (r.Top < top)
                    top = r.Top;
                if (r.Right > right)
                    right = r.Right;
                if (r.Bottom > bottom)
                    bottom = r.Bottom;
            }

            return new Rectangle(
                left,
                top,
                right - left,
                bottom - top
                );
        }

		#endregion
	}
}
