﻿using System;
using System.Drawing;
using System.Windows.Forms;

namespace WobbrockLib.Windows
{
	/// <summary>
	/// <para>
	/// An instance of a CursorWrapper wraps the system's mouse cursor such that it can be guaranteed
	/// to be shown or hidden as desired, without the client having to worry about keeping track of the
	/// show/hide count for the cursor. An assumption is that when this class is created, the cursor
	/// has not been shown or hidden already (its count is zero).
	/// </para>
	/// <para>
	/// A CursorWrapper instance also allows a client to easily adjust the position and clipping rect
	/// of the system cursor. Convenience methods are provided to freeze the system cursor at a position,
	/// to unfreeze it, and to check if it is frozen.
	/// </para>
	/// </summary>
	public class CursorWrapper
	{
		#region Fields

		private Cursor _current; // the current cursor type shown
		private int _count; // the show/hide count, where zero is visible

		#endregion

		#region Constructor

		/// <summary>
		/// Creates an instance of a VisibleCursor, which wraps the system cursor and helps manage 
		/// its shown or hidden state.
		/// </summary>
		public CursorWrapper()
		{
			Cursor.Current = Cursors.Default;
			_current = Cursors.Default;
			_count = 0;
		}

		/// <summary>
		/// Destructor.
		/// </summary>
		~CursorWrapper()
		{
			Cursor.Current = Cursors.Default;
			Cursor.Clip = Rectangle.Empty;
		}

		#endregion

		#region Appearance

		/// <summary>
		/// Gets or sets the current cursor type. Use Cursors.[type] to set.
		/// </summary>
		public Cursor Current
		{
			get { return _current; }
			set
			{
				Cursor.Current = value;
				_current = value;
			}
		}

		/// <summary>
		/// The one property for the VisibleCursor, where setting this property to true will guarantee
		/// that the cursor is shown, and setting it to false will guarantee that it is hidden.
		/// </summary>
		public bool Visible
		{
			get
			{
				return _count >= 0;
			}
			set
			{
				if (value) // setting the cursor visible
				{
					while (_count < 0)
					{
						Cursor.Show();
						_count++;
					}
				}
				else // setting the cursor hidden
				{
					while (_count >= 0)
					{
						Cursor.Hide();
						_count--; 
					}
				}
			}
		}

		#endregion

		#region Position

		/// <summary>
		/// Convenience property for getting or setting the system cursor's position.
		/// </summary>
		public Point Position
		{
			get { return Cursor.Position; }
			set { Cursor.Position = value; }
		}

		#endregion

		#region Clipping

		/// <summary>
		/// Convenience property for getting or setting the system cursor's clipping rectangle.
		/// </summary>
		public Rectangle Clip
		{
			get { return Cursor.Clip; }
			set { Cursor.Clip = value; }
		}

		/// <summary>
		/// Resets the system cursor's clipping rectangle to be the default bounds.
		/// </summary>
		public void ResetClip()
		{
			this.Clip = Rectangle.Empty;
		}

		#endregion

		#region Freezing

		/// <summary>
		/// Freeze the pointer at the current screen point. This is the same as setting the
		/// clipping region to be at a given point with a size of one pixel.
		/// </summary>
		public void Freeze()
		{
			Point pt = this.Position;
			this.Clip = new Rectangle(pt.X, pt.Y, 1, 1);
		}

		/// <summary>
		/// Freeze the pointer at the given screen point. This is the same as setting the
		/// clipping region to be at a given point with a size of one pixel.
		/// </summary>
		/// <param name="pt">The screen coordinate at which to freeze the pointer.</param>
		public void Freeze(Point pt)
		{
			this.Position = pt;
			this.Clip = new Rectangle(pt.X, pt.Y, 1, 1);
		}

		/// <summary>
		/// Unfreeze a frozen mouse pointer. This is functionally the same as resetting the
		/// clipping region.
		/// </summary>
		public void Unfreeze()
		{
			this.Clip = Rectangle.Empty;
		}

		/// <summary>
		/// Unfreeze a frozen mouse pointer ensuring it is unfrozen at a given screen point.
		/// This is functionally the same as resetting the clipping region.
		/// </summary>
		/// <param name="pt">The screen coordinate at which to unfreeze the pointer.</param>
		public void Unfreeze(Point pt)
		{
			this.Position = pt;
			this.Clip = Rectangle.Empty;
		}

		/// <summary>
		/// Gets whether the system cursor is currently frozen, which is defined as having a
		/// 1 px by 1 px clipping rectangle.
		/// </summary>
		public bool Frozen
		{
			get
			{
				Rectangle r = this.Clip;
				return (r.Width == 1 && r.Height == 1);
			}
		}

		#endregion
	}
}
