﻿using System;
using System.Diagnostics;
using System.Text;
using System.Runtime.InteropServices;
using System.Windows.Forms;

namespace WobbrockLib.Windows
{
	/// <summary>
	/// Utility class that enables interaction with Win32-style menus on the Microsoft Windows operating system.
	/// This utility class does not presume to work with newer MenuStrip-style menus. A good example of Win32
	/// menus is in the Notepad application that ships with Microsoft Windows.
	/// </summary>
	public static class Menu32Ex
	{
		#region Keys Affecting Menus

		/// <summary>
		/// Gets whether the given key data could open or access a menu or menu bar by default on 
		/// Microsoft Windows.
		/// </summary>
		/// <param name="data">The KeyData field from a KeyEventArgs.</param>
		/// <returns>True if the key or key combination could open a menu; false otherwise.</returns>
		public static bool KeysOpenMenu(Keys data)
		{
			Keys code = (Keys.KeyCode & data);
			bool ctrl = (data & Keys.Control) != 0;
			bool alt = (data & Keys.Alt) != 0;
			bool shift = (data & Keys.Shift) != 0;

			return
				(alt && code != Keys.None) ||           // Alt + any key (mnemonic)
				(alt && code == Keys.Space) ||          // Alt + Space
				code == Keys.Space ||					// Space (if Alt was previously pressed)
				code == Keys.Enter ||                   // Enter (if Alt was previously pressed)
				code == Keys.Down ||                    // Down arrow (if Alt was previously pressed)
				code == Keys.Up ||                      // Up arrow (if Alt was previously pressed)
				code == Keys.Apps ||                    // Menu logo key
				(ctrl && code == Keys.Escape) ||		// Ctrl + Escape also opens Start Menu
				(shift && code == Keys.F10) ||          // Shift + F10
				(ctrl && shift && code == Keys.F10);	// Ctrl + Shift + F10
		}

		/// <summary>
		/// Gets whether the given key data could put the focus on the menu bar by default on
		/// Microsoft Windows.
		/// </summary>
		/// <param name="data">The KeyData field from a KeyEventArgs.</param>
		/// <returns>True if the key or key combination could access a menu; false otherwise.</returns>
		public static bool KeysAccessMenu(Keys data)
		{
			Keys code = (Keys.KeyCode & data);
			bool ctrl = (data & Keys.Control) != 0;
			bool alt = (data & Keys.Alt) != 0;
			bool shift = (data & Keys.Shift) != 0;

			return
				(!ctrl && code == Keys.Menu) ||
				(!ctrl && code == Keys.LMenu) ||
				(!ctrl && code == Keys.RMenu) ||
				code == Keys.F10;
		}

		/// <summary>
		/// Gets whether the given key data could close a menu by default on Microsoft Windows.
		/// </summary>
		/// <param name="data">The KeyData field from a KeyEventArgs.</param>
		/// <returns>True if the key or key combination could close a menu; false otherwise.</returns>
		public static bool KeysCloseMenu(Keys data)
		{
			Keys code = (Keys.KeyCode & data);
			bool shift = (data & Keys.Shift) != 0;
			bool ctrl = (data & Keys.Control) != 0;
			bool alt = (data & Keys.Alt) != 0;

			return
				code == Keys.Escape ||					// Esc
				code == Keys.Enter ||					// Enter
				code == Keys.Menu ||					// Alt
				code == Keys.LMenu ||					// Left Alt
				code == Keys.RMenu ||					// Right Alt
				code == Keys.LWin ||                    // Left Windows logo key
				code == Keys.RWin ||					// Right Windows logo key
				code == Keys.Apps ||					// Menu logo key
				(ctrl && code == Keys.Escape) ||		// Ctrl + Escape
				code == Keys.F10 ||						// F10
				(shift && code == Keys.F10) ||          // Shift + F10
				(ctrl && shift && code == Keys.F10);    // Ctrl + Shift + F10
		}

		/// <summary>
		/// Gets whether the given key data could navigate the selection highlight atop a menu or
		/// menu bars on Microsoft Windows.
		/// </summary>
		/// <param name="data">The KeyData field from a KeyEventArgs.</param>
		/// <returns>True if the key or key combination could navigate a menu; false otherwise.</returns>
		public static bool KeysNavigateMenu(Keys data)
		{
			Keys code = (Keys.KeyCode & data);
			bool shift = (data & Keys.Shift) != 0;
			bool ctrl = (data & Keys.Control) != 0;
			bool alt = (data & Keys.Alt) != 0;

			return
				data == Keys.Left ||
				data == Keys.Right ||
				data == Keys.Up ||
				data == Keys.Down ||
				data == Keys.Home ||
				data == Keys.End;
		}

		#endregion

		#region Currently Open Menus

		/// <summary>
		/// Determines whether there is a menu open on the main GUI thread of whatever the active application 
		/// is right now. This function only works for "old style" menus like those found in Microsoft Notepad.
		/// </summary>
		/// <returns>True if there is a menu active; false otherwise.</returns>
		public static bool IsMenuOpen()
		{
			return IsMenuOpen(false);
		}

		/// <summary>
		/// Determines whether there is a menu active on the main GUI thread of whatever the active application 
		/// is right now. This function only works for "old style" menus like those found in Microsoft Notepad.
		/// </summary>
		/// <param name="verbose">If true, output is written to the Debug console.</param>
		/// <returns>True if there is a menu active; false otherwise.</returns>
		public static bool IsMenuOpen(bool verbose)
		{
			bool active = false;
			Win32.GUITHREADINFO gui = Win32.GUITHREADINFO.Empty;
			gui.cbSize = (uint) Marshal.SizeOf(gui);

			if (Win32.GetGUIThreadInfo(0u, ref gui)) // get what main UI thread is doing now
			{
				bool hasOwner = (gui.hwndMenuOwner != IntPtr.Zero);
				bool inMenuMode = ((gui.flags & Win32.GUI.INMENUMODE) != 0); // regular menubar menu
				bool popupMenuMode = ((gui.flags & Win32.GUI.POPUPMENUMODE) != 0); // popup context menu
				bool systemMenuMode = ((gui.flags & Win32.GUI.SYSTEMMENUMODE) != 0); // system menu at top-left of a window
				active = hasOwner || inMenuMode || popupMenuMode || systemMenuMode;

				if (verbose)
				{
					Debug.WriteLine("hwndOwner={0}, GUI_INMENUMODE={1}, GUI_POPUPMENUMODE={2}, GUI_SYSTEMMENUMODE={3}",
						gui.hwndMenuOwner, inMenuMode, popupMenuMode, systemMenuMode);
				}
			}
			else if (verbose)
			{
				Debug.WriteLine("Win32 call to GetGUIThreadInfo failed.");
			}
			return active;
		}

		/// <summary>
		/// Gets the window handle of the window with the active menu on the foreground thread. If no window
		/// has an active menu, or if the Win32 call to GetGUIThreadInfo fails, then IntPtr.Zero is returned.
		/// </summary>
		/// <returns>The window handle to the foreground window that is currently showing a menu; if none,
		/// then IntPtr.Zero is returned.</returns>
		public static IntPtr GetWindowWithOpenMenu()
		{
			return GetWindowWithOpenMenu(false);
		}

		/// <summary>
		/// Gets the window handle of the window with the active menu on the foreground thread. If no window
		/// has an active menu, or if the Win32 call to GetGUIThreadInfo fails, then IntPtr.Zero is returned.
		/// </summary>
		/// <param name="verbose">If true, output is written to the Debug console.</param>
		/// <returns>The window handle to the foreground window that is currently showing a menu; if none,
		/// then IntPtr.Zero is returned.</returns>
		public static IntPtr GetWindowWithOpenMenu(bool verbose)
		{
			IntPtr hWnd = IntPtr.Zero;
			Win32.GUITHREADINFO gui = Win32.GUITHREADINFO.Empty;
			gui.cbSize = (uint) Marshal.SizeOf(gui);

			if (Win32.GetGUIThreadInfo(0u, ref gui)) // get what main UI thread is doing now
			{
				hWnd = gui.hwndMenuOwner;
				if (verbose)
				{
					Debug.WriteLine("hwndOwner={0}", hWnd);
				}
			}
			else if (verbose)
			{
				Debug.WriteLine("Win32 call to GetGUIThreadInfo failed.");
			}
			return hWnd;
		}

		#endregion

		#region Mnemonics & Shortcuts

		/// <summary>
		/// Gets the position of the given mnemonic character on the given menu. A character mnemonic
		/// is the character following an ampersand when the menu item is named. Windows renders such
		/// a character underlined when displaying it in a menu.
		/// </summary>
		/// <param name="hMenu">A handle to a menu.</param>
		/// <param name="mnemonic">The character mnemonic to test.</param>
		/// <returns>The zero-based index of the first menu item to contain the given mnemonic; or -1
		/// if no such mnemonic exists on the given menu.</returns>
		public static int GetMnemonicPos(IntPtr hMenu, char mnemonic)
		{
			char[] mnemonics = GetMnemonics(hMenu);
			if (mnemonics != null)
			{
				for (int i = 0; i < mnemonics.Length; i++)
				{
					if (mnemonics[i] == mnemonic)
						return i;
				}
			}
			return -1;
		}

		/// <summary>
		/// Gets the mnemonic characters for each item in a given menu.
		/// </summary>
		/// <param name="hMenu">The menu from which to retrieve mnemonic characters.</param>
		/// <returns>
		/// An array of characters corresponding to the mnemonics present in the given menu.
		/// </returns>
		public static char[] GetMnemonics(IntPtr hMenu)
		{
			if (!Win32.IsMenu(hMenu))
				return null;

			int count = Win32.GetMenuItemCount(hMenu);
			char[] mnemonics = new char[count];
			for (uint i = 0; i < count; i++)
			{
				int length = Win32.GetMenuString(hMenu, i, null, 0, (uint) Win32.MF.BYPOSITION);
				if (length > 0)
				{
					StringBuilder sb = new StringBuilder(length + 1);
					int copied = Win32.GetMenuString(hMenu, i, sb, sb.Capacity, (uint) Win32.MF.BYPOSITION);
					if (copied == length) // successfully copied the menu string
					{
						string s = sb.ToString();
						int pos = s.IndexOf('&');
						if (pos != -1 && pos < s.Length - 1)
						{
							mnemonics[i] = s[pos + 1];
						}
					}
				}
			}
			return mnemonics;
		}

		/// <summary>
		/// Gets the position of the given shortcut string on the given menu. A shortcut string is
		/// the string following a tab character ('\t') in the menu item text. Windows renders such a
		/// string as the shortcut after the menu item name.
		/// </summary>
		/// <param name="hMenu">A handle to a menu.</param>
		/// <param name="shortcut">The shortcut string to look for, e.g., "Ctrl+S" or "Ctrl+Shift+1".</param>
		/// <returns>The zero-based index of the first menu item to contain the given shortcut string; 
		/// or -1 if no such shortcut string exists on the given menu.</returns>
		public static int GetShortcutPos(IntPtr hMenu, string shortcut)
		{
			string[] shortcuts = GetShortcuts(hMenu);
			if (shortcuts != null)
			{
				for (int i = 0; i < shortcuts.Length; i++)
				{
					if (shortcuts[i] == shortcut)
						return i;
				}
			}
			return -1;
		}

		/// <summary>
		/// Gets the shortcut key string for each item in a given menu.
		/// </summary>
		/// <param name="hMenu">The menu from which to retrieve the shortcuts.</param>
		/// <returns>
		/// An array of strings corresponding to the shortcuts present in the given menu.
		/// </returns>
		public static string[] GetShortcuts(IntPtr hMenu)
		{
			if (!Win32.IsMenu(hMenu))
				return null;

			int count = Win32.GetMenuItemCount(hMenu);
			string[] shortcuts = new string[count];
			for (uint i = 0; i < count; i++)
			{
				int length = Win32.GetMenuString(hMenu, i, null, 0, (uint) Win32.MF.BYPOSITION);
				if (length > 0)
				{
					StringBuilder sb = new StringBuilder(length + 1);
					int copied = Win32.GetMenuString(hMenu, i, sb, sb.Capacity, (uint) Win32.MF.BYPOSITION);
					if (copied == length) // successfully copied the menu string
					{
						string s = sb.ToString();
						int pos = s.IndexOf('\t');
						if (pos != -1 && pos < s.Length - 1)
						{
							shortcuts[i] = s.Substring(pos + 1);
						}
					}
				}
			}
			return shortcuts;
		}

		#endregion

		#region Menu Item Properties

		/// <summary>
		/// Gets whether the given menu item is disabled.
		/// </summary>
		/// <param name="hMenu">A handle to a menu.</param>
		/// <param name="pos">The zero-based index of the menu item to test.</param>
		/// <returns>True if the given item exists and is disabled; false otherwise.</returns>
		public static bool ItemDisabled(IntPtr hMenu, uint pos)
		{
			if (!Win32.IsMenu(hMenu))
				return false;

			uint flags = Win32.GetMenuState(hMenu, pos, (uint) Win32.MF.BYPOSITION);
			return (flags & (uint) Win32.MF.DISABLED) != 0u;
		}

		/// <summary>
		/// Gets whether the given menu item is grayed.
		/// </summary>
		/// <param name="hMenu">A handle to a menu.</param>
		/// <param name="pos">The zero-based index of the menu item to test.</param>
		/// <returns>True if the given item exists and is grayed; false otherwise.</returns>
		public static bool ItemGrayed(IntPtr hMenu, uint pos)
		{
			if (!Win32.IsMenu(hMenu))
				return false;

			uint flags = Win32.GetMenuState(hMenu, pos, (uint) Win32.MF.BYPOSITION);
			return (flags & (uint) Win32.MF.GRAYED) != 0u;
		}

		/// <summary>
		/// Gets whether the given menu item is checked.
		/// </summary>
		/// <param name="hMenu">A handle to a menu.</param>
		/// <param name="pos">The zero-based index of the menu item to test.</param>
		/// <returns>True if the given item exists and is checked; false otherwise.</returns>
		public static bool ItemChecked(IntPtr hMenu, uint pos)
		{
			if (!Win32.IsMenu(hMenu))
				return false;

			uint flags = Win32.GetMenuState(hMenu, pos, (uint) Win32.MF.BYPOSITION);
			return (flags & (uint) Win32.MF.CHECKED) != 0u;
		}

		/// <summary>
		/// Gets whether the given menu item is highlighted.
		/// </summary>
		/// <param name="hMenu">A handle to a menu.</param>
		/// <param name="pos">The zero-based index of the menu item to test.</param>
		/// <returns>True if the given item exists and is highlighted; false otherwise.</returns>
		public static bool ItemHighlighted(IntPtr hMenu, uint pos)
		{
			if (!Win32.IsMenu(hMenu))
				return false;

			uint flags = Win32.GetMenuState(hMenu, pos, (uint) Win32.MF.BYPOSITION);
			return (flags & (uint) Win32.MF.HILITE) != 0u;
		}

		/// <summary>
		/// Gets whether the given menu item is a horizontal line separator.
		/// </summary>
		/// <param name="hMenu">A handle to a menu.</param>
		/// <param name="pos">The zero-based index of the menu item to test.</param>
		/// <returns>True if the given item exists and is a separator; false otherwise.</returns>
		public static bool ItemIsSeparator(IntPtr hMenu, uint pos)
		{
			if (!Win32.IsMenu(hMenu))
				return false;

			uint flags = Win32.GetMenuState(hMenu, pos, (uint) Win32.MF.BYPOSITION);
			return (flags & (uint) Win32.MF.SEPARATOR) != 0u;
		}

		#endregion

		#region System Menu

		/// <summary>
		/// Delets the system menu for a given window by removing all of the system menu's items. 
		/// WARNING: Doing this makes a window undraggable, unsizable, and uncloseable from its
		/// titlebar.
		/// </summary>
		/// <param name="hWnd">A handle to the window whose system menu to delete.</param>
		public static void DeleteSystemMenu(IntPtr hWnd)
		{
			IntPtr hMenu = Win32.GetSystemMenu(hWnd, false);
			int count = Win32.GetMenuItemCount(hMenu);
			if (count > 0)
			{
				for (uint i = (uint) count - 1; i >= 0; i--)
				{
					Win32.DeleteMenu(hMenu, i, (uint) Win32.MF.BYPOSITION);
				}
			}
		}

		#endregion
	}
}
