using System;
using System.Text;
using System.Runtime.InteropServices;

namespace WobbrockLib.Windows
{
    /// <summary>
    /// See http://www.pinvoke.net/default.aspx/.
    /// </summary>
    public static class Win32
    {
        #region x86 or x64?

        /// <summary>
        /// Convenience property to tell whether the platform target being used is x86 or x64.
        /// </summary>
        /// <returns>
        /// Either 32 or 64 is returned as the number of bits in an IntPtr on this system.
        /// The former, 32, corresponds to use on an x86 build; the latter, 64, corresponds to
        /// use on an x64 build.
        /// </returns>
        /// <remarks>
        /// <para>
        /// The WobbrockLib.DLL generally should be built with "Any CPU" as its platform target.
        /// This allows it to be used by both x86 applications or x64 applications. If either of the 
        /// explicit platform targets are used instead of "Any CPU," then the application using WobbrockLib.DLL
        /// must also be built with the same platform target, or a loading error occurs.
        /// </para>
        /// <para>
        /// Most applications can ignore this issue except when using Win32 structs that have 
        /// "LayoutKind.Explicit" attributes and IntPtrs in them, because now the offsets will depend on whether 
        /// the size of an IntPtr (or UIntPtr) is 32 bits or 64 bits. (Note that even on x64 platforms, the size of
        /// a C# integer type is still just 32 bits. It is only IntPtr and UIntPtr that change.)
        /// </para>
        /// <para>
        /// Windows x86 requires integer data types and struct fields to start on 4-byte boundaries; Windows x64
        /// requires them to start on 8-byte boundaries. Therefore, a second cause of needing different structures
        /// depending on x86 or x64 is if "LayoutKind.Explicit" attributes are used to reproduce C unions and 
        /// layout boundaries come into play.
        /// </para>
        /// <para>
        /// Therefore, this Win32 class provides both 32-bit and 64-bit versions of all structures that use
        /// "LayoutKind.Explicit" and have (U)IntPtrs in them or have layout boundary issues. If using one of 
        /// these structures, it is up to the developer to use the correct one. This helper property should 
        /// be used to determine at runtime which structure to use.
        /// </para>
        /// </remarks>
        public static int SizeOfIntPtr
        {
            get
            {
                const int BitsInByte = 8;
                return BitsInByte * IntPtr.Size;
            }
        }

        #endregion

        #region Windows Messages

        //
        // Windows Messages
        //

		// For HWND_BROADCAST, use new IntPtr(Win32.HWND.BROADCAST).
        [DllImport("user32.dll")]
        public static extern int SendMessage(IntPtr hWnd, uint Msg, uint wParam, int lParam);

        [DllImport("user32.dll")]
        public static extern bool PostMessage(IntPtr hWnd, uint Msg, uint wParam, int lParam);

        public enum WM : uint // Windows messages (WM_*)
        {
            //
            // Standard Windows meassages
            //
            SETFOCUS = 0x0007,
            KILLFOCUS = 0x0008,
            CLOSE = 0x0010,
            ERASEBKGND = 0x0014,
            CANCELMODE = 0x001F, // cancel modal loop
            NCHITTEST = 0x0084, // nonclient hit test
            NCMOUSEMOVE = 0x00A0, // nonclient mouse move
            NCLBUTTONDOWN = 0x00A1,
            NCLBUTTONUP = 0x00A2,
            NCLBUTTONDBLCLK = 0x00A3,
            NCRBUTTONDOWN = 0x00A4,
            NCRBUTTONUP = 0x00A5,
            NCRBUTTONDBLCLK = 0x00A6,
            NCMBUTTONDOWN = 0x00A7,
            NCMBUTTONUP = 0x00A8,
            NCMBUTTONDBLCLK = 0x00A9,
            INPUT = 0x00FF, // for receiving raw device input
            KEYDOWN = 0x0100,
            KEYUP = 0x0101,
            SYSKEYDOWN = 0x0104,
            SYSKEYUP = 0x0105,
            INITDIALOG = 0x0110,
            COMMAND = 0x0111,
            SYSCOMMAND = 0x0112,
            TIMER = 0x0113,
            HSCROLL = 0x0114,
            VSCROLL = 0x0115,
            INITMENU = 0x0116,
            INITMENUPOPUP = 0x0117,
            MENUSELECT = 0x011F,
            MENUCHAR = 0x0120,
            ENTERIDLE = 0x0121,
            MENURBUTTONUP = 0x0122,
            MENUDRAG = 0x0123,
            MENUGETOBJECT = 0x0124,
            UNINITMENUPOPUP = 0x0125,
            MENUCOMMAND = 0x0126,
            QUERYUISTATE = 0x0129,
            MOUSEMOVE = 0x0200,
            LBUTTONDOWN = 0x0201,
            LBUTTONUP = 0x0202,
            LBUTTONDBLCLK = 0x0203,
            RBUTTONDOWN = 0x0204,
            RBUTTONUP = 0x0205,
            RBUTTONDBLCLK = 0x0206,
            MBUTTONDOWN = 0x0207,
            MBUTTONUP = 0x0208,
            MBUTTONDBLCLK = 0x0209,
            MOUSEWHEEL = 0x020A,
            XBUTTONDOWN = 0x020B,
            XBUTTONUP = 0x020C,
            XBUTTONDBLCLK = 0x020D,
            NCXBUTTONDOWN = 0x00AB,
            NCXBUTTONUP = 0x00AC,
            NCXBUTTONDBLCLK = 0x00AD,
            ENTERMENULOOP = 0x0211,
            EXITMENULOOP = 0x0212,
            USER = 0x0400,

            //
            // WebCam *.avi Capture Windows Messages
            // http://msdn.microsoft.com/en-us/library/windows/desktop/dd757695%28v=vs.85%29.aspx
            //
            CAP_SET_CALLBACK_FRAME = 0x0405, // sets a preview callback function in the application; AVICap calls this procedure when the capture window captures preview frames
            CAP_DRIVER_CONNECT = 0x040A, // connect to the video capture device
            CAP_DRIVER_DISCONNECT = 0x040B, // disconnect from the video capture device
            CAP_FILE_SET_CAPTURE_FILE = 0x0414, // set the capture file
            CAP_FILE_GET_CAPTURE_FILE = 0x0415, // get the capture file
            CAP_EDIT_COPY = 0x041E, // copy video capture to the clipboard
            CAP_SET_AUDIOFORMAT = 0x0423, // sets the audio format to use when performing streaming
            CAP_GET_AUDIOFORMAT = 0x0424, // obtains the audio format
            CAP_DLG_VIDEOFORMAT = 0x0429, // shows a dialog to adjust video format options
            CAP_DLG_VIDEOSOURCE = 0x042A, // shows a dialog to adjust the video source
            CAP_DLG_VIDEODISPLAY = 0x042B, // shows a dialog to adjust the video output
            CAP_GET_VIDEOFORMAT = 0x042C, // retrieves a copy of the video format in use
            CAP_SET_VIDEOFORMAT = 0x042D, // sets the format of captured video data
            CAP_DLG_VIDEOCOMPRESSION = 0x042E, // shows a dialog to set the compression option for use during capture
            CAP_SET_PREVIEW = 0x0432, // enables or disables preview mode; in preview mode, frames are transferred from the capture hardware to system memory and then displayed in the capture window using GDI functions
            CAP_SET_PREVIEWRATE = 0x0434, // sets the frame display rate in preview mode
            CAP_SET_SCALE = 0x0435, // enables or disables scaling of the preview video images; if scaling is enabled, the captured video frame is stretched to the dimensions of the capture window
            CAP_GET_STATUS = 0x0436, // retrieves the status of the capture window
            CAP_GRAB_FRAME = 0x043C, // retrieves and displays a single frame from the capture driver; after capture, overlay and preview are disabled
            CAP_GRAB_FRAME_NOSTOP = 0x043D, // fills the frame buffer with a single uncompressed frame from the capture device and displays it; unlike with the WM_CAP_GRAB_FRAME message, the state of overlay or preview is not altered by this message
            CAP_SEQUENCE = 0x043E, // start streaming the video to an *.avi file
            CAP_SET_SEQUENCE_SETUP = 0x0440, // set the parameters for streaming capture
            CAP_GET_SEQUENCE_SETUP = 0x0441, // get the parameters for streaming capture
            CAP_STOP = 0x0444 // stop sreaming the video to an *.avi file
        }

        public enum HT : int  // hit testing
        {
            ERROR = -2,
            TRANSPARENT = -1,
            NOWHERE = 0,
            CLIENT = 1,
            CAPTION = 2,
            SYSMENU = 3,
            GROWBOX = 4,
            SIZE = 4,
            MENU = 5,
            HSCROLL = 6,
            VSCROLL = 7,
            MINBUTTON = 8,
            MAXBUTTON = 9,
            LEFT = 10,
            RIGHT = 11,
            TOP = 12,
            TOPLEFT = 13,
            TOPRIGHT = 14,
            BOTTOM = 15,
            BOTTOMLEFT = 16,
            BOTTOMRIGHT = 17,
            BORDER = 18
        }

        public enum SC : uint
        {
            SIZE = 0xF000,
            SEPARATOR = 0xF00F,
            MOVE = 0xF010,
            MINIMIZE = 0xF020,
            MAXIMIZE = 0xF030,
            NEXTWINDOW = 0xF040,
            PREVWINDOW = 0xF050,
            CLOSE = 0xF060,
            VSCROLL = 0xF070,
            HSCROLL = 0xF080,
            MOUSEMENU = 0xF090,
            KEYMENU = 0xF100,
            ARRANGE = 0xF110,
            RESTORE = 0xF120,
            TASKLIST = 0xF130,
            SCREENSAVE = 0xF140,
            HOTKEY = 0xF150,
            DEFAULT = 0xF160,
            MONITORPOWER = 0xF170,
            CONTEXTHELP = 0xF180
        }

        #endregion

        #region Mouse and Keyboard Input

        //
        // SendInput function
        //
        [Flags]
        public enum MOUSEEVENTF : uint
        {
            MOVE = 0x01,
            LEFTDOWN = 0x02,
            LEFTUP = 0x04,
            RIGHTDOWN = 0x08,
            RIGHTUP = 0x10,
            MIDDLEDOWN = 0x20,
            MIDDLEUP = 0x40,
            XDOWN = 0x80,
            XUP = 0x100,
            WHEEL = 0x800,
            VIRTUALDESK = 0x4000,
            ABSOLUTE = 0x8000,
            XBUTTON1 = 0x01,
            XBUTTON2 = 0x02
        }

		/// <summary>
        /// Under x86, this structure is 24 bytes. Under x64, it is 32 bytes, although it
        /// could pack to 28 bytes, but it must end on a 8-byte boundary as x64 requires.
        /// </summary>
        [StructLayout(LayoutKind.Sequential)]
        public struct MOUSEINPUT
        {
			public const uint WHEEL_DELTA = 120;
			public static readonly MOUSEINPUT Empty;
            public int dx;
            public int dy;
            public uint mouseData;
            public MOUSEEVENTF dwFlags; // uint
            public uint time;
            public UIntPtr dwExtraInfo;
        }

		/// <summary>
		/// Virtual key codes
		/// </summary>
		/// <see cref="https://docs.microsoft.com/en-us/windows/win32/inputdev/virtual-key-codes"/>
		public enum VK : byte
		{
			// 0x00 : Unassigned

			/// <summary>
			/// Left mouse button
			/// </summary>
			LBUTTON = 0x01,

			/// <summary>
			/// Right mouse button
			/// </summary>
			RBUTTON = 0x02,

			/// <summary>
			/// Control-break processing
			/// </summary>
			CANCEL = 0x03,

			/// <summary>
			/// Middle mouse button (three-button mouse) - NOT contiguous with LBUTTON and RBUTTON
			/// </summary>
			MBUTTON = 0x04,

			/// <summary>
			/// Windows 2000/XP: X1 mouse button - NOT contiguous with LBUTTON and RBUTTON
			/// </summary>
			XBUTTON1 = 0x05,

			/// <summary>
			/// Windows 2000/XP: X2 mouse button - NOT contiguous with LBUTTON and RBUTTON
			/// </summary>
			XBUTTON2 = 0x06,

			// 0x07 : Undefined

			/// <summary>
			/// BACKSPACE key
			/// </summary>
			BACK = 0x08,

			/// <summary>
			/// TAB key
			/// </summary>
			TAB = 0x09,

			// 0x0A - 0x0B : Reserved

			/// <summary>
			/// CLEAR key
			/// </summary>
			CLEAR = 0x0C,

			/// <summary>
			/// ENTER key
			/// </summary>
			RETURN = 0x0D,

			// 0x0E - 0x0F : Undefined

			/// <summary>
			/// SHIFT key
			/// </summary>
			SHIFT = 0x10,

			/// <summary>
			/// CTRL key
			/// </summary>
			CONTROL = 0x11,

			/// <summary>
			/// ALT key
			/// </summary>
			MENU = 0x12,

			/// <summary>
			/// PAUSE key
			/// </summary>
			PAUSE = 0x13,

			/// <summary>
			/// CAPS LOCK key
			/// </summary>
			CAPITAL = 0x14,

			/// <summary>
			/// Input Method Editor (IME) Kana mode
			/// </summary>
			KANA = 0x15,

			/// <summary>
			/// IME Hanguel mode (maintained for compatibility; use HANGUL)
			/// </summary>
			HANGEUL = 0x15,

			/// <summary>
			/// IME Hangul mode
			/// </summary>
			HANGUL = 0x15,

			// 0x16 : Undefined

			/// <summary>
			/// IME Junja mode
			/// </summary>
			JUNJA = 0x17,

			/// <summary>
			/// IME final mode
			/// </summary>
			FINAL = 0x18,

			/// <summary>
			/// IME Hanja mode
			/// </summary>
			HANJA = 0x19,

			/// <summary>
			/// IME Kanji mode
			/// </summary>
			KANJI = 0x19,

			// 0x1A : Undefined

			/// <summary>
			/// ESC key
			/// </summary>
			ESCAPE = 0x1B,

			/// <summary>
			/// IME convert
			/// </summary>
			CONVERT = 0x1C,

			/// <summary>
			/// IME nonconvert
			/// </summary>
			NONCONVERT = 0x1D,

			/// <summary>
			/// IME accept
			/// </summary>
			ACCEPT = 0x1E,

			/// <summary>
			/// IME mode change request
			/// </summary>
			MODECHANGE = 0x1F,

			/// <summary>
			/// SPACEBAR
			/// </summary>
			SPACE = 0x20,

			/// <summary>
			/// PAGE UP key
			/// </summary>
			PRIOR = 0x21,

			/// <summary>
			/// PAGE DOWN key
			/// </summary>
			NEXT = 0x22,

			/// <summary>
			/// END key
			/// </summary>
			END = 0x23,

			/// <summary>
			/// HOME key
			/// </summary>
			HOME = 0x24,

			/// <summary>
			/// LEFT ARROW key
			/// </summary>
			LEFT = 0x25,

			/// <summary>
			/// UP ARROW key
			/// </summary>
			UP = 0x26,

			/// <summary>
			/// RIGHT ARROW key
			/// </summary>
			RIGHT = 0x27,

			/// <summary>
			/// DOWN ARROW key
			/// </summary>
			DOWN = 0x28,

			/// <summary>
			/// SELECT key
			/// </summary>
			SELECT = 0x29,

			/// <summary>
			/// PRINT key
			/// </summary>
			PRINT = 0x2A,

			/// <summary>
			/// EXECUTE key
			/// </summary>
			EXECUTE = 0x2B,

			/// <summary>
			/// PRINT SCREEN key
			/// </summary>
			SNAPSHOT = 0x2C,

			/// <summary>
			/// INS key
			/// </summary>
			INSERT = 0x2D,

			/// <summary>
			/// DEL key
			/// </summary>
			DELETE = 0x2E,

			/// <summary>
			/// HELP key
			/// </summary>
			HELP = 0x2F,

			/// <summary>
			/// 0 key
			/// </summary>
			D0 = 0x30,

			/// <summary>
			/// 1 key
			/// </summary>
			D1 = 0x31,

			/// <summary>
			/// 2 key
			/// </summary>
			D2 = 0x32,

			/// <summary>
			/// 3 key
			/// </summary>
			D3 = 0x33,

			/// <summary>
			/// 4 key
			/// </summary>
			D4 = 0x34,

			/// <summary>
			/// 5 key
			/// </summary>
			D5 = 0x35,

			/// <summary>
			/// 6 key
			/// </summary>
			D6 = 0x36,

			/// <summary>
			/// 7 key
			/// </summary>
			D7 = 0x37,

			/// <summary>
			/// 8 key
			/// </summary>
			D8 = 0x38,

			/// <summary>
			/// 9 key
			/// </summary>
			D9 = 0x39,

			// 0x3A - 0x40 : Undefined

			/// <summary>
			/// A key
			/// </summary>
			A = 0x41,

			/// <summary>
			/// B key
			/// </summary>
			B = 0x42,

			/// <summary>
			/// C key
			/// </summary>
			C = 0x43,

			/// <summary>
			/// D key
			/// </summary>
			D = 0x44,

			/// <summary>
			/// E key
			/// </summary>
			E = 0x45,

			/// <summary>
			/// F key
			/// </summary>
			F = 0x46,

			/// <summary>
			/// G key
			/// </summary>
			G = 0x47,

			/// <summary>
			/// H key
			/// </summary>
			H = 0x48,

			/// <summary>
			/// I key
			/// </summary>
			I = 0x49,

			/// <summary>
			/// J key
			/// </summary>
			J = 0x4A,

			/// <summary>
			/// K key
			/// </summary>
			K = 0x4B,

			/// <summary>
			/// L key
			/// </summary>
			L = 0x4C,

			/// <summary>
			/// M key
			/// </summary>
			M = 0x4D,

			/// <summary>
			/// N key
			/// </summary>
			N = 0x4E,

			/// <summary>
			/// O key
			/// </summary>
			O = 0x4F,

			/// <summary>
			/// P key
			/// </summary>
			P = 0x50,

			/// <summary>
			/// Q key
			/// </summary>
			Q = 0x51,

			/// <summary>
			/// R key
			/// </summary>
			R = 0x52,

			/// <summary>
			/// S key
			/// </summary>
			S = 0x53,

			/// <summary>
			/// T key
			/// </summary>
			T = 0x54,

			/// <summary>
			/// U key
			/// </summary>
			U = 0x55,

			/// <summary>
			/// V key
			/// </summary>
			V = 0x56,

			/// <summary>
			/// W key
			/// </summary>
			W = 0x57,

			/// <summary>
			/// X key
			/// </summary>
			X = 0x58,

			/// <summary>
			/// Y key
			/// </summary>
			Y = 0x59,

			/// <summary>
			/// Z key
			/// </summary>
			Z = 0x5A,

			/// <summary>
			/// Left Windows key (Microsoft Natural keyboard)
			/// </summary>
			LWIN = 0x5B,

			/// <summary>
			/// Right Windows key (Natural keyboard)
			/// </summary>
			RWIN = 0x5C,

			/// <summary>
			/// Applications key (Natural keyboard)
			/// </summary>
			APPS = 0x5D,

			// 0x5E : Reserved

			/// <summary>
			/// Computer Sleep key
			/// </summary>
			SLEEP = 0x5F,

			/// <summary>
			/// Numeric keypad 0 key
			/// </summary>
			NUMPAD0 = 0x60,

			/// <summary>
			/// Numeric keypad 1 key
			/// </summary>
			NUMPAD1 = 0x61,

			/// <summary>
			/// Numeric keypad 2 key
			/// </summary>
			NUMPAD2 = 0x62,

			/// <summary>
			/// Numeric keypad 3 key
			/// </summary>
			NUMPAD3 = 0x63,

			/// <summary>
			/// Numeric keypad 4 key
			/// </summary>
			NUMPAD4 = 0x64,

			/// <summary>
			/// Numeric keypad 5 key
			/// </summary>
			NUMPAD5 = 0x65,

			/// <summary>
			/// Numeric keypad 6 key
			/// </summary>
			NUMPAD6 = 0x66,

			/// <summary>
			/// Numeric keypad 7 key
			/// </summary>
			NUMPAD7 = 0x67,

			/// <summary>
			/// Numeric keypad 8 key
			/// </summary>
			NUMPAD8 = 0x68,

			/// <summary>
			/// Numeric keypad 9 key
			/// </summary>
			NUMPAD9 = 0x69,

			/// <summary>
			/// Multiply key
			/// </summary>
			MULTIPLY = 0x6A,

			/// <summary>
			/// Add key
			/// </summary>
			ADD = 0x6B,

			/// <summary>
			/// Separator key
			/// </summary>
			SEPARATOR = 0x6C,

			/// <summary>
			/// Subtract key
			/// </summary>
			SUBTRACT = 0x6D,

			/// <summary>
			/// Decimal key
			/// </summary>
			DECIMAL = 0x6E,

			/// <summary>
			/// Divide key
			/// </summary>
			DIVIDE = 0x6F,

			/// <summary>
			/// F1 key
			/// </summary>
			F1 = 0x70,

			/// <summary>
			/// F2 key
			/// </summary>
			F2 = 0x71,

			/// <summary>
			/// F3 key
			/// </summary>
			F3 = 0x72,

			/// <summary>
			/// F4 key
			/// </summary>
			F4 = 0x73,

			/// <summary>
			/// F5 key
			/// </summary>
			F5 = 0x74,

			/// <summary>
			/// F6 key
			/// </summary>
			F6 = 0x75,

			/// <summary>
			/// F7 key
			/// </summary>
			F7 = 0x76,

			/// <summary>
			/// F8 key
			/// </summary>
			F8 = 0x77,

			/// <summary>
			/// F9 key
			/// </summary>
			F9 = 0x78,

			/// <summary>
			/// F10 key
			/// </summary>
			F10 = 0x79,

			/// <summary>
			/// F11 key
			/// </summary>
			F11 = 0x7A,

			/// <summary>
			/// F12 key
			/// </summary>
			F12 = 0x7B,

			/// <summary>
			/// F13 key
			/// </summary>
			F13 = 0x7C,

			/// <summary>
			/// F14 key
			/// </summary>
			F14 = 0x7D,

			/// <summary>
			/// F15 key
			/// </summary>
			F15 = 0x7E,

			/// <summary>
			/// F16 key
			/// </summary>
			F16 = 0x7F,

			/// <summary>
			/// F17 key
			/// </summary>
			F17 = 0x80,

			/// <summary>
			/// F18 key
			/// </summary>
			F18 = 0x81,

			/// <summary>
			/// F19 key
			/// </summary>
			F19 = 0x82,

			/// <summary>
			/// F20 key
			/// </summary>
			F20 = 0x83,

			/// <summary>
			/// F21 key
			/// </summary>
			F21 = 0x84,

			/// <summary>
			/// F22 key
			/// </summary>
			F22 = 0x85,

			/// <summary>
			/// F23 key
			/// </summary>
			F23 = 0x86,

			/// <summary>
			/// F24 key
			/// </summary>
			F24 = 0x87,

			// 0x88 - 0x8F : Unassigned

			/// <summary>
			/// NUM LOCK key
			/// </summary>
			NUMLOCK = 0x90,

			/// <summary>
			/// SCROLL LOCK key
			/// </summary>
			SCROLL = 0x91,

			// 0x92 - 0x96 : OEM Specific

			// 0x97 - 0x9F : Unassigned

			//
			// L* & R* - left and right Alt, Ctrl and Shift virtual keys. Used only as parameters to 
			// GetAsyncKeyState() and GetKeyState(). No other API or message will distinguish left and 
			// right keys in this way.
			//

			/// <summary>
			/// Left SHIFT key - Used only as parameters to GetAsyncKeyState() and GetKeyState()
			/// </summary>
			LSHIFT = 0xA0,

			/// <summary>
			/// Right SHIFT key - Used only as parameters to GetAsyncKeyState() and GetKeyState()
			/// </summary>
			RSHIFT = 0xA1,

			/// <summary>
			/// Left CONTROL key - Used only as parameters to GetAsyncKeyState() and GetKeyState()
			/// </summary>
			LCONTROL = 0xA2,

			/// <summary>
			/// Right CONTROL key - Used only as parameters to GetAsyncKeyState() and GetKeyState()
			/// </summary>
			RCONTROL = 0xA3,

			/// <summary>
			/// Left MENU key - Used only as parameters to GetAsyncKeyState() and GetKeyState()
			/// </summary>
			LMENU = 0xA4,

			/// <summary>
			/// Right MENU key - Used only as parameters to GetAsyncKeyState() and GetKeyState()
			/// </summary>
			RMENU = 0xA5,

			/// <summary>
			/// Windows 2000/XP: Browser Back key
			/// </summary>
			BROWSER_BACK = 0xA6,

			/// <summary>
			/// Windows 2000/XP: Browser Forward key
			/// </summary>
			BROWSER_FORWARD = 0xA7,

			/// <summary>
			/// Windows 2000/XP: Browser Refresh key
			/// </summary>
			BROWSER_REFRESH = 0xA8,

			/// <summary>
			/// Windows 2000/XP: Browser Stop key
			/// </summary>
			BROWSER_STOP = 0xA9,

			/// <summary>
			/// Windows 2000/XP: Browser Search key
			/// </summary>
			BROWSER_SEARCH = 0xAA,

			/// <summary>
			/// Windows 2000/XP: Browser Favorites key
			/// </summary>
			BROWSER_FAVORITES = 0xAB,

			/// <summary>
			/// Windows 2000/XP: Browser Start and Home key
			/// </summary>
			BROWSER_HOME = 0xAC,

			/// <summary>
			/// Windows 2000/XP: Volume Mute key
			/// </summary>
			VOLUME_MUTE = 0xAD,

			/// <summary>
			/// Windows 2000/XP: Volume Down key
			/// </summary>
			VOLUME_DOWN = 0xAE,

			/// <summary>
			/// Windows 2000/XP: Volume Up key
			/// </summary>
			VOLUME_UP = 0xAF,

			/// <summary>
			/// Windows 2000/XP: Next Track key
			/// </summary>
			MEDIA_NEXT_TRACK = 0xB0,

			/// <summary>
			/// Windows 2000/XP: Previous Track key
			/// </summary>
			MEDIA_PREV_TRACK = 0xB1,

			/// <summary>
			/// Windows 2000/XP: Stop Media key
			/// </summary>
			MEDIA_STOP = 0xB2,

			/// <summary>
			/// Windows 2000/XP: Play/Pause Media key
			/// </summary>
			MEDIA_PLAY_PAUSE = 0xB3,

			/// <summary>
			/// Windows 2000/XP: Start Mail key
			/// </summary>
			LAUNCH_MAIL = 0xB4,

			/// <summary>
			/// Windows 2000/XP: Select Media key
			/// </summary>
			LAUNCH_MEDIA_SELECT = 0xB5,

			/// <summary>
			/// Windows 2000/XP: Start Application 1 key
			/// </summary>
			LAUNCH_APP1 = 0xB6,

			/// <summary>
			/// Windows 2000/XP: Start Application 2 key
			/// </summary>
			LAUNCH_APP2 = 0xB7,

			// 0xB8 - 0xB9 : Reserved

			/// <summary>
			/// Used for miscellaneous characters; it can vary by keyboard. Windows 2000/XP: For the US standard keyboard, 
			/// the ';:' key 
			/// </summary>
			OEM_1 = 0xBA,

			/// <summary>
			/// Windows 2000/XP: For any country/region, the '+' key
			/// </summary>
			OEM_PLUS = 0xBB,

			/// <summary>
			/// Windows 2000/XP: For any country/region, the ',' key
			/// </summary>
			OEM_COMMA = 0xBC,

			/// <summary>
			/// Windows 2000/XP: For any country/region, the '-' key
			/// </summary>
			OEM_MINUS = 0xBD,

			/// <summary>
			/// Windows 2000/XP: For any country/region, the '.' key
			/// </summary>
			OEM_PERIOD = 0xBE,

			/// <summary>
			/// Used for miscellaneous characters; it can vary by keyboard. Windows 2000/XP: For the US standard keyboard, the '/?' key 
			/// </summary>
			OEM_2 = 0xBF,

			/// <summary>
			/// Used for miscellaneous characters; it can vary by keyboard. Windows 2000/XP: For the US standard keyboard, the '`~' key 
			/// </summary>
			OEM_3 = 0xC0,

			// 0xC1 - 0xD7 : Reserved

			// 0xD8 - 0xDA : Unassigned

			/// <summary>
			/// Used for miscellaneous characters; it can vary by keyboard. Windows 2000/XP: For the US standard 
			/// keyboard, the '[{' key
			/// </summary>
			OEM_4 = 0xDB,

			/// <summary>
			/// Used for miscellaneous characters; it can vary by keyboard. Windows 2000/XP: For the US standard 
			/// keyboard, the '\|' key
			/// </summary>
			OEM_5 = 0xDC,

			/// <summary>
			/// Used for miscellaneous characters; it can vary by keyboard. Windows 2000/XP: For the US standard 
			/// keyboard, the ']}' key
			/// </summary>
			OEM_6 = 0xDD,

			/// <summary>
			/// Used for miscellaneous characters; it can vary by keyboard. Windows 2000/XP: For the US standard 
			/// keyboard, the 'single-quote/double-quote' key
			/// </summary>
			OEM_7 = 0xDE,

			/// <summary>
			/// Used for miscellaneous characters; it can vary by keyboard.
			/// </summary>
			OEM_8 = 0xDF,

			// 0xE0 : Reserved

			// 0xE1 : OEM Specific

			/// <summary>
			/// Windows 2000/XP: Either the angle bracket key or the backslash key on the RT 102-key keyboard
			/// </summary>
			OEM_102 = 0xE2,

			// (0xE3-E4) : OEM specific

			/// <summary>
			/// Windows 95/98/Me, Windows NT 4.0, Windows 2000/XP: IME PROCESS key
			/// </summary>
			PROCESSKEY = 0xE5,

			// 0xE6 : OEM specific

			/// <summary>
			/// Windows 2000/XP: Used to pass Unicode characters as if they were keystrokes. The PACKET key is the 
			/// low word of a 32-bit Virtual Key value used for non-keyboard input methods. For more information, 
			/// see Remark in KEYBDINPUT, SendInput, WM_KEYDOWN, and WM_KEYUP
			/// </summary>
			PACKET = 0xE7,

			// 0xE8 : Unassigned

			// 0xE9-F5 : OEM specific

			/// <summary>
			/// Attn key
			/// </summary>
			ATTN = 0xF6,

			/// <summary>
			/// CrSel key
			/// </summary>
			CRSEL = 0xF7,

			/// <summary>
			/// ExSel key
			/// </summary>
			EXSEL = 0xF8,

			/// <summary>
			/// Erase EOF key
			/// </summary>
			EREOF = 0xF9,

			/// <summary>
			/// Play key
			/// </summary>
			PLAY = 0xFA,

			/// <summary>
			/// Zoom key
			/// </summary>
			ZOOM = 0xFB,

			/// <summary>
			/// Reserved
			/// </summary>
			NONAME = 0xFC,

			/// <summary>
			/// PA1 key
			/// </summary>
			PA1 = 0xFD,

			/// <summary>
			/// Clear key
			/// </summary>
			OEM_CLEAR = 0xFE

			// 0xFF : Unassigned
		}

		[Flags]
        public enum KEYEVENTF : uint
        {
            EXTENDEDKEY = 0x01,
            KEYUP = 0x02,
            UNICODE = 0x04,
            SCANCODE = 0x08
        }

        /// <summary>
        /// Under x86, this structure is 16 bytes. Under x64, it is 24 bytes, even though
        /// it would pack to 20 bytes; but it must end on a 8-byte boundary as x64 requires. 
        /// </summary>
        [StructLayout(LayoutKind.Sequential)]
        public struct KEYBDINPUT
        {
            public static readonly KEYBDINPUT Empty;
            public ushort wVk;
            public ushort wScan;
            public KEYEVENTF dwFlags; // uint
            public uint time;
            public UIntPtr dwExtraInfo;
        }

        /// <summary>
        /// This structure is 8 bytes on both x86 and x64, as it uses no IntPtrs or UIntPtrs.
        /// </summary>
        [StructLayout(LayoutKind.Sequential)]
        public struct HARDWAREINPUT
        {
            public static readonly HARDWAREINPUT Empty;
            public uint uMsg;
            public ushort wParamL;
            public ushort wParamH;
        }

        [Flags]
        public enum INPUTF : uint
        {
            MOUSE = 0x00,
            KEYBOARD = 0x01,
            HARDWARE = 0x02
        }

        [StructLayout(LayoutKind.Explicit, Size = 28)]
        public struct INPUTx86 // a C union http://msdn.microsoft.com/en-us/library/ms646270%28v=vs.85%29.aspx
        {
            public static readonly INPUTx86 Empty;	            // 0 bytes (static)
            [FieldOffset(00)] public INPUTF type;				// 4 bytes (uint)
            // <union>
            [FieldOffset(04)] public MOUSEINPUT mi;				// 24 bytes (biggest in union)
            [FieldOffset(04)] public KEYBDINPUT ki;				// 16 bytes
            [FieldOffset(04)] public HARDWAREINPUT hi;			// 8 bytes
            // </union>
        }

        /// <summary>
        /// For the x64 version of the INPUT structure, even though the INPUTF flag is effectively a uint and therefore
        /// only 4 bytes even on x64, the offset of the unioned structs must be at 08, not 04, so that they begin on an
        /// 8-byte boundary as required by x64.
        /// </summary>
        [StructLayout(LayoutKind.Explicit, Size = 40)]
        public struct INPUTx64 // a C union http://msdn.microsoft.com/en-us/library/ms646270%28v=vs.85%29.aspx
        {
            public static readonly INPUTx64 Empty;	            // 0 bytes (static)
            [FieldOffset(00)] public INPUTF type;				// 4 bytes (uint)
            // <union>
            [FieldOffset(08)] public MOUSEINPUT mi;				// 32 bytes (biggest in union)
            [FieldOffset(08)] public KEYBDINPUT ki;				// 24 bytes
            [FieldOffset(08)] public HARDWAREINPUT hi;			// 8 bytes
            // </union>
        }

        // for the cbSize parameter, use System.Runtime.InteropServices.Marshal.SizeOf(i)
        [DllImport("user32.dll")]
        public static extern uint SendInput(uint nInputs, ref INPUTx86 pInputs, int cbSize); // x86

		[DllImport("user32.dll")]
		public static extern uint SendInput(uint nInputs, INPUTx86[] pInputs, int cbSize); // x86

		[DllImport("user32.dll")]
        public static extern uint SendInput(uint nInputs, ref INPUTx64 pInputs, int cbSize); // x64

		[DllImport("user32.dll")]
		public static extern uint SendInput(uint nInputs, INPUTx64[] pInputs, int cbSize); // x64

		[DllImport("user32.dll")]
        public static extern void keybd_event(byte bVk, byte bScan, uint dwFlags, UIntPtr dwExtraInfo); // old Win16 call

        [DllImport("user32.dll")]
        public static extern void mouse_event(uint dwFlags, uint dx, uint dy, uint dwData, UIntPtr dwExtraInfo); // old Win16 call

		[DllImport("user32")]
		public static extern int GetDoubleClickTime();

		[DllImport("user32")]
		public static extern int ToAscii(uint uVirtKey, uint uScanCode, byte[] lpKeyState, out short lpChar, uint uFlags);

		[DllImport("user32")]
		public static extern bool GetKeyboardState(byte[] lpKeyState);

		[DllImport("user32.dll")]
		public static extern short GetKeyState(int nVirtKey);

		[DllImport("user32.dll")]
        public static extern bool BlockInput(bool fBlockIt);

        [DllImport("user32.dll")]
        public static extern int ShowCursor(bool bShow);

		[DllImport("user32.dll")]
		public static extern bool SetCursorPos(int x, int y);

		[DllImport("user32.dll")]
		public static extern bool GetCursorPos(out POINT lpPoint);

		[DllImport("user32.dll")]
		public static extern bool ClipCursor(ref RECT lpRect);

		#endregion

		#region Windows Hooks

		//
		// Windows Hooks
		//
		public enum WH : int // http://www.cs.unc.edu/Research/assist/doc/pyhook/public/pyHook.HookManager.HookConstants-class.html
        {
            CALLWNDPROC = 4,
            CALLWNDPROCRET = 12,
            CBT = 5,
            DEBUG = 9,
            FOREGROUNDIDLE = 11,
            GETMESSAGE = 3,
            HARDWARE = 8,
            JOURNALPLAYBACK = 1,
            JOURNALRECORD = 0,
            KEYBOARD = 2,
            KEYBOARD_LL = 13,
            MAX = 15,
            MIN = -1,
            MOUSE = 7,
            MOUSE_LL = 14,
            MSGFILTER = -1,
            SHELL = 10,
            SYSMSGFILTER = 6
        }

        [Flags]
        public enum MK : uint // key state masks for mouse messages
        {
            LBUTTON = 0x01,
            RBUTTON = 0x02,
            SHIFT = 0x04,
            CONTROL = 0x08,
            MBUTTON = 0x10,
            XBUTTON1 = 0x20,
            XBUTTON2 = 0x40
        }

        [Flags]
        public enum KF : uint // keyboard flags
        {
            EXTENDED = 0x100,
            ALTDOWN = 0x2000,
            DLGMODE = 0x00, // TODO
            MENUMODE = 0x00, // TODO
            REPEAT = 0x00, // TODO
            UP = 0x8000
        }

        [Flags]
        public enum LLMHF : uint // low level mouse hook flags
        {
            INJECTED = 0x0001
        }

        [Flags]
        public enum LLKHF : uint // low level keyboard hook flags
        {
            EXTENDED = (KF.EXTENDED >> 8), // 0x0001,
            INJECTED = 0x0010,             // 0x0010,
            ALTDOWN = (KF.ALTDOWN >> 8),   // 0x0020,
            UP = (KF.UP >> 8)              // 0x0080,
        }

        [StructLayout(LayoutKind.Sequential)]
        public struct POINT
        {
            public static readonly POINT Empty;
            public int x;
            public int y;
        }

        [StructLayout(LayoutKind.Sequential)]
        public struct MSLLHOOKSTRUCT // http://msdn.microsoft.com/en-us/library/ms644970%28VS.85%29.aspx
        {
            public static readonly MSLLHOOKSTRUCT Empty;
            public POINT pt;
            public uint mouseData;
            public uint flags; // LLMHF
            public uint time;
            public UIntPtr dwExtraInfo;
        }

        [StructLayout(LayoutKind.Sequential)]
        public struct KBDLLHOOKSTRUCT // http://msdn.microsoft.com/en-us/library/ms644967%28VS.85%29.aspx
        {
            public static readonly KBDLLHOOKSTRUCT Empty;
            public uint vkCode;
            public uint scanCode;
            public uint flags; // LLKHF
            public uint time;
            public UIntPtr dwExtraInfo;
        }

        public delegate IntPtr WindowsHookProc(int nCode, int wParam, IntPtr lParam);

        [DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern IntPtr GetModuleHandle(string lpModuleName);

        [DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern IntPtr SetWindowsHookEx(int idHook, WindowsHookProc lpfn, IntPtr hMod, uint dwThreadId);

        [DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern bool UnhookWindowsHookEx(IntPtr hhk);

        [DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern IntPtr CallNextHookEx(IntPtr hhk, int nCode, int wParam, IntPtr lParam);

        #endregion

        #region Raw Input Devices

        //
        // Raw Input (http://msdn.microsoft.com/en-us/library/ms645543%28v=VS.85%29.aspx)
        //
        [Flags]
        public enum RIDEV : uint
        {
            APPKEYS = 0x00000400,
            CAPTUREMOUSE = 0x00000200,
            DEVNOTIFY = 0x00002000,
            EXCLUDE = 0x00000010,
            EXINPUTSINK = 0x00001000,
            INPUTSINK = 0x00000100,
            NOHOTKEYS = 0x00000200,
            NOLEGACY = 0x00000030,
            PAGEONLY = 0x00000020,
            REMOVE = 0x00000001
        }

        /// <summary>
        /// On x86, this structure is 12 bytes; on x64, it is 16 bytes.
        /// </summary>
        [StructLayout(LayoutKind.Sequential)]
        public struct RAWINPUTDEVICE // http://msdn.microsoft.com/en-us/library/ms645565%28v=VS.85%29.aspx
        {
            public static readonly RAWINPUTDEVICE Empty; // 0 bytes
            public ushort usUsagePage; // 2 bytes
            public ushort usUsage; // 2 bytes
            public uint dwFlags; // RIDEV, 4 bytes
            public IntPtr hwndTarget; // 4 byes (x86) or 8 bytes (x64)
        }

        [DllImport("user32.dll")]
        public static extern bool RegisterRawInputDevices(RAWINPUTDEVICE[] pRawInputDevices, uint uiNumDevices, uint cbSize);

        [DllImport("user32.dll")]
        public static extern uint GetRegisteredRawInputDevices(IntPtr pRawInputDevices, ref uint puiNumDevices, uint cbSize);

        public enum RIDI : uint // http://msdn.microsoft.com/en-us/library/ms645597%28v=VS.85%29.aspx
        {
            PREPARSEDDATA = 0x20000005,
            DEVICENAME =    0x20000007,
            DEVICEINFO =    0x2000000B
        }

        /// <summary>
        /// For the mouse, the Usage Page is 1 and the Usage is 2. On x86 platforms and x64 platforms, this structure
        /// is 16 bytes, as it must end on a 4-byte or 8-byte boundary, respectively.
        /// </summary>
        [StructLayout(LayoutKind.Sequential)]
        public struct RID_DEVICE_INFO_MOUSE // http://msdn.microsoft.com/en-us/library/ms645589%28v=VS.85%29.aspx
        {
            public static readonly RID_DEVICE_INFO_MOUSE Empty; // 0 bytes
            public uint dwId; // 4 bytes
            public uint dwNumberOfButtons; // 4 bytes
            public uint dwSampleRate; // 4 bytes
            public bool fHasHorizontalWheel; // 1 byte
        }

        /// <summary>
        /// For the keyboard, the Usage Page is 1 and the Usage is 6. On x86 platforms and x64 platforms, this structure
        /// is 24 bytes, which ends on both a 4-byte and 8-byte boundary.
        /// </summary>
        [StructLayout(LayoutKind.Sequential)]
        public struct RID_DEVICE_INFO_KEYBOARD // http://msdn.microsoft.com/en-us/library/ms645587%28v=VS.85%29.aspx
        {
            public static readonly RID_DEVICE_INFO_KEYBOARD Empty; // 0 bytes
            public uint dwType; // 4 bytes
            public uint dwSubType; // 4 bytes
            public uint dwKeyboardMode; // 4 bytes
            public uint dwNumberOfFunctionKeys; // 4 bytes
            public uint dwNumberOfIndicators; // 4 bytes
            public uint dwNumberOfKeysTotal; // 4 bytes
        }

        /// <summary>
        /// For joysticks, the Usage Page is 1 and the Usage is 4; for gamepads, it is 1 & 5; for keypads, it is 1 & 7; for a 
        /// multi-axis controller, it is 1 & 8. See http://www.usb.org/developers/devclass_docs/Hut1_12v2.pdf page 26 for more.
        /// </summary>
        [StructLayout(LayoutKind.Sequential)]
        public struct RID_DEVICE_INFO_HID // http://msdn.microsoft.com/en-us/library/ms645584%28v=VS.85%29.aspx
        {
            public static readonly RID_DEVICE_INFO_HID Empty; // 0 bytes
            public uint dwVendorId; // 4 bytes
            public uint dwProductId; // 4 bytes
            public uint dwVersionNumber; // 4 bytes
            public ushort usUsagePage; // 2 bytes
            public ushort usUsage; // 2 bytes
        }

        [Flags]
        public enum RIM : uint // http://msdn.microsoft.com/en-us/library/ms645581%28v=VS.85%29.aspx
        {
            INPUT = 0, // from WM_INPUT
            INPUTSINK = 1, // from WM_INPUT
            TYPEMOUSE = 0,
            TYPEKEYBOARD = 1,
            TYPEHID = 2
        }

        /// <summary>
        /// This structure works as defined on both x86 and x64.
        /// </summary>
        [StructLayout(LayoutKind.Explicit, Size = 32)]
        public struct RID_DEVICE_INFO // a C union http://msdn.microsoft.com/en-us/library/ms645581%28v=VS.85%29.aspx
        {
            public static readonly RID_DEVICE_INFO Empty;               // 0 bytes (static)
            [FieldOffset(0)] public uint cbSize;                        // 4 bytes
            [FieldOffset(4)] public uint dwType;                        // RIM
            // <union>
            [FieldOffset(8)] public RID_DEVICE_INFO_MOUSE mouse;	    // 13 bytes (padded to 16)
            [FieldOffset(8)] public RID_DEVICE_INFO_KEYBOARD keyboard;	// 24 bytes (biggest in union)
            [FieldOffset(8)] public RID_DEVICE_INFO_HID hid;			// 16 bytes
            // </union>
        }

        [DllImport("user32.dll")]
        public static extern uint GetRawInputDeviceInfo(IntPtr hDevice, uint uiCommand, IntPtr pData, ref uint pcbSize);

        [StructLayout(LayoutKind.Sequential)]
        public struct RAWINPUTDEVICELIST
        {
            public static readonly RAWINPUTDEVICELIST Empty;
            public IntPtr hDevice;
            public uint dwType; // RIM
        }

        [DllImport("user32.dll")]
        public static extern uint GetRawInputDeviceList(RAWINPUTDEVICELIST[] pRawInputDeviceList, ref uint puiNumDevices, uint cbSize);

        public enum RID : uint // http://msdn.microsoft.com/en-us/library/ms645596%28v=VS.85%29.aspx
        {
            HEADER = 0x10000005, // Get the header information from the RAWINPUT structure.
            INPUT = 0x10000003 // Get the raw data from the RAWINPUT structure.
        }

        /// <summary>
        /// Under x86, this structure is 16 bytes. Under x64, it is 24 bytes so that it ends on an 8-byte boundary.
        /// That is, 4 bytes of padding are added by the system at the end of the structure.
        /// </summary>
        [StructLayout(LayoutKind.Sequential)]
        public struct RAWINPUTHEADER // http://msdn.microsoft.com/en-us/library/ms645571%28v=VS.85%29.aspx
        {
            public static readonly RAWINPUTHEADER Empty;
            public uint dwType; // RIM
            public uint dwSize; // The size, in bytes, of the entire input packet of data. This includes RAWINPUT plus possible extra input reports in the RAWHID variable length array.
            public IntPtr hDevice; // A handle to the device generating the raw input data. 
            public uint wParam; // The value passed in the wParam parameter of the WM_INPUT message.
        }

        [Flags]
        public enum MOUSE : uint // http://msdn.microsoft.com/en-us/library/ms645578%28v=VS.85%29.aspx
        {
            MOVE_RELATIVE = 0x00,
            MOVE_ABSOLUTE = 0x01,
            VIRTUAL_DESKTOP = 0x02,
            ATTRIBUTES_CHANGED = 0x04
        }

        [Flags]
        public enum RI_MOUSE : ushort  // http://msdn.microsoft.com/en-us/library/ms645578%28v=VS.85%29.aspx
        {
            LEFT_BUTTON_DOWN = 0x0001,
            LEFT_BUTTON_UP = 0x0002,
            MIDDLE_BUTTON_DOWN = 0x0010,
            MIDDLE_BUTTON_UP = 0x0020,
            RIGHT_BUTTON_DOWN = 0x0004,
            RIGHT_BUTTON_UP = 0x0008,
            BUTTON_1_DOWN = LEFT_BUTTON_DOWN,
            BUTTON_1_UP = LEFT_BUTTON_UP,
            BUTTON_2_DOWN = RIGHT_BUTTON_DOWN,
            BUTTON_2_UP = RIGHT_BUTTON_UP,
            BUTTON_3_DOWN = MIDDLE_BUTTON_DOWN,
            BUTTON_3_UP = MIDDLE_BUTTON_UP,
            BUTTON_4_DOWN = 0x0040, // XBUTTON1
            BUTTON_4_UP = 0x0080, // XBUTTON1
            BUTTON_5_DOWN = 0x100, // XBUTTON2
            BUTTON_5_UP = 0x0200, // XBUTTON2
            WHEEL = 0x0400, // The wheel delta is stored in usButtonData of RAWMOUSE.
            HORIZONTAL_WHEEL = 0x0800 // The wheel delta is stored in usButtonData of RAWMOUSE.
        }

        /// <summary>
        /// This structure works as defined on both x86 and x64.
        /// </summary>
        [StructLayout(LayoutKind.Explicit, Size = 24)]
        public struct RAWMOUSE // a C union
        {
            public static readonly RAWMOUSE Empty; // 0 bytes (static)
            [FieldOffset(00)] public ushort usFlags; // 2 bytes but next field needs to be on 4-byte boundary
            
            // <union>
            [FieldOffset(04)] public uint ulButtons; // 4 bytes (reserved)
            // <struct>
            [FieldOffset(04)] public ushort usButtonFlags; // 2 bytes (RI_MOUSE)
            [FieldOffset(06)] public ushort usButtonData;  // 2 bytes
            // </struct>
            // </union>

            [FieldOffset(08)] public uint ulRawButtons; // 4 bytes
            [FieldOffset(12)] public int lLastX; // 4 bytes
            [FieldOffset(16)] public int lLastY; // 4 bytes
            [FieldOffset(20)] public uint ulExtraInformation; // 4 bytes
        }

        [Flags]
        public enum RI_KEY : ushort // http://msdn.microsoft.com/en-us/library/ms645575%28v=VS.85%29.aspx
        {
            MAKE = 0, // The key is down.
            BREAK = 1, // The key is up.
            E0 = 2, // This is the left version of Alt, Shift, or Ctrl.
            E1 = 4 // This is the right version of Alt, Shift, or Ctrl.
        }

        /// <summary>
        /// Under x86 and x64, both, this structure is 16 bytes.
        /// </summary>
        [StructLayout(LayoutKind.Sequential)]
        public struct RAWKEYBOARD
        {
            public static readonly RAWKEYBOARD Empty;
            public ushort MakeCode;
            public ushort Flags; // RI_KEY
            public ushort Reserved;
            public ushort VKey; // VK virtual key codes
            public uint Message; // The corresponding window message, for example WM_KEYDOWN, WM_SYSKEYDOWN, and so forth.
            public uint ExtraInformation;
        }

        /// <summary>
        /// Under x86, this structure is 12 bytes. Under x64, it is 16 bytes.
        /// </summary>
        [StructLayout(LayoutKind.Sequential)]
        public struct RAWHID
        {
            public static readonly RAWHID Empty;
            public uint dwSizeHid;
            public uint dwCount;
            public IntPtr bRawData;
        }

        [StructLayout(LayoutKind.Explicit, Size = 40)]
        public struct RAWINPUTx86 // a C union http://msdn.microsoft.com/en-us/library/ms645562%28v=VS.85%29.aspx
        {
            public static readonly RAWINPUTx86 Empty; // 0 bytes
            [FieldOffset(00)] public RAWINPUTHEADER header; // 16 bytes
            // <union>
            [FieldOffset(16)] public RAWMOUSE mouse; // 24 bytes
            [FieldOffset(16)] public RAWKEYBOARD keyboard; // 16 bytes
            [FieldOffset(16)] public RAWHID hid; // 12 bytes
            // </union>
        }

        /// <summary>
        /// On x64, due to padding, the size of this structure is 72 bytes, even though it fits in 68.
        /// </summary>
        [StructLayout(LayoutKind.Explicit, Size = 48)]
        public struct RAWINPUTx64 // a C union http://msdn.microsoft.com/en-us/library/ms645562%28v=VS.85%29.aspx
        {
            public static readonly RAWINPUTx64 Empty; // 0 bytes
            [FieldOffset(00)] public RAWINPUTHEADER header; // 24 bytes
            // <union>
            [FieldOffset(24)] public RAWMOUSE mouse; // 24 bytes
            [FieldOffset(24)] public RAWKEYBOARD keyboard; // 16 bytes
            [FieldOffset(24)] public RAWHID hid; // 16 bytes
            // </union>
        }

        [DllImport("user32.dll")]
        public static extern uint GetRawInputData(IntPtr hRawInput, uint uiCommand, IntPtr pData, ref uint pcbSize, uint cbSizeHeader);

        [DllImport("user32.dll")]
        public static extern uint DefRawInputProc(IntPtr paRawInput, int nInput, uint cbSizeHeader);

        #endregion

        #region Animation

        //
        // Animation
        //
        public enum IDANI
        {
            OPEN = 0x01,
            CLOSE = 0x02,
            CAPTION = 0x03
        }

        [StructLayout(LayoutKind.Sequential)]
        public struct RECT
        {
            public static readonly RECT Empty;
            public int left;
            public int top;
            public int right;
            public int bottom;
        }

        [DllImport("user32.dll")]
        public static extern bool IsRectEmpty(ref RECT lprc);

        [DllImport("user32.dll")]
        public static extern bool DrawAnimatedRects(IntPtr hWnd, int idAni, ref RECT lprcFrom, ref RECT lprcTo);

        [Flags]
        public enum AW : uint
        {
            HOR_POSITIVE = 0x01,
            HOR_NEGATIVE = 0x02,
            VER_POSITIVE = 0x04,
            VER_NEGATIVE = 0x08,
            CENTER = 0x10,
            HIDE = 0x10000,
            ACTIVATE = 0x20000,
            SLIDE = 0x40000,
            BLEND = 0x80000
        }

        [DllImport("user32.dll")]
        public static extern bool AnimateWindow(IntPtr hWnd, uint dwTime, uint dwFlags);

        #endregion

        #region Toolbars

        //
        // Toolbars
        //
        public enum TB : uint
        {
            BUTTONCOUNT = WM.USER + 24,
            GETITEMRECT = WM.USER + 29,
            GETBUTTONTEXT = WM.USER + 45,
            GETRECT = WM.USER + 51
        }

		#endregion

		#region TrackBars

		/// <summary>
		/// TrackBar control messages. 
		/// Values obtained from https://github.com/tpn/winsdk-10/blob/master/Include/10.0.16299.0/um/CommCtrl.h.
		/// See https://docs.microsoft.com/en-us/windows/win32/controls/bumper-trackbar-control-reference-messages.
		/// </summary>
		public enum TBM : uint
		{
			GETPOS = WM.USER + 0,
			GETRANGEMIN = WM.USER + 1,
			GETRANGEMAX = WM.USER + 2,
			GETTIC = WM.USER + 3,
			SETTIC = WM.USER + 4,
			SETPOS = WM.USER + 5,
			SETRANGE = WM.USER + 6,
			SETRANGEMIN = WM.USER + 7,
			SETRANGEMAX = WM.USER + 8,
			CLEARTICS = WM.USER + 9,
			SETSEL = WM.USER + 10,
			SETSELSTART = WM.USER + 11,
			SETSELEND = WM.USER + 12,
			GETPTICS = WM.USER + 14,
			GETTICPOS = WM.USER + 15,
			GETNUMTICS = WM.USER + 16,
			GETSELSTART = WM.USER + 17,
			GETSELEND = WM.USER + 18,
			CLEARSEL = WM.USER + 19,
			SETTICFREQ = WM.USER + 20,
			SETPAGESIZE = WM.USER + 21,
			GETPAGESIZE = WM.USER + 22,
			SETLINESIZE = WM.USER + 23,
			GETLINESIZE = WM.USER + 24,
			GETTHUMBRECT = WM.USER + 25,
			GETCHANNELRECT = WM.USER + 26,
			SETTHUMBLENGTH = WM.USER + 27,
			GETTHUMBLENGTH = WM.USER + 28,
			SETTOOLTIPS = WM.USER + 29,
			GETTOOLTIPS = WM.USER + 30,
			SETTIPSIDE = WM.USER + 31,
			SETBUDDY = WM.USER + 32,
			GETBUDDY = WM.USER + 33,
			SETPOSNOTIFY = WM.USER + 34
		}

		/// <summary>
		/// For use with TBM_GETTHUMBRECT and TBM_GETCHANNELRECT. Pass 0u for wParam.
		/// See https://docs.microsoft.com/en-us/windows/win32/controls/tbm-getthumbrect.
		/// </summary>
		[DllImport("user32.dll")]
		public static extern int SendMessage(IntPtr hWnd, uint Msg, uint wParam, ref RECT lParam);

		#endregion
		
		#region Windows

		//
		// Window Management
		//
		public enum GW : uint
        {
            HWNDFIRST = 0,
            HWNDLAST = 1,
            HWNDNEXT = 2,
            HWNDPREV = 3,
            OWNER = 4,
            CHILD = 5,
            ENABLEDPOPUP = 6,
            MAX = 6
        }

        [DllImport("user32.dll")]
        public static extern IntPtr GetWindow(IntPtr hWnd, uint uCmd);

        [DllImport("user32.dll")]
        public static extern IntPtr GetNextWindow(IntPtr hWnd, uint wCmd);

		public enum GA : uint
		{
			PARENT = 1,
			ROOT = 2,
			ROOTOWNER = 3
		}

		[DllImport("user32.dll")]
		public static extern IntPtr GetAncestor(IntPtr hWnd, uint gaFlags);

		[DllImport("user32.dll")]
		public static extern IntPtr GetParent(IntPtr hWnd);

		[DllImport("user32.dll")]
		public static extern IntPtr GetDesktopWindow();

		[DllImport("user32.dll")]
		public static extern IntPtr GetTopWindow(IntPtr hWnd);

		[DllImport("user32.dll")]
        public static extern bool IsWindow(IntPtr hWnd);

        [DllImport("user32.dll")]
        public static extern bool IsIconic(IntPtr hWnd); // minimized

        [DllImport("user32.dll")]
        public static extern bool IsZoomed(IntPtr hWnd); // maximized

        [DllImport("user32.dll")]
        public static extern bool DestroyWindow(IntPtr hWnd);

        public enum SW : int
        {
            HIDE = 0,
            SHOWNORMAL = 1,
            NORMAL = 1,
            SHOWMINIMIZED = 2,
            SHOWMAXIMIZED = 3,
            MAXIMIZE = 3,
            SHOWNOACTIVATE = 4,
            SHOW = 5,
            MINIMIZE = 6,
            SHOWMINNOACTIVE = 7,
            SHOWNA = 8,
            RESTORE = 9,
            SHOWDEFAULT = 10,
            FORCEMINIMIZE = 11,
            MAX = 11
        }

        [DllImport("user32.dll")]
        public static extern bool ShowWindow(IntPtr hWnd, int nCmdShow);

        [DllImport("User32.dll")]
        public static extern bool ShowWindowAsync(IntPtr hWnd, int nCmdShow);

        [DllImport("user32.dll")]
        public static extern IntPtr FindWindow(string lpClassName, string lpWindowName);

        [DllImport("user32.dll")]
        public static extern IntPtr FindWindowEx(IntPtr hWndParent, IntPtr hWndChildAfter, string lpszClass, string lpszWindow);

		[DllImport("user32.dll")]
        public static extern IntPtr WindowFromPoint(POINT point);

		[DllImport("user32.dll")]
        public static extern IntPtr ChildWindowFromPoint(IntPtr hWndParent, POINT point);

        [DllImport("user32.dll")]
        public static extern IntPtr RealChildWindowFromPoint(IntPtr hWndParent, POINT ptParentClientCoords);

		[DllImport("user32.dll")]
		public static extern bool IsWindowVisible(IntPtr hWnd);

		[DllImport("user32.dll")]
        public static extern int GetWindowText(IntPtr hWnd, StringBuilder lpString, int nMaxCount);

		[DllImport("user32.dll")]
		public static extern int GetWindowTextLength(IntPtr hWnd);

		[DllImport("user32.dll")]
        public static extern bool GetWindowRect(IntPtr hWnd, ref RECT lpRect);


		//
		// Window Position and Order
		//
		
		/// <summary>
		/// Defines HWND_BOTTOM, HWND_NOTOPMOST, etc. for use as the second parameter to SetWindowPos. 
		/// Pass these as the hWndInsertAfter parameter if an actual window handle is not used there. 
		/// To pass these integer constants, use an IntPtr instance, e.g., new IntPtr((int) Win32.HWND.TOPMOST).
		/// </summary>
		public enum HWND : int // http://msdn.microsoft.com/en-us/library/ms633545%28VS.85%29.aspx
		{
			NOTOPMOST = -2,
			TOPMOST = -1,
			TOP = 0,
			BOTTOM = 1,
			BROADCAST = 0xFFFF
		}

		[Flags]
		public enum SWP : uint // http://msdn.microsoft.com/en-us/library/ms633545%28VS.85%29.aspx
		{
			ASYNCWINDOWPOS = 0x4000,
			DEFERERASE = 0x2000,
			DRAWFRAME = 0x0020,
			FRAMECHANGED = 0x0020,
			HIDEWINDOW = 0x0080,
			NOACTIVATE = 0x0010,
			NOCOPYBITS = 0x0100,
			NOMOVE = 0x0002,
			NOOWNERZORDER = 0x0200,
			NOREDRAW = 0x0008,
			NOREPOSITION = 0x0200,
			NOSENDCHANGING = 0x0400,
			NOSIZE = 0x0001,
			NOZORDER = 0x0004,
			SHOWWINDOW = 0x0040,
		}

		[DllImport("user32.dll")]
		public static extern bool SetWindowPos(IntPtr hWnd, IntPtr hWndInsertAfter, int X, int Y, int cx, int cy, uint uFlags);

		[DllImport("user32.dll")]
		public static extern bool BringWindowToTop(IntPtr hWnd);

		[DllImport("user32.dll")]
		public static extern bool SetForegroundWindow(IntPtr hWnd);

		[DllImport("user32.dll")]
		public static extern IntPtr GetForegroundWindow();

		[DllImport("user32.dll")]
		public static extern IntPtr SetActiveWindow(IntPtr hWnd);

		[DllImport("user32.dll")]
		public static extern IntPtr GetActiveWindow();

		[DllImport("user32.dll")]
		public static extern IntPtr SetFocus(IntPtr hWnd);

		[DllImport("user32.dll")]
		public static extern IntPtr GetFocus();

		[DllImport("user32.dll")]
		public static extern IntPtr SetCapture(IntPtr hWnd);

		[DllImport("user32.dll")]
		public static extern IntPtr GetCapture();

		[DllImport("user32.dll")]
		public static extern bool ReleaseCapture();


		//
		// Window Styles
		//
		[Flags]
        public enum WS : uint // https://docs.microsoft.com/en-us/windows/win32/winmsg/window-styles?redirectedfrom=MSDN
		{
			BORDER = 0x00800000,
			CAPTION = 0x00C00000,
			CHILD = 0x40000000,
			CHILDWINDOW = 0x40000000,
			CLIPCHILDREN = 0x02000000,
			CLIPSIBLINGS = 0x04000000,
			DISABLED = 0x08000000,
			DLGFRAME = 0x00400000,
			GROUP = 0x00020000,
			HSCROLL = 0x00100000,
			ICONIC = 0x20000000,
			MAXIMIZE = 0x01000000,
			MAXIMIZEBOX = 0x00010000,
			MINIMIZE = 0x20000000,
			MINIMIZEBOX = 0x00020000,
			OVERLAPPED = 0x00000000,
			OVERLAPPEDWINDOW = (OVERLAPPED | CAPTION | SYSMENU | THICKFRAME | MINIMIZEBOX | MAXIMIZEBOX),
			POPUP = 0x80000000,
			POPUPWINDOW = (POPUP | BORDER | SYSMENU),
			SIZEBOX = 0x00040000,
			SYSMENU = 0x00080000,
			TABSTOP = 0x00010000,
			THICKFRAME = 0x00040000,
			TILED = 0x00000000,
			TILEDWINDOW = (OVERLAPPED | CAPTION | SYSMENU | THICKFRAME | MINIMIZEBOX | MAXIMIZEBOX),
			VISIBLE = 0x10000000,
			VSCROLL = 0x00200000
		}

		//
		// Extended Window Styles
		//
		[Flags]
		public enum WS_EX : uint // https://docs.microsoft.com/en-us/windows/win32/winmsg/extended-window-styles
		{
			ACCEPTFILES = 0x00000010,
			APPWINDOW = 0x00040000,
			CLIENTEDGE = 0x00000200,
			COMPOSITED = 0x02000000,
			CONTEXTHELP = 0x00000400,
			CONTROLPARENT = 0x00010000,
			DLGMODALFRAME = 0x00000001,
			LAYERED = 0x00080000,
			LAYOUTRTL = 0x00400000,
			LEFT = 0x00000000,
			LEFTSCROLLBAR = 0x00004000,
			LTRREADING = 0x00000000,
			MDICHILD = 0x00000040,
			NOACTIVATE = 0x08000000,
			NOINHERITLAYOUT = 0x00100000,
			NOPARENTNOTIFY = 0x00000004,
			NOREDIRECTIONBITMAP = 0x00200000,
			OVERLAPPEDWINDOW = (WINDOWEDGE | CLIENTEDGE),
			PALETTEWINDOW = (WINDOWEDGE | TOOLWINDOW | TOPMOST),
			RIGHT = 0x00001000,
			RIGHTSCROLLBAR = 0x00000000,
			RTLREADING = 0x00002000,
			STATICEDGE = 0x00020000,
			TOOLWINDOW = 0x00000080,
			TOPMOST = 0x00000008,
			TRANSPARENT = 0x00000020,
			WINDOWEDGE = 0x00000100
		}

        public enum GWL : int
        {
            EXSTYLE = -20,
			HINSTANCE = -6,
			HWNDPARENT = -8,
			ID = -12,
			STYLE = -16,
			USERDATA = -21,
			WNDPROC = -4
		}

		[DllImport("user32.dll")]
		public static extern uint GetWindowLong(IntPtr hWnd, int nIndex);

		[DllImport("user32.dll")]
        public static extern uint SetWindowLong(IntPtr hWnd, int nIndex, uint dwNewLong);

		[DllImport("user32.dll")]
		public static extern IntPtr GetWindowLongPtr(IntPtr hWnd, int nIndex);

		[DllImport("user32.dll")]
		public static extern IntPtr SetWindowLongPtr(IntPtr hWnd, int nIndex, IntPtr dwNewLong);

		[Flags]
		public enum CS : uint
		{
			BYTEALIGNCLIENT = 0x1000,
			BYTEALIGNWINDOW = 0x2000,
			CLASSDC = 0x0040,
			DBLCLKS = 0x0008,
			DROPSHADOW = 0x00020000,
			GLOBALCLASS = 0x4000,
			HREDRAW = 0x0002,
			NOCLOSE = 0x0200,
			OWNDC = 0x0020,
			PARENTDC = 0x0080,
			SAVEBITS = 0x0800,
			VREDRAW = 0x0001
		}

		public enum GCL : int
		{
			ATOM = -32,
			CBCLSEXTRA = -20,
			CBWNDEXTRA = -18,
			HBRBACKGROUND = -10,
			HCURSOR = -12,
			HICON = -14,
			HICONSM = -34,
			HMODULE = -16,
			MENUNAME = -8,
			STYLE = -26,
			WNDPROC = -24
		}

		[DllImport("user32.dll")]
		public static extern uint GetClassLong(IntPtr hWnd, int nIndex);

		[DllImport("user32.dll")]
		public static extern uint SetClassLong(IntPtr hWnd, int nIndex, uint dwNewLong);


		//
		// Scroll Bars
		//
		public enum SB : int
		{
			/// <summary>
			/// Retrieves the position of the scroll box in a window's standard horizontal 
			/// scroll bar.
			/// </summary>
			HORZ = 0,

			/// <summary>
			/// Retrieves the position of the scroll box in a window's standard vertical 
			/// scroll bar.
			/// </summary>
			VERT = 1,

			/// <summary>
			/// Retrieves the position of the scroll box in a scroll bar control. The hWnd 
			/// parameter must be the handle to the scroll bar control.
			/// </summary>
			CTL = 2,

			/// <summary>
			/// Retrieves the position of both the horizontal and vertical scroll bars.
			/// </summary>
			BOTH = 3
		}

		[StructLayout(LayoutKind.Sequential)]
		public struct SCROLLINFO
		{
			public static readonly SCROLLINFO Empty;
			public uint cbSize;
			public uint fMask;
			public int nMin;
			public int nMax;
			public uint nPage;
			public int nPos;
			public int nTrackPos;
		}

		/// <summary>
		/// Use for the fMask parameter in SCROLLINFO.
		/// </summary>
		public enum SIF : uint
		{
			RANGE = 0x1,
			PAGE = 0x2,
			POS = 0x4,
			DISABLENOSCROLL = 0x8,
			TRACKPOS = 0x10,
			ALL = (PAGE | POS | RANGE | TRACKPOS)
		}

		[DllImport("user32.dll")]
		public static extern int GetScrollPos(IntPtr hWnd, int nBar);

		[DllImport("user32.dll")]
		public static extern int SetScrollPos(IntPtr hWnd, int nBar, int nPos, bool bRedraw);

		[DllImport("user32.dll")]
		public static extern bool GetScrollRange(IntPtr hWnd, int nBar, out int lpMinPos, out int lpMaxPos);

		[DllImport("user32.dll")]
		public static extern bool SetScrollRange(IntPtr hWnd, int nBar, int nMinPos, int nMaxPos, bool bRedraw);

		/// <summary>
		/// Before calling, set the cbSize member of lpsi to the size of lpsi, and set the fMask member
		/// to indicate the parameters to get.
		/// </summary>
		[DllImport("user32.dll")]
		public static extern bool GetScrollInfo(IntPtr hwnd, int nBar, ref SCROLLINFO lpsi);

		/// <summary>
		/// Before calling, set the cbSize member of lpsi to the size of lpsi, set the fMask member to 
		/// indicate the parameters to set, and specify the new parameter values in the appropriate members.
		/// </summary>
		[DllImport("user32.dll")]
		public static extern int SetScrollInfo(IntPtr hwnd, int nBar, ref SCROLLINFO lpsi, bool redraw);

		#endregion

		#region Sound

		//
		// Sound
		//
		[DllImport("kernel32.dll")]
        public static extern bool Beep(uint dwFreq, uint dwDuration);

        [Flags]
        public enum SND : uint
        {
            SYNC = 0x0,		    // can be used with sndPlaySound
            ASYNC = 0x1,		// can be used with sndPlaySound
            NODEFAULT = 0x2,	// can be used with sndPlaySound
            MEMORY = 0x4,		// can be used with sndPlaySound
            LOOP = 0x8,		    // can be used with sndPlaySound
            NOSTOP = 0x10,		// can be used with sndPlaySound
            PURGE = 0x40,
            APPLICATION = 0x80,
            NOWAIT = 0x2000,
            RESOURCE = 0x40004,
            ALIAS = 0x10000,
            FILENAME = 0x20000,
            ALIAS_ID = 0x110000
        }

        [DllImport("winmm.dll")]
        public static extern bool PlaySound(byte[] lpszSound, IntPtr hmod, uint fdwSound);

        [DllImport("winmm.dll")]
        public static extern bool sndPlaySound(byte[] lpszSound, uint fuSound);

		#endregion

		#region Menus

		//
		// Menus
		//

		/// <summary>
		/// Menu Flags
		/// https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-getmenustate
		/// </summary>
		[Flags]
        public enum MF : uint
        {
			BYCOMMAND = 0x00,
			STRING = 0x00,
            GRAYED = 0x01,
			DISABLED = 0x02,
			BITMAP = 0x04,
			CHECKED = 0x08,
			POPUP = 0x10,
			MENUBARBREAK = 0x20,
			MENUBREAK = 0x40,
			HILITE = 0x80,
			OWNERDRAW = 0x100,
			BYPOSITION = 0x400,
			SEPARATOR = 0x800
		}

        [DllImport("user32.dll")]
        public static extern IntPtr GetSystemMenu(IntPtr hWnd, bool bRevert);

		[DllImport("user32.dll")]
		public static extern IntPtr GetMenu(IntPtr hWnd);

		[DllImport("user32.dll")]
		public static extern bool IsMenu(IntPtr hMenu);

		[DllImport("user32.dll")]
		public static extern int GetMenuItemCount(IntPtr hMenu);

		[DllImport("user32.dll")]
		public static extern int GetMenuString(IntPtr hMenu, uint uIDItem, StringBuilder lpString, int cchMax, uint flags);

		[DllImport("user32.dll")]
		public static extern uint GetMenuState(IntPtr hMenu, uint uId, uint uFlags);

		[DllImport("user32.dll")]
        public static extern bool AppendMenu(IntPtr hMenu, uint uFlags, uint uIDNewItem, string lpNewItem);

        [DllImport("user32.dll")]
        public static extern bool InsertMenu(IntPtr hMenu, uint uPosition, uint uFlags, uint uIDNewItem, string lpNewItem);

        [DllImport("user32.dll")]
        public static extern bool RemoveMenu(IntPtr hMenu, uint uPosition, uint uFlags);

        [DllImport("user32.dll")]
        public static extern bool DeleteMenu(IntPtr hMenu, uint uPosition, uint uFlags);

        [DllImport("user32.dll")]
        public static extern bool EndMenu();

		[Flags]
        public enum TPM : uint
        {
            LEFTBUTTON = 0x0000,
            RECURSE = 0x0001,
            RIGHTBUTTON = 0x0002,
            LEFTALIGN = 0x0000,
            CENTERALIGN = 0x0004,
            RIGHTALIGN = 0x0008,
            TOPALIGN = 0x0000,
            VCENTERALIGN = 0x0010,
            BOTTOMALIGN = 0x0020,
            HORIZONTAL = 0x0000,
            VERTICAL = 0x0040,
            NONOTIFY = 0x0080,
            RETURNCMD = 0x0100,
            HORPOSANIMATION = 0x0400,
            HORNEGANIMATION = 0x0800,
            VERPOSANIMATION = 0x1000,
            VERNEGANIMATION = 0x2000,
            NOANIMATION = 0x4000,
            LAYOUTRTL = 0x8000
        }

        [DllImport("user32.dll")]
        public static extern uint TrackPopupMenu(IntPtr hMenu, uint uFlags, int x, int y, int nReserved, IntPtr hWnd, IntPtr prcRect);

        [DllImport("user32.dll")]
        public static extern uint TrackPopupMenuEx(IntPtr hmenu, uint fuFlags, int x, int y, IntPtr hWnd, IntPtr lptpm);

        [DllImport("kernel32.dll")]
        public static extern bool CloseHandle(IntPtr hObject);

        #endregion

        #region Message Boxes

        //
        // Message Box
        //
        [Flags]
        public enum MB : uint
        {
            ABORTRETRYIGNORE = 0x0002,
            CANCELTRYCONTINUE = 0x0006,
            HELP = 0x4000,
            OK = 0x0000,
            OKCANCEL = 0x0001,
            RETRYCANCEL = 0x0005,
            YESNO = 0x0004,
            YESNOCANCEL = 0x0003,
            ICONEXCLAMATION = 0x0030,
            ICONWARNING = 0x0030,
            ICONINFORMATION = 0x0040,
            ICONASTERISK = 0x0040,
            ICONQUESTION = 0x0020,
            ICONSTOP = 0x0010,
            ICONERROR = 0x0010,
            ICONHAND = 0x0010,
            DEFBUTTON1 = 0x0000,
            DEFBUTTON2 = 0x0100,
            DEFBUTTON3 = 0x0200,
            DEFBUTTON4 = 0x0300,
            APPLMODAL = 0x0000,
            SYSTEMMODAL = 0x1000,
            TASKMODAL = 0x2000,
            DEFAULT_DESKTOP_ONLY = 0x20000,
            RIGHT = 0x80000,
            RTLREADING = 0x100000,
            SETFOREGROUND = 0x10000,
            TOPMOST = 0x40000,
            SERVICE_NOTIFICATION = 0x200000,
            SERVICE_NOTIFICATION_NT3X = 0x40000
        }

        // MessageBox result constants
        public enum ID : int
        {
            OK = 1,
            CANCEL = 2,
            ABORT = 3,
            RETRY = 4,
            IGNORE = 5,
            YES = 6,
            NO = 7,
            CLOSE = 8,
            HELP = 9,
            TRYAGAIN = 10,
            CONTINUE = 11,
            TIMEOUT = 32000
        }

		[DllImport("user32.dll")]
		public static extern int MessageBox(IntPtr hWnd, string lpText, string lpCaption, uint uType);

		#endregion

		#region Threads and Processes

		//
		// Threads and Processes
		//
		[DllImport("user32.dll")]
        public static extern uint WaitForInputIdle(IntPtr hProcess, uint dwMilliseconds);

        [DllImport("user32.dll")]
        public static extern bool AttachThreadInput(uint idAttach, uint idAttachTo, bool fAttach);

        [DllImport("kernel32.dll")]
        public static extern uint GetCurrentThreadId();

        [DllImport("user32.dll")]
        public static extern uint GetWindowThreadProcessId(IntPtr hWnd, ref uint lpdwProcessId);

		[Flags]
		public enum GUI : uint
		{
			CARETBLINKING = 0x01,
			INMENUMODE = 0x04,
			INMOVESIZE = 0x02,
			POPUPMENUMODE = 0x10,
			SYSTEMMENUMODE = 0x08
		}

		[StructLayout(LayoutKind.Sequential)]
		public struct GUITHREADINFO
		{
			public static readonly GUITHREADINFO Empty;
			public uint cbSize;
			public GUI flags;
			public IntPtr hwndActive;
			public IntPtr hwndFocus;
			public IntPtr hwndCapture;
			public IntPtr hwndMenuOwner;
			public IntPtr hwndMoveSize;
			public IntPtr hwndCaret;
			public RECT rcCaret;
		};

		[DllImport("user32.dll")]
		public static extern bool GetGUIThreadInfo(uint idThread, ref GUITHREADINFO pgui);

		#endregion

		#region System Parameters

		// https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-systemparametersinfoa
		public enum SPI : uint
        {
            GETMOUSE = 0x0003,
            SETMOUSE = 0x0004,
            GETFILTERKEYS = 0x0032,
            SETFILTERKEYS = 0x0033,
            GETMOUSESPEED = 0x0070,
            SETMOUSESPEED = 0x0071,
			SETDRAGWIDTH = 0x004C,
			SETDRAGHEIGHT = 0x004D,
			GETACTIVEWNDTRKTIMEOUT = 0x2002,
			SETACTIVEWNDTRKTIMEOUT = 0x2003
		}

        [Flags]
        public enum SPIF : uint
        {
            UPDATEINIFILE = 0x01,
            SENDCHANGE = 0x02,
            SENDWININICHANGE = 0x02
        }

        [Flags]
        public enum FKF : uint
        {
            FILTERKEYSON = 0x0001,
            AVAILABLE = 0x0002,
            HOTKEYACTIVE = 0x0004,
            CONFIRMHOTKEY = 0x0008,
            HOTKEYSOUND = 0x0010,
            INDICATOR = 0x0020,
            CLICKON = 0x0040
        }

        [StructLayout(LayoutKind.Sequential)]
        public struct FILTERKEYS
        {
            public static readonly FILTERKEYS Empty;
            public uint cbSize;
            public uint dwFlags;
            public uint iWaitMSec;
            public uint iDelayMSec;
            public uint iRepeatMSec;
            public uint iBounceMSec;
        };

        [DllImport("user32.dll")] // for getting and setting FILTERKEYS
        public static extern bool SystemParametersInfo(uint uiAction, uint uiParam, ref FILTERKEYS pvParam, uint fWinIni);

        [DllImport("user32.dll")] // for getting values and for setting with reference types (non-arrays)
        public static extern bool SystemParametersInfo(uint uiAction, uint uiParam, ref uint pvParam, uint fWinIni);

        [DllImport("user32.dll")] // for getting and setting arrays
        public static extern bool SystemParametersInfo(uint uiAction, uint uiParam, uint[] pvParam, uint fWinIni);

        [DllImport("user32.dll")] // for setting single values
        public static extern bool SystemParametersInfo(uint uiAction, uint uiParam, uint pvParam, uint fWinIni);


		// https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-getsystemmetrics
		public enum SM : int
		{
			CXSCREEN = 0,
			CYSCREEN = 1,
			CXDRAG = 68,
			CYDRAG = 69
		}

		[DllImport("user32.dll")] // for getting system metrics; there is no setter analog, as SystemParametersInfo is used
		public static extern int GetSystemMetrics(int nIndex);

		#endregion

		#region Device Capabilities

		// http://pinvoke.net/default.aspx/gdi32/GetDeviceCaps.html
		public enum DEVCAPS : int
		{
			DRIVERVERSION = 0, // Device driver version
			TECHNOLOGY = 2, // Device classification
			HORZSIZE = 4, // Horizontal size in millimeters
			VERTSIZE = 6, // Vertical size in millimeters
			HORZRES = 8, // Horizontal width in pixels
			VERTRES = 10, // Vertical height in pixels
			BITSPIXEL = 12, // Number of bits per pixel
			PLANES = 14, // Number of planes
			NUMBRUSHES = 16, // Number of brushes the device has
			NUMPENS = 18, // Number of pens the device has
			NUMMARKERS = 20, // Number of markers the device has
			NUMFONTS = 22, // Number of fonts the device has
			NUMCOLORS = 24, // Number of colors the device supports
			PDEVICESIZE = 26, // Size required for device descriptor 
			CURVECAPS = 28, //  Curve capabilities
			LINECAPS = 30, // Line capabilities
			POLYGONALCAPS = 32, // Polygonal capabilities
			TEXTCAPS = 34, // Text capabilities
			CLIPCAPS = 36, // Clipping capabilities
			RASTERCAPS = 38, // Bitblt capabilities
			ASPECTX = 40, // Length of the X leg
			ASPECTY = 42, // Length of the Y leg
			ASPECTXY = 44, // Length of the hypotenuse
			SHADEBLENDCAPS = 45, // Shading and Blending caps

			LOGPIXELSX = 88, // Logical pixels inch in X
			LOGPIXELSY = 90, // Logical pixels inch in Y

			SIZEPALETTE = 104, // Number of entries in physical palette
			NUMRESERVED = 106, // Number of reserved entries in palette
			COLORRES = 108, // Actual color resolution

			PHYSICALWIDTH = 110, // Physical Width in device units
			PHYSICALHEIGHT = 111, // Physical Height in device units
			PHYSICALOFFSETX = 112, // Physical Printable Area x margin
			PHYSICALOFFSETY = 113, // Physical Printable Area y margin
			SCALINGFACTORX = 114, // Scaling factor x
			SCALINGFACTORY = 115, // Scaling factor y

			VREFRESH = 116, // Current vertical refresh rate of the display device (for displays only) in Hz
			DESKTOPVERTRES = 117, //  Vertical height of entire desktop in pixels
			DESKTOPHORZRES = 118, // Horizontal width of entire desktop in pixels
			BLTALIGNMENT = 119 // Preferred blt alignment
		}

		// https://docs.microsoft.com/en-us/windows/win32/api/wingdi/nf-wingdi-getdevicecaps
		[DllImport("gdi32.dll")]
		public static extern int GetDeviceCaps(IntPtr hdc, int index);

		/// <summary>
		/// This struct is used with EnumDisplaySettings, but it simply won't load into memory as
		/// laid out. The issue is "dmFormName," which should be at offset 70 and of length 32.
		/// Only by putting it at offset 72, and length 30, will this structure load into memory
		/// at runtime, but that risks "dmFormName" returning with an incorrect value. I do not
		/// know why this struct cannot be used, when its non-union-containing version, below,
		/// loads fine and, when inspected with 'Marshal.OffsetOf(typeof(Win32.DEVMODE), "dmFormName")',
		/// it shows that field at offset 70, just as it should be. 
		/// </summary>
		/// <see cref="https://www.pinvoke.net/default.aspx/Structures.DEVMODE"/>
		/// <see cref="https://docs.microsoft.com/en-us/windows/win32/api/wingdi/ns-wingdi-devmodea"/>
		/// <see cref="https://stackoverflow.com/questions/60872044/how-to-get-scaling-factor-for-each-monitor-e-g-1-1-25-1-5"/>
		[StructLayout(LayoutKind.Explicit, CharSet = CharSet.Ansi, Size = 156)]
		public struct DEVMODE__DO_NOT_USE
		{
			[FieldOffset(0), MarshalAs(UnmanagedType.ByValTStr, SizeConst = 32)]
			public string dmDeviceName;
			
			[FieldOffset(32)] public ushort dmSpecVersion;
			[FieldOffset(34)] public ushort dmDriverVersion;
			[FieldOffset(36)] public ushort dmSize;
			[FieldOffset(38)] public ushort dmDriverExtra;
			[FieldOffset(40)] public uint dmFields;

			// <union>
			[FieldOffset(44)] public short dmOrientation;
			[FieldOffset(46)] public short dmPaperSize;
			[FieldOffset(48)] public short dmPaperLength;
			[FieldOffset(50)] public short dmPaperWidth;
			[FieldOffset(52)] public short dmScale;
			[FieldOffset(54)] public short dmCopies;
			[FieldOffset(56)] public short dmDefaultSource;
			[FieldOffset(58)] public short dmPrintQuality;
			// </union>

			// <union>
			[FieldOffset(44)] public int dmPositionX;
			[FieldOffset(48)] public int dmPositionY;
			[FieldOffset(52)] public uint dmDisplayOrientation;
			[FieldOffset(56)] public uint dmDisplayFixedOutput;
			// </union>

			[FieldOffset(60)] public short dmColor;
			[FieldOffset(62)] public short dmDuplex;
			[FieldOffset(64)] public short dmYResolution;
			[FieldOffset(66)] public short dmTTOption;
			[FieldOffset(68)] public short dmCollate;

			[FieldOffset(70), MarshalAs(UnmanagedType.ByValTStr, SizeConst = 32)]
			//[FieldOffset(72), MarshalAs(UnmanagedType.ByValTStr, SizeConst = 30)]  // will make work, but "dmFormName" might be wrong?
			public string dmFormName;
			
			[FieldOffset(102)] public ushort dmLogPixels;
			[FieldOffset(104)] public uint dmBitsPerPel;
			[FieldOffset(108)] public uint dmPelsWidth; 
			[FieldOffset(112)] public uint dmPelsHeight;
			
			// <union>
			[FieldOffset(116)] public uint dmDisplayFlags;
			[FieldOffset(116)] public uint dmNup;
			// </union>
			
			[FieldOffset(120)] public uint dmDisplayFrequency;
			[FieldOffset(124)] public uint dmICMMethod;
			[FieldOffset(128)] public uint dmICMIntent;
			[FieldOffset(132)] public uint dmMediaType; 
			[FieldOffset(136)] public uint dmDitherType; 
			[FieldOffset(140)] public uint dmReserved1;
			[FieldOffset(144)] public uint dmReserved2; 
			[FieldOffset(148)] public uint dmPanningWidth;
			[FieldOffset(152)] public uint dmPanningHeight;
		}

		/// <summary>
		/// By avoiding the explicit layout and C-style unions, this version of DEVMODE loads
		/// fine into memory, and "dmFormName" is indeed at offset 70, as it should be. But
		/// the overlapping unions cannot be supported with this version.
		/// </summary>
		[StructLayout(LayoutKind.Sequential)]
		public struct DEVMODE
		{
			[MarshalAs(UnmanagedType.ByValTStr, SizeConst = 32)]
			public string dmDeviceName;

			public ushort dmSpecVersion;
			public ushort dmDriverVersion;
			public ushort dmSize;
			public ushort dmDriverExtra;
			public uint dmFields;
			public int dmPositionX;
			public int dmPositionY;
			public uint dmDisplayOrientation;
			public uint dmDisplayFixedOutput;
			public short dmColor;
			public short dmDuplex;
			public short dmYResolution;
			public short dmTTOption;
			public short dmCollate;

			[MarshalAs(UnmanagedType.ByValTStr, SizeConst = 32)]
			public string dmFormName;

			public ushort dmLogPixels;
			public uint dmBitsPerPel;
			public uint dmPelsWidth;
			public uint dmPelsHeight;
			public uint dmDisplayFlags;
			public uint dmDisplayFrequency;
			public uint dmICMMethod;
			public uint dmICMIntent;
			public uint dmMediaType;
			public uint dmDitherType;
			public uint dmReserved1;
			public uint dmReserved2;
			public uint dmPanningWidth;
			public uint dmPanningHeight;
		}

		// https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-enumdisplaysettingsa
		[DllImport("user32.dll")]
		public static extern bool EnumDisplaySettings(string lpszDeviceName, int iModeNum, ref DEVMODE lpDevMode);

		#endregion

		#region Graphics (GDI+)

		/// <summary>
		/// https://docs.microsoft.com/en-us/windows/win32/api/wingdi/nf-wingdi-setrop2
		/// https://www.pinvoke.net/default.aspx/gdi32.setrop2
		/// </summary>
		public enum R2 : int
		{
			BLACK = 1,			// Pixel is always 0.
			NOTMERGEPEN = 2,	// Pixel is the inverse of the R2_MERGEPEN color.
			MASKNOTPEN = 3,		// Pixel is a combination of the colors common to both the screen and the inverse of the pen.
			NOTCOPYPEN = 4,		// Pixel is the inverse of the pen color.
			MASKPENNOT = 5,		// Pixel is a combination of the colors common to both the pen and the inverse of the screen.
			NOT = 6,			// Pixel is the inverse of the screen color.
			XORPEN = 7,			// Pixel is a combination of the colors in the pen and in the screen, but not in both.
			NOTMASKPEN = 8,		// Pixel is the inverse of the R2_MASKPEN color.
			MASKPEN = 9,		// Pixel is a combination of the colors common to both the pen and the screen.
			NOTXORPEN = 10,		// Pixel is the inverse of the R2_XORPEN color.
			NOP = 11,			// Pixel remains unchanged.
			MERGENOTPEN = 12,	// Pixel is a combination of the screen color and the inverse of the pen color.
			COPYPEN = 13,		// Pixel is the pen color.
			MERGEPENNOT = 14,	// Pixel is a combination of the pen color and the inverse of the screen color.
			MERGEPEN = 15,		// Pixel is a combination of the pen color and the screen color.
			WHITE = 16			// Pixel is always 1.
		}

		[DllImport("gdi32.dll")]
		public extern static int SetROP2(IntPtr hdc, int rop2);

		[DllImport("user32.dll")]
		public extern static IntPtr GetDC(IntPtr hWnd);

		[DllImport("user32.dll")]
		public extern static int ReleaseDC(IntPtr hWnd, IntPtr hDC);

		[DllImport("gdi32.dll")]
		public static extern IntPtr CreateCompatibleDC(IntPtr hdc);

		[DllImport("gdi32.dll")]
		public extern static bool MoveToEx(IntPtr hdc, int x, int y, ref POINT lppt);

		[DllImport("gdi32.dll")]
		public extern static bool LineTo(IntPtr hdc, int x, int y);

		// https://www.pinvoke.net/default.aspx/user32/drawtext.html
		public enum DT : uint
		{
			TOP = 0x00000000,
			LEFT = 0x00000000,
			CENTER = 0x00000001,
			RIGHT = 0x00000002,
			VCENTER = 0x00000004,
			BOTTOM = 0x00000008,
			WORDBREAK = 0x00000010,
			SINGLELINE = 0x00000020,
			EXPANDTABS = 0x00000040,
			TABSTOP = 0x00000080,
			NOCLIP = 0x00000100,
			EXTERNALLEADING = 0x00000200,
			CALCRECT = 0x00000400,
			NOPREFIX = 0x00000800,
			INTERNAL = 0x00001000,
			EDITCONTROL = 0x00002000,
			PATH_ELLIPSIS = 0x00004000,
			END_ELLIPSIS = 0x00008000,
			MODIFYSTRING = 0x00010000,
			RTLREADING = 0x00020000,
			WORD_ELLIPSIS = 0x00040000,
			NOFULLWIDTHCHARBREAK = 0x00080000,
			HIDEPREFIX = 0x00100000,
			PREFIXONLY = 0x00200000
		}

		// http://pinvoke.net/default.aspx/Structures/DRAWTEXTPARAMS.html
		// https://docs.microsoft.com/en-us/windows/win32/api/winuser/ns-winuser-drawtextparams
		[StructLayout(LayoutKind.Sequential)]
		public struct DRAWTEXTPARAMS
		{
			public uint cbSize;
			public int iTabLength;
			public int iLeftMargin;
			public int iRightMargin;
			public uint uiLengthDrawn;
		}

		// https://docs.microsoft.com/en-us/windows/win32/api/wingdi/nf-wingdi-setbkmode
		public enum BM : int
		{
			TRANSPARENT = 1,
			OPAQUE = 2
		}

		// https://www.pinvoke.net/default.aspx/Enums/TernaryRasterOperations.html
		public enum BLT : uint
		{
			SRCCOPY = 0x00CC0020,		// dest = source
			SRCPAINT = 0x00EE0086,		// dest = source OR dest
			SRCAND = 0x008800C6,		//dest = source AND dest
			SRCINVERT = 0x00660046,		// dest = source XOR dest
			SRCERASE = 0x00440328,		// dest = source AND (NOT dest)
			NOTSRCCOPY = 0x00330008,	// dest = (NOT source)
			NOTSRCERASE = 0x001100A6,	// dest = (NOT src) AND (NOT dest)
			MERGECOPY = 0x00C000CA,		// dest = (source AND pattern)
			MERGEPAINT = 0x00BB0226,	// dest = (NOT source) OR dest
			PATCOPY = 0x00F00021,		// dest = pattern
			PATPAINT = 0x00FB0A09,		// dest = DPSnoo
			PATINVERT = 0x005A0049,		// dest = pattern XOR dest
			DSTINVERT = 0x00550009,		// dest = (NOT dest)
			BLACKNESS = 0x00000042,		// dest = BLACK
			WHITENESS = 0x00FF0062,		// dest = WHITE
			CAPTUREBLT = 0x40000000		// Capture window as seen on screen. This includes layered windows such as WPF windows with AllowsTransparency="true".
		}

		// https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-drawtext
		[DllImport("user32.dll")]
		public static extern int DrawText(IntPtr hdc, StringBuilder lpchText, int cchText, ref RECT lprc, uint format);

		// https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-drawtextexa
		[DllImport("user32.dll")]
		public static extern int DrawTextEx(IntPtr hdc, StringBuilder lpchText, int cchText, ref RECT lprc, uint format, ref DRAWTEXTPARAMS lpdtp);

		// https://docs.microsoft.com/en-us/windows/win32/api/wingdi/nf-wingdi-textouta
		[DllImport("gdi32.dll")]
		public static extern bool TextOut(IntPtr hdc, int x, int y, string lpString, int c);

		// https://docs.microsoft.com/en-us/windows/win32/api/wingdi/nf-wingdi-settextcolor
		[DllImport("gdi32.dll")]
		public static extern uint SetTextColor(IntPtr hdc, uint color);

		// https://docs.microsoft.com/en-us/windows/win32/api/wingdi/nf-wingdi-setbkmode
		[DllImport("gdi32.dll")]
		public static extern int SetBkMode(IntPtr hdc, int mode);

		// https://docs.microsoft.com/en-us/windows/win32/api/wingdi/nf-wingdi-getbkmode
		[DllImport("gdi32.dll")]
		public static extern int GetBkMode(IntPtr hdc);

		// https://docs.microsoft.com/en-us/windows/win32/api/wingdi/nf-wingdi-beginpath
		[DllImport("gdi32.dll")]
		public static extern bool BeginPath(IntPtr hdc);

		// https://docs.microsoft.com/en-us/windows/win32/api/wingdi/nf-wingdi-endpath
		[DllImport("gdi32.dll")]
		public static extern bool EndPath(IntPtr hdc);

		// https://docs.microsoft.com/en-us/windows/win32/api/wingdi/nf-wingdi-fillpath
		[DllImport("gdi32.dll")]
		public static extern bool FillPath(IntPtr hdc);

		// https://docs.microsoft.com/en-us/windows/win32/api/wingdi/nf-wingdi-createsolidbrush
		// For the color parameter, pass ColorTranslator.ToWin32(color).
		[DllImport("gdi32.dll")]
		public static extern IntPtr CreateSolidBrush(uint color);

		// https://docs.microsoft.com/en-us/windows/win32/api/wingdi/nf-wingdi-selectobject
		[DllImport("gdi32.dll")]
		public static extern IntPtr SelectObject(IntPtr hdc, IntPtr h);

		// https://docs.microsoft.com/en-us/windows/win32/api/wingdi/nf-wingdi-deleteobject
		[DllImport("gdi32.dll")]
		public static extern bool DeleteObject(IntPtr ho);

		// https://docs.microsoft.com/en-us/windows/win32/api/wingdi/nf-wingdi-patblt
		// For the rop parameter, pass a BLT value.
		[DllImport("gdi32.dll")]
		public static extern bool PatBlt(IntPtr hdc, int x, int y, int w, int h, uint rop);

		// https://docs.microsoft.com/en-us/windows/win32/api/wingdi/nf-wingdi-bitblt
		// For the rop parameter, pass a BLT value.
		[DllImport("gdi32.dll")]
		public static extern bool BitBlt(IntPtr hdc, int x, int y, int cx, int cy, IntPtr hdcSrc, int x1, int y1, uint rop);

		#endregion

		#region Locale

		public enum LOCALE : uint
        {
            NOUSEROVERRIDE = 0x80000000,            // do not use user overrides
            RETURN_NUMBER = 0x20000000,             // return number instead of string

            // Modifier for genitive names
            RETURN_GENITIVE_NAMES = 0x10000000,     // flag for returning the Genitive forms of month names

            //
            //  The following LCTypes are mutually exclusive in that they may NOT
            //  be used in combination with each other.
            //

            // These are the various forms of the name of the locale:
            SLOCALIZEDDISPLAYNAME = 0x00000002,     // localized name of locale, e.g., "German (Germany)" in UI language
            SENGLISHDISPLAYNAME = 0x00000072,       // Display name (language + country usually) in English, e.g., "German (Germany)"
            SNATIVEDISPLAYNAME = 0x00000073,        // Display name in native locale language, e.g., "Deutsch (Deutschland)

            SLOCALIZEDLANGUAGENAME = 0x0000006f,    // Language Display Name for a language, e.g., "German" in UI language
            SENGLISHLANGUAGENAME = 0x00001001,      // English name of language, e.g., "German"
            SNATIVELANGUAGENAME = 0x00000004,       // native name of language, e.g., "Deutsch"

            SLOCALIZEDCOUNTRYNAME = 0x00000006,     // localized name of country, e.g., "Germany" in UI language
            SENGLISHCOUNTRYNAME = 0x00001002,       // English name of country, e.g., "Germany"
            SNATIVECOUNTRYNAME = 0x00000008,        // native name of country, e.g., "Deutschland"

            // Additional LCTYPEs
            SABBREVLANGNAME = 0x00000003,           // abbreviated language name

            ICOUNTRY = 0x00000005,                  // country code
            SABBREVCTRYNAME = 0x00000007,           // abbreviated country name
            IGEOID = 0x0000005B,                    // geographical location id

            IDEFAULTLANGUAGE = 0x00000009,          // default language id
            IDEFAULTCOUNTRY = 0x0000000A,           // default country code
            IDEFAULTCODEPAGE = 0x0000000B,          // default oem code page
            IDEFAULTANSICODEPAGE = 0x00001004,      // default ansi code page
            IDEFAULTMACCODEPAGE = 0x00001011,       // default mac code page

            SLIST = 0x0000000C,                     // list item separator
            IMEASURE = 0x0000000D,                  // 0 = metric, 1 = US

            SDECIMAL = 0x0000000E,                  // decimal separator
            STHOUSAND = 0x0000000F,                 // thousand separator
            SGROUPING = 0x00000010,                 // digit grouping
            IDIGITS = 0x00000011,                   // number of fractional digits
            ILZERO = 0x00000012,                    // leading zeros for decimal
            INEGNUMBER = 0x00001010,                // negative number mode
            SNATIVEDIGITS = 0x00000013,             // native digits for 0-9

            SCURRENCY = 0x00000014,                 // local monetary symbol
            SINTLSYMBOL = 0x00000015,               // uintl monetary symbol
            SMONDECIMALSEP = 0x00000016,            // monetary decimal separator
            SMONTHOUSANDSEP = 0x00000017,           // monetary thousand separator
            SMONGROUPING = 0x00000018,              // monetary grouping
            ICURRDIGITS = 0x00000019,               // # local monetary digits
            IINTLCURRDIGITS = 0x0000001A,           // # uintl monetary digits
            ICURRENCY = 0x0000001B,                 // positive currency mode
            INEGCURR = 0x0000001C,                  // negative currency mode

            SDATE = 0x0000001D,                     // date separator (derived from LOCALE_SSHORTDATE, use that instead)
            STIME = 0x0000001E,                     // time separator (derived from LOCALE_STIMEFORMAT, use that instead)
            SSHORTDATE = 0x0000001F,                // short date format string
            SLONGDATE = 0x00000020,                 // long date format string
            STIMEFORMAT = 0x00001003,               // time format string
            IDATE = 0x00000021,                     // short date format ordering (derived from LOCALE_SSHORTDATE, use that instead)
            ILDATE = 0x00000022,                    // long date format ordering (derived from LOCALE_SLONGDATE, use that instead)
            ITIME = 0x00000023,                     // time format specifier (derived from LOCALE_STIMEFORMAT, use that instead)
            ITIMEMARKPOSN = 0x00001005,             // time marker position (derived from LOCALE_STIMEFORMAT, use that instead)
            ICENTURY = 0x00000024,                  // century format specifier (short date, LOCALE_SSHORTDATE is preferred)
            ITLZERO = 0x00000025,                   // leading zeros in time field (derived from LOCALE_STIMEFORMAT, use that instead)
            IDAYLZERO = 0x00000026,                 // leading zeros in day field (short date, LOCALE_SSHORTDATE is preferred)
            IMONLZERO = 0x00000027,                 // leading zeros in month field (short date, LOCALE_SSHORTDATE is preferred)
            S1159 = 0x00000028,                     // AM designator
            S2359 = 0x00000029,                     // PM designator

            ICALENDARTYPE = 0x00001009,             // type of calendar specifier
            IOPTIONALCALENDAR = 0x0000100B,         // additional calendar types specifier
            IFIRSTDAYOFWEEK = 0x0000100C,           // first day of week specifier
            IFIRSTWEEKOFYEAR = 0x0000100D,          // first week of year specifier

            SDAYNAME1 = 0x0000002A,                 // long name for Monday
            SDAYNAME2 = 0x0000002B,                 // long name for Tuesday
            SDAYNAME3 = 0x0000002C,                 // long name for Wednesday
            SDAYNAME4 = 0x0000002D,                 // long name for Thursday
            SDAYNAME5 = 0x0000002E,                 // long name for Friday
            SDAYNAME6 = 0x0000002F,                 // long name for Saturday
            SDAYNAME7 = 0x00000030,                 // long name for Sunday
            SABBREVDAYNAME1 = 0x00000031,           // abbreviated name for Monday
            SABBREVDAYNAME2 = 0x00000032,           // abbreviated name for Tuesday
            SABBREVDAYNAME3 = 0x00000033,           // abbreviated name for Wednesday
            SABBREVDAYNAME4 = 0x00000034,           // abbreviated name for Thursday
            SABBREVDAYNAME5 = 0x00000035,           // abbreviated name for Friday
            SABBREVDAYNAME6 = 0x00000036,           // abbreviated name for Saturday
            SABBREVDAYNAME7 = 0x00000037,           // abbreviated name for Sunday
            SMONTHNAME1 = 0x00000038,               // long name for January
            SMONTHNAME2 = 0x00000039,               // long name for February
            SMONTHNAME3 = 0x0000003A,               // long name for March
            SMONTHNAME4 = 0x0000003B,               // long name for April
            SMONTHNAME5 = 0x0000003C,               // long name for May
            SMONTHNAME6 = 0x0000003D,               // long name for June
            SMONTHNAME7 = 0x0000003E,               // long name for July
            SMONTHNAME8 = 0x0000003F,               // long name for August
            SMONTHNAME9 = 0x00000040,               // long name for September
            SMONTHNAME10 = 0x00000041,              // long name for October
            SMONTHNAME11 = 0x00000042,              // long name for November
            SMONTHNAME12 = 0x00000043,              // long name for December
            SMONTHNAME13 = 0x0000100E,              // long name for 13th month (if exists)
            SABBREVMONTHNAME1 = 0x00000044,         // abbreviated name for January
            SABBREVMONTHNAME2 = 0x00000045,         // abbreviated name for February
            SABBREVMONTHNAME3 = 0x00000046,         // abbreviated name for March
            SABBREVMONTHNAME4 = 0x00000047,         // abbreviated name for April
            SABBREVMONTHNAME5 = 0x00000048,         // abbreviated name for May
            SABBREVMONTHNAME6 = 0x00000049,         // abbreviated name for June
            SABBREVMONTHNAME7 = 0x0000004A,         // abbreviated name for July
            SABBREVMONTHNAME8 = 0x0000004B,         // abbreviated name for August
            SABBREVMONTHNAME9 = 0x0000004C,         // abbreviated name for September
            SABBREVMONTHNAME10 = 0x0000004D,        // abbreviated name for October
            SABBREVMONTHNAME11 = 0x0000004E,        // abbreviated name for November
            SABBREVMONTHNAME12 = 0x0000004F,        // abbreviated name for December
            SABBREVMONTHNAME13 = 0x0000100F,        // abbreviated name for 13th month (if exists)

            SPOSITIVESIGN = 0x00000050,             // positive sign
            SNEGATIVESIGN = 0x00000051,             // negative sign
            IPOSSIGNPOSN = 0x00000052,              // positive sign position (derived from INEGCURR)
            INEGSIGNPOSN = 0x00000053,              // negative sign position (derived from INEGCURR)
            IPOSSYMPRECEDES = 0x00000054,           // mon sym precedes pos amt (derived from ICURRENCY)
            IPOSSEPBYSPACE = 0x00000055,            // mon sym sep by space from pos amt (derived from ICURRENCY)
            INEGSYMPRECEDES = 0x00000056,           // mon sym precedes neg amt (derived from INEGCURR)
            INEGSEPBYSPACE = 0x00000057,            // mon sym sep by space from neg amt (derived from INEGCURR)

            FONTSIGNATURE = 0x00000058,             // font signature
            SISO639LANGNAME = 0x00000059,           // ISO abbreviated language name
            SISO3166CTRYNAME = 0x0000005A,          // ISO abbreviated country name

            IDEFAULTEBCDICCODEPAGE = 0x00001012,    // default ebcdic code page
            IPAPERSIZE = 0x0000100A,                // 1 = letter, 5 = legal, 8 = a3, 9 = a4
            SENGCURRNAME = 0x00001007,              // english name of currency
            SNATIVECURRNAME = 0x00001008,           // native name of currency
            SYEARMONTH = 0x00001006,                // year month format string
            SSORTNAME = 0x00001013,                 // sort name
            IDIGITSUBSTITUTION = 0x00001014,        // 0 = context, 1 = none, 2 = national

            SNAME = 0x0000005c,                     // locale name (with sort info) (ie: de-DE_phoneb)
            SDURATION = 0x0000005d,                 // time duration format
            SKEYBOARDSTOINSTALL = 0x0000005e,       // (windows only) keyboards to install
            SSHORTESTDAYNAME1 = 0x00000060,         // Shortest day name for Monday
            SSHORTESTDAYNAME2 = 0x00000061,         // Shortest day name for Tuesday
            SSHORTESTDAYNAME3 = 0x00000062,         // Shortest day name for Wednesday
            SSHORTESTDAYNAME4 = 0x00000063,         // Shortest day name for Thursday
            SSHORTESTDAYNAME5 = 0x00000064,         // Shortest day name for Friday
            SSHORTESTDAYNAME6 = 0x00000065,         // Shortest day name for Saturday
            SSHORTESTDAYNAME7 = 0x00000066,         // Shortest day name for Sunday
            SISO639LANGNAME2 = 0x00000067,          // 3 character ISO abbreviated language name
            SISO3166CTRYNAME2 = 0x00000068,         // 3 character ISO country name
            SNAN = 0x00000069,                      // Not a Number
            SPOSINFINITY = 0x0000006a,              // + Infinity
            SNEGINFINITY = 0x0000006b,              // - Infinity
            SSCRIPTS = 0x0000006c,                  // Typical scripts in the locale
            SPARENT = 0x0000006d,                   // Fallback name for resources
            SCONSOLEFALLBACKNAME = 0x0000006e,      // Fallback name for within the console
        }

        [DllImport("kernel32.dll")]
        public static extern int GetLocaleInfoEx(String lpLocaleName, LOCALE LCType, StringBuilder lpLCData, int cchData);

        [DllImport("kernel32.dll")]
        public static extern bool SetLocaleInfo(uint locale, LOCALE LCType, string lpLCData);

        #endregion 

        #region Timers

        //
        // Timers
        //
        [DllImport("user32.dll")]
        public static extern uint SetTimer(IntPtr hWnd, uint nIDEvent, uint uElapse, IntPtr lpTimerFunc);

        [DllImport("user32.dll")]
        public static extern bool KillTimer(IntPtr hWnd, uint uIDEvent);

        #endregion

        #region Webcam

        //
        // See http://msdn.microsoft.com/en-us/library/windows/desktop/dd757696%28v=vs.85%29.aspx
        //
        [DllImport("avicap32.dll")]
        public static extern IntPtr capCreateCaptureWindow(string lpszWindowName, int dwStyle, int x, int y, int nWidth, int nHeight, IntPtr hwndParent, int nID);

        /// <summary>
        /// Bitmap information values. See http://msdn.microsoft.com/en-us/library/windows/desktop/dd183376%28v=vs.85%29.aspx.
        /// </summary>
        public enum BI : uint
        {
            RGB = 0,
            RLE8,
            RLE4,
            BITFIELDS,
            JPEG,
            PNG
        }

        /// <summary>
        /// Contained in the BITMAPINFO structure.
        /// See http://msdn.microsoft.com/en-us/library/windows/desktop/dd318229%28v=vs.85%29.aspx.
        /// </summary>
        [StructLayout(LayoutKind.Sequential)]
        public struct BITMAPINFOHEADER
        {
            public static readonly BITMAPINFOHEADER Empty;
            public uint biSize;
            public int biWidth;
            public int biHeight;
            public ushort biPlanes;
            public ushort biBitCount;
            public uint biCompression;
            public uint biSizeImage;
            public int biXPelsPerMeter;
            public int biYPelsPerMeter;
            public uint biClrUsed;
            public uint biClrImportant;
        }

        /// <summary>
        /// Contained in the BITMAPINFO structure.
        /// See http://msdn.microsoft.com/en-us/library/dd162938.
        /// </summary>
        [StructLayout(LayoutKind.Sequential)]
        public struct RGBQUAD
        {
            public static readonly RGBQUAD Empty;
            public byte rgbBlue;
            public byte rgbGreen;
            public byte rgbRed;
            public byte rgbReserved;
        }

        /// <summary>
        /// For use with WM_CAP_GET_VIDEOFORMAT and WM_CAP_SET_VIDEOFORMAT. 
        /// See http://msdn.microsoft.com/en-us/library/dd183375.
        /// </summary>
        [StructLayout(LayoutKind.Sequential)]
        public struct BITMAPINFO
        {
            public static readonly BITMAPINFO Empty;
            public BITMAPINFOHEADER bmiHeader;
            [MarshalAs(UnmanagedType.ByValArray, SizeConst=256)]
            public RGBQUAD[] bmiColors;
        }

        // For use with WM_CAP_GET_VIDEOFORMAT and WM_CAP_SET_VIDEOFORMAT. 
        // See http://msdn.microsoft.com/en-us/library/windows/desktop/dd743908%28v=vs.85%29.aspx 
        // and http://msdn.microsoft.com/en-us/library/windows/desktop/dd757737%28v=vs.85%29.aspx.
        // Pass the size of the lParam structure as the wParam.
        [DllImport("user32.dll")]
        public static extern int SendMessage(IntPtr hWnd, uint Msg, uint wParam, ref BITMAPINFO lParam);

        /// <summary>
        /// For use with WM_CAP_GET_AUDIOFORMAT and WM_CAP_SET_AUTOFORMAT.
        /// See http://msdn.microsoft.com/en-us/library/windows/desktop/dd757713%28v=vs.85%29.aspx.
        /// </summary>
        public struct WAVEFORMATEX
        {
            public static readonly WAVEFORMATEX Empty;
            public ushort wFormatTag;
            public ushort nChannels;
            public uint nSamplesPerSec;
            public uint nAvgBytesPerSec;
            public ushort nBlockAlign;
            public ushort wBitsPerSample;
            public ushort cbSize;
        }

        // For use with WM_CAP_GET_AUDIOFORMAT and WM_CAP_SET_AUDIOFORMAT. 
        // See http://msdn.microsoft.com/en-us/library/windows/desktop/dd743908%28v=vs.85%29.aspx 
        // and http://msdn.microsoft.com/en-us/library/windows/desktop/dd757737%28v=vs.85%29.aspx.
        // Pass the size of the lParam structure as the wParam.
        [DllImport("user32.dll")]
        public static extern int SendMessage(IntPtr hWnd, uint Msg, uint wParam, ref WAVEFORMATEX lParam);

        /// <summary>
        /// Structure to use with WM_CAP_GET_STATUS message.
        /// See http://msdn.microsoft.com/en-us/library/windows/desktop/dd756939(v=vs.85).aspx.
        /// </summary>
        [StructLayout(LayoutKind.Sequential)]
        public struct CAPSTATUS
        {
            public static readonly CAPSTATUS Empty;
            public uint uiImageWidth;
            public uint uiImageHeight;
            public bool fLiveWindow;
            public bool fOverlayWindow;
            public bool fScale;
            public POINT ptScroll;
            public bool fUsingDefaultPalette;
            public bool fAudioHardware;
            public bool fCapFileExists;
            public uint dwCurrentVideoFrame;
            public uint dwCurrentVideoFramesDropped;
            public uint dwCurrentWaveSamples;
            public uint dwCurrentTimeElapsedMS;
            public IntPtr hPalCurrent; // HPALETTE -- handle to current palette
            public bool fCapturingNow;
            public uint dwReturn;
            public uint wNumVideoAllocated;
            public uint wNumAudioAllocated;
        };

        // For use with WM_CAP_GET_STATUS. 
        // See http://msdn.microsoft.com/en-us/library/windows/desktop/dd743906(v=vs.85).aspx.
        // Pass the size of the lParam structure as the wParam.
        [DllImport("user32.dll")]
        public static extern int SendMessage(IntPtr hWnd, uint Msg, uint wParam, ref CAPSTATUS lParam);

        /// <summary>
        /// Structure for use with the WM_CAP_GET_SEQUENCE_SETUP and WM_CAP_SET_SEQUENCE_SETUP messages. Used
        /// to configure streaming video capture to an *.avi file. See http://msdn.microsoft.com/en-us/library/windows/desktop/dd756942(v=vs.85).aspx.
        /// </summary>
        [StructLayout(LayoutKind.Sequential)]
        public struct CAPTUREPARMS 
        {
            public static readonly CAPTUREPARMS Empty;
            public uint dwRequestMicroSecPerFrame;
            public bool fMakeUserHitOKToCapture;
            public uint wPercentDropForError;
            public bool fYield;
            public uint dwIndexSize;
            public uint wChunkGranularity;
            public bool fUsingDOSMemory;
            public uint wNumVideoRequested;
            public bool fCaptureAudio;
            public uint wNumAudioRequested;
            public uint vKeyAbort;
            public bool fAbortLeftMouse;
            public bool fAbortRightMouse;
            public bool fLimitEnabled;
            public uint wTimeLimit;
            public bool fMCIControl;
            public bool fStepMCIDevice;
            public uint dwMCIStartTime;
            public uint dwMCIStopTime;
            public bool fStepCaptureAt2x;
            public uint wStepCaptureAverageFrames;
            public uint dwAudioBufferSize;
            public bool fDisableWriteCache;
            public uint AVStreamMaster;
        };

        // For use with WM_CAP_GET_SEQUENCE_SETUP and WM_CAP_SET_SEQUENCE_SETUP. 
        // See http://msdn.microsoft.com/en-us/library/windows/desktop/dd743905(v=vs.85).aspx and
        // http://msdn.microsoft.com/en-us/library/windows/desktop/dd757733(v=vs.85).aspx.
        // Pass the size of the lParam structure as the wParam.
        [DllImport("user32.dll")]
        public static extern int SendMessage(IntPtr hWnd, uint Msg, uint wParam, ref CAPTUREPARMS lParam);

        // For use with WM_CAP_FILE_GET_CAPTURE_FILE and WM_CAP_FILE_SET_CAPTURE_FILE.
        // See http://msdn.microsoft.com/en-us/library/windows/desktop/dd743898(v=vs.85).aspx and
        // http://msdn.microsoft.com/en-us/library/windows/desktop/dd743901(v=vs.85).aspx.
        // For getting, pass the buffer size for wParam and the string to hold the filename.
        // For setting, pass 0 for wParam and the string for the filename.
        [DllImport("user32.dll")]
        public static extern int SendMessage(IntPtr hWnd, uint Msg, uint wParam, string lParam);

        #endregion
    }
}
