﻿using System;
using System.Text;
using System.Drawing;
using System.Diagnostics;

namespace WobbrockLib.Windows
{
	/// <summary>
	/// Utility class that enables interaction with windows on the Microsoft Windows operating 
	/// system. It heavily relies on the Win32 utility class.
	/// </summary>
	public static class WindowEx
	{
		#region Focus Function Flags

		/// <summary>
		/// These flags are used with the FocusWindow call to indicate what focusing
		/// functions will be carried out. Each of these flags corresponds to a Win32
		/// function that enables some focusing-like behavior to take place.
		/// </summary>
		[Flags]
		public enum FocusFns
		{
			/// <summary>
			/// No focus operations will be carried out. Makes the FocusWindow call a no-op
			/// unless it is combined with other flags, in which case, it is superfluous.
			/// </summary>
			None = 0x00,

			/// <summary>
			/// The Win32 function BringWindowToTop will be called. This function brings the given window to the top of the Z-order.
			/// </summary>
			/// <see cref="https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-bringwindowtotop"/>
			BringToTop = 0x01,

			/// <summary>
			/// The Win32 function SetForegroundWindow will be called. This function sets the given window in the foreground.
			/// </summary>
			/// <see cref="https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-setforegroundwindow"/>
			Foreground = 0x02,

			/// <summary>
			/// The Win32 function SetActiveWindow will be called. This function activates the given window.
			/// </summary>
			/// <see cref="https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-setactivewindow"/>
			Activate = 0x04,

			/// <summary>
			/// The Win32 function SetFocus will be called. This function directs the keyboard input focus to the given window.
			/// </summary>
			/// <see cref="https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-setfocus"/>
			Focus = 0x08,

			/// <summary>
			/// The Win32 function SetCapture will be called. This function captures the mouse. Only a visible portion of a
			/// window, usually the foreground window, can capture the mouse. Capturing ensures that mouse events are sent
			/// to the window.
			/// </summary>
			/// <see cref="https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-setcapture"/>
			Capture = 0x10,

			/// <summary>
			/// The Win32 function AttachThreadInput will be called. It is necessary to couple this flag with SetFocus when
			/// one application is setting the input focus to a window not created on its thread. If used, the AttachThreadInput
			/// call for attaching, and then again for detaching, will surround any other flags' calls.
			/// </summary>
			/// <see cref="https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-attachthreadinput"/>
			AttachInput = 0x20,

			/// <summary>
			/// Encodes all of the FocusFlags together in a bitwise combination.
			/// </summary>
			All = (BringToTop | Foreground | Activate | Focus | Capture | AttachInput)
		}

		#endregion

		#region Window Focus

		/// <summary>
		/// Performs focusing-type behaviors on the given window, dictated by the given focus flags.
		/// </summary>
		/// <param name="hWnd">The handle to the window on which to set the focus.</param>
		/// <param name="fns">The flags indicating what kinds of focusing functions to perform.</param>
		/// <returns>True if, for whatever focus flags are given, all functions succeed; false otherwise.</returns>
		public static bool FocusWindow(IntPtr hWnd, FocusFns fns)
		{
			return FocusWindow(hWnd, fns, false);
		}

		/// <summary>
		/// Performs focusing-type behaviors on the given window, dictated by the given focus flags.
		/// </summary>
		/// <param name="hWnd">The handle to the window on which to set the focus.</param>
		/// <param name="fns">The flags indicating what kinds of focusing functions to perform.</param>
		/// <param name="verbose">If true, results of each function call will be printed to the Console.</param>
		/// <returns>True if, for whatever focus flags are given, all functions succeed; false otherwise.</returns>
		public static bool FocusWindow(IntPtr hWnd, FocusFns fns, bool verbose)
		{
			if (hWnd == IntPtr.Zero)
				return false; // trivial fail
			if (fns == FocusFns.None)
				return true; // trivial success

			// success outcomes
			bool s01 = true;        // BringWindowToTop
			bool s02 = true;        // SetForegroundWindow
			bool s04 = true;        // SetActiveWindow
			bool s08 = true;        // SetFocus
			bool s10 = true;        // SetCapture
			bool s20a = true;       // AttachThreadInput (attach)
			bool s20b = true;       // AttachThreadInput (detach)

			if (verbose)
			{
				string title = GetWindowTitle(hWnd);
				Debug.WriteLine(String.Format("<FocusWindow title=\"{0}\">", title));
			}

			uint procId = 0u;
			uint idAttach = 0u;
			uint idAttachTo = 0u;

			if ((fns & FocusFns.AttachInput) != 0)
			{
				idAttach = Win32.GetCurrentThreadId();
				idAttachTo = Win32.GetWindowThreadProcessId(hWnd, ref procId);
				s20a = Win32.AttachThreadInput(idAttach, idAttachTo, true); // attach
				if (verbose)
					Debug.WriteLine("\tAttachThreadInput: {0}", s20a);
			}

			if ((fns & FocusFns.BringToTop) != 0)
			{
				s01 = Win32.BringWindowToTop(hWnd);
				if (verbose)
					Debug.WriteLine("\tBringWindowToTop: {0}", s01);
			}

			if ((fns & FocusFns.Foreground) != 0)
			{
				s02 = Win32.SetForegroundWindow(hWnd);
				if (verbose)
					Debug.WriteLine("\tSetForegroundWindow: {0}", s02);
			}

			if ((fns & FocusFns.Activate) != 0)
			{
				IntPtr prevHwnd = Win32.SetActiveWindow(hWnd);
				s04 = (prevHwnd != IntPtr.Zero);
				if (verbose)
					Debug.WriteLine("\tSetActiveWindow: {0}", s04);
			}

			if ((fns & FocusFns.Focus) != 0)
			{
				IntPtr prevHwnd = Win32.SetFocus(hWnd);
				s08 = (prevHwnd != IntPtr.Zero);
				if (verbose)
					Debug.WriteLine("\tSetFocus: {0}", s08);
			}

			if ((fns & FocusFns.Capture) != 0)
			{
				IntPtr prevHwnd = Win32.SetCapture(hWnd);
				s10 = (prevHwnd != IntPtr.Zero);
				if (verbose)
					Debug.WriteLine("\tSetCapture: {0}", s10);
			}

			if ((fns & FocusFns.AttachInput) != 0)
			{
				s20b = Win32.AttachThreadInput(idAttach, idAttachTo, false); // detach
				if (verbose)
					Debug.WriteLine("\tDetachThreadInput: {0}", s20b);
			}

			if (verbose)
			{
				Debug.WriteLine("</FocusWindow>");
			}

			// all that we tried must be successful
			return (s01 && s02 && s04 && s08 && s10 && s20a && s20b);
		}

		/// <summary>
		/// Sets the focus on the top-level window at the given screen coordinates.
		/// </summary>
		/// <param name="pt">The screen coordinate where to focus the window.</param>
		/// <param name="fns">The flags indicating what kinds of focusing functions to perform.</param>
		/// <param name="topLevel">If true, only gets a top level window, not a child control window.</param>
		/// <returns>True if, for whatever focus flags are given, all functions succeed; false otherwise.</returns>
		public static bool FocusWindowAt(Point pt, FocusFns fns, bool topLevel)
		{
			return FocusWindowAt(pt, fns, topLevel, false);
		}

		/// <summary>
		/// Sets the focus on the top-level window at the given screen coordinates.
		/// </summary>
		/// <param name="pt">The screen coordinate where to focus the window.</param>
		/// <param name="fns">The flags indicating what kinds of focusing functions to perform.</param>
		/// <param name="topLevel">If true, only gets a top level window, not a child control window.</param>
		/// <param name="verbose">If true, results of each function call will be printed to the Console.</param>
		/// <returns>True if, for whatever focus flags are given, all functions succeed; false otherwise.</returns>
		public static bool FocusWindowAt(Point pt, FocusFns fns, bool topLevel, bool verbose)
		{
			IntPtr hwnd = GetWindowAt(pt, topLevel);
			return FocusWindow(hwnd, fns, verbose);
		}

		#endregion

		#region Getting Windows

		/// <summary>
		/// Gets the window handle at the given screen coordinate. Because Windows regards controls as 
		/// "windows," if the screen point falls on a widget, we need to ask for the root ancestor window, 
		/// which will be the top window in the UI hierarchy for that widget.
		/// </summary>
		/// <param name="pt">The screen coordinate where to find the window.</param>
		/// <param name="topLevel">If true, only gets a top level window, not a child control window.</param>
		/// <returns>The handle to the window at the given screen coordinate.</returns>
		public static IntPtr GetWindowAt(Point pt, bool topLevel)
		{
			Win32.POINT point;
			point.x = pt.X;
			point.y = pt.Y;
			IntPtr hWnd = Win32.WindowFromPoint(point);
			return topLevel ? Win32.GetAncestor(hWnd, (uint) Win32.GA.ROOT) : hWnd;
		}

		/// <summary>
		/// Gets the top-level window that contains the given window handle. If the window handle is a
		/// top-level window already, then its own handle is returned.
		/// </summary>
		/// <param name="hWnd">The handle of the window or control for which to get its top-level
		/// window.</param>
		/// <returns>The handle of a top-level window, or IntPtr.Zero if hWnd is invalid or is the
		/// desktop window.</returns>
		public static IntPtr GetTopLevelWindow(IntPtr hWnd)
		{
			if (hWnd != IntPtr.Zero)
			{
				return Win32.GetAncestor(hWnd, (uint) Win32.GA.ROOT);
			}
			return IntPtr.Zero;
		}

		/// <summary>
		/// Gets the title of the window with the given handle. Can be useful for ensuring the focus
		/// is sent to the desired window.
		/// </summary>
		/// <param name="hWnd">The window handle whose title to retrieve.</param>
		/// <returns>The window title, if any, from the given window.</returns>
		public static string GetWindowTitle(IntPtr hWnd)
		{
			int len = Win32.GetWindowTextLength(hWnd);
			if (len > 0)
			{
				StringBuilder sb = new StringBuilder(len + 1);
				Win32.GetWindowText(hWnd, sb, sb.Capacity);
				return sb.ToString();
			}
			return String.Empty;
		}

		#endregion

		#region Window Styles

		/// <summary>
		/// Gets whether the given window is topmost or not. This is akin to testing whether it has the
		/// WS_EX_TOPMOST extended window style.
		/// </summary>
		/// <param name="hWnd">The window to test for topmost-ness.</param>
		/// <returns>True if the given window has the WS_EX_TOPMOST style; false otherwise.</returns>
		public static bool GetTopMost(IntPtr hWnd)
		{
			uint exStyle = Win32.GetWindowLong(hWnd, (int) Win32.GWL.EXSTYLE);
			return ((exStyle & (uint) Win32.WS_EX.TOPMOST) != 0u);
		}

		/// <summary>
		/// Sets a window to be topmost or not topmost. This is akin to giving it the WS_EX_TOPMOST extended 
		/// window style.
		/// </summary>
		/// <param name="hWnd">A handle to the window to set topmost or not.</param>
		/// <param name="topMost">If true, the window will be made topmost; if false, it will be made non-topmost.</param>
		/// <returns>True if the setting was successful; false otherwise. If the given hWnd already had
		/// the requested setting, then true is returned as a trivial success.</returns>
		public static bool SetTopMost(IntPtr hWnd, bool topMost)
		{
			uint retval = 1u; // non-zero is success; zero is failure
			uint exStyle = Win32.GetWindowLong(hWnd, (int) Win32.GWL.EXSTYLE);
			if (topMost && ((exStyle & (uint) Win32.WS_EX.TOPMOST) == 0u)) // set topmost and we're not topmost
			{
				retval = Win32.SetWindowLong(hWnd, (int) Win32.GWL.EXSTYLE, exStyle | (uint) Win32.WS_EX.TOPMOST);
			}
			else if (!topMost && ((exStyle & (uint) Win32.WS_EX.TOPMOST) != 0u)) // set non-topmost and we are topmost
			{
				retval = Win32.SetWindowLong(hWnd, (int) Win32.GWL.EXSTYLE, exStyle & ~((uint) Win32.WS_EX.TOPMOST));
			}
			return (retval != 0u);
		}

		/// <summary>
		/// Indicates whether the given window or control has the WS_EX_TRANSPARENT extended window style. 
		/// This is NOT visual transparency. Rather, it is "input transparency." For example, clicks 
		/// will pass through the window to whatever window underlies it. Also, asking for the window 
		/// that lies on a point within its area will give us the window below this one.
		/// </summary>
		/// <param name="hWnd">The window handle for the window to check for transparency.</param>
		/// <returns>True if the given window is set to "input transparent;" false otherwise.</returns>
		public static bool GetInputTransparent(IntPtr hWnd)
		{
			uint exStyle = Win32.GetWindowLong(hWnd, (int) Win32.GWL.EXSTYLE);
			return ((exStyle & (uint) Win32.WS_EX.TRANSPARENT) != 0u);
		}

		/// <summary>
		/// Sets whether the given window is given the WS_EX_TRANSPARENT extended window style. This
		/// is NOT visual transparency. Rather, it is "input transparency." For example, clicks will
		/// pass through the window to whatever window underlies it. Also, asking for the window that
		/// lies on a point within its area will give us the window below this one.
		/// <param name="hWnd">The window handle for the window to turn transparency on or off.</param>
		/// <param name="transparent">If true, the window will be set to "input transparent;" if false, 
		/// the window will be set to not have such transparency. The desired setting is only applied if
		/// it is not already currently set.</param>
		/// <returns>True if the setting was successful; false otherwise. If the given hWnd already had
		/// the requested setting, then true is returned as a trivial success.</returns>
		/// <see cref="https://docs.microsoft.com/en-us/windows/win32/winmsg/extended-window-styles"/>
		public static bool SetInputTransparent(IntPtr hWnd, bool transparent)
		{
			uint retval = 1u; // non-zero is success; zero is failure
			uint exStyle = Win32.GetWindowLong(hWnd, (int) Win32.GWL.EXSTYLE);
			if (transparent && ((exStyle & (uint) Win32.WS_EX.TRANSPARENT) == 0u)) // set transparent and we're not transparent
			{
				retval = Win32.SetWindowLong(hWnd, (int) Win32.GWL.EXSTYLE, exStyle | (uint) Win32.WS_EX.TRANSPARENT);
			}
			else if (!transparent && ((exStyle & (uint) Win32.WS_EX.TRANSPARENT) != 0u)) // set non-transparent and we are transparent
			{
				retval = Win32.SetWindowLong(hWnd, (int) Win32.GWL.EXSTYLE, exStyle & ~((uint) Win32.WS_EX.TRANSPARENT));
			}
			return (retval != 0u);
		}

		#endregion
	}
}
