﻿using System;
using System.Text;
using System.Collections.Generic;

namespace WobbrockLib.Windows
{
	public abstract class UIAMenus
	{
		#region Types

		/// <summary>
		/// Type indicating a menu action: whether it is opened, closed, or both.
		/// </summary>
		[Flags]
		public enum MenuEvent
		{
			None = 0x00,
			Open = 0x01,
			Close = 0x02,
			OpenClose = (Open | Close)
		}

		/// <summary>
		/// Type indicating the user interface tree that will be registered.
		/// </summary>
		public enum UITree
		{
			None = 0,       // TreeScope.None
			Self = 1,       // TreeScope.Element => self only
			Kids = 2,       // TreeScope.Children => immediate children
			SelfKids = 3,   // TreeScope.Element | TreeScope.Children
			Subtree = 7     // TreeScope.Subtree => self plus all descendants
		}

		#endregion

		#region Events

		/// <summary>
		/// Event for when a menu is opened on any registered window.
		/// </summary>
		public event EventHandler<IntPtr> MenuOpened;

		/// <summary>
		/// Event for when a menu is closed on any registered window.
		/// </summary>
		public event EventHandler<IntPtr> MenuClosed;

		#endregion

		#region Fields

		/// <summary>
		/// Keep track of registered window handles for menu events and prevent the same
		/// window from being registered more than once with the same window handle.
		/// </summary>
		protected readonly List<KeyValuePair<IntPtr, MenuEvent>> _listeners;

		#endregion

		#region Constructor

		/// <summary>
		/// Default constructor.
		/// </summary>
		public UIAMenus()
		{
			_listeners = new List<KeyValuePair<IntPtr, MenuEvent>>();
		}

		#endregion

		#region Abstract

		/// <summary>
		/// Register for MenuOpened and MenuClosed events through UI Automation.
		/// </summary>
		/// <param name="hWnd">A handle to the window or control on which to listen for menu events.</param>
		/// <param name="what">What menu events to listen for: opened, closed, or both.</param>
		/// <param name="where">Where to listen for the menu events in the UI tree.</param>
		/// <returns>True if we successfully register for the events; false otherwise.</returns>
		public abstract bool RegisterListener(IntPtr hWnd, MenuEvent what, UITree where);

		/// <summary>
		/// Unregister all of the event handlers that have been registered.
		/// </summary>
		public abstract void UnregisterListeners();

		#endregion

		#region Raising Events

		/// <summary>
		/// Raises the MenuOpened event. Called by subclasses.
		/// </summary>
		/// <param name="e">The arguments for this event, in this case, an IntPtr to a window handle
		/// on which the menu was opened.</param>
		protected virtual void OnMenuOpened(IntPtr e)
		{
			this.MenuOpened?.BeginInvoke(this, e, null, null);
		}

		/// <summary>
		/// Raises the MenuClosed event. Called by subclasses.
		/// </summary>
		/// <param name="e">The arguments for this event, in this case, an IntPtr to a window handle
		/// on which the menu was closed.</param>
		protected virtual void OnMenuClosed(IntPtr e)
		{
			this.MenuClosed?.BeginInvoke(this, e, null, null);
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets the number of menu event listeners currently active.
		/// </summary>
		public int Listeners
		{
			get { return _listeners.Count; }
		}

		/// <summary>
		/// Gets a string showing current menu listeners.
		/// </summary>
		public string NowListening
		{
			get
			{
				StringBuilder sb = new StringBuilder("Listeners: ");
				if (_listeners?.Count > 0)
				{
					foreach (KeyValuePair<IntPtr, MenuEvent> kvp in _listeners)
					{
						sb.AppendFormat("<{0}:{1}>; ", kvp.Value, kvp.Key);
					}
				}
				else // null or zero
				{
					sb.Append("None.");
				}
				return sb.ToString();
			}
		}

		#endregion

		#region Methods

		/// <summary>
		/// Tests whether this instance is currently listening for a menu event on the given window 
		/// handle.
		/// </summary>
		/// <param name="hWnd">A handle to the window to test.</param>
		/// <param name="what">The menu event type to listen for. This can be Open, Close, or OpenClose.</param>
		/// <returns>True if the event or events are being listened for on the given window; false otherwise.
		/// Also, false is returned if evt is None.</returns>
		public bool IsListeningTo(IntPtr hWnd, MenuEvent what)
		{
			if (what == MenuEvent.None)
				return false;

			bool wantOpen = ((what & MenuEvent.Open) != 0);
			bool wantClose = ((what & MenuEvent.Close) != 0);

			bool hasOpen = false;
			bool hasClose = false;

			foreach (KeyValuePair<IntPtr, MenuEvent> kvp in _listeners)
			{
				if (wantOpen && !hasOpen)
				{
					if (kvp.Key == hWnd && kvp.Value == MenuEvent.Open)
						hasOpen = true;
				}
				if (wantClose && !hasClose)
				{
					if (kvp.Key == hWnd && kvp.Value == MenuEvent.Close)
						hasClose = true;
				}
				if (hasOpen && hasClose)
					break; // no need to continue
			}

			if (wantOpen && !wantClose)
				return hasOpen;
			else if (wantClose && !wantOpen)
				return hasClose;
			else if (wantOpen && wantClose)
				return hasOpen && hasClose;
			else
				return false; // can't happen, for compiler
		}

		#endregion
	}
}
